<?php
/* Copyright start
  Copyright (C) 2008 - 2020 Fortinet Inc.
  All rights reserved.
  FORTINET CONFIDENTIAL & FORTINET PROPRIETARY SOURCE CODE
  Copyright end */

namespace Cybersponse\ActorsBundle\Serializer;


use Cybersponse\ActorsBundle\Entity\Actor;
use Cybersponse\ActorsBundle\Provider\AvatarProvider;
use Cybersponse\AuthorizationBundle\Entity\Role;
use Cybersponse\AuthorizationBundle\Entity\Team;
use Cybersponse\CoreBundle\Provider\RuntimeContextProvider;
use Cybersponse\CoreBundle\Serializer\EfficientItemNormalizer;
use Cybersponse\SettingsBundle\Entity\UserSettings;
use Dunglas\ApiBundle\Api\IriConverterInterface;
use Dunglas\ApiBundle\Api\ResourceCollectionInterface;
use Symfony\Component\Security\Core\Authorization\AuthorizationCheckerInterface;
use Symfony\Component\Security\Core\Exception\AccessDeniedException;
use Symfony\Component\Serializer\Normalizer\NormalizerInterface;
use Symfony\Component\Serializer\Serializer;

class ActorNormalizer extends EfficientItemNormalizer implements NormalizerInterface
{
    /** @var AuthorizationCheckerInterface */
    protected $authorizationChecker;

    /** @var IriConverterInterface */
    protected $iriConverter;

    /** @var ResourceCollectionInterface */
    protected $resourceCollection;

    /** @var RuntimeContextProvider */
    protected $runtimeContextProvider;

    /**
     * @inheritdoc
     */
    public function supportsNormalization($data, $format = null)
    {
        return self::FORMAT === $format && ($data instanceof Actor);
    }

    /**
     * @inheritdoc
     */
    public function supportsDenormalization($data, $type, $format = null)
    {
        return is_subclass_of($type, Actor::class);
    }

    /**
     * @param Actor $actor
     * @inheritdoc
     */
    public function normalize($actor, $format = null, array $context = [])
    {
        $resource                                       = $this->guessResource($actor);
        $context['resource']                            = $resource;
        $context['ignore_authorization']                = true;
        $context['ignore_field_authorization']          = true;
        $context['ignore_relationship_authorization']   = true;
        $context['ignore_depth']                        = false;
        $result                                         = parent::normalize($actor, $format, $context);

        if (is_array($result)) {
            $userSettingsResource = $this->resourceCollection->getResourceForEntity(UserSettings::class);
            $userSettingsIri = $this->iriConverter->getIriFromResource($userSettingsResource);
            $result['@settings'] = sprintf('%s/%s', $userSettingsIri, $actor->getUuid());
        }

        return $result;
    }

    /**
     * @inheritdoc
     */
    public function denormalize($data, $class, $format = null, array $context = [])
    {
        // TODO: Check for perm:security.create && !array_key_exists(object_to_populate, $context) ?
        $context['ignore_field_authorization'] = true;

        if (!$this->canUpdateSecurityModels() && array_key_exists('object_to_populate', $context)) {
            /** @var Actor $previousActor */
            $previousActor = $context['object_to_populate'];

            if (array_key_exists('roles', $data)) {
                $previousRoles = $previousActor->getRoles();
                $decodedRoles = $data['roles'];
                $this->denyAccessIfChanged($decodedRoles, $previousRoles, Role::class);
            }

            if (array_key_exists('teams', $data)) {
                $previousTeams = $previousActor->getTeams()->toArray();
                $decodedTeams = $data['teams'];
                $this->denyAccessIfChanged($decodedTeams, $previousTeams, Team::class);
            }
        }

        return parent::denormalize($data, $class, $format, $context);
    }

    /**
     * @param mixed $avatarProvider
     */
    public function setAvatarProvider($avatarProvider)
    {
        $this->avatarProvider = $avatarProvider;
    }

    /**
     * @return bool
     */
    protected function canUpdateSecurityModels()
    {
        return $this->runtimeContextProvider->isCli() || $this->authorizationChecker->isGranted('perm:security.update');
    }

    /**
     * @param AuthorizationCheckerInterface $authorizationChecker
     */
    public function setAuthorizationChecker($authorizationChecker)
    {
        $this->authorizationChecker = $authorizationChecker;
    }

    /**
     * @param IriConverterInterface $iriConverter
     */
    public function setIriConverter($iriConverter)
    {
        $this->iriConverter = $iriConverter;
    }

    /**
     * @param ResourceCollectionInterface $resourceCollection
     */
    public function setResourceCollection($resourceCollection)
    {
        $this->resourceCollection = $resourceCollection;
    }

    /**
     * @param RuntimeContextProvider $runtimeContextProvider
     */
    public function setRuntimeContextProvider($runtimeContextProvider)
    {
        $this->runtimeContextProvider = $runtimeContextProvider;
    }

    /**
     * @param $decodedEntities
     * @param $previousEntities
     * @param $className
     */
    protected function denyAccessIfChanged($decodedEntities, $previousEntities, $className)
    {
        $newEntities = [];
        foreach ($decodedEntities as $decodedNewEntity) {
            if (is_string($decodedNewEntity)) {
                $newEntities[] = $this->iriConverter->getItemFromIri($decodedNewEntity);
            } else {
                /** @var Serializer $serializer */
                $serializer = $this->serializer;
                $newEntities[] = $serializer->denormalize($decodedNewEntity, $className);
            }
        }

        if (count($newEntities) !== count($previousEntities)) {
            throw new AccessDeniedException();
        }

        foreach ($newEntities as $newEntity) {
            if (!in_array($newEntity, $previousEntities)) {
                throw new AccessDeniedException();
            }
        }
    }
}
