<?php
/* Copyright start
  Copyright (C) 2008 - 2020 Fortinet Inc.
  All rights reserved.
  FORTINET CONFIDENTIAL & FORTINET PROPRIETARY SOURCE CODE
  Copyright end */

namespace Cybersponse\AuthenticationBundle\Hmac;


use Cybersponse\CoreBundle\Provider\CacheableInterface;
use Cybersponse\ProxyBundle\GuzzleHttp\HmacClient;
use Cybersponse\ProxyBundle\Provider\SecretKeyProviderInterface;
use GuzzleHttp\Exception\ClientException;
use GuzzleHttp\Psr7\Request as GuzzleRequest;

class DasHmacSecretKeyProvider implements SecretKeyProviderInterface
{
    const UUID = 'uuid';
    const SECRET_KEY = 'secretKey';
    const DAS_SECRET_KEY = 'pkey';
    const DAS_KEY_IDENTIFIER = 'pubkey';

    /** @var HmacClient */
    protected $dasClient;

    /** @var CacheableInterface */
    protected $cache;

    /** @var string */
    protected $uuid;

    /** @var string */
    protected $secretKey;

    /** @var bool */
    protected $isInitialized = false;

    /**
     * @var int
     */
    protected $cacheLifeTime;

    /**
     * @param HmacClient $dasClient
     */
    public function setDasClient($dasClient)
    {
        $this->dasClient = $dasClient;
    }

    /**
     * @param CacheableInterface $cache
     */
    public function setCache($cache)
    {
        $this->cache = $cache;
    }

    /**
     * @param int $cacheLifeTime
     */
    public function setCacheLifeTime($cacheLifeTime)
    {
        $this->cacheLifeTime = $cacheLifeTime;
    }

    /**
     * @param string $keyIdentifier
     * @return string
     * @throws \Exception
     */
    public function getSecretKey($keyIdentifier)
    {
        if (!$this->isInitialized) {
            $this->initializeCredentials($keyIdentifier);
        }

        return $this->secretKey;
    }

    /**
     * @return string
     * @throws \Exception
     */
    public function getUserId()
    {
        if (!$this->isInitialized) {
            throw new \Exception('Cannot load API user\'s UUID without having first retrieved the secret key.');
        }

        return $this->uuid;
    }

    /**
     * @param string $keyIdentifier
     * @return bool
     * @throws \Exception
     */
    protected function initializeCredentials($keyIdentifier)
    {
        // Only necessary once
        if ($this->isInitialized) {
            return true;
        }

        // Check cache first
        if ($this->cache->cacheContains($keyIdentifier)) {
            $cachedCredentials = $this->cache->getFromCache($keyIdentifier);

            $this->secretKey = $cachedCredentials[self::SECRET_KEY];
            $this->uuid = $cachedCredentials[self::UUID];

            $this->isInitialized = true;
            return true;
        }

        // Reach out to DAS
        $publicKeyRequest = [
            self::DAS_KEY_IDENTIFIER => $keyIdentifier
        ];

        $apikeysUri = sprintf('%s/%s/', $this->dasClient->getTargetUri(), 'apikeys');
        $encodedPublicKeyBody = json_encode($publicKeyRequest);
        $request = new GuzzleRequest('PUT', $apikeysUri, [], $encodedPublicKeyBody);

        try {
            $response = $this->dasClient->send($request);
            $responseBody = json_decode($response->getBody(), true);
            $this->secretKey = $responseBody[self::DAS_SECRET_KEY];
            $this->uuid = $responseBody[self::UUID];

            $credentials = [
                self::UUID => $this->uuid,
                self::SECRET_KEY => $this->secretKey
            ];

            $this->cache->persistToCache($keyIdentifier, $credentials, $this->cacheLifeTime);
            $this->isInitialized = true;
            return true;
            
        } catch (\Exception $clientException) {
            error_log($clientException->getMessage());
            throw new \Exception('Unable to load API credentials from cache or DAS.', 0, $clientException);
        }

    }
}
