<?php
/* Copyright start
  Copyright (C) 2008 - 2020 Fortinet Inc.
  All rights reserved.
  FORTINET CONFIDENTIAL & FORTINET PROPRIETARY SOURCE CODE
  Copyright end */

namespace Cybersponse\AuthorizationBundle;


use Cybersponse\CoreBundle\Api\ModelAwareResource;
use Dunglas\ApiBundle\Api\ResourceInterface;
use Symfony\Component\Security\Core\Authentication\Token\Storage\TokenStorageInterface;
use Symfony\Component\Security\Core\Authorization\AuthorizationCheckerInterface;
use Symfony\Component\Security\Core\Exception\AccessDeniedException;

class AuthorizationProvider
{
    /** @var TokenStorageInterface */
    protected $tokenStorage;

    /** @var AuthorizationCheckerInterface */
    protected $authorizationChecker;

    /**
     * @param AuthorizationCheckerInterface $authorizationChecker
     * @param TokenStorageInterface $tokenStorage
     */
    function __construct(AuthorizationCheckerInterface $authorizationChecker, TokenStorageInterface $tokenStorage)
    {
        $this->authorizationChecker = $authorizationChecker;
        $this->tokenStorage = $tokenStorage;
    }

    /**
     * @param ResourceInterface $resource
     * @param array|string $actions
     * @param mixed $object
     *
     * @throws AccessDeniedException
     */
    public function denyAccessUnlessGranted($resource, $actions, $object = null)
    {
        if (!$this->isGranted($resource, $actions, $object)) {
            throw new AccessDeniedException();
        }
    }

    /**
     * @param ResourceInterface $resource
     * @param string $field
     * @param string|array $actions
     */
    public function denyFieldAccessUnlessGranted($resource, $field, $actions)
    {
        if (is_array($actions)) {
            foreach ($actions as $action) {
                $this->denyAccessUnlessGranted($resource, [sprintf('%s.%s', $field, $action)]);
            }
        } else {
            $this->denyAccessUnlessGranted($resource, [sprintf('%s.%s', $field, $actions)]);
        }
    }

    /**
     * @param ResourceInterface $resource
     * @param array|string $actions
     * @param mixed $object
     *
     * @return bool
     */
    public function isGranted($resource, $actions, $object = null)
    {
        if (!is_array($actions)) {
            $actions = [$actions];
        }

        $permissionsRequired = [];

        foreach ($actions as $action) {
            $permission = $this->buildPermission($action, $resource);
            $permissionsRequired[] = $permission;
        }

        return $this->authorizationChecker->isGranted($permissionsRequired, $object);
    }

    /**
     * @param string $action
     * @param ResourceInterface $resource
     * @return string
     *
     * @throws \Exception
     */
    protected function buildPermission($action, $resource)
    {
        $moduleName = $this->getModuleNameFromResource($resource);

        return sprintf('perm:%s.%s', $moduleName, $action);
    }

    /**
     * @param ResourceInterface $resource
     * @return string
     * @throws \Exception
     */
    public static function getModuleNameFromResource($resource)
    {
        if (!$resource instanceof ModelAwareResource) {
            throw new \Exception(sprintf('Resource: %s is not instance of ModelAwareResource.', $resource->getShortName()));
        }
        return $resource->getModuleType();
    }
}
