<?php
/* Copyright start
  Copyright (C) 2008 - 2020 Fortinet Inc.
  All rights reserved.
  FORTINET CONFIDENTIAL & FORTINET PROPRIETARY SOURCE CODE
  Copyright end */

namespace Cybersponse\AuthorizationBundle\Permissions;


use Cybersponse\AuthorizationBundle\Entity\Role;

class PermissionsDictionaryBuilder
{
    const FIELD_PERMS = 'fields';
    const PERM_CRUD_CREATE = 'create';
    const PERM_CRUD_READ = 'read';
    const PERM_CRUD_UPDATE = 'update';
    const PERM_CRUD_DELETE = 'delete';
    const PERM_CRUD_EXECUTE = 'execute';

    const PERM_FIELD_WRITE = 'write';
    const PERM_FIELD_READ = 'read';

    const MERGE_STRATEGY_UNION = 0;
    const MERGE_STRATEGY_TAKE_LEFT = -1;
    const MERGE_STRATEGY_TAKE_RIGHT = 1;

    /**
     * @param Role[] $roles
     * @return array
     */
    public function buildPermissionsDictionary($roles)
    {
        $permissionsDictionary = [];

        foreach ($roles as $role) {
            $rolePermissions = [];
            foreach ($role->getModulePermissions() as $modulePermission) {
                $module = $modulePermission->getModule();
                $moduleType = $module->getType();

                $modulePermissions = $this->buildDefaultPermissions();
                $modulePermissions[static::PERM_CRUD_CREATE] = $modulePermission->getCanCreate();
                $modulePermissions[static::PERM_CRUD_READ] = $modulePermission->getCanRead();
                $modulePermissions[static::PERM_CRUD_UPDATE] = $modulePermission->getCanUpdate();
                $modulePermissions[static::PERM_CRUD_DELETE] = $modulePermission->getCanDelete();
                $modulePermissions[static::PERM_CRUD_EXECUTE] = $modulePermission->getCanExecute();

                $allFieldPermissions = [];
                foreach ($modulePermission->getFieldPermissions() as $fieldPermission) {
                    $field = $fieldPermission->getField();

                    $fieldPermissions = array_key_exists($field, $allFieldPermissions) ? $allFieldPermissions[$field] : [];

                    $readPermission = array_key_exists(static::PERM_FIELD_READ, $fieldPermissions) ? $fieldPermissions[static::PERM_FIELD_READ] : null;
                    if (!$readPermission) {
                        $readPermission = $fieldPermission->getCanRead();
                    }
                    $writePermission = array_key_exists(static::PERM_FIELD_READ, $fieldPermissions) ? $fieldPermissions[static::PERM_FIELD_READ] : null;
                    if (!$writePermission) {
                        $writePermission = $fieldPermission->getCanWrite();
                    }

                    if (!is_null($readPermission)) {
                        $fieldPermissions[static::PERM_FIELD_READ] = $readPermission;
                    }
                    if (!is_null($writePermission)) {
                        $fieldPermissions[static::PERM_FIELD_WRITE] = $writePermission;
                    }

                    if (count($fieldPermissions) !== 0) {
                        $allFieldPermissions[$field] = $fieldPermissions;
                    }
                }

                $modulePermissions[static::FIELD_PERMS] = $allFieldPermissions;
                $rolePermissions[$moduleType] = $modulePermissions;
            }
            $permissionsDictionary = $this->mergePermissions($permissionsDictionary, $rolePermissions);
        }

        return $permissionsDictionary;
    }

    /**
     * @param array $array1
     * @param array $array2
     * @return array
     */
    public function mergePermissions($array1, $array2)
    {
        $mergedPermissions = [];

        $array2Modules = array_keys($array2);
        $array1Modules = array_keys($array1);
        $sharedKeys = array_intersect($array1Modules, $array2Modules);

        $array1OnlyModules = array_diff($array1Modules, $sharedKeys);
        $array2OnlyModules = array_diff($array2Modules, $sharedKeys);

        // merge first
        foreach ($sharedKeys as $sharedModule) {
            $array1ModulePermissions = $array1[$sharedModule];
            $array2ModulePermissions = $array2[$sharedModule];

            $array1FieldPermissions = $array1ModulePermissions[static::FIELD_PERMS];
            $array2FieldPermissions = $array2ModulePermissions[static::FIELD_PERMS];

            // field read perms
            $array1FieldReadPermissions = $array1ModulePermissions[static::PERM_CRUD_READ] ? $this->getFieldPermissionsByType(static::PERM_FIELD_READ, $array1FieldPermissions) : [];
            $array2FieldReadPermissions = $array2ModulePermissions[static::PERM_CRUD_READ] ? $this->getFieldPermissionsByType(static::PERM_FIELD_READ, $array2FieldPermissions) : [];

            if (!$array1ModulePermissions[static::PERM_CRUD_READ]) {
                $fieldReadPermissions = $array2FieldReadPermissions;
            } else if (!$array2ModulePermissions[static::PERM_CRUD_READ]) {
                $fieldReadPermissions = $array1FieldReadPermissions;
            } else {
                $fieldReadPermissions = $this->mergeFieldPermissions(static::PERM_FIELD_READ, $array1FieldReadPermissions, $array2FieldReadPermissions);
            }

            // field write perms
            $array1FieldWritePermissions = $array1ModulePermissions[static::PERM_CRUD_UPDATE] ? $this->getFieldPermissionsByType(static::PERM_FIELD_WRITE, $array1FieldPermissions) : [];
            $array2FieldWritePermissions = $array2ModulePermissions[static::PERM_CRUD_UPDATE] ? $this->getFieldPermissionsByType(static::PERM_FIELD_WRITE, $array2FieldPermissions) : [];

            if (!$array1ModulePermissions[static::PERM_CRUD_UPDATE]) {
                $fieldWritePermissions = $array2FieldWritePermissions;
            } else if (!$array2ModulePermissions[static::PERM_CRUD_UPDATE]) {
                $fieldWritePermissions = $array1FieldWritePermissions;
            } else {
                $fieldWritePermissions = $this->mergeFieldPermissions(static::PERM_FIELD_WRITE, $array1FieldWritePermissions, $array2FieldWritePermissions);
            }

            $fieldPermissions = array_merge_recursive($fieldReadPermissions, $fieldWritePermissions);

            $mergedModulePermissions = [
                static::PERM_CRUD_CREATE => $array1ModulePermissions[static::PERM_CRUD_CREATE] || $array2ModulePermissions[static::PERM_CRUD_CREATE],
                static::PERM_CRUD_READ => $array1ModulePermissions[static::PERM_CRUD_READ] || $array2ModulePermissions[static::PERM_CRUD_READ],
                static::PERM_CRUD_UPDATE => $array1ModulePermissions[static::PERM_CRUD_UPDATE] || $array2ModulePermissions[static::PERM_CRUD_UPDATE],
                static::PERM_CRUD_DELETE => $array1ModulePermissions[static::PERM_CRUD_DELETE] || $array2ModulePermissions[static::PERM_CRUD_DELETE],
                static::PERM_CRUD_EXECUTE => $array1ModulePermissions[static::PERM_CRUD_EXECUTE] || $array2ModulePermissions[static::PERM_CRUD_EXECUTE],
                static::FIELD_PERMS => $fieldPermissions
            ];

            $mergedPermissions[$sharedModule] = $mergedModulePermissions;
        }

        // take exclusive array1 modules
        foreach ($array1OnlyModules as $module) {
            $mergedPermissions[$module] = $array1[$module];
        }
        // take exclusive array2 modules
        foreach ($array2OnlyModules as $module) {
            $mergedPermissions[$module] = $array2[$module];
        }

        return $mergedPermissions;
    }

    /**
     * @param string $permissionType
     * @param array $fieldPermissions
     * @return array
     */
    public function getFieldPermissionsByType($permissionType, $fieldPermissions)
    {
        $fieldPermissionsByType = [];
        foreach ($fieldPermissions as $field => $permissions) {
            if (array_key_exists($permissionType, $permissions) && !is_null($permissions[$permissionType])) {
                $fieldPermissionsByType[$field] = [$permissionType => $permissions[$permissionType]];
            }
        }
        return $fieldPermissionsByType;
    }

    /**
     * @param string $permissionType
     * @param array $fieldPermissions1
     * @param array $fieldPermissions2
     * @return array
     */
    public function mergeFieldPermissions($permissionType, $fieldPermissions1, $fieldPermissions2)
    {
        $sharedFields = array_intersect(array_keys($fieldPermissions1), array_keys($fieldPermissions2));
        $mergedFieldPermissions = [];

        foreach ($sharedFields as $field) {
            $permission1 = array_key_exists($permissionType, $fieldPermissions1[$field]) ? $fieldPermissions1[$field][$permissionType] : null;
            $permission2 = array_key_exists($permissionType, $fieldPermissions2[$field]) ? $fieldPermissions2[$field][$permissionType] : null;

            $mergedFieldPermissions[$field] = [
                $permissionType => $permission1 || $permission2
            ];
        }

        return $mergedFieldPermissions;
    }

    /**
     * @return array
     */
    protected function buildDefaultPermissions()
    {
        return [
            static::PERM_CRUD_CREATE => false,
            static::PERM_CRUD_READ => false,
            static::PERM_CRUD_UPDATE => false,
            static::PERM_CRUD_DELETE => false,
            static::PERM_CRUD_EXECUTE => false
        ];
    }
}
