<?php

namespace Cybersponse\AuthorizationBundle\Subscriber;

use Doctrine\ORM\EntityManager;
use Doctrine\ORM\EntityRepository;
use Dunglas\ApiBundle\Event\Events;
use GuzzleHttp\Psr7\Request as GuzzleRequest;
use Cybersponse\TenantBundle\Entity\Prod\Agent;
use Cybersponse\AuthorizationBundle\Entity\Role;
use Cybersponse\AuthorizationBundle\Entity\Team;
use Cybersponse\TenantBundle\Entity\Prod\Tenant;
use Cybersponse\TenantBundle\Entity\Prod\Appliance;
use Cybersponse\WorkflowBundle\Subscriber\DataEvent;
use Symfony\Component\EventDispatcher\EventSubscriberInterface;

class AgentSubscriber implements EventSubscriberInterface
{
    const POSTMAN_CONFIG_DIRECTORY = '/opt/cyops-routing-agent/postman/configs/agents/';
    const AGENT_ROLE_UUID = '3006d2f2-2f3a-4ae8-a36c-03ae75bf17fb';

    /** @var EntityManager */
    protected $entityManager;

    /** @var DasClient */
    protected $dasClient;

    /** @var ContainerInterface */
    protected $container;

    public static function getSubscribedEvents()
    {
        return [
            Events::PRE_UPDATE => ['handleUpdate', 100],
            Events::PRE_CREATE => ['handleCreate', 100]
        ];
    }

    protected function updateApplianceTeamsOnAgentOwnersChange($previousName, $teams, $agentName, $previousTeam)
    {
        $applianceRepo = $this->entityManager->getRepository(Appliance::class);
        $appliance = $applianceRepo->findBy(array("name" => $previousName));
        if ($appliance) {
            $appliance = $appliance[0];
            foreach($previousTeam as $oldTeam) {
                $appliance->removeTeam($oldTeam);
            }
            foreach ($teams as $team) {
                $appliance->addTeam($team);
            }
            $appliance->setName($agentName);
            $this->entityManager->persist($appliance);
            $this->entityManager->flush();
        }
    }

    protected function propagateAgentTeamToTenant($teams, $agentId, $previousOwner)
    {
        $tenantRepo = $this->entityManager->getRepository(Tenant::class);
        $tenant = $tenantRepo->findBy(array("tenantId" => $agentId));
        if ($tenant) {
            $tenant = $tenant[0];
            foreach ($previousOwner as $oldOwner) {
                $tenant->removeOwner($oldOwner);
            }
            foreach ($teams as $team) {
                $tenant->addOwner($team);
            }
            $this->entityManager->persist($tenant);
            $this->entityManager->flush();
        }
    }

    protected function propagateTenantTeamtoAgent($tenantId, $teams, $tenantName, $agentName, $agentPreviousName, $previousOwner)
    {
        $agentRepo = $this->entityManager->getRepository(Agent::class);
        $agent = $agentRepo->findBy(array("agentId" => $tenantId));

        if ($agent) {
            $agent = $agent[0];
            foreach($previousOwner as $oldOwner) {
                $agent->removeOwner($oldOwner);
            }
            foreach ($teams as $team) {
                $agent->addOwner($team);
            }
            $agent->setName($tenantName);
            $this->entityManager->persist($agent);
            $this->entityManager->flush();
            $this->updateApplianceTeamsOnAgentOwnersChange($agentPreviousName, $teams, $agentName, $previousOwner);
        }
    }

    protected function getTeamPersistedObject($ownerIris) {
        $ownerUuid = [];
        foreach ($ownerIris as $owner) {
            array_push($ownerUuid, end(explode('/', $owner)));
        }
        $TeamRepo = $this->entityManager->getRepository(Team::class);
        return $TeamRepo->findBy(array("uuid" => $ownerUuid));
    }

    /**
     * @param DataEvent $event
     * @throws \Exception
     */
    public function handleUpdate(DataEvent $event)
    {
        $data = $event->getData();
        if ($event->getResource()->getModelType() == 'agents') {
            $previousData = $event->getPrevious();
            $previousName = 'Agent' . '-' . $previousData['name'];
            $previousOwner = $this->getTeamPersistedObject($previousData['owners']);

            $agentId = $previousData['agentId'];

            $name = $data->getName();
            $agentName = 'Agent' . '-' . $name;
            $teams = $data->getOwners();

            $this->updateApplianceTeamsOnAgentOwnersChange($previousName, $teams, $agentName, $previousOwner);
            $this->propagateAgentTeamToTenant($teams, $agentId, $previousOwner);
        }
        if ($event->getResource()->getModelType() == 'tenants') {
            $previousData = $event->getPrevious();
            $tenantId = $previousData['tenantId'];
            $previousName = 'Agent' . '-' . $previousData['name'];
            $previousOwner = $this->getTeamPersistedObject($previousData['owners']);
            $tenantName = $data->getName();
            $name = $data->getName();
            $agentName = 'Agent' . '-' . $name;
            $teams = $data->getOwners();
            $this->propagateTenantTeamtoAgent($tenantId, $teams, $tenantName, $agentName, $previousName, $previousOwner);
        }
    }

    public function handleCreate(DataEvent $event)
    {
        if ($event->getResource()->getModelType() == 'agents') {
            $data = $event->getData();
            $agentId = $data->getAgentId();
            $agentName = 'Agent' . '-' . $data->getName();
            $applianceRepo = $this->entityManager->getRepository(Appliance::class);
            $appliance = $applianceRepo->findBy(array("name" => $agentName));
            if (!$appliance) {
                $owners = $data->getOwners();
                $dasUri = $this->container->getParameter('das.uri');
                $headers = array(
                    "Content-type" => "application/json"
                );
                $applianceData = array(
                    "domain" => $_SERVER['HTTP_HOST'],
                    "type" => 3,
                    "nfa" => [],
                    "system" => true,
                    "tenant" => $agentId . '_nXGkvPBdcR_' . $agentName,
                    "private_key_path" => self::POSTMAN_CONFIG_DIRECTORY . $agentId . '/APPLIANCE_PRIVATE_KEY',
                    "public_key_path" => self::POSTMAN_CONFIG_DIRECTORY . $agentId . '/APPLIANCE_PUBLIC_KEY',
                    "status" => 1,
                    "bind_name" => null,
                    "credentials" => array(
                        "loginid" => null
                    )
                );
                $dasRequests = new GuzzleRequest('post', $dasUri . '/users', $headers, json_encode($applianceData));
                try {
                    $dasResponse = $this->dasClient->send($dasRequests);
                    $resCode = $dasResponse->getStatusCode();
                    if ($resCode == 201) {
                        $response = json_decode($dasResponse->getBody());
                        $this->write_keys_to_postman($agentId, $response);
                        $this->addApplianceDetails($agentName, $response->uuid, $owners);
                    }
                } catch (\Exception $error) {
                    error_log($error->getMessage());
                }
            }
        }
    }

    protected function addApplianceDetails($agentName, $userId, $teams)
    {
        $applianceData = new Appliance();
        $applianceData->setName($agentName);

        $roleRepo = $this->entityManager->getRepository(Role::class);
        $role = $roleRepo->findBy(array("uuid" => self::AGENT_ROLE_UUID));
        if ($role) {
            $role_obj = $role[0];
            $applianceData->addRole($role_obj);
        }
        $applianceData->setUserId($userId);
        foreach ($teams as $team) {
            $applianceData->addTeam($team);
        }
        $this->entityManager->persist($applianceData);
        $this->entityManager->flush();
    }

    public function write_keys_to_postman($agentId, $response)
    {
        try {
            mkdir(self::POSTMAN_CONFIG_DIRECTORY . $agentId, 0755);
        } catch (\Exception $error) {
            error_log($error->getMessage());
            return false;
        }
        $publicKey = fopen(self::POSTMAN_CONFIG_DIRECTORY . $agentId . '/APPLIANCE_PUBLIC_KEY', "w");
        fwrite($publicKey, $response->pubkey);
        fclose($publicKey);
        $privateKey = fopen(self::POSTMAN_CONFIG_DIRECTORY . $agentId . '/APPLIANCE_PRIVATE_KEY', "w");
        fwrite($privateKey, $response->pkey);
        fclose($privateKey);
    }

    /**
     * @param ContainerInterface $container
     */
    public function setContainer($container)
    {
        $this->container = $container;
    }

    /**
     * @param EntityRepository $tagManager
     */
    public function setDasClient($dasClient)
    {
        $this->dasClient = $dasClient;
    }

    /**
     * @param EntityRepository $tagManager
     */
    public function setEntityManager($entityManager)
    {
        $this->entityManager = $entityManager;
    }
}
