<?php
/* Copyright start
  Copyright (C) 2008 - 2020 Fortinet Inc.
  All rights reserved.
  FORTINET CONFIDENTIAL & FORTINET PROPRIETARY SOURCE CODE
  Copyright end */

namespace Cybersponse\AuthorizationBundle\Tests\Unit\Permissions\PermissionsDictionaryBuilder;


use Cybersponse\AuthorizationBundle\Permissions\PermissionsDictionaryBuilder;

class MergePermissionsMethodTest extends \PHPUnit_Framework_TestCase
{
    protected function executeSystemUnderTest($permissions1, $permissions2)
    {
        $serializer = new PermissionsDictionaryBuilder();
        return $serializer->mergePermissions($permissions1, $permissions2);
    }

    // region General Module Tests

    public function test_GivenNonOverlappingModules_ShouldReturnAllModules()
    {
        $expectedModule1 = 'expectedModule1';
        $expectedModule2 = 'expectedModule2';
        $permissions1 = [
            $expectedModule1 => $this->createModulePermissions(true)
        ];
        $permissions2 = [
            $expectedModule2 => $this->createModulePermissions(true)
        ];

        $mergedPermissions = $this->executeSystemUnderTest($permissions1, $permissions2);

        $this->assertArrayHasKey($expectedModule1, $mergedPermissions);
        $this->assertArrayHasKey($expectedModule2, $mergedPermissions);
    }

    // endregion

    // region Module Permission Precedence

    public function test_GivenFirstModuleCreatePermissionTrue_GivenSecondModuleCreatePermissionTrue_ShouldReturnModuleCreatePermissionTrue()
    {
        $expectedModule = 'expectedModule';
        $permissions1 = [
            $expectedModule => $this->createModulePermissions(true)
        ];
        $permissions2 = [
            $expectedModule => $this->createModulePermissions(true)
        ];

        $mergedPermissions = $this->executeSystemUnderTest($permissions1, $permissions2);

        $this->assertTrue($mergedPermissions[$expectedModule][PermissionsDictionaryBuilder::PERM_CRUD_CREATE]);
    }

    public function test_GivenFirstModuleCreatePermissionTrue_GivenSecondModuleCreatePermissionFalse_ShouldReturnModuleCreatePermissionTrue()
    {
        $expectedModule = 'expectedModule';
        $permissions1 = [
            $expectedModule => $this->createModulePermissions(true)
        ];
        $permissions2 = [
            $expectedModule => $this->createModulePermissions(false)
        ];

        $mergedPermissions = $this->executeSystemUnderTest($permissions1, $permissions2);

        $this->assertTrue($mergedPermissions[$expectedModule][PermissionsDictionaryBuilder::PERM_CRUD_CREATE]);
    }

    public function test_GivenFirstModuleCreatePermissionFalse_GivenSecondModuleCreatePermissionTrue_ShouldReturnModuleCreatePermissionTrue()
    {
        $expectedModule = 'expectedModule';
        $permissions1 = [
            $expectedModule => $this->createModulePermissions(false)
        ];
        $permissions2 = [
            $expectedModule => $this->createModulePermissions(true)
        ];

        $mergedPermissions = $this->executeSystemUnderTest($permissions1, $permissions2);

        $this->assertTrue($mergedPermissions[$expectedModule][PermissionsDictionaryBuilder::PERM_CRUD_CREATE]);
    }

    public function test_GivenFirstModuleCreatePermissionFalse_GivenSecondModuleCreatePermissionFalse_ShouldReturnModuleCreatePermissionFalse()
    {
        $expectedModule = 'expectedModule';
        $permissions1 = [
            $expectedModule => $this->createModulePermissions(false)
        ];
        $permissions2 = [
            $expectedModule => $this->createModulePermissions(false)
        ];

        $mergedPermissions = $this->executeSystemUnderTest($permissions1, $permissions2);

        $this->assertFalse($mergedPermissions[$expectedModule][PermissionsDictionaryBuilder::PERM_CRUD_CREATE]);
    }

    // endregion

    // region Field Permission Precedence

    public function test_GivenAllOverlappingFieldPermissions_ShouldReturnAllFields()
    {
        $expectedModule = 'expectedModule';
        $expectedField1 = 'expectedField1';
        $expectedField2 = 'expectedField2';
        $permissions1 = [
            $expectedModule => $this->createModulePermissions(true, true, true, true, [
                $expectedField1 => [PermissionsDictionaryBuilder::PERM_FIELD_READ => true],
                $expectedField2 => [PermissionsDictionaryBuilder::PERM_FIELD_READ => true]
            ])
        ];
        $permissions2 = [
            $expectedModule => $this->createModulePermissions(true, true, true, true, [
                $expectedField1 => [PermissionsDictionaryBuilder::PERM_FIELD_READ => false],
                $expectedField2 => [PermissionsDictionaryBuilder::PERM_FIELD_READ => false]
            ])
        ];

        $mergedPermissions = $this->executeSystemUnderTest($permissions1, $permissions2);

        $this->assertArrayHasKey($expectedField1, $mergedPermissions[$expectedModule][PermissionsDictionaryBuilder::FIELD_PERMS]);
        $this->assertArrayHasKey($expectedField2, $mergedPermissions[$expectedModule][PermissionsDictionaryBuilder::FIELD_PERMS]);
    }

    public function test_GivenPartiallyOverlappingFieldPermissions_ShouldReturnOverlappingFields()
    {
        $expectedModule = 'expectedModule';
        $expectedField = 'expectedField';
        $permissions1 = [
            $expectedModule => $this->createModulePermissions(true, true, true, true, [
                $expectedField => [PermissionsDictionaryBuilder::PERM_FIELD_READ => true],
                'unexpectedField' => [PermissionsDictionaryBuilder::PERM_FIELD_READ => true]
            ])
        ];
        $permissions2 = [
            $expectedModule => $this->createModulePermissions(true, true, true, true, [
                $expectedField => [PermissionsDictionaryBuilder::PERM_FIELD_READ => false]
            ])
        ];

        $mergedPermissions = $this->executeSystemUnderTest($permissions1, $permissions2);

        $this->assertArrayHasKey($expectedField, $mergedPermissions[$expectedModule][PermissionsDictionaryBuilder::FIELD_PERMS]);
    }

    public function test_GivenPartiallyOverlappingFieldPermissions_ShouldNotReturnNonOverlappingFields()
    {
        $expectedModule = 'expectedModule';
        $expectedField = 'expectedField';
        $unexpectedField = 'unexpectedField';
        $permissions1 = [
            $expectedModule => $this->createModulePermissions(true, true, true, true, [
                $expectedField => [PermissionsDictionaryBuilder::PERM_FIELD_READ => true],
                $unexpectedField => [PermissionsDictionaryBuilder::PERM_FIELD_READ => true]
            ])
        ];
        $permissions2 = [
            $expectedModule => $this->createModulePermissions(true, true, true, true, [
                $expectedField => [PermissionsDictionaryBuilder::PERM_FIELD_READ => false]
            ])
        ];

        $mergedPermissions = $this->executeSystemUnderTest($permissions1, $permissions2);

        $this->assertArrayNotHasKey($unexpectedField, $mergedPermissions[$expectedModule][PermissionsDictionaryBuilder::FIELD_PERMS]);
    }

    public function test_GivenNonOverlappingFieldPermissions_ShouldReturnNoFields()
    {
        $expectedModule = 'expectedModule';
        $permissions1 = [
            $expectedModule => $this->createModulePermissions(true, true, true, true, [
                'field1' => [PermissionsDictionaryBuilder::PERM_FIELD_READ => true]
            ])
        ];
        $permissions2 = [
            $expectedModule => $this->createModulePermissions(true, true, true, true, [
                'field2' => [PermissionsDictionaryBuilder::PERM_FIELD_READ => false]
            ])
        ];

        $mergedPermissions = $this->executeSystemUnderTest($permissions1, $permissions2);

        $this->assertEmpty($mergedPermissions[$expectedModule][PermissionsDictionaryBuilder::FIELD_PERMS]);
    }

    public function test_GivenOverlappingFieldPermissions_GivenBothFieldReadPermissionTrue_ShouldReturnFieldReadPermissionTrue()
    {
        $expectedModule = 'expectedModule';
        $expectedField = 'expectedField';
        $permissions1 = [
            $expectedModule => $this->createModulePermissions(true, true, true, true, [
                $expectedField => [PermissionsDictionaryBuilder::PERM_FIELD_READ => true]
            ])
        ];
        $permissions2 = [
            $expectedModule => $this->createModulePermissions(true, true, true, true, [
                $expectedField => [PermissionsDictionaryBuilder::PERM_FIELD_READ => true]
            ])
        ];

        $mergedPermissions = $this->executeSystemUnderTest($permissions1, $permissions2);

        $this->assertTrue($mergedPermissions[$expectedModule][PermissionsDictionaryBuilder::FIELD_PERMS][$expectedField][PermissionsDictionaryBuilder::PERM_FIELD_READ]);
    }

    public function test_GivenOverlappingFieldPermissions_GivenFirstFieldReadPermissionTrue_ShouldReturnFieldReadPermissionTrue()
    {
        $expectedModule = 'expectedModule';
        $expectedField = 'expectedField';
        $permissions1 = [
            $expectedModule => $this->createModulePermissions(true, true, true, true, [
                $expectedField => [PermissionsDictionaryBuilder::PERM_FIELD_READ => true]
            ])
        ];
        $permissions2 = [
            $expectedModule => $this->createModulePermissions(true, true, true, true, [
                $expectedField => [PermissionsDictionaryBuilder::PERM_FIELD_READ => false]
            ])
        ];

        $mergedPermissions = $this->executeSystemUnderTest($permissions1, $permissions2);

        $this->assertTrue($mergedPermissions[$expectedModule][PermissionsDictionaryBuilder::FIELD_PERMS][$expectedField][PermissionsDictionaryBuilder::PERM_FIELD_READ]);
    }

    public function test_GivenOverlappingFieldPermissions_GivenSecondFieldReadPermissionTrue_ShouldReturnFieldReadPermissionTrue()
    {
        $expectedModule = 'expectedModule';
        $expectedField = 'expectedField';
        $permissions1 = [
            $expectedModule => $this->createModulePermissions(true, true, true, true, [
                $expectedField => [PermissionsDictionaryBuilder::PERM_FIELD_READ => false]
            ])
        ];
        $permissions2 = [
            $expectedModule => $this->createModulePermissions(true, true, true, true, [
                $expectedField => [PermissionsDictionaryBuilder::PERM_FIELD_READ => true]
            ])
        ];

        $mergedPermissions = $this->executeSystemUnderTest($permissions1, $permissions2);

        $this->assertTrue($mergedPermissions[$expectedModule][PermissionsDictionaryBuilder::FIELD_PERMS][$expectedField][PermissionsDictionaryBuilder::PERM_FIELD_READ]);
    }

    public function test_GivenOverlappingFieldPermissions_GivenBothFieldReadPermissionFalse_ShouldReturnFieldReadPermissionFalse()
    {
        $expectedModule = 'expectedModule';
        $expectedField = 'expectedField';
        $permissions1 = [
            $expectedModule => $this->createModulePermissions(true, true, true, true, [
                $expectedField => [PermissionsDictionaryBuilder::PERM_FIELD_READ => false]
            ])
        ];
        $permissions2 = [
            $expectedModule => $this->createModulePermissions(true, true, true, true, [
                $expectedField => [PermissionsDictionaryBuilder::PERM_FIELD_READ => false]
            ])
        ];

        $mergedPermissions = $this->executeSystemUnderTest($permissions1, $permissions2);

        $this->assertFalse($mergedPermissions[$expectedModule][PermissionsDictionaryBuilder::FIELD_PERMS][$expectedField][PermissionsDictionaryBuilder::PERM_FIELD_READ]);
    }

    public function test_GivenOverlappingFieldPermissions_GivenBothFieldReadPermissionNull_ShouldNotReturnFieldPermission()
    {
        $expectedModule = 'expectedModule';
        $unexpectedField = 'unexpectedField';
        $permissions1 = [
            $expectedModule => $this->createModulePermissions(true, true, true, true, [
                $unexpectedField => [PermissionsDictionaryBuilder::PERM_FIELD_READ => null]
            ])
        ];
        $permissions2 = [
            $expectedModule => $this->createModulePermissions(true, true, true, true, [
                $unexpectedField => [PermissionsDictionaryBuilder::PERM_FIELD_READ => null]
            ])
        ];

        $mergedPermissions = $this->executeSystemUnderTest($permissions1, $permissions2);

        $this->assertArrayNotHasKey($unexpectedField, $mergedPermissions[$expectedModule][PermissionsDictionaryBuilder::FIELD_PERMS]);
    }

    public function test_GivenFirstModulePermissionFalse_GivenSecondModulePermissionTrue_GivenSecondFieldPermissions_ShouldReturnSecondFieldPermissions()
    {
        $expectedModule = 'expectedModule';
        $expectedField = 'expectedField';
        $permissions1 = [
            $expectedModule => $this->createModulePermissions(true, false, true, true, [])
        ];
        $permissions2 = [
            $expectedModule => $this->createModulePermissions(true, true, true, true, [
                $expectedField => [PermissionsDictionaryBuilder::PERM_FIELD_READ => false]
            ])
        ];

        $mergedPermissions = $this->executeSystemUnderTest($permissions1, $permissions2);

        $this->assertArrayHasKey($expectedField, $mergedPermissions[$expectedModule][PermissionsDictionaryBuilder::FIELD_PERMS]);
    }

    // endregion

    protected function createModulePermissions($create = false, $read = false, $update = false, $delete = false, $fieldPermissions = [])
    {
        return [
            PermissionsDictionaryBuilder::PERM_CRUD_CREATE => $create,
            PermissionsDictionaryBuilder::PERM_CRUD_READ => $read,
            PermissionsDictionaryBuilder::PERM_CRUD_UPDATE => $update,
            PermissionsDictionaryBuilder::PERM_CRUD_DELETE => $delete,
            PermissionsDictionaryBuilder::FIELD_PERMS => $fieldPermissions
        ];
    }
}
