<?php
/* Copyright start
  Copyright (C) 2008 - 2020 Fortinet Inc.
  All rights reserved.
  FORTINET CONFIDENTIAL & FORTINET PROPRIETARY SOURCE CODE
  Copyright end */

namespace Cybersponse\CoreBundle\Api;

use Doctrine\Common\Inflector\Inflector;
use Dunglas\ApiBundle\Api\Operation\Operation;
use Dunglas\ApiBundle\Api\ResourceInterface;
use Symfony\Component\Routing\Route;

class OperationFactory
{
    const ROUTE_NAME_PREFIX = 'api_';
    const DEFAULT_CONTROLLER = 'CybersponseCoreBundle:AuthorizationResource';

    /**
     * Creates collection operation.
     *
     * @param ResourceInterface $resource
     * @param string|array      $methods
     * @param string|null       $path
     * @param null              $controller
     * @param null              $routeName
     * @param array             $context
     *
     * @return Operation
     */
    public function createCollectionOperation(
        ResourceInterface $resource,
        $methods,
        $path = null,
        $controller = null,
        $routeName = null,
        array $context = []
    ) {
        return $this->createOperation($resource, true, $methods, $path, $controller, $routeName, $context);
    }

    /**
     * Creates item operation.
     *
     * @param ResourceInterface $resource
     * @param string|array      $methods
     * @param string|null       $path
     * @param null              $controller
     * @param null              $routeName
     * @param array             $context
     *
     * @return Operation
     */
    public function createItemOperation(
        ResourceInterface $resource,
        $methods,
        $path = null,
        $controller = null,
        $routeName = null,
        array $context = [],
        $requirements = []
    ) {
        return $this->createOperation($resource, false, $methods, $path, $controller, $routeName, $context, false, $requirements);
    }

    /**
     * Creates an item field operation.
     *
     * @param ResourceInterface $resource
     * @param string|array      $methods
     * @param string|null       $path
     * @param null              $controller
     * @param null              $routeName
     * @param array             $context
     *
     * @return Operation
     */
    public function createItemFieldOperation(
        ResourceInterface $resource,
        $methods,
        $path = null,
        $controller = null,
        $routeName = null,
        array $context = []
    ) {
        return $this->createOperation($resource, false, $methods, $path, $controller, $routeName, $context, true);
    }

    /**
     * Creates operation.
     *
     * @param ResourceInterface $resource
     * @param bool              $collection
     * @param string|array      $methods
     * @param string|null       $path
     * @param string|null       $controller
     * @param string|null       $routeName
     * @param array             $context
     * @param bool              $field
     *
     * @return Operation
     */
    private function createOperation(
        ResourceInterface $resource,
        $collection,
        $methods,
        $path = null,
        $controller = null,
        $routeName = null,
        array $context = [],
        $field = false,
        $requirements = []
    ) {
        $shortName = $resource->getShortName();

        if ($resource instanceof ModelAwareResource) {
            $type = $resource->getModelType();
        } else {
            $type = Inflector::pluralize(Inflector::tableize($shortName));
        }

        // Populate path
        if (null === $path) {
            $path = '/' . $type;

            if (!$collection) {
                $path .= '/{id}';
            }
            if ($field) {
                $path .= '/{field}';
                if (in_array($methods, ['DELETE'])) {
                    $path .= '/{fieldId}';
                }
            }
        }

        // Guess default method
        if (is_array($methods)) {
            $defaultMethod = $methods[0];
        } else {
            $defaultMethod = $methods;
        }

        // Populate controller
        if (null === $controller) {
            $defaultAction = strtolower($defaultMethod);

            if ($collection) {
                $defaultAction = 'c' . $defaultAction;
            }

            if ($field) {
                $defaultAction = $defaultAction . 'Field';
            }

            $controller = self::DEFAULT_CONTROLLER . ':' . $defaultAction;

            // Populate route name
            if (null === $routeName) {
                $routeName = self::ROUTE_NAME_PREFIX . $type . '_' . $defaultAction;
            }
        }

        return new Operation(
            new Route(
                $path,
                [
                    '_controller' => $controller,
                    '_resource' => $shortName,
                ],
                $requirements,
                [],
                '',
                [],
                $methods
            ),
            $routeName,
            $context
        );
    }
}
