<?php
/* Copyright start
  Copyright (C) 2008 - 2020 Fortinet Inc.
  All rights reserved.
  FORTINET CONFIDENTIAL & FORTINET PROPRIETARY SOURCE CODE
  Copyright end */

namespace Cybersponse\CoreBundle\Command;

use Cybersponse\CoreBundle\Entity\ExportJob;
use Cybersponse\CoreBundle\Entity\ExportTemplate;
use Cybersponse\CoreBundle\Service\ExportService;
use Cybersponse\FilesBundle\Entity\File;
use Cybersponse\FilesBundle\Entity\PgFile;
use Cybersponse\UtilsBundle\Utils;
use Ramsey\Uuid\Uuid;
use Symfony\Bundle\FrameworkBundle\Command\ContainerAwareCommand;
use Symfony\Component\Console\Input\InputInterface;
use Symfony\Component\Console\Input\InputOption;
use Symfony\Component\Console\Output\OutputInterface;
use Symfony\Component\Console\Style\SymfonyStyle;
use Symfony\Component\Filesystem\Filesystem;

/**
 * Class ExportJobCommand
 */
class ExportJobCommand extends ContainerAwareCommand
{
    const COMPONENT = 'component';
    const ENTITY_TYPE = 'entityType';
    const HTTP_METHOD = 'operation';
    const REMOTE_ADDR = 'source';
    const DELTA_DATA = 'data';
    const TRANSACTION_DATE = 'transactionDate';
    const UUID = 'entityUuid';
    const USER_UUID = 'userId';
    const USER = 'user';

    /** @var Filesystem $filesystem */
    protected $filesystem;

    protected function configure()
    {
        $this
            ->setName('fortisoar:jobs:export')
            ->setDescription('Exports records and configuration into a zip file')
            ->setDefinition([
                new InputOption(
                    'job-uuid',
                    'j',
                    InputOption::VALUE_REQUIRED | InputOption::VALUE_IS_ARRAY,
                    'The uuid of the export job that will be started.'
                ),
                new InputOption(
                    'user-token',
                    't',
                    InputOption::VALUE_OPTIONAL,
                    'The user token for the export context. If empty, the appliance user will be used.'
                ),
                new InputOption(
                    'token-type',
                    'tt',
                    InputOption::VALUE_OPTIONAL,
                    'Type of token HMAC or JWT.'
                )
            ]);
        
        $this->filesystem = new Filesystem();
    }

    protected function execute(InputInterface $input, OutputInterface $output)
    {
        $io = new SymfonyStyle($input, $output);

        $jobUuids = $input->getOption('job-uuid');
        if (count($jobUuids) == 0) {
            $io->error('Please specify a job uuid');
            return;
        }
        $jobUuid = $jobUuids[0];
        $rawToken = $input->getOption('user-token');
        $tokenType = $input->getOption('token-type');
        if ($rawToken && $tokenType) {
            Utils::setRefreshedToken($this->getContainer(), $rawToken, $tokenType);
        }
        /** @var ExportJob $exportJob */
        $exportJob = Utils::findRecordWithClassName($this->getContainer(), ExportJob::class, $jobUuid);
        $doctrine = $this->getContainer()->get('doctrine');

        $status = 'In Progress';
        $this->updateExportJob($exportJob, $status, null, null);

        /** @var ExportService $exportService */
        $exportService = $this->getContainer()->get('cybersponse.core.service.export_service');
        try {
            $configExportData = $exportService->exportConfiguration($exportJob, $tokenType);
        } catch (\Exception $e) {
            $this->updateExportJob($exportJob, 'Error', null, null, $e->getMessage());
            $io->error(sprintf('Error while exporting file: %s', $e->getMessage()));
        }
        if ($configExportData) {
            $jsonData = json_encode($configExportData, JSON_PRETTY_PRINT);
            $filename = $exportJob->getFileName();
            $filename = $filename ? $filename : 'config.json';
            $mimetype = 'application/json';
            $timestamp = new \DateTime('NOW', new \DateTimeZone('UTC'));

            $pgFile = new PgFile();
            $pgFileEntityManager = $doctrine->getManagerForClass(PgFile::class);
            /** @var AbstractFile $file */
            $pgFile->setFile(base64_encode($jsonData));
            $pgFile->setUuid(Uuid::uuid4()->toString());
            $pgFileEntityManager->persist($pgFile);
            $pgFileEntityManager->flush();

            $fileEntityManager = $doctrine->getManagerForClass(File::class);
            $file = new File();
            $file->setFile("/api/3/pgfile/" . $pgFile->getUuid());
            $file->setFilename($filename);
            $file->setMimeType($mimetype);
            $file->setUploadDate($timestamp);
            $file->setSize(strlen($jsonData));
            $file->setUuid(Uuid::uuid4()->toString());
            $fileEntityManager->persist($file);
            $fileEntityManager->flush();

            $status = 'Export Complete';
            $this->updateExportJob($exportJob, $status, null, "/api/3/files/" . $file->getUuid());

            $entityManager = $this->getContainer()->get('doctrine')->getManagerForClass(ExportTemplate::class);
            $timestamp = new \DateTime('NOW', new \DateTimeZone('UTC'));
            $exportTemplate = $exportJob->getTemplate();
            $exportTemplate->setLastExportDate($timestamp);
            $entityManager->persist($exportTemplate);
            $entityManager->flush();

            $io->success(sprintf('Export completed successfully.'));
        }
    }

    /**
     * @return array
     */
    protected function prepareChangeData(array $oldData, array $newData, string $entityId)
    {
        $deltaData = [];
        $deltaData['oldData'] = $oldData;
        $deltaData['newData'] = $newData;
        $tokenStorage = $this->getContainer()->get('security.token_storage');
        $user = $tokenStorage->getToken() ? $tokenStorage->getToken()->getUser()->getUuid(): 'undefined';
        $userName = is_string($user) ? 'undefined': Utils::extractUserName($user);

        return [
            self::COMPONENT => 'crudhub',
            self::ENTITY_TYPE => 'ExportJob',
            self::UUID => $entityId,
            self::USER_UUID => $user,
            self::USER => $userName,
            self::HTTP_METHOD => 'Update',
            self::TRANSACTION_DATE => (int)($_SERVER['REQUEST_TIME_FLOAT']*1000),
            self::DELTA_DATA => $deltaData
        ];
    }

    private function updateExportJob(ExportJob $exportJob, $status = null, $progressPercent = null, $file = "", $errorMessage = null)
    {
        $entityManager = $this->getContainer()->get('doctrine')->getManagerForClass(ExportJob::class);
        $oldData = [
            '@id' => '/api/3/export_jobs/' . $exportJob->getUuid()
        ];
        $newData = [
            '@id' => '/api/3/export_jobs/' . $exportJob->getUuid()
        ];
        if ($status !== null) {
            $oldStatus = $exportJob->getStatus();
            $exportJob->setStatus($status);
            $oldData['status'] = $oldStatus;
            $newData['status'] = $status;
        }
        if ($progressPercent !== null) {
            $oldProgressPercent = $exportJob->getProgressPercent();
            $exportJob->setProgressPercent($progressPercent);
            $oldData['progressPercent'] = $oldProgressPercent;
            $newData['progressPercent'] = $progressPercent;
        }
        if ($file !== "") {
            $oldFile = $exportJob->getFile();
            $exportJob->setFile($file);
            $oldData['file'] = $oldFile;
            $newData['file'] = $file;
        }
        if ($errorMessage !== null) {
            $oldErrorMessage = $exportJob->getErrorMessage();
            $exportJob->setErrorMessage($errorMessage);
            $oldData['errorMessage'] = $oldErrorMessage;
            $newData['errorMessage'] = $errorMessage;
        }
        $timestamp = new \DateTime('NOW', new \DateTimeZone('UTC'));
        $exportJob->setModifyDate($timestamp);
        $entityManager->persist($exportJob);
        $entityManager->flush();
        $body = $this->prepareChangeData($oldData, $newData, $exportJob->getUuid());
        $utilsObj = new Utils();
        $utilsObj->sendData($this, [$body]);
    }
    
    public function get($service)
    {
        return $this->getContainer()->get($service);
    }
}
