<?php
/* Copyright start
  Copyright (C) 2008 - 2020 Fortinet Inc.
  All rights reserved.
  FORTINET CONFIDENTIAL & FORTINET PROPRIETARY SOURCE CODE
  Copyright end */

namespace Cybersponse\CoreBundle\Command;

use Cybersponse\CoreBundle\Entity\ImportJob;
use Cybersponse\CoreBundle\Service\ImportService;
use Cybersponse\FilesBundle\Entity\AbstractFile;
use Cybersponse\FilesBundle\Entity\File;
use Cybersponse\FilesBundle\Entity\PgFile;
use Cybersponse\UtilsBundle\Utils;
use Symfony\Bundle\FrameworkBundle\Command\ContainerAwareCommand;
use Symfony\Component\Console\Input\InputInterface;
use Symfony\Component\Console\Input\InputOption;
use Symfony\Component\Console\Output\OutputInterface;
use Symfony\Component\Console\Style\SymfonyStyle;
use Symfony\Component\Filesystem\Filesystem;
use ZipArchive;
use Exception;

/**
 * Class ImportJobCommand
 */
class ImportJobCommand extends ContainerAwareCommand
{
    const COMPONENT = 'component';
    const ENTITY_TYPE = 'entityType';
    const HTTP_METHOD = 'operation';
    const REMOTE_ADDR = 'source';
    const DELTA_DATA = 'data';
    const TRANSACTION_DATE = 'transactionDate';
    const UUID = 'entityUuid';
    const USER_UUID = 'userId';
    const USER = 'user';

    const STATUS = array(
        "DRAFT" => "Draft",
        "GENERATING_OPTIONS" => "Generating Options",
        "OPTIONS_GENERATED" => "Reviewing",
        "IMPORTING" => "In Progress",
        "PUBLISHING" => "Publishing Modules",
        "ERROR" => "Error",
        "IMPORT_COMPLETE" => "Import Complete"
    );

    /** @var Filesystem $filesystem */
    protected $filesystem;

    protected function configure()
    {
        $this
            ->setName('fortisoar:jobs:import')
            ->setDescription('Imports records and configuration from a Zip/JSON file')
            ->setDefinition([
                new InputOption(
                    'job-uuid',
                    'j',
                    InputOption::VALUE_REQUIRED | InputOption::VALUE_IS_ARRAY,
                    'The uuid of the import job that will be started.'
                ),
                new InputOption(
                    'generate-options',
                    'o',
                    InputOption::VALUE_NONE,
                    'Will update the Import Job record with an options object for user input.'
                ),
                new InputOption(
                    'user-token',
                    'ut',
                    InputOption::VALUE_OPTIONAL,
                    'The user token for the import context. If empty, the appliance user will be used.'
                ),
                new InputOption(
                    'token-type',
                    'tt',
                    InputOption::VALUE_OPTIONAL,
                    'Type of token HMAC or JWT.'
                )
            ]);
        
        $this->filesystem = new Filesystem();
    }

    protected function execute(InputInterface $input, OutputInterface $output)
    {
        $io = new SymfonyStyle($input, $output);

        $jobUuids = $input->getOption('job-uuid');
        $generateOptions = $input->getOption('generate-options');
        if (count($jobUuids) == 0) {
            $io->error('Please specify a job uuid');
            return;
        }
        $jobUuid = $jobUuids[0];
        $rawToken = $input->getOption('user-token');
        $tokenType = $input->getOption('token-type');
        if ($rawToken && $tokenType) {
            Utils::setRefreshedToken($this->getContainer(), $rawToken, $tokenType);
        }
        /** @var ImportJob $importJob */
        $importJob = Utils::findRecordWithClassName($this->getContainer(), ImportJob::class, $jobUuid);
        if (!$importJob) {
            throw new Exception(sprintf('Import Job %s does not exist', $jobUuid));
        }

        try {
            list($filePath, $folderPath) = $this->loadFolderIntoTmp($io, $importJob);
            $configJsonFilePath = sprintf('%s/config.json', $folderPath);
            $jsonData = file_get_contents($configJsonFilePath);
            $jsonObj = json_decode($jsonData, true);
            if (is_null($jsonObj)) {
                throw new Exception('Config JSON is empty');
            }
            if ($generateOptions) {
                $this->generateImportJobOptions($jsonObj, $importJob, $tokenType);
                $io->success(sprintf('Import job options populated successfully'));
            } else {
                $output = $this->importConfigJson($jsonObj, $importJob, $folderPath, $tokenType);
                $io->success(sprintf('Import completed successfully'));
            }
            // Clean up
            $this->filesystem->remove([$filePath, $folderPath]);
        } catch (Exception $e) {
            $this->updateImportJob($importJob, self::STATUS['ERROR'], null, null, $e->getMessage());
            $io->error(sprintf('Error while importing file: %s', $e->getMessage()));
            return;
        }
    }

    private function loadFolderIntoTmp(SymfonyStyle $io, ImportJob $importJob)
    {
        /** @var string $fileIri */
        $fileIri = $importJob->getFile();
        $explodedIRI = explode('/', $fileIri);
        $fileUuid = end($explodedIRI);
        
        /** @var AbstractFile $file */
        $file = Utils::findRecordWithClassName($this->getContainer(), File::class, $fileUuid);

        /** @var string $fileIri */
        $pgFileIri = $file->getFile();
        $explodedPgIRI = explode('/', $pgFileIri);
        $pgFileUuid = end($explodedPgIRI);

        /** @var AbstractFile $file */
        $pgFile = Utils::findRecordWithClassName($this->getContainer(), PgFile::class, $pgFileUuid);
        $folderPath = sprintf('/tmp/%s', $pgFileUuid);

        // Support single JSON files
        if ($file->getMimeType() == 'application/json') {
            $jsonFilePath = sprintf('%s/config.json', $folderPath);
            $this->filesystem->mkdir($folderPath);
            $this->filesystem->dumpFile($jsonFilePath, base64_decode(stream_get_contents($pgFile->getFile())));
            return [$jsonFilePath, $folderPath];
        }

        $zipFilePath = sprintf('/tmp/%s.zip', $pgFileUuid);
        $this->filesystem->dumpFile($zipFilePath, base64_decode(stream_get_contents($pgFile->getFile())));

        $zip = new ZipArchive;
        $res = $zip->open($zipFilePath);
        if ($res === true) {
            $zip->extractTo($folderPath);
            $zip->close();
        } else {
            throw new Exception('Unable to unzip the archive file');
            return null;
        }

        return [$zipFilePath, $folderPath];
    }

    private function generateImportJobOptions(array $config, ImportJob $importJob, $tokenType)
    {
        // Clear them first
        $this->updateImportJob($importJob, null, 0, [], null);
        // Generate options for each key
        /** @var ContextBuilder $contextBuilder */
        $importOptionsGenerator = $this->getContainer()->get('cybersponse.core.service.import_options_generator');
        
        $this->updateImportJob($importJob, self::STATUS['GENERATING_OPTIONS'], 10, null, null, 'starting');

        $options = $importOptionsGenerator->generate($config, $importJob, $tokenType);

        // Save options to entity
        $this->updateImportJob($importJob, self::STATUS['OPTIONS_GENERATED'], 100, $options);
    }

    public function updateImportJob(ImportJob $importJob, $status = null, $progressPercent = null, $options = null, $errorMessage = null, $currentlyImporting = null)
    {
        $entityManager = $this->getContainer()->get('doctrine')->getManagerForClass(ImportJob::class);
        $oldData = [
            '@id' => '/api/3/import_jobs/' . $importJob->getUuid()
        ];
        $newData = [
            '@id' => '/api/3/import_jobs/' . $importJob->getUuid()
        ];
        if ($status !== null) {
            $oldStatus = $importJob->getStatus();
            $importJob->setStatus($status);
            $oldData['status'] = $oldStatus;
            $newData['status'] = $status;
        }
        if ($progressPercent !== null) {
            $oldProgressPercent = $importJob->getProgressPercent();
            $importJob->setProgressPercent($progressPercent);
            $oldData['progressPercent'] = $oldProgressPercent;
            $newData['progressPercent'] = $progressPercent;
        }
        if ($errorMessage !== null) {
            $oldErrorMessage = $importJob->getErrorMessage();
            $importJob->setErrorMessage($errorMessage);
            $oldData['errorMessage'] = $oldErrorMessage;
            $newData['errorMessage'] = $errorMessage;
        }
        if ($options !== null) {
            $oldOptions = $importJob->getOptions();
            $importJob->setOptions($options);
            $oldData['options'] = $oldOptions;
            $newData['options'] = $options;
        }
        if ($currentlyImporting !== null) {
            $old = $importJob->getCurrentlyImporting();
            $importJob->setCurrentlyImporting($currentlyImporting);
            $oldData['currentlyImporting'] = $old;
            $newData['currentlyImporting'] = $currentlyImporting;
        }
        $timestamp = new \DateTime('NOW', new \DateTimeZone('UTC'));
        $importJob->setModifyDate($timestamp);
        $entityManager->persist($importJob);
        $entityManager->flush();
        $body = $this->prepareChangeData($oldData, $newData, $importJob->getUuid());
        $utilsObj = new Utils();
        $utilsObj->sendData($this, [$body]);
    }

    private function importConfigJson(array $config, ImportJob $importJob, string $folderPath, $tokenType='JWT')
    {
        /** @var ImportService $importService */
        $importService = $this->getContainer()->get('cybersponse.core.service.import_service');

        $this->updateImportJob($importJob, self::STATUS['IMPORTING']);
        $configImportResult = $importService->importConfiguration($config, $importJob, $this, $tokenType);
        if (!$configImportResult['success']) {
            throw new Exception('Configuration import failed.');
        }
        $this->updateImportJob($importJob, self::STATUS['IMPORT_COMPLETE'], null, null, null, 'completed');
    }

    /**
     * @return array
     */
    protected function prepareChangeData(array $oldData, array $newData, string $entityId)
    {
        $deltaData = [];
        $deltaData['oldData'] = $oldData;
        $deltaData['newData'] = $newData;
        $tokenStorage = $this->getContainer()->get('security.token_storage');
        $user = $tokenStorage->getToken() ? $tokenStorage->getToken()->getUser()->getUuid(): 'undefined';
        $userName = is_string($user) ? 'undefined': Utils::extractUserName($user);

        $changeData = [
            self::COMPONENT => 'crudhub',
            self::ENTITY_TYPE => 'ImportJob',
            self::UUID => $entityId,
            self::HTTP_METHOD => 'Update',
            self::USER_UUID => $user,
            self::USER => $userName,
            self::TRANSACTION_DATE => (int)($_SERVER['REQUEST_TIME_FLOAT']*1000),
            self::DELTA_DATA => $deltaData
        ];

        return $changeData;
    }
    
    public function get($service)
    {
        return $this->getContainer()->get($service);
    }
}
