<?php
/* Copyright start
  Copyright (C) 2008 - 2020 Fortinet Inc.
  All rights reserved.
  FORTINET CONFIDENTIAL & FORTINET PROPRIETARY SOURCE CODE
  Copyright end */
namespace Cybersponse\CoreBundle\Controller;

use Cybersponse\AuthorizationBundle\Permissions\PermissionsDictionaryBuilder;
use Cybersponse\CoreBundle\Api\ApiEvents;
use Cybersponse\UtilsBundle\Utils;
use Cybersponse\WorkflowBundle\Subscriber\DataEvent;
use Dunglas\ApiBundle\Event\Events;
use Dunglas\ApiBundle\Exception\DeserializationException;
use Dunglas\ApiBundle\JsonLd\Response;
use Symfony\Component\HttpFoundation\JsonResponse;
use Symfony\Component\HttpFoundation\Request;
use Symfony\Component\HttpKernel\Exception\HttpException;
use Symfony\Component\HttpKernel\Exception\NotFoundHttpException;

/**
 * Class ProxyController
 * @package Cybersponse\CoreBundle\Controller
 */
class ProxyController extends AuthorizationResourceController
{
    const MODULE_TYPE = "proxy";
    const ENCODE_DECODE_FIELDS = ['proxies'];
    const PROXY_UUID = 'c90e8c1b-cu11-as33-3sk-717199a2e182';
    const PROXY_ROUTING_KEY = 'key.crudhub.proxy';

    /**
     * @param Request $request
     * @param $id
     * @return Response
     */
    public function getOnIdAction(Request $request, $id)
    {
        return $this->getAction($request);
    }

    /**
     * @param Request $request
     * @return Response|static
     */
    public function getAction(Request $request)
    {
        $resource = $this->identifyResourceByShortName(self::MODULE_TYPE);
        $this->denyAccessUnlessGranted($resource, PermissionsDictionaryBuilder::PERM_CRUD_READ);
        $this->get('event_dispatcher')->dispatch(ApiEvents::PRE_API_RETRIEVE_LIST, new DataEvent($resource, null));
        try {
            $object = $this->findOrThrowNotFound($resource, self::PROXY_UUID);
        } catch (NotFoundHttpException $e) {
            return JsonResponse::create((object)null);
        }
        $serializer = $this->get('serializer');
        $updatedObject = $serializer->normalize($object, self::NORMALIZER_METHOD, $resource->getNormalizationContext());
        return $this->sendResponse($updatedObject);
    }

    /**
     * @param $dataObject
     * @param bool $publish_data
     * @return \Symfony\Component\HttpFoundation\Response|static
     */
    public function sendResponse($dataObject, $publish_data = false)
    {
        $cloneDataObject = $dataObject;
        $isLocalhost = in_array($_SERVER['REMOTE_ADDR'], ['127.0.0.1', 'localhost', '::1']);
        $dataObject['proxies'] = json_decode($dataObject['proxies']);
        $cloneDataObject['proxies'] = json_decode($cloneDataObject['proxies']);

        foreach ($cloneDataObject['proxies'] as $ot_key => $fields) {
            $fields->value = $this->buildURL($fields);
        }
        if ($publish_data) {
            $utilsObj = new Utils();
            $utilsObj->sendReplicationData($this, [$cloneDataObject], self::PROXY_ROUTING_KEY);
        }
        
        if (!$isLocalhost) {
            foreach ($dataObject['proxies'] as $ot_key => $fields) {
                foreach ($fields as $in_key => $value) {
                    if ($in_key == 'password') {
                        $dataObject['proxies']->$ot_key->$in_key = null;
                    }
                }
            }
            return JsonResponse::create($dataObject);
        }
        return JsonResponse::create($cloneDataObject);
    }

    /**
     * @param $data
     * @return string
     */
    protected function buildURL($data)
    {
        $stringValue = '';
        $protocol = 'https://';
        if (!empty($data->url)) {
            if (substr($data->url, 0, 7) === "http://") {
                $url = substr($data->url, 7, strlen($data->url));
                $protocol = 'http://';
            } else if (substr($data->url, 0, 8) === "https://") {
                $url = substr($data->url, 8, strlen($data->url));
                $protocol = 'https://';
            } else {
                $url = $data->url;
            }
            if (!empty($data->username) && !empty($data->password)) {
                $stringValue = $protocol . $data->username . ':' . $data->password . '@' . $url;
            } else {
                $stringValue = $protocol . $url;
            }
            if (!empty($data->port)) {
                $stringValue .= ':' . $data->port;
            }
        }
        return $stringValue;
    }

    /**
     * @param Request $request
     * @return Response
     * @throws DeserializationException
     */
    public function postAction(Request $request)
    {
        $resource = $this->identifyResourceByShortName(self::MODULE_TYPE);
        $this->denyAccessUnlessGranted($resource, PermissionsDictionaryBuilder::PERM_CRUD_CREATE);
        $this->get('event_dispatcher')->dispatch(ApiEvents::PRE_API_CREATE, new DataEvent($resource, null));
        $decodedRequestData = json_decode($request->getContent(), true);
        $request_data = $this->prePostAdd($decodedRequestData);
        $object = $this->createObject($request_data, $resource, $resource->getDenormalizationContext());
        $this->get('event_dispatcher')->dispatch(Events::PRE_CREATE_VALIDATION, new DataEvent($resource, $object));
        $violations = $this->get('validator')->validate($object, null, $resource->getValidationGroups());
        if (0 === count($violations)) {
            $this->get('event_dispatcher')->dispatch(Events::PRE_CREATE, new DataEvent($resource, $object));
            $serializer = $this->get('serializer');
            $updatedObject = $serializer->normalize($object, self::NORMALIZER_METHOD, $resource->getNormalizationContext());
            return $this->sendResponse($updatedObject, true);
        }
        return $this->getErrorResponse($violations);
    }

    /**
     * @param $decodedRequestData
     * @return mixed
     */
    public function prePostAdd($decodedRequestData)
    {
        $this->validateProxy($decodedRequestData);
        $decodedRequestData['proxies'] = json_encode($decodedRequestData['proxies']);
        $decodedRequestData['uuid'] = self::PROXY_UUID;
        array_push($decodedRequestData['noProxy'], 'localhost', '127.0.0.1');
        $decodedRequestData['noProxy'] = array_unique($decodedRequestData['noProxy']);
        return $decodedRequestData;
    }

    /**
     * @param $proxy_body
     */
    private function validateProxy($proxy_body)
    {
        if ($proxy_body === null) {
            throw new HttpException(Response::HTTP_BAD_REQUEST, "Proxy data cannot be null");
        }
        if (empty($proxy_body['proxies'])) {
            throw new HttpException(Response::HTTP_BAD_REQUEST, "HTTP and HTTPS Proxy cannot be null");
        }
    }

    /**
     * @param Request $request
     * @param string $id
     * @return ProxyController|Response|\Symfony\Component\HttpFoundation\Response|static
     * @throws DeserializationException
     */
    public function putAction(Request $request, $id)
    {
        $resource = $this->identifyResourceByShortName(self::MODULE_TYPE);
        $this->denyAccessUnlessGranted($resource, PermissionsDictionaryBuilder::PERM_CRUD_UPDATE);
        $this->get('event_dispatcher')->dispatch(ApiEvents::PRE_API_UPDATE, new DataEvent($resource, null));
        $object = $this->findOrThrowNotFound($resource, self::PROXY_UUID);
        $context = $resource->getDenormalizationContext();
        $context['object_to_populate'] = $object;
        $serializer = $this->get('serializer');
        $objectData = json_decode($request->getContent(), true);
        $preUpdatedObject = $serializer->normalize($object, self::NORMALIZER_METHOD, $resource->getNormalizationContext());
        $request_data = $this->prePutAdd($objectData, $preUpdatedObject);
        $object = $this->createObject($request_data, $resource, $context);
        $this->get('event_dispatcher')->dispatch(Events::PRE_UPDATE_VALIDATION, new DataEvent($resource, $object));
        $violations = $this->get('validator')->validate($object, null, $resource->getValidationGroups());
        if (0 === count($violations)) {
            // Validation succeed
            $this->get('event_dispatcher')->dispatch(Events::PRE_UPDATE, new DataEvent($resource, $object, $preUpdatedObject));
            $updatedObject = $serializer->normalize($object, self::NORMALIZER_METHOD, $resource->getNormalizationContext());
            return $this->sendResponse($updatedObject, true);
        }
        return $this->getErrorResponse($violations);
    }

    /**
     * @param $objectData
     * @param $preUpdatedObject
     * @return mixed
     */
    protected function prePutAdd($objectData, $preUpdatedObject)
    {
        $proxies = json_decode($preUpdatedObject['proxies']);
        $objectData['proxies'] = $this->merge_existing_proxy($proxies, $objectData['proxies']);
        if (isset($objectData['noProxy'])) {
            array_push($objectData['noProxy'], 'localhost', '127.0.0.1');
            $objectData['noProxy'] = array_unique($objectData['noProxy']);
        }
        return $objectData;
    }

    /**
     * @param $exiting_proxy
     * @param $new_proxy
     * @return string
     */
    protected function merge_existing_proxy($exiting_proxy, $new_proxy)
    {
        if($new_proxy) {
            foreach ($new_proxy as $ot_key => $field) {
                if (!array_key_exists("password", $field)) {
                    if (array_key_exists('password', $exiting_proxy->$ot_key)) {
                        $new_proxy[$ot_key]['password'] = $exiting_proxy->$ot_key->password;
                    }
                }
            }
            return json_encode($new_proxy);
        }
        return json_encode($exiting_proxy);
    }

    /**
     * @param Request $request
     * @param string $id
     */
    public function deleteAction(Request $request, $id)
    {
        throw new HttpException(Response::HTTP_METHOD_NOT_ALLOWED, "Method Not Implemented");
    }

    /**
     * @param Request $request
     */
    public function bulkAction(Request $request)
    {
        throw new HttpException(Response::HTTP_METHOD_NOT_ALLOWED, "Bulk Operation on Proxy Module is not allowed");
    }
}
