<?php
/* Copyright start
  Copyright (C) 2008 - 2020 Fortinet Inc.
  All rights reserved.
  FORTINET CONFIDENTIAL & FORTINET PROPRIETARY SOURCE CODE
  Copyright end */
namespace Cybersponse\CoreBundle\Controller;

use Cybersponse\AuthorizationBundle\Permissions\PermissionsDictionaryBuilder;
use Cybersponse\CoreBundle\Api\ApiEvents;
use Cybersponse\WorkflowBundle\Subscriber\DataEvent;
use Dunglas\ApiBundle\Api\ResourceInterface;
use Dunglas\ApiBundle\Event\Events;
use Dunglas\ApiBundle\Exception\DeserializationException;
use Dunglas\ApiBundle\JsonLd\Response;
use Symfony\Component\HttpFoundation\Request;
use Symfony\Component\HttpKernel\Exception\HttpException;

/**
 * Class SecretController
 * @package Cybersponse\CoreBundle\Controller
 */
class SecretController extends AuthorizationResourceController
{
    const MODULE_TYPE = "secrets";


    /**
     * @param Request $request
     * @return Response
     */
    public function cgetAction(Request $request)
    {
        $resource = $this->getResource($request);
        $this->denyAccessUnlessGranted($resource, PermissionsDictionaryBuilder::PERM_CRUD_READ);
        $this->get(self::EVENT_DISPATCHER)->dispatch(ApiEvents::PRE_API_RETRIEVE_LIST, new DataEvent($resource, null));
        $data = $this->getCollectionData($resource, $request);
        if ($request->get($this->container->getParameter('api.collection.pagination.page_parameter_name')) &&
            0 === count($data)
        ) {
            throw $this->createNotFoundException();
        }
        $this->get(self::EVENT_DISPATCHER)->dispatch(Events::RETRIEVE_LIST, new DataEvent($resource, $data));

        return $this->getSuccessResponse($resource, $data, Response::HTTP_OK, [], ['request_uri' => $request->getRequestUri()]);
    }

    /**
     * @param ResourceInterface $resource
     * @param array|object $data
     * @param int $status
     * @param array $headers
     * @param array $additionalContext
     * @param array $__selectFields
     * @return Response
     */
    protected function getSuccessResponse(
        ResourceInterface $resource,
        $data,
        $status = 200,
        array $headers = [],
        array $additionalContext = [],
        array $__selectFields = []
    )
    {
        $normalized_data = $this->get('serializer')->normalize(
            $data,
            'json-ld',
            $resource->getNormalizationContext() + $additionalContext
        );
        $isLocalhost = in_array($_SERVER['REMOTE_ADDR'], ['127.0.0.1', 'localhost', '::1']);
        if (!$isLocalhost) {
            $__selectFields = ['@id', '@type', 'name', '@context'];
            $normalized_data = $this->trimOutput($normalized_data, $__selectFields);
        }

        $response = new Response(
            $normalized_data,
            $status,
            $headers
        );
        if ($status == Response::HTTP_CREATED) {
            $iri = $this->get('api.iri_converter')->getIriFromItem($data);
            $response->headers->set('Location', $iri);
        }
        return $response;
    }


    /**
     * @param Request $request
     * @param int $id
     * @return Response
     */
    public function getAction(Request $request, $id)
    {
        $resource = $this->identifyResourceByShortName(self::MODULE_TYPE);
        $this->denyAccessUnlessGranted($resource, PermissionsDictionaryBuilder::PERM_CRUD_READ);
        $object = $this->findOrThrowNotFound($resource, $id, $request);
        $this->denyAccessUnlessGranted($resource, PermissionsDictionaryBuilder::PERM_CRUD_READ, $object);
        return $this->getSuccessResponse($resource, $object);
    }


    /**
     * @param Request $request
     * @return Response
     * @throws DeserializationException
     */
    public function postAction(Request $request)
    {
        $resource = $this->identifyResourceByShortName(self::MODULE_TYPE);
        $this->denyAccessUnlessGranted($resource, PermissionsDictionaryBuilder::PERM_CRUD_CREATE);
        $this->get(self::EVENT_DISPATCHER)->dispatch(ApiEvents::PRE_API_CREATE, new DataEvent($resource, null));
        $decodedRequestData = json_decode($request->getContent(), true);
        $object = $this->createObject($decodedRequestData, $resource, $resource->getDenormalizationContext());
        $this->get(self::EVENT_DISPATCHER)->dispatch(Events::PRE_CREATE_VALIDATION, new DataEvent($resource, $object));
        $violations = $this->get('validator')->validate($object, null, $resource->getValidationGroups());
        if (0 === count($violations)) {
            $this->get(self::EVENT_DISPATCHER)->dispatch(Events::PRE_CREATE, new DataEvent($resource, $object));
            return $this->getSuccessResponse($resource, $object);
        }
        return $this->getErrorResponse($violations);
    }

    /**
     * @param Request $request
     * @param string $id
     * @return Response
     * @throws DeserializationException
     */
    public function putAction(Request $request, $id)
    {
        $resource = $this->identifyResourceByShortName(self::MODULE_TYPE);
        $this->denyAccessUnlessGranted($resource, PermissionsDictionaryBuilder::PERM_CRUD_UPDATE);
        $this->get(self::EVENT_DISPATCHER)->dispatch(ApiEvents::PRE_API_UPDATE, new DataEvent($resource, null));
        $object = $this->findOrThrowNotFound($resource, $id);
        $context = $resource->getDenormalizationContext();
        $context['object_to_populate'] = $object;
        $serializer = $this->get('serializer');
        $objectData = json_decode($request->getContent(), true);
        $preUpdatedObject = $serializer->normalize($object, 'json-ld', $resource->getNormalizationContext());
        $object = $this->createObject($objectData, $resource, $context);
        $this->get(self::EVENT_DISPATCHER)->dispatch(Events::PRE_UPDATE_VALIDATION, new DataEvent($resource, $object));
        $violations = $this->get('validator')->validate($object, null, $resource->getValidationGroups());
        if (0 === count($violations)) {
            // Validation succeed
            $this->get(self::EVENT_DISPATCHER)->dispatch(Events::PRE_UPDATE, new DataEvent($resource, $object, $preUpdatedObject));
            return $this->getSuccessResponse($resource, $object);
        }
        return $this->getErrorResponse($violations);
    }

    public function bulkAction(Request $request)
    {
        throw new HttpException(Response::HTTP_METHOD_NOT_ALLOWED, "Bulk Operation on Secret Module is not allowed");
    }
}
