<?php
/* Copyright start
  Copyright (C) 2008 - 2020 Fortinet Inc.
  All rights reserved.
  FORTINET CONFIDENTIAL & FORTINET PROPRIETARY SOURCE CODE
  Copyright end */

namespace Cybersponse\CoreBundle\DependencyInjection;


use Cybersponse\ActorsBundle\Entity\Actor;
use Cybersponse\CoreBundle\Serializer\EfficientCollectionNormalizer;
use Cybersponse\CoreBundle\Serializer\EfficientItemNormalizer;
use Cybersponse\MetadataBundle\Generation\Annotations\Classes\ModelType\ModelType;
use Cybersponse\MetadataBundle\Generation\Annotations\Classes\ModuleType\ModuleType;
use Cybersponse\MetadataBundle\Generation\Annotations\Classes\Operations\Operations;
use Cybersponse\ViewTemplateBundle\Entity\SystemViewTemplate;
use Dunglas\ApiBundle\Api\Operation\Operation;
use ReflectionClass;
use Symfony\Component\DependencyInjection\Compiler\CompilerPassInterface;
use Symfony\Component\DependencyInjection\ContainerBuilder;
use Symfony\Component\DependencyInjection\Definition;
use Symfony\Component\DependencyInjection\Reference;
use Symfony\Component\Finder\Finder;
use Symfony\Component\Finder\SplFileInfo;

class DunglasApiMetadataCompilerPass implements CompilerPassInterface {

    /**
     * You can modify the container here before it is dumped to PHP code.
     *
     * @param ContainerBuilder $container
     *
     * @api
     */
    public function process(ContainerBuilder $container)
    {
        $apiPropertyInfo = $container->getDefinition('api.property_info');

        $argument = $apiPropertyInfo->getArgument(0);
        $apiPropertyInfo->replaceArgument(0, $argument);

        $configurationHandlerMap = [
            Configuration::CLASS_TYPE => __NAMESPACE__ . '\ApiClassConfigurationHandler',
            Configuration::NAMESPACE_TYPE => __NAMESPACE__. '\ApiNamespaceConfigurationHandler'
        ];

        $apiConfigurations = $container->getParameter('cybersponse.core.configurations');

        foreach($apiConfigurations as $apiConfiguration) {
            /** @var ApiConfigurationHandlerInterface $configurationHandler */
            $configurationHandler = new $configurationHandlerMap[$apiConfiguration['type']]($container);

            $configurationHandler->handle($apiConfiguration['path'], $apiConfiguration['parent']);
        }

        $itemNormalizerDefinition = $container->getDefinition('api.json_ld.normalizer.item');
        $itemNormalizerDefinition->setClass(EfficientItemNormalizer::class);

        $collectionNormalizerDefinition = $container->getDefinition('api.hydra.normalizer.collection');
        $collectionNormalizerDefinition->setClass(EfficientCollectionNormalizer::class);

        //Move arguments around for compatibility
        $nameConverterArgument = $itemNormalizerDefinition->getArgument(5);
        $itemNormalizerDefinition->replaceArgument(5, new Reference('request_stack'));
        $itemNormalizerDefinition->addArgument($nameConverterArgument);
        $traitFlagProviderReference = new Reference('cybersponse.core.traits.trait_flag_provider');
        $itemNormalizerDefinition->addArgument($traitFlagProviderReference);
        $authorizationProviderReference = new Reference('cybersponse.authorization.authorization_provider');
        $itemNormalizerDefinition->addMethodCall('setAuthorizationProvider', [$authorizationProviderReference]);
        $permissionResolverReference = new Reference('cybersponse.authorization.permissions.permission_resolver');
        $itemNormalizerDefinition->addMethodCall('setPermissionResolver', [$permissionResolverReference]);
    }
}

interface ApiConfigurationHandlerInterface {
    public function handle($path, $parent);
}

abstract class ApiConfigurationHandler implements ApiConfigurationHandlerInterface {
    /** @var ContainerBuilder */
    protected $container;

    function __construct($container)
    {
        $this->container = $container;
    }

    protected function registerClass($className, $parent) {
        $explodedClassName = explode('\\', $className);
        $baseName = end($explodedClassName);

        $resourceServiceName = 'resource.' . strtolower($baseName);

        if ($this->container->hasDefinition($resourceServiceName)) {
            return;
        }

        if (!is_null($parent) && !$this->extendsOrImplements($className, $parent)) {
            return;
        }

        // Todo this should all be moved to cybersponse_core:api configuration some day
        $defaultOperationConfig = [
            'enabled' => true,
            'requirements' => []
        ];

        $itemOperationConfigs = [
            'GET' => $defaultOperationConfig,
            'PUT' => $defaultOperationConfig,
            'DELETE' => $defaultOperationConfig
        ];
        $collectionOperationConfigs = [
            'GET' => $defaultOperationConfig,
            'POST' => $defaultOperationConfig
        ];
        $itemFieldOperationConfigs = [
            'GET' => $defaultOperationConfig,
            'POST' => $defaultOperationConfig,
            'DELETE' => $defaultOperationConfig
        ];

        $itemOperations = [];

        /*
         * @todo We need to take this check out and do it at the metadata level,
         * @todo which means we should move this part of compiler pass into the actor bundle
         * @todo which also means we should probably fire off an event so the actor bundle can catch it
         */

        $reflectionClass = new ReflectionClass($className);
        $annotationReader = $this->container->get('annotation_reader');

        /** @var Operations $operationsClassAnnotation */
        $operationsClassAnnotation = $annotationReader->getClassAnnotation($reflectionClass, Operations::class);

        if ($operationsClassAnnotation) {
            $itemOperationConfigs['GET']['enabled'] = $operationsClassAnnotation->read;
            $itemOperationConfigs['PUT']['enabled'] = $operationsClassAnnotation->update;
            $itemOperationConfigs['DELETE']['enabled'] = $operationsClassAnnotation->delete;

            $collectionOperationConfigs['GET']['enabled'] = $operationsClassAnnotation->read;
            $collectionOperationConfigs['POST']['enabled'] = $operationsClassAnnotation->create;

            //Field operations are all updates
            $itemFieldOperationConfigs['GET']['enabled'] = $operationsClassAnnotation->update;
            $itemFieldOperationConfigs['POST']['enabled'] = $operationsClassAnnotation->update;
            $itemFieldOperationConfigs['DELETE']['enabled'] = $operationsClassAnnotation->update;
        }

        if ($className == Actor::class) {
            $itemOperations[] = new Reference('resource.actor.current_item_operation.get');
            $itemOperations[] = new Reference('resource.actor.current_item_operation.put');
        }

        if ($this->extendsOrImplements($className, SystemViewTemplate::class)) {
            $itemOperationConfigs['GET']['requirements'] = [ 'id' => '.*' ];
            $itemOperationConfigs['PUT']['requirements'] = [ 'id' => '.*' ];
            $itemOperationConfigs['DELETE']['requirements'] = [ 'id' => '.*' ];
            $collectionOperationConfigs['GET']['requirements'] = [ 'id' => '.*' ];
            $collectionOperationConfigs['POST']['requirements'] = [ 'id' => '.*' ];
            $itemFieldOperationConfigs['GET']['requirements'] = [ 'id' => '.*' ];
        }

        $collectionOperations = $this->buildOperations($collectionOperationConfigs, $baseName, 'collection');
        $itemOperations = array_merge($itemOperations, $this->buildOperations($itemOperationConfigs, $baseName, 'item'));
        $itemFieldOperations = $this->buildOperations($itemFieldOperationConfigs, $baseName, 'itemField');

        $itemOperations = array_merge($itemOperations, $itemFieldOperations);

        $resourceDefinition = new Definition();

        /** @var ModelType $modelTypeClassAnnotation */
        $modelTypeClassAnnotation = $annotationReader->getClassAnnotation($reflectionClass, ModelType::class);

        if ($modelTypeClassAnnotation) {
            $resourceDefinition->addMethodCall('setModelType', [$modelTypeClassAnnotation->modelType]);

            //This will be the same by default unless overridden by annotation later
            $resourceDefinition->addMethodCall('setModuleType', [$modelTypeClassAnnotation->modelType]);
        }

        /** @var ModuleType $moduleTypeClassAnnotation */
        $moduleTypeClassAnnotation = $annotationReader->getClassAnnotation($reflectionClass, ModuleType::class);

        if ($moduleTypeClassAnnotation) {
            $resourceDefinition->addMethodCall('setModuleType', [$moduleTypeClassAnnotation->moduleType]);
        }

        $resourceDefinition
            ->setClass($this->container->getParameter('cybersponse.core.model_aware_resource.class'))
            ->addArgument($className)
            ->addMethodCall('initFilters', [[
                new Reference('cybersponse.query.orm.venom_query_filter'),
                new Reference('cybersponse.query.orm.venom_search_filter')
            ]])
            ->addMethodCall('initItemOperations', [$itemOperations])
            ->addMethodCall('initCollectionOperations', [$collectionOperations])
            ->addTag('api.resource');

        $this->container->setDefinition($resourceServiceName, $resourceDefinition);
    }

    /**
     * @param $operationConfigs
     * @param $baseName
     * @param $operationType String can be either 'item' or 'collection' or 'itemField'
     * @return array
     */
    protected function buildOperations($operationConfigs, $baseName, $operationType)
    {
        $operations = [];

        foreach ($operationConfigs as $operationName => $operationConfig) {
            $arguments = $this->buildArgumentsFromConfig($baseName, $operationConfig, $operationName);

            if (is_null($arguments)) {
                continue;
            }

            $operationDefinitionName = 'resource.' . $baseName . '.' . $operationType . '_operation.' . strtolower($operationName);
            $operationDefinition = new Definition();
            $operationDefinition
                ->setClass(Operation::class)
                ->setPublic(false)
                ->setFactory([new Reference('api.operation_factory'), 'create' . ucfirst($operationType) . 'Operation'])
                ->setArguments($arguments);

            $this->container->setDefinition($operationDefinitionName, $operationDefinition);
            $operations[] = new Reference($operationDefinitionName);
        }

        return $operations;
    }

    /**
     * @param $baseName
     * @param $operationConfig
     * @param $operationName
     * @return array|null
     */
    private function buildArgumentsFromConfig($baseName, $operationConfig, $operationName)
    {
        if (!$operationConfig['enabled']) {
            return null;
        }

        $arguments = [
            new Reference('resource.' . $baseName), //Resource service name
            strtoupper($operationName), // Method
            null, // Path
            null, // Controller
            null, // Route Name
            [], // Additional Context
        ];

        if ($operationConfig['requirements']) {
            $arguments[] = $operationConfig['requirements'];
        }
        return $arguments;
    }

    private function extendsOrImplements($className, $parent) {
        $class = new ReflectionClass($className);
        if($class === false) {
            return false;
        }

        do {
            $name = $class->getName();
            if($parent == $name) {
                return true;
            }
            $interfaces = $class->getInterfaceNames();
            if(is_array($interfaces) && in_array($parent, $interfaces)) {
                return true;
            }
            $class = $class->getParentClass();
        } while($class !== false);

        return false;
    }
}

class ApiNamespaceConfigurationHandler extends ApiConfigurationHandler {

    public function handle($path, $parent)
    {
        $finder = new Finder();

        $finder->files()->in(__DIR__ . '/../../../' . str_replace('\\', '/', $path))->depth(0);

        /** @var SplFileInfo $file */
        foreach($finder as $file) {
            $baseName = $file->getBasename('.php');

            $className = $path . '\\' . $baseName;

            $this->registerClass($className, $parent);
        }
    }
}

class ApiClassConfigurationHandler extends ApiConfigurationHandler {
    public function handle($path, $parent)
    {
        $this->registerClass($path, $parent);
    }
}
