<?php
/* Copyright start
  Copyright (C) 2008 - 2020 Fortinet Inc.
  All rights reserved.
  FORTINET CONFIDENTIAL & FORTINET PROPRIETARY SOURCE CODE
  Copyright end */

namespace Cybersponse\CoreBundle\Doctrine\Orm;


use Doctrine\Common\Cache\Cache;
use Doctrine\Common\Collections\ArrayCollection;
use Doctrine\Common\Inflector\Inflector;
use Doctrine\Common\Persistence\Mapping\ReflectionService;
use Doctrine\Common\Persistence\Mapping\RuntimeReflectionService;
use Doctrine\Common\Reflection\RuntimePublicReflectionProperty;
use Doctrine\Common\Util\ClassUtils;
use Doctrine\ORM\EntityManager;
use Doctrine\ORM\Mapping\ClassMetadata;
use Doctrine\ORM\Mapping\ClassMetadataFactory;
use Doctrine\ORM\Mapping\ClassMetadataInfo;
use Doctrine\ORM\Mapping\MappingException;
use Doctrine\ORM\PersistentCollection;
use ReflectionClass;
use ReflectionProperty;

class DisconnectedClassMetadataFactory extends ClassMetadataFactory
{
    /**
     * Salt used by specific Object Manager implementation.
     *
     * @var string
     */
    protected $cacheSalt = '$CRUDHUB-CLASSMETADATA';

    /**
     * @var \Doctrine\Common\Cache\Cache|null
     */
    protected $cacheDriver;

    /**
     * @var ClassMetadata[]
     */
    protected $loadedMetadata = array();

    /**
     * @var bool
     */
    protected $initialized = false;

    /**
     * @var ReflectionService|null
     */
    protected $reflectionService = null;

    /**
     * @return \Doctrine\Common\Persistence\Mapping\StaticReflectionService
     */
    public function getReflectionService()
    {
        return new SemiStaticReflectionService();
    }

    /**
     * @return string
     */
    public function getCacheSalt()
    {
        return $this->cacheSalt;
    }

    /**
     * @inheritdoc
     */
    public function getLoadedMetadata()
    {
        return $this->loadedMetadata;
    }

    /**
     * @inheritdoc
     */
    public function setCacheDriver(Cache $cacheDriver = null)
    {
        $this->cacheDriver = $cacheDriver;
    }

    /**
     * @inheritdoc
     */
    public function getMetadataFor($className)
    {
        if (is_array($className)) {
            $className = $className[0];
        }

        if (isset($this->loadedMetadata[$className])) {
            return $this->loadedMetadata[$className];
        }

        // Check for namespace alias
        if (strpos($className, ':') !== false) {
            list($namespaceAlias, $simpleClassName) = explode(':', $className, 2);

            $realClassName = $this->getFqcnFromAlias($namespaceAlias, $simpleClassName);
        } else {
            $realClassName = ClassUtils::getRealClass($className);
        }

        if (isset($this->loadedMetadata[$realClassName])) {
            // We do not have the alias name in the map, include it
            return $this->loadedMetadata[$className] = $this->loadedMetadata[$realClassName];
        }

        $loadingException = null;

        try {
            if ($this->cacheDriver) {
                if (($cached = $this->cacheDriver->fetch($realClassName . $this->cacheSalt)) !== false) {
                    $this->loadedMetadata[$realClassName] = $cached;

                    $this->wakeupReflection($cached, $this->getReflectionService());
                } else {
                    foreach ($this->loadMetadata($realClassName) as $loadedClassName) {
                        $this->cacheDriver->save(
                            $loadedClassName . $this->cacheSalt,
                            $this->loadedMetadata[$loadedClassName],
                            null
                        );
                    }
                }
            } else {
                $this->loadMetadata($realClassName);
            }
        } catch (MappingException $loadingException) {
            if (!$fallbackMetadataResponse = $this->onNotFoundMetadata($realClassName)) {
                throw $loadingException;
            }

            $this->loadedMetadata[$realClassName] = $fallbackMetadataResponse;
        }

        if ($className !== $realClassName) {
            // We do not have the alias name in the map, include it
            $this->loadedMetadata[$className] = $this->loadedMetadata[$realClassName];
        }

        return $this->loadedMetadata[$className];
    }

    /**
     * @inheritdoc
     */
    public function hasMetadataFor($className)
    {
        return isset($this->loadedMetadata[$className]);
    }

    /**
     * @inheritdoc
     */
    public function setMetadataFor($className, $class)
    {
        $this->loadedMetadata[$className] = $class;
    }

    /**
     * Invalidates the class metadata
     *
     * @param $className
     */
    public function invalidateMetadata($className)
    {
        unset($this->loadedMetadata[$className]);
        $this->cacheDriver->delete($className . $this->cacheSalt);
    }

    /**
     * Saves the class metadata
     *
     * @param ClassMetadataInfo $metadata
     */
    public function recacheMetadata($metadata)
    {
        $className = $metadata->getName();
        $this->invalidateMetadata($className);
        $this->cacheDriver->save($className . $this->cacheSalt, $metadata);

        $reflService = $this->getReflectionService();
        $this->wakeupReflection($metadata, $reflService);

        $this->setMetadataFor($className, $metadata);
    }

    /**
     * @inheritdoc
     */
    protected function loadMetadata($name)
    {
        if (!$this->initialized) {
            $this->initialize();
        }

        $loaded = array();

        $parentClasses = $this->getParentClasses($name);
        $parentClasses[] = $name;

        // Move down the hierarchy of parent classes, starting from the topmost class
        $parent = null;
        $rootEntityFound = false;
        $visited = array();
        $reflService = $this->getReflectionService();
        foreach ($parentClasses as $className) {
            if (isset($this->loadedMetadata[$className])) {
                $parent = $this->loadedMetadata[$className];
                if ($this->isEntity($parent)) {
                    $rootEntityFound = true;
                    array_unshift($visited, $className);
                }
                continue;
            }

            $class = $this->newClassMetadataInstance($className);
            $this->initializeReflection($class, $reflService);

            $this->doLoadMetadata($class, $parent, $rootEntityFound, $visited);

            $this->loadedMetadata[$className] = $class;

            $this->resolveInversedRelationships($class);

            $parent = $class;

            if ($this->isEntity($class)) {
                $rootEntityFound = true;
                array_unshift($visited, $className);
            }

            $this->wakeupReflection($class, $reflService);

            $loaded[] = $className;
        }

        array_map([$this, 'resolveDiscriminatorValue'], array_map([$this, 'getMetadataFor'], $loaded));

        return $loaded;
    }

    /**
     * @inheritdoc
     */
    protected function resolveDiscriminatorValue(ClassMetadata $metadata)
    {
        if ($metadata->discriminatorValue
            || !$metadata->discriminatorMap
            || $metadata->isMappedSuperclass
            || !$metadata->reflClass
            || $metadata->reflClass->isAbstract()) {
            return;
        }

        // minor optimization: avoid loading related metadata when not needed
        foreach ($metadata->discriminatorMap as $discriminatorValue => $discriminatorClass) {
            if ($discriminatorClass === $metadata->name) {
                $metadata->discriminatorValue = $discriminatorValue;

                return;
            }
        }

        // iterate over discriminator mappings and resolve actual referenced classes according to existing metadata
        foreach ($metadata->discriminatorMap as $discriminatorValue => $discriminatorClass) {
            if ($metadata->name === $this->getMetadataFor($discriminatorClass)->getName()) {
                $metadata->discriminatorValue = $discriminatorValue;

                return;
            }
        }

        throw MappingException::mappedClassNotPartOfDiscriminatorMap($metadata->name, $metadata->rootEntityName);
    }

    /**
     * @param ClassMetadata $metadata
     *
     * @return void
     */
    private function resolveInversedRelationships(ClassMetadata $metadata)
    {
        if ($metadata->reflClass->isAbstract()) {
            return;
        }

        $className = $metadata->getName();
        $classParts = explode('\\', $className);
        $classShortName = end($classParts);

        //Should we create a field that we know they won't use?
        $classAsFieldName = Inflector::camelize(Inflector::pluralize($classShortName));

        //Find any missing relationships for the inversedByMap
        foreach ($metadata->associationMappings as &$associationMapping) {
            if ($associationMapping['type'] & ClassMetadata::TO_MANY) {

                if (!is_null($associationMapping['mappedBy'])) {
                    continue;
                }

                if (!is_null($associationMapping['inversedBy']) && !$associationMapping['isOwningSide']) {
                    continue;
                }

                //Add inversed relationships to all other class if not exists
                $inversedByClassName = $associationMapping['targetEntity'];

                if ($inversedByClassName != $className) {
                    //Clone
                    $inversedAssociationMapping = $associationMapping;
                    $inversedAssociationMapping['fieldName'] = $classAsFieldName;
                    $inversedAssociationMapping['targetEntity'] = $className;
                    $inversedAssociationMapping['isOwningSide'] = false;
                    $inversedAssociationMapping['mappedBy'] = $associationMapping['fieldName'];
                    $inversedAssociationMapping['sourceEntity'] = $inversedByClassName;

                    //Set inversed by
                    $associationMapping['inversedBy'] = $classAsFieldName;
                    $associationMapping['mappedBy'] = null;

                    /** @var ClassMetadata $inversedMetadata */
                    $inversedMetadata = $this->getMetadataFor($inversedByClassName);

                    if ($inversedMetadata->hasAssociation($classAsFieldName)) {
                        continue;
                    }

                    if ($inversedAssociationMapping['type'] == ClassMetadata::MANY_TO_MANY) {
                        $inversedMetadata->mapManyToMany($inversedAssociationMapping);
                    }
                    if ($inversedAssociationMapping['type'] == ClassMetadata::ONE_TO_MANY) {
                        $inversedMetadata->mapOneToMany($inversedAssociationMapping);
                    }

                    $this->recacheMetadata($inversedMetadata);
                }
            }
        }
    }
}

class SemiStaticReflectionService extends RuntimeReflectionService
{
    public function getClass($class)
    {
        return new ReflectionClass($class);
    }

    public function getAccessibleProperty($class, $property)
    {
        try {
            $reflectionProperty = new ReflectionProperty($class, $property);

            if ($reflectionProperty->isPublic()) {
                $reflectionProperty = new RuntimePublicReflectionProperty($class, $property);
            }

            $reflectionProperty->setAccessible(true);

            return $reflectionProperty;
        } catch (\Exception $exception) {
            return new FakeReflectionProperty($class, $property);
        }
    }
}

class FakeReflectionProperty extends ReflectionProperty
{
    protected $fakeClass;
    protected $fakeName;

    public function __construct($class, $name)
    {
        $this->fakeClass = $class;
        $this->fakeName = $name;
        return;
    }

    public function getFakeClass()
    {
        return $this->fakeClass;
    }

    public function getFakeName()
    {
        return $this->fakeName;
    }

    public function getValue($object = null)
    {
        $reflectionClass = new \ReflectionClass(EntityManager::class);

        /** @var EntityManager $entityManager */
        $entityManager = $reflectionClass->newInstanceWithoutConstructor();

        $collection = new ArrayCollection([]);

        $persistentCollection = new PersistentCollection($entityManager, $this->getFakeClass(), $collection);
        $persistentCollection->setOwner($object, [
            'inversedBy' => $this->getFakeName()
        ]);

        return $persistentCollection;
    }

    public function setValue($object, $value = null)
    {
        return null;
    }
}
