<?php
/* Copyright start
  Copyright (C) 2008 - 2020 Fortinet Inc.
  All rights reserved.
  FORTINET CONFIDENTIAL & FORTINET PROPRIETARY SOURCE CODE
  Copyright end */

namespace Cybersponse\CoreBundle\Serializer;


use Doctrine\Common\Collections\Collection;
use Dunglas\ApiBundle\Api\Filter\FilterInterface;
use Dunglas\ApiBundle\Api\ResourceCollectionInterface;
use Dunglas\ApiBundle\Api\ResourceInterface;
use Dunglas\ApiBundle\Api\ResourceResolverTrait;
use Dunglas\ApiBundle\Hydra\Serializer\CollectionNormalizer;
use Dunglas\ApiBundle\JsonLd\ContextBuilder;
use Dunglas\ApiBundle\JsonLd\Serializer\ContextTrait;
use Dunglas\ApiBundle\Model\PaginatorInterface;
use Symfony\Component\Security\Core\Exception\AccessDeniedException;
use Symfony\Component\Serializer\Normalizer\NormalizerInterface;

class EfficientCollectionNormalizer extends CollectionNormalizer
{
    use ResourceResolverTrait;
    use ContextTrait;

    /**
     * @var string
     */
    const HYDRA_COLLECTION = 'hydra:Collection';
    /**
     * @var string
     */
    const HYDRA_PAGED_COLLECTION = 'hydra:PagedCollection';

    /**
     * @var ContextBuilder
     */
    private $contextBuilder;
    /**
     * @var string
     */
    private $pageParameterName;

    public function __construct(
        ResourceCollectionInterface $resourceCollection,
        ContextBuilder $contextBuilder,
        $pageParameterName
    ) {
        parent::__construct($resourceCollection, $contextBuilder, $pageParameterName);

        $this->resourceCollection = $resourceCollection;
        $this->contextBuilder = $contextBuilder;
        $this->pageParameterName = $pageParameterName;
    }


    public function normalize($object, $format = null, array $context = [])
    {
        if (!$this->serializer instanceof NormalizerInterface) {
            throw new \Exception(sprintf('Expected class NormalizerInterface, got %s', get_class($this->serializer)));
        }

        $resource = $this->guessResource($object, $context);

        if (isset($context['json_ld_sub_level'])) {
            $data = [];
            $includeKeys = !isset($object[0]) && !($object instanceof Collection);

            foreach ($object as $index => $obj) {
                try {
                    $item = $this->serializer->normalize($obj, $format, $context);
                } catch (AccessDeniedException $accessDeniedException) {
                    continue;
                }

                if ($includeKeys) {
                    $data[$index] = $item;
                } else {
                    $data[] = $item;
                }
            }
        } else {
            $context = $this->createContext($resource, $context);
            $contextId = isset($context['request_uri']) ? $context['request_uri'] : null;
            $data = [
                '@context' => $this->contextBuilder->getResourceContext($resource, $context),
                '@id' => $contextId
            ];
            list($parts, $parameters) = $this->parseRequestUri($contextId);

            if ($object instanceof PaginatorInterface) {
                $data['@type'] = self::HYDRA_PAGED_COLLECTION;

                $currentPage = $object->getCurrentPage();
                $lastPage = $object->getLastPage();

                if (1. !== $currentPage) {
                    $previousPage = $currentPage - 1.;
                    $data['hydra:previousPage'] = $this->getPageUrl($parts, $parameters, $previousPage);
                }

                if ($currentPage !== $lastPage) {
                    $data['hydra:nextPage'] = $this->getPageUrl($parts, $parameters, $currentPage + 1.);
                }

                $data['hydra:totalItems'] = $object->getTotalItems();
                $data['hydra:itemsPerPage'] = $object->getItemsPerPage();
                $data['hydra:firstPage'] = $this->getPageUrl($parts, $parameters, 1.);
                $data['hydra:lastPage'] = $this->getPageUrl($parts, $parameters, $lastPage);
            } else {
                $data['@type'] = self::HYDRA_COLLECTION;
            }

            $data['hydra:member'] = [];
            foreach ($object as $obj) {
                $data['hydra:member'][] = $this->serializer->normalize($obj, $format, $context);
            }

            $filters = $resource->getFilters();
            if (!empty($filters)) {
                $data['hydra:search'] = $this->getSearch($resource, $parts, $filters);
            }
        }

        return $data;
    }

    /**
     * Parse and standardize the request URI.
     *
     * @param string $requestUri
     *
     * @return array
     */
    private function parseRequestUri($requestUri)
    {
        $parts = parse_url($requestUri);
        if (false === $parts) {
            throw new \InvalidArgumentException(sprintf('The request URI "%s" is malformed.', $requestUri));
        }

        $parameters = [];
        if (isset($parts['query'])) {
            parse_str($parts['query'], $parameters);

            // Remove existing page parameter
            if (isset($parameters[$this->pageParameterName])) {
                unset($parameters[$this->pageParameterName]);
            }
        }

        return [$parts, $parameters];
    }

    /**
     * Gets a collection URL for the given page.
     *
     * @param array $parts
     * @param array $parameters
     * @param float $page
     *
     * @return string
     */
    private function getPageUrl(array $parts, array $parameters, $page)
    {
        if (1. !== $page) {
            $parameters[$this->pageParameterName] = $page;
        }

        $parts['query'] = http_build_query($parameters);

        $url = $parts['path'];

        if ('' !== $parts['query']) {
            $url .= '?' . $parts['query'];
        }

        return $url;
    }

    /**
     * Returns the content of the Hydra search property.
     *
     * @param ResourceInterface $resource
     * @param array $parts
     * @param FilterInterface[] $filters
     *
     * @return array
     */
    private function getSearch(ResourceInterface $resource, array $parts, array $filters)
    {
        $variables = [];
        $mapping = [];
        foreach ($filters as $filter) {
            foreach ($filter->getDescription($resource) as $variable => $data) {
                $variables[] = $variable;
                $mapping[] = [
                    '@type' => 'IriTemplateMapping',
                    'variable' => $variable,
                    'property' => $data['property'],
                    'required' => $data['required'],
                ];
            }
        }

        return [
            '@type' => 'hydra:IriTemplate',
            'hydra:template' => sprintf('%s{?%s}', $parts['path'], implode(',', $variables)),
            'hydra:variableRepresentation' => 'BasicRepresentation',
            'hydra:mapping' => $mapping,
        ];
    }
}
