<?php
/* Copyright start
  Copyright (C) 2008 - 2020 Fortinet Inc.
  All rights reserved.
  FORTINET CONFIDENTIAL & FORTINET PROPRIETARY SOURCE CODE
  Copyright end */

namespace Cybersponse\CoreBundle\Serializer;


use Cybersponse\AuthorizationBundle\AuthorizationProvider;
use Cybersponse\AuthorizationBundle\Permissions\PermissionResolver;
use Cybersponse\AuthorizationBundle\Permissions\PermissionsDictionaryBuilder;
use Cybersponse\CoreBundle\Api\ModelAwareResource;
use Cybersponse\CoreBundle\Traits\TraitFlagProvider;
use Cybersponse\MetadataBundle\Generation\Metadata\ModelAttributeMetadata;
use Cybersponse\QueryBundle\Constants\AttributeLevelOperators;
use Doctrine\Common\Inflector\Inflector;
use Doctrine\Common\Proxy\Proxy;
use Dunglas\ApiBundle\Api\IriConverterInterface;
use Dunglas\ApiBundle\Api\ResourceCollectionInterface;
use Dunglas\ApiBundle\Api\ResourceInterface;
use Dunglas\ApiBundle\Api\ResourceResolverTrait;
use Dunglas\ApiBundle\Exception\InvalidArgumentException;
use Dunglas\ApiBundle\Exception\RuntimeException;
use Dunglas\ApiBundle\JsonLd\ContextBuilder;
use Dunglas\ApiBundle\JsonLd\Response;
use Dunglas\ApiBundle\JsonLd\Serializer\ContextTrait;
use Dunglas\ApiBundle\Mapping\AttributeMetadataInterface;
use Dunglas\ApiBundle\Mapping\ClassMetadataFactoryInterface;
use Dunglas\ApiBundle\Mapping\ClassMetadataInterface;
use Symfony\Component\Debug\Exception\ContextErrorException;
use Symfony\Component\HttpFoundation\Request;
use Symfony\Component\HttpFoundation\RequestStack;
use Symfony\Component\HttpKernel\Exception\HttpException;
use Symfony\Component\PropertyAccess\Exception\NoSuchPropertyException;
use Symfony\Component\PropertyAccess\PropertyAccessorInterface;
use Symfony\Component\Security\Core\Exception\AccessDeniedException;
use Symfony\Component\Serializer\NameConverter\NameConverterInterface;
use Symfony\Component\Serializer\Normalizer\AbstractNormalizer;
use Symfony\Component\Serializer\Normalizer\DenormalizerInterface;
use Symfony\Component\Serializer\Normalizer\NormalizerInterface;

class EfficientItemNormalizer extends AbstractNormalizer
{
    use ResourceResolverTrait;
    use ContextTrait;

    /** @var string */
    const FORMAT = 'json-ld';

    /** @var IriConverterInterface */
    private $iriConverter;

    /** @var ClassMetadataFactoryInterface */
    private $apiClassMetadataFactory;

    /** @var PropertyAccessorInterface */
    private $propertyAccessor;

    /** @var ContextBuilder */
    private $contextBuilder;

    /** @var Request */
    protected $request = null;

    /** @var TraitFlagProvider */
    protected $traitFlagProvider;

    /** @var AuthorizationProvider */
    protected $authorizationProvider;

    /** @var PermissionResolver */
    protected $permissionResolver;

    /** @var ResourceInterface[] */
    protected $authorizedResourceMap = [];

    const AS_IRI = 'as_iri';
    const IGNORE_AUTHORIZATION = 'ignore_authorization';
    const IGNORE_FIELD_AUTH = 'ignore_field_authorization';
    const DEPTH = 'depth';
    const DEFAULT_VISIBLE_FORM_TYPE = ["multiselectpicklist"];
    const DEFAULT_VISIBLE_EXPORTED_FORM_TYPE = ["tags"];


    public function __construct(
        ResourceCollectionInterface $resourceCollection,
        IriConverterInterface $iriConverter,
        ClassMetadataFactoryInterface $apiClassMetadataFactory,
        ContextBuilder $contextBuilder,
        PropertyAccessorInterface $propertyAccessor,
        RequestStack $requestStack = null,
        NameConverterInterface $nameConverter = null,
        TraitFlagProvider $traitFlagProvider = null
    ) {

        parent::__construct(null, $nameConverter);

        $this->resourceCollection = $resourceCollection;
        $this->iriConverter = $iriConverter;
        $this->apiClassMetadataFactory = $apiClassMetadataFactory;
        $this->contextBuilder = $contextBuilder;
        $this->propertyAccessor = $propertyAccessor;
        $this->traitFlagProvider = $traitFlagProvider;

        $this->setCircularReferenceHandler(function ($object) {
            return $this->iriConverter->getIriFromItem($object);
        });

        if (!is_null($requestStack)) {
            $this->request = $requestStack->getCurrentRequest();
        }
    }

    /**
     * @param AuthorizationProvider $authorizationProvider
     */
    public function setAuthorizationProvider($authorizationProvider)
    {
        $this->authorizationProvider = $authorizationProvider;
    }

    /**
     * @param PermissionResolver $permissionResolver
     */
    public function setPermissionResolver($permissionResolver)
    {
        $this->permissionResolver = $permissionResolver;
    }

    /**
     * @inheritdoc
     */
    public function supportsNormalization($data, $format = null)
    {
        return self::FORMAT === $format && is_object($data);
    }

    /**
     * @inheritdoc
     */
    public function supportsDenormalization($data, $type, $format = null)
    {
        return self::FORMAT === $format;
    }

    /**
     * @inheritdoc
     */
    public function normalize($object, $format = null, array $context = [])
    {
        // Todo these should be passed in only through context and not the Request object
        $showToMany              = $this->getOption('relationships', $context); //Backwards compatibility
        $showToMany              = $this->getOption('showToMany', $context, $showToMany);
        $showToOne               = $this->getOption('showToOne', $context, true);
        $indexOnlyKey            = $this->getOption('indexOnlyKey', $context, null, false);
        $ignoreDepth             = $this->getOption('ignoreDepth', $context);
        $onlyShowScalars         = $this->getOption('onlyShowScalars', $context, false);
        $asJson                  = $this->getOption('asJson', $context, false);
        $export                  = $this->getOption('export', $context, false);
        $removeDefaults          = $this->getOption('removeDefaults', $context, false);
        $blockPassword           = $this->getOption('blockPassword', $context, true);
        $required_attributes     = $this->getOption('required_attributes', $context, [], false);
        $required_m2m_attributes = $this->getOption('required_m2m_attributes', $context, ["recordTags"], false);
        $count_only              = $this->getOption('count_only', $context, false);

        //Prevent from using the request object to set this
        $ignoreAuthorization = false;
        $ignoreRelationshipAuthorization = false;
        if (array_key_exists(self::IGNORE_AUTHORIZATION, $context)) {
            $ignoreAuthorization = $context[self::IGNORE_AUTHORIZATION];
            $ignoreRelationshipAuthorization = true;
        }
        if (array_key_exists('ignore_relationship_authorization', $context)) {
            $ignoreRelationshipAuthorization = $context['ignore_relationship_authorization'];
        }
        if (!array_key_exists('permission', $context)) {
            $context['permission'] = [PermissionsDictionaryBuilder::PERM_CRUD_READ];
        }
        $ignoreFieldAuthorization = array_key_exists(self::IGNORE_FIELD_AUTH, $context) ? $context[self::IGNORE_FIELD_AUTH] : false;

        if (!array_key_exists(self::AS_IRI, $context)) {
            $context[self::AS_IRI] = false;
        }

        if (!array_key_exists(self::DEPTH, $context)) {
            $context[self::DEPTH] = 0;
        }

        $context[self::DEPTH]++;

        if (!$this->serializer instanceof NormalizerInterface) {
            throw new RuntimeException('The serializer must implement the NormalizerInterface.');
        }

        if (is_object($object) && $this->isCircularReference($object, $context)) {
            return $this->handleCircularReference($object);
        }

        // Strict is set to false because we really don't care which class it serializes it to
        // The iri will always be the correct resource
        $resource = $this->guessResource($object);

        $data = [];
        if (!isset($context['json_ld_has_context'])) {
            if (!$export && !$asJson) {
                $data['@context'] = $this->contextBuilder->getResourceContext($resource, $context);
            }
        }

        $context = $this->createContext($resource, $context);

        $classMetadata = $this->getMetadata($resource, $context);

        /** @var ModelAttributeMetadata[] $attributesMetadata */
        $attributesMetadata = $classMetadata->getAttributes();

        if (!$export) { // Removed as $json for performance fixes
            $data['@id'] = $this->iriConverter->getIriFromItem($object);
        }

        if ($context[self::AS_IRI]) {
            return $this->iriConverter->getIriFromItem($object);
        }

        // Removed Condition as its required to be send in both $export as well as $json
        $data['@type'] = ($iri = $classMetadata->getIri()) ? $iri : $resource->getShortName();

        $defaultObject = null;

        if ($removeDefaults) {
            $class = get_class($object);
            $defaultObject = new $class;
        }

        if (!$ignoreAuthorization) {
            $this->authorizationProvider->denyAccessUnlessGranted($resource, $context['permission'], $object);
        }

        foreach ($attributesMetadata as $attributeMetadata) {

            if ($blockPassword && $attributeMetadata->getFormType() == 'password') {
                continue;
            }

            if ($required_attributes && count(array_intersect($required_attributes, [$attributeMetadata->getName()])) == 0) {
                continue;
            }

            if (!$ignoreFieldAuthorization && !$this->authorizationProvider->isGranted($resource, [sprintf('%s.%s', $attributeMetadata->getName(), PermissionsDictionaryBuilder::PERM_FIELD_READ)])) {
                continue;
            }

            $attributeAsIri = $attributeMetadata->isNormalizationLink();
            $attributeName = $attributeMetadata->getName();
            $propertyPath = sprintf('[%s]', $attributeName);

            if (!$attributeMetadata->isReadable()) {
                if (!$indexOnlyKey) {
                    continue;
                }
                $propertyPath = sprintf('[%s]%s', $indexOnlyKey, $propertyPath);
            }

            if (!$attributeMetadata->isWritable() && $export) { // remove $asJson for Performance Improvement
                continue;
            }

            if ($attributeMetadata->isIdentifier() && !$export) {
                continue;
            }

            if ($export || $asJson) { // $asJson is added extra for Performance Improvement
                $attributeAsIri = !$attributeMetadata->isExportable();
            }

            $attributeValue = $this->propertyAccessor->getValue($object, $attributeName);

            if ($attributeMetadata->isHtmlEscape()) {
                $attributeValue = htmlentities(htmlspecialchars_decode($attributeValue));
            }

            if ($removeDefaults) {
                $defaultValue = $this->propertyAccessor->getValue($defaultObject, $propertyPath);
                if ($attributeValue === $defaultValue) {
                    continue;
                }
            }

            if ($this->nameConverter) {
                $attributeName = $this->nameConverter->normalize($attributeName);
            }

            if ($indexOnlyKey && is_string($attributeValue)) {
                $strippedString = strip_tags($attributeValue);
                $strippedString = str_replace('&nbsp;', '', $strippedString);
                $this->propertyAccessor->setValue($data, $propertyPath, $strippedString);
                continue;
            }

            if ($onlyShowScalars) {
                if (is_scalar($attributeValue)) {
                    $this->propertyAccessor->setValue($data, $propertyPath, $attributeValue);
                }
                continue;
            }

            // TODO Turn this in to a function like: evaluateAssociation()
            // TODO return a value or throw exception if value should not be added
            if (isset($attributeMetadata->getTypes()[0])) {
                $type = $attributeMetadata->getTypes()[0];
                $relationshipContext = $context;
                $relationshipContext[self::AS_IRI] = $attributeAsIri;

                $subResource = null;
                $showRelationship = false;

                if ($attributeValue) {
                    if ($type->isCollection()) {
                        // To Many
                        $collectionType = $type->getCollectionType();
                        if ($collectionType) {
                            $subResource = $this->getResourceFromType($collectionType);
                        }
                        $showRelationship = $showToMany;

                        if ($subResource && !$ignoreDepth && $context[self::DEPTH] > 1) {
                            //Todo this should potentially be a count object
                            continue;
                        }
                    } else {
                        // To One
                        $subResource = $this->getResourceFromType($type);
                        $showRelationship = $showToOne;
                    }
                }

                if ($subResource) {
                    if ($showRelationship === false) {
                        if (in_array($attributeMetadata->getFormType(), self::DEFAULT_VISIBLE_FORM_TYPE)) {
                            $relationshipContext[self::AS_IRI] = $asJson;
                        } else if (in_array($attributeMetadata->getFormType(), self::DEFAULT_VISIBLE_EXPORTED_FORM_TYPE)){
                            $relationshipContext[self::AS_IRI] = true;
                        } else
                        {
                            if (count(array_intersect($required_m2m_attributes, [$attributeMetadata->getName()])) == 0) {
                                continue;
                            } else {
                                $relationshipContext[self::AS_IRI] = true;
                            }
                        }
                    }

                    if (is_array($showRelationship)) {
                        if (!array_key_exists($attributeName, $showRelationship)) {
                            continue;
                        }
                        $relationshipContext = array_merge($relationshipContext, $showRelationship[$attributeName]);
                        if ($ignoreRelationshipAuthorization) {
                            $relationshipContext = array_merge($relationshipContext, [self::IGNORE_AUTHORIZATION => true]);
                        }
                    }

                    if (!$ignoreRelationshipAuthorization) {
                        try {
                            $subResourceName = $subResource->getShortName();
                            if (!array_key_exists($subResourceName, $this->authorizedResourceMap)) {
                                $this->authorizationProvider->denyFieldAccessUnlessGranted($resource, $subResourceName, [PermissionsDictionaryBuilder::PERM_CRUD_READ]);
                                $this->authorizedResourceMap[$subResourceName] = true;
                            }
                        } catch (AccessDeniedException $accessDeniedException) {
                            continue;
                        }
                    }

                    try {
                        if ($attributeMetadata->getFormType() == 'tags'){
                            $value = [];
                            foreach($attributeValue as $item) {
                                array_push($value, $item->getUuid());
                            }
                        }else{
                            $value = $this->normalizeRelation($attributeValue, $subResource, $relationshipContext);
                        }
                        if ($count_only && $type->isCollection()) {
                            $value = count($value);
                        }

                        $this->propertyAccessor->setValue(
                            $data,
                            $propertyPath,
                            $value
                        );
                    } catch (AccessDeniedException $accessDeniedException) {
                        continue;
                    }
                    continue;
                }
            }

            $this->propertyAccessor->setValue(
                $data,
                $propertyPath,
                $this->serializer->normalize($attributeValue, self::FORMAT, $context)
            );
        }

        $context[self::DEPTH]--;

        return $data;
    }

    /**
     * @param ClassMetadataInterface $classMetadata
     * @param AttributeMetadataInterface $attributeMetadata
     * @return bool
     */
    protected function isWritable($classMetadata, $attributeMetadata)
    {
        $className = $classMetadata->getName();
        $traits = class_uses($className);
        $attributeName = $attributeMetadata->getName();

        foreach ($traits as $trait) {
            $traitInstance = new \ReflectionClass($trait);
            // TODO - not sure if this has to check private...doing it anyway
            if ($traitInstance->hasProperty($attributeName) && !$traitInstance->getProperty($attributeName)->isPrivate()) {
                return $this->traitFlagProvider->ignoreTrait($trait) || $attributeMetadata->isWritable();
            }
        }

        return $attributeMetadata->isWritable() || $this->isCreate();
    }


    /**
     * @return bool
     */
    protected function isCreate()
    {
        if (array_key_exists('REQUEST_METHOD', $_SERVER)) {
            $method = $_SERVER['REQUEST_METHOD'];
            if ($method === 'POST') {
                return true;
            }
        }
        return false;
    }


    /**
     * @inheritdoc
     */
    public function denormalize($data, $class, $format = null, array $context = [])
    {
        $parseToMany = $this->getOption('parseToMany', $context, true);
        $parseToOne = $this->getOption('parseToOne', $context, true);

        //TODO : This is temporary fix. Please remove this code and fix normalizer issues properly.
        if ($this->request !== null && (strpos($this->request->getRequestUri(), '/avatars/current') !== false || strpos($this->request->getRequestUri(), '/actors/current') !== false)) {
            $context[self::IGNORE_FIELD_AUTH] = true;
        }

        $ignoreFieldAuthorization = array_key_exists(self::IGNORE_FIELD_AUTH, $context) ? $ignoreFieldAuthorization = $context[self::IGNORE_FIELD_AUTH] : false;

        if (!$this->serializer instanceof DenormalizerInterface) {
            throw new RuntimeException('The serializer must implement the DenormalizerInterface to denormalize relations.');
        }

        /** @var ModelAwareResource $resource */
        $resource = $this->guessResource($data, $context, true);
        $normalizedData = $this->prepareForDenormalization($data);

        $context = $this->createContext($resource, $context);

        $classMetadata = $this->getMetadata($resource, $context);
        $attributesMetadata = $classMetadata->getAttributes();

        $allowedAttributes = [];
        foreach ($attributesMetadata as $attributeName => $attributeMetadata) {
            if ($this->isWritable($classMetadata, $attributeMetadata)) {
                // TODO: Check camel case!
                $requiredPermission = sprintf('%s.%s.%s', $resource->getModuleType(), $attributeMetadata->getName(), PermissionsDictionaryBuilder::PERM_FIELD_WRITE);
                if ($ignoreFieldAuthorization || $this->permissionResolver->resolvePermission($requiredPermission)) {
                    $allowedAttributes[] = $attributeName;
                }
            }
        }

        $overrideClass = false;

        if (isset($data['@id']) && !isset($context['object_to_populate'])) {
            $context['object_to_populate'] = $this->iriConverter->getItemFromIri($data['@id']);

            // Avoid issues with proxies if we populated the object
            $overrideClass = true;
        }

        // It's possible to get a proxy object when it's loaded during user entity hydration
        if (isset($context['object_to_populate']) && $context['object_to_populate'] instanceof Proxy) {
            $overrideClass = true;
        }

        $instanceClass = $overrideClass ? get_class($context['object_to_populate']) : $class;
        $reflectionClass = new \ReflectionClass($instanceClass);
        if ($reflectionClass->isAbstract()) {
            throw new InvalidArgumentException(
                sprintf(
                    'Cannot create an instance of %s from serialized data because it is an abstract resource',
                    $instanceClass
                )
            );
        }

        $object = $this->instantiateObject(
            $normalizedData,
            $instanceClass,
            $context,
            $reflectionClass,
            $allowedAttributes
        );

        foreach ($normalizedData as $attributeName => $attributeValue) {
            // Ignore JSON-LD special attributes
            if ('@' === $attributeName[0]) {
                continue;
            }

            if (!$attributeName) {
                throw new \Exception(sprintf('Unexpected value=>%s, please check the corresponding resource type against the value', $attributeValue));
            }

            if ($this->nameConverter) {
                $attributeName = $this->nameConverter->denormalize($attributeName);
            }

            if (!in_array($attributeName, $allowedAttributes) || in_array($attributeName, $this->ignoredAttributes)) {
                continue;
            }

            $types = $attributesMetadata[$attributeName]->getTypes();
            if (isset($types[0])) {
                $type = $types[0];

                // apply showToMany and showToOne flags
                if (!is_null($type->getClass()) && (($type->isCollection() && !$parseToMany) || (!$type->isCollection() && !$parseToOne))) {
                    continue;
                }

                if ($type->isCollection() && ($collectionType = $type->getCollectionType()) && ($class = $collectionType->getClass())) {
                    if (!is_array($attributeValue)) {
                        if (is_string($attributeValue) && strlen($attributeValue) > 0) {
                            $attributeValue = array_map('trim', explode(',', $attributeValue));
                        } else if (is_object($attributeValue)) {
                            $attributeValue = [$attributeValue];
                        } else {
                            continue;
                        }
                    }

                    $values = [];
                    foreach ($attributeValue as $index => $obj) {
                        $values[$index] = $this->denormalizeRelation(
                            $resource,
                            $attributesMetadata[$attributeName],
                            $class,
                            $obj,
                            $context
                        );
                    }

                    $this->setValue($object, $attributeName, $values);

                    continue;
                }

                if ($attributeValue && ($class = $type->getClass())) {
                    $this->setValue(
                        $object,
                        $attributeName,
                        $this->denormalizeRelation(
                            $resource,
                            $attributesMetadata[$attributeName],
                            $class,
                            $attributeValue,
                            $context
                        )
                    );

                    continue;
                }
            }

            $this->setValue($object, $attributeName, $attributeValue);
        }

        return $object;
    }

    /**
     * Denormalizes a relation.
     *
     * @param ResourceInterface          $currentResource
     * @param AttributeMetadataInterface $attributeMetadata
     * @param string                     $class
     * @param mixed                      $value
     * @param array                      $context
     *
     * @return object|null
     *
     * @throws InvalidArgumentException
     * @throws \Exception
     */
    private function denormalizeRelation(
        ResourceInterface $currentResource,
        AttributeMetadataInterface $attributeMetadata,
        $class,
        $value,
        array $context
    ) {
        if (!$this->serializer instanceof DenormalizerInterface) {
            throw new \Exception(sprintf('Expected class DenormalizerInterface, got %s', get_class($this->serializer)));
        }

        if ('DateTime' === $class) {
            return $this->serializer->denormalize($value, $class ?: null, self::FORMAT, $context);
        }

        $attributeName = $attributeMetadata->getName();

        $resource = $this->resourceCollection->getResourceForEntity($class);
        if (!$resource) {
            throw new HttpException(Response::HTTP_BAD_REQUEST, sprintf(
                'Type not supported (found "%s" in attribute "%s" of "%s")',
                $class,
                $attributeName,
                $currentResource->getEntityClass()
            ));
        }

        // Always allow IRI to be compliant with the Hydra spec
        if (is_string($value)) {
            try {
                return $this->iriConverter->getItemFromIri($value);
            } catch (InvalidArgumentException $e) {
                try {
                    $iri = $this->iriConverter->getIriFromResource($resource);
                    $valueIRI = sprintf('%s/%s', $iri, $value);
                    return $this->iriConverter->getItemFromIri($valueIRI);
                } catch (InvalidArgumentException $e) {
                    throw new HttpException(Response::HTTP_BAD_REQUEST, (sprintf(
                        'IRI not supported (found "%s" in "%s" of "%s")',
                        $value,
                        $attributeName,
                        $currentResource->getEntityClass()
                    )), $e);
                }
            }
        }

        if (!$attributeMetadata->isDenormalizationLink()) {
            return $this->serializer->denormalize(
                $value,
                $class,
                self::FORMAT,
                $this->createRelationContext($resource, $context)
            );
        }

        throw new InvalidArgumentException(sprintf(
            'Nested objects for attribute "%s" of "%s" are not enabled. Use serialization groups to change that behavior.',
            $attributeName,
            $currentResource->getEntityClass()
        ));
    }

    /**
     * @param Request $request
     */
    public function setRequest($request)
    {
        $this->request = $request;
    }

    /**
     * @param object $object
     * @param array $context
     * @return bool
     */
    protected function isCircularReference($object, &$context)
    {
        $ignoreDepth = array_key_exists('ignore_depth', $context) ? $context['ignore_depth'] : false;

        return parent::isCircularReference($object, $context) || (!$ignoreDepth && $context[self::DEPTH] > 2);
    }

    /**
     * @param $optionName
     * @param array $context
     * @param bool $default
     * @param bool $validateBoolean
     * @return bool|mixed
     */
    protected function getOption($optionName, array $context, $default = false, $validateBoolean = true)
    {
        $option = $default;

        if (!is_null($this->request)) {
            $key = AttributeLevelOperators::OPERATOR_TOKEN . $optionName;
            $rawOption = $this->request->query->get($key);
            if (is_null($rawOption)) {
                $rawOption = $this->request->request->get($key, $default);
            }
            $option = $rawOption;
            if ($validateBoolean) {
                $option = filter_var($rawOption, FILTER_VALIDATE_BOOLEAN);
            }
        }

        $contextName = Inflector::tableize($optionName);

        $option = array_key_exists($contextName, $context)
            ? $context[$contextName]
            : $option;

        return $option;
    }

    /**
     * Sets a value of the object using the PropertyAccess component.
     *
     * @param object $object
     * @param string $attributeName
     * @param mixed  $value
     */
    private function setValue($object, $attributeName, $value)
    {
        try {
            $this->propertyAccessor->setValue($object, $attributeName, $value);
        } catch (NoSuchPropertyException $exception) {
            // Properties not found are ignored
        } catch (ContextErrorException $exception1) {
            // If user sends empty string instead of null should be ignored.
            // Fix for bug#13173
        }
    }

    /**
     * Normalizes a relation as an URI if is a Link or as a JSON-LD object.
     *
     * @param mixed                      $relatedObject
     * @param ResourceInterface          $resource
     * @param array                      $context
     *
     * @return string|array
     *
     * @throws \Exception
     */
    private function normalizeRelation(
        $relatedObject,
        ResourceInterface $resource,
        array $context
    ) {
        if (!$this->serializer instanceof NormalizerInterface) {
            throw new \Exception(sprintf('Expected class NormalizerInterface, got %s', get_class($this->serializer)));
        }
        unset($context['required_attributes']);
        return $this->serializer->normalize(
            $relatedObject,
            self::FORMAT,
            $this->createRelationContext($resource, $context)
        );
    }

    /**
     * Gets metadata for the given resource with the current context.
     *
     * Fallback to the resource own groups if no context is provided.
     *
     * @param ResourceInterface $resource
     * @param array             $context
     *
     * @return ClassMetadataInterface
     */
    private function getMetadata(ResourceInterface $resource, array $context)
    {
        return $this->apiClassMetadataFactory->getMetadataFor(
            $resource->getEntityClass(),
            isset($context['json_ld_normalization_groups']) ? $context['json_ld_normalization_groups'] : $resource->getNormalizationGroups(),
            isset($context['json_ld_denormalization_groups']) ? $context['json_ld_denormalization_groups'] : $resource->getDenormalizationGroups(),
            isset($context['json_ld_validation_groups']) ? $context['json_ld_validation_groups'] : $resource->getValidationGroups()
        );
    }
}
