<?php
/* Copyright start
  Copyright (C) 2008 - 2020 Fortinet Inc.
  All rights reserved.
  FORTINET CONFIDENTIAL & FORTINET PROPRIETARY SOURCE CODE
  Copyright end */

namespace Cybersponse\CoreBundle\Service;

use Cybersponse\CoreBundle\Provider\CacheableInterface;
use Cybersponse\MetadataBundle\Generation\Metadata\ModelAttributeMetadata;
use Dunglas\ApiBundle\Api\ResourceCollectionInterface;
use Dunglas\ApiBundle\Api\ResourceInterface;
use Dunglas\ApiBundle\Hydra\ApiDocumentationBuilder as DunglasApiDocumentationBuilder;
use Dunglas\ApiBundle\JsonLd\ContextBuilder;
use Dunglas\ApiBundle\Mapping\ClassMetadataFactoryInterface;
use Dunglas\ApiBundle\Mapping\ClassMetadataInterface;
use Symfony\Component\HttpKernel\Exception\NotFoundHttpException;
use Symfony\Component\Routing\RouterInterface;
use Symfony\Component\Serializer\Serializer;

/**
 * @
 *
 * Class ApiDocumentationBuilder
 * @package Cybersponse\CoreBundle\Service
 */
class ApiDocumentationBuilder extends DunglasApiDocumentationBuilder
{
    const API_DOC_REDIS_KEY = 'api:doc';
    const API_DOC_MAP_REDIS_KEY = 'api:doc:map';

    /** @var array */
    protected $documentationMap;

    /** @var CacheableInterface */
    protected $cache;

    /** @var ResourceCollectionInterface */
    protected $newResourceCollection;

    /** @var ClassMetadataFactoryInterface */
    protected $newClassMetadataFactory;

    /** @var Serializer */
    protected $serializer;

    public function __construct(
        ResourceCollectionInterface $resourceCollection,
        ContextBuilder $contextBuilder,
        RouterInterface $router,
        ClassMetadataFactoryInterface $classMetadataFactory,
        $title,
        $description
    ) {
        parent::__construct($resourceCollection, $contextBuilder, $router, $classMetadataFactory, $title, $description);
        $this->newResourceCollection = $resourceCollection;
        $this->newClassMetadataFactory = $classMetadataFactory;
    }

    /**
     * @param $entityName
     * @return null|array
     */
    public function getApiDocumentationForEntity($entityName)
    {
        $documentationMap = $this->getApiDocumentationAsMap();

        if (!array_key_exists($entityName, $documentationMap)) {
            throw new NotFoundHttpException();
        }

        return $documentationMap[$entityName];
    }

    /**
     * @param Serializer $serializer
     */
    public function setSerializer($serializer)
    {
        $this->serializer = $serializer;
    }

    /**
     * @param CacheableInterface $cache
     */
    public function setCache($cache)
    {
        $this->cache = $cache;
    }

    /**
     * Returns the updated documentation based on metadata fields
     *
     * @return array
     */
    private function getUpdatedApiDocumentation()
    {
        $documentation = parent::getApiDocumentation();
        $this->modifyApiDocumentation($documentation, $this->documentationMap);

        return $documentation;
    }

    /**
     * Returns the documentation from redis cache
     *
     * @return array
     */
    public function getApiDocumentation()
    {
        $serializedDocumentation = $this->cache->getFromCache(self::API_DOC_REDIS_KEY);

        if ($serializedDocumentation) {
            $documentation = $serializedDocumentation;
        } else {
            $documentation = $this->getUpdatedApiDocumentation();
            $this->cache->persistToCache(self::API_DOC_REDIS_KEY, $documentation);
        }

        return $documentation;
    }

    /**
     * Returns the documentation entity map from redis cache
     *
     * @return array
     */
    public function getApiDocumentationAsMap()
    {
        $serializedDocumentationMap = $this->cache->getFromCache(self::API_DOC_MAP_REDIS_KEY);

        if ($serializedDocumentationMap) {
            $this->documentationMap = $serializedDocumentationMap;
        } else {
            $this->getApiDocumentation();
            $this->cache->persistToCache(self::API_DOC_MAP_REDIS_KEY, $this->documentationMap);
        }

        return $this->documentationMap;
    }

    /**
     * Modifies the documentation and creates a documentation map
     *
     * @param $documentation
     * @return array
     */
    private function modifyApiDocumentation(&$documentation, &$documentationMap)
    {
        //Create property metadata map
        $resourceMetadataMap = [];

        /** @var ResourceInterface $resource */
        foreach ($this->newResourceCollection as $resource) {
            $key = $resource->getShortName();

            $resourceMetadataMap[$key] = $this->newClassMetadataFactory->getMetadataFor(
                $resource->getEntityClass(),
                $resource->getNormalizationGroups(),
                $resource->getDenormalizationGroups(),
                $resource->getValidationGroups()
            );
        }

        foreach ($documentation['hydra:supportedClass'] as &$supportedClass) {
            $classShortName = ltrim($supportedClass['@id'], '#');

            if (!array_key_exists($classShortName, $resourceMetadataMap)) {
                continue;
            }

            /** @var ClassMetadataInterface $resourceMetadata */
            $resourceMetadata = $resourceMetadataMap[$classShortName];

            $supportedClass['cybersponse:metadata'] = $this->serializer->normalize($resourceMetadata, null, ['groups' => ['metadata']]);

            foreach ($supportedClass['hydra:supportedProperty'] as &$supportedProperty) {

                $property = $supportedProperty['hydra:property'];

                $explodedPropertyName = explode('/', $property['@id']);

                $propertyName = end($explodedPropertyName);

                /** @var ModelAttributeMetadata $attributeMetadata */
                foreach ($resourceMetadata->getAttributes() as $attributeMetadata) {
                    if ($attributeMetadata->getName() == $propertyName) {
                        $supportedProperty['cybersponse:metadata'] = $this->serializer->normalize($attributeMetadata, null, ['groups' => ['metadata']]);
                    }
                }
            }

            $documentationMap[$classShortName] = $supportedClass;
        }
    }
}
