<?php
/* Copyright start
  Copyright (C) 2008 - 2020 Fortinet Inc.
  All rights reserved.
  FORTINET CONFIDENTIAL & FORTINET PROPRIETARY SOURCE CODE
  Copyright end */

namespace Cybersponse\CoreBundle\Service\Config;

use Cybersponse\AuthorizationBundle\Entity\Role;
use Cybersponse\ViewTemplateBundle\Entity\Dashboard;

/**
 * Class DashboardConfig
 * @package Cybersponse\CoreBundle\Service\Config
 */
class DashboardConfig extends PortingConfig
{
    /**
     * 
     * @param array $data List of dashboards
     * @return array
     */
    public function getImportOptions(array $data, $tokenType) {
        $authorizationChecker = $this->container->get('security.authorization_checker');
        $resourceCollection = $this->container->get('api.resource_collection');
        $resource = $resourceCollection->getResourceForEntity(Dashboard::class);
        if (!$authorizationChecker->isGranted(['perm:dashboard.create'], $resource) || !$authorizationChecker->isGranted(['perm:dashboard.update'], $resource) || !$authorizationChecker->isGranted(['perm:dashboard.read'], $resource)) {
            return [];
        }
        $repository = $this->loadRepository(Dashboard::class);
        $roleRepository = $this->loadRepository(Role::class);
        $options = [];
        foreach ($data as $dashboard) {
            $existingUuids = $repository->findOneBy(['uuid' => $dashboard['uuid']]);
            $existingDisplayNames = $repository->findOneBy(['displayName' => $dashboard['displayName']]);
            if (array_key_exists('roleIds', $dashboard)) {
                $roleUuids = $dashboard['roleIds'];
            } else {
                $roleUuids = [];
                foreach ($dashboard['roles'] as $roleIri) {
                    $parts = explode('/', $roleIri);
                    $roleUuid = end($parts);
                    $roleUuids[] = $roleUuid;
                }
            }

            $roles = $roleRepository->findBy(['uuid' => $roleUuids]);
            $foundRoles = [];
            foreach ($roles as $role) {
                $foundRoles[] = [
                    '@id' => '/api/3/roles/' . $role->getUuid(),
                    'name' => $role->getName()
                ];
            }
            
            $individualOptions = [];
            $individualOptions['uuid'] = $existingDisplayNames ? $existingDisplayNames->getUuid() : $dashboard['uuid'];
            $individualOptions['name'] = $dashboard['displayName'];
            $individualOptions['roles'] = $foundRoles;
            $individualOptions['include'] = true;
            $individualOptions['exists'] = $existingUuids || $existingDisplayNames ? true : false;

            $options[] = $individualOptions;
        }
        return $options;
    }

    /**
     * Imports the config data using the specified options
     *
     * @param array $data
     * @param array $options
     * @return mixed
     */
    public function import(array $data, array $options, $tokenType) {
        $dashboardsToAdd = [];
        $dashboardsToUpdate = [];
        foreach ($options as $dashboardOption) {
            if (!$dashboardOption['include']) {
                // Do not merge skipped items
                continue;
            }
            foreach ($data as $dashboard) {
                if ($dashboardOption['uuid'] == $dashboard['uuid']) {
                    unset($dashboard['userId']);
                    $dashboard['roles'] = array_column($dashboardOption['roles'], '@id');
                    $dashboard['filterArray'] = isset($dashboard['config']['filterArray']) ? $dashboard['config']['filterArray'] : $dashboard['filterArray'];
                    unset($dashboard['roleIds']);
                    if (!$dashboardOption['exists']) {
                        $dashboardsToAdd[] = $dashboard;
                        break;
                    }
                    $dashboard['@id'] = sprintf('/api/3/dashboard/%s', $dashboard['uuid']);
                    unset($dashboard['uuid']);
                    $dashboardsToUpdate[] = $dashboard;
                    break;
                }
            }
        }
        $results = [];
        foreach ($dashboardsToAdd as $dashboard) {
            $fullURI = sprintf('%s/api/3/dashboard', $this->container->getParameter('crudhub.uri'));
            $result = $this->sendRequest($fullURI, 'post', $dashboard, $tokenType);
            $results[] = $result;
        }
        if (count($dashboardsToUpdate) > 0) {
            $fullURI = sprintf('%s/api/3/update/dashboard', $this->container->getParameter('crudhub.uri'));
            $body = [
              "data" => $dashboardsToUpdate
            ];
            $result = $this->sendRequest($fullURI, 'put', $body, $tokenType, $tokenType);
            $results[] = $result;
        }
        
        if (in_array(false, $results)) {
            return false;
        }
        return true;
    }

    /**
     * Exports the config using the specified options
     *
     * @param array $options
     * @return array
     */
    public function export(array $options, $tokenType='JWT') {
        $authorizationChecker = $this->container->get('security.authorization_checker');
        $resourceCollection = $this->container->get('api.resource_collection');
        $resource = $resourceCollection->getResourceForEntity(Dashboard::class);
        if (!$authorizationChecker->isGranted(['perm:dashboard.read'], $resource)) {
            return [];
        }
        return $this->getResourceFromIRI($options, Dashboard::class, 'resource.dashboard', true);
    }
}
