<?php
/* Copyright start
  Copyright (C) 2008 - 2020 Fortinet Inc.
  All rights reserved.
  FORTINET CONFIDENTIAL & FORTINET PROPRIETARY SOURCE CODE
  Copyright end */

namespace Cybersponse\CoreBundle\Service\Config;

use Cybersponse\MetadataBundle\Command\SystemUpdate;
use Cybersponse\MetadataBundle\Entity\StagingModelMetadata;
use Symfony\Bridge\Doctrine\ManagerRegistry;
use Symfony\Component\Config\Resource\ResourceInterface;

/**
 * Class ModuleConfig
 * @package Cybersponse\CoreBundle\Service\Config
 */
class ModuleConfig extends PortingConfig
{

    /**
     * 
     * @param array $data List of modules
     * @return array
     */
    public function getImportOptions(array $data, $tokenType)
    {
        $moduleNames = array_column($data, 'type');
        $existingModules = $this->requestMmds($moduleNames, false);
        $repository = $this->loadRepository(StagingModelMetadata::class);
        $options = [];
        foreach ($data as $module) {
            /**
             * $module is like 
             * {
             *   "@type": "StagingModelMetadata",
             *   "attributes": [...],
             *   "type": "alerts",
             *   "tableName": "alerts",
             *   ...
             * }
             */

            $existingModule = self::findBy($existingModules, 'type', $module['type'], true);
            if (!$existingModule) {
                $existingModule = self::findBy($existingModules, 'tableName', $module['tableName'], true);
            }
            $moduleOptions = [];
            $moduleOptions['include'] = true;
            $systemModuleTables = ['widgets', 'workflows', 'rules', 'reporting', 'file'];
            $moduleOptions['isValid'] = $module['type'] && $module['tableName'] && $module['descriptions'] && $module['descriptions']['singular'] && $module['descriptions']['plural'] && $module['displayName'] && !in_array($module['tableName'], $systemModuleTables);
            $moduleOptions['type'] = $module['type'];
            $moduleOptions['name'] = $moduleOptions['isValid'] ? $module['descriptions']['plural'] : $module['type'];
            $moduleOptions['include'] = $moduleOptions['isValid'];
            $moduleOptions['exists'] = $existingModule ? true : false;
            $uniqueConstraintAllowed = $this->isUniqueConstraintAllowed($module, $existingModule);
            if ($uniqueConstraintAllowed && $module['uniqueConstraint'] && count($module['uniqueConstraint']) > 0) {
                $uniqueColumns = array_map('strtolower', $module['uniqueConstraint'][0][$module['type'] . '_unique']['columns']);
            } else if ($existingModule && $existingModule['uniqueConstraint'] && count($existingModule['uniqueConstraint']) > 0) {
                $uniqueColumns = array_map('strtolower', $existingModule['uniqueConstraint'][0][$existingModule['type'] . '_unique']['columns']);
            }
            $moduleOptions['attributes'] = [];
            $newAttributeNames = [];
            foreach ($module['attributes'] as $attribute) {
                $existingAttribute = $existingModule ? self::findBy($existingModule['attributes'], 'name', $attribute['name']) : null;
                $attributeOptions = [];
                $attributeOptions['name'] = $attribute['name'];
                $attributeOptions['inUniqueConstraint'] = in_array(strtolower($attribute['name']), $uniqueColumns);
                $attributeOptions['title'] = $attribute['descriptions']['singular'];
                $attributeOptions['type'] = $attribute['formType'];
                $attributeOptions['exists'] = $existingAttribute ? true : false;
                if ($existingAttribute && $existingAttribute['system']) {
                    $attributeOptions['include'] = false;
                }
                if ($existingAttribute) {
                    $attributeOptions['@id'] = $existingAttribute['@id'];
                    $attributeOptions['system'] = $existingAttribute['system'];
                    if (!$existingAttribute['system']) {
                        list($changes, $conflicts) = $this->compareModuleAttributes($attribute, $existingAttribute);
                        if (count($changes) > 0) {
                            $attributeOptions['changes'] = $changes;
                        }
                        if (count($conflicts) > 0) {
                            $attributeOptions['conflicts'] = $conflicts;
                            $moduleOptions['has_conflicts'] = true;
                        }
                    }
                }
                // If field is manyToMany or manyToOne and points to a non-existent module, skip it
                if (!$attributeOptions['conflicts'] && ($attribute['formType'] == 'manyToMany' || $attribute['formType'] == 'manyToOne' || $attribute['formType'] == 'oneToMany')) {
                    $foundModule = self::findBy($data, 'type', $attribute['dataSource']['model'], true);
                    if (!$foundModule) {
                        // Doesn't exist in imported modules - check existing modules
                        $foundModule = $repository->findOneBy(['type' => $attribute['dataSource']['model']]);
                    }
                    if (!$foundModule) {
                        $attributeOptions['conflicts'] = [
                            [
                                'field' => 'type',
                                'existingValue' => null,
                                'newValue' => $attribute['dataSource']['model'],
                                'useExisting' => true,
                                'message' => 'Correlated module does not exist'
                            ]
                        ];
                        $attributeOptions['include'] = false;
                    }
                }
                $newAttributeNames[] = strtolower($attribute['name']);
                $moduleOptions['attributes'][] = $attributeOptions;
            }

            if ($existingModule) {
                $changes = $this->compareModuleMetadata($module, $existingModule, $uniqueConstraintAllowed);
                $moduleOptions['changes'] = $changes;
                $moduleOptions['@id'] = $existingModule['@id'];
                // Check for attributes that may be removed if replaced
                foreach ($existingModule['attributes'] as $existingAttribute) {
                    if (array_search(strtolower($existingAttribute['name']), $newAttributeNames) === false) {
                        $foundAttribute = null;
                        foreach ($module['attributes'] as $attribute) {
                            $bothManyToMany = $existingAttribute['formType'] == 'manyToMany' && $attribute['formType'] == 'manyToMany';
                            $pointingToSameModel = isset($existingAttribute['dataSource']) && isset($attribute['dataSource']) && $existingAttribute['dataSource']['model'] == $attribute['dataSource']['model'];
                            if ($bothManyToMany && $pointingToSameModel) {
                                $foundAttribute = $attribute;
                            }
                        }
                        if ($foundAttribute) {
                            // If another field is a many to many pointing to the same model, we consider them the same
                            continue;
                        }
                        $attributeOptions = [];
                        $attributeOptions['name'] = $existingAttribute['name'];
                        $attributeOptions['@id'] = $existingAttribute['@id'];
                        $attributeOptions['include'] = true;
                        $attributeOptions['title'] = $existingAttribute['descriptions']['singular'];
                        $attributeOptions['type'] = $existingAttribute['formType'];
                        $attributeOptions['skipped_in_import'] = true;
                        $attributeOptions['inUniqueConstraint'] = in_array(strtolower($existingAttribute['name']), $uniqueColumns);
                        $moduleOptions['attributes'][] = $attributeOptions;
                    }
                }
            }

            $options[] = $moduleOptions;
        }
        return $options;
    }

    private function isUniqueConstraintAllowed(array $module, $existingModule) {
        $uniquenessAllowed = true;
        try {
            $columns = [];
            $oldColumn = [];

            $uniqueness = array_key_exists('uniqueConstraint', $module) && is_array($module['uniqueConstraint']) && count($module['uniqueConstraint']) > 0;
            $uniquenessOld = $existingModule && array_key_exists('uniqueConstraint', $existingModule) && is_array($existingModule['uniqueConstraint']) && count($existingModule['uniqueConstraint']) > 0;

            if ($uniquenessOld) {
                $oldColumn = $existingModule['uniqueConstraint'][0][$existingModule['type'] . '_unique']['columns'];
            }

            if ($uniqueness) {
                $columns = $module['uniqueConstraint'][0][$module['type'] . '_unique']['columns'];
                if ($oldColumn && $oldColumn === $columns) {
                    return true;
                }

                if ($existingModule) {
                    $className = $this->getClassNameFromSingularDescription($module['descriptions']['singular']);
                    $entityManager = $this->container->get('doctrine')->getManagerForClass($className);
                    $queryBuilder = $entityManager->getRepository($className)->createQueryBuilder('a');
                    $queryBuilder->select('count(a) as the_count');
                    foreach ($columns as $column) {
                        $attribute = self::findBy($module['attributes'], 'name', $column, true);
                        if (!$attribute) {
                            // New attribute - ignore
                            continue;
                        }
                        if ($attribute['dataSource'] && $attribute['dataSource']['model']) {
                            $selectText = sprintf('IDENTITY(a.%s)', $column);
                        } else {
                            $selectText = sprintf('a.%s', $column);
                        }
                        $queryBuilder->addSelect($selectText);
                        $queryBuilder->andWhere($selectText . ' IS NOT NULL');
                        $queryBuilder->addGroupBy(sprintf('a.%s', $column));
                    }
                    $query = $queryBuilder
                        ->having('count(a) > 1')
                        ->getQuery();
                    $result = $query->getResult();
                    $uniquenessAllowed = count($result) == 0; // todo: determine if there are duplicates preventing change to $columns
                }
            }
        } catch (\Exception $e) {
            error_log("Exception occurred while checking for unique constraint " . $e);
        }
        return $uniquenessAllowed;
    }

    /**
     * Determines changes and conflicts between attributes
     *
     * @param array $attribute
     * @param array $existingAttribute
     * @return array
     */
    private function compareModuleMetadata(array $module, array $existingModule, $uniquenessAllowed)
    {
        $changes = [];
        unset($module['attributes']);
        unset($existingModule['attributes']);
        $diffs = array_diff($module, $existingModule);
        $diffs['descriptions'] = array_diff($module['descriptions'], $existingModule['descriptions']);

        $existingModule['defaultSort'] = $existingModule['defaultSort'] ? $existingModule['defaultSort'] : [];
        $module['defaultSort'] = $module['defaultSort'] ? $module['defaultSort'] : [];
        if (array_diff($module['defaultSort'], $existingModule['defaultSort']) || array_diff($existingModule['defaultSort'], $module['defaultSort'])) {
            $defaultSorts = array_column($module['defaultSort'], 'field');
            $existingDefaultSorts = array_column($existingModule['defaultSort'], 'field');
            $changes[] = [
                'field' => 'defaultSort',
                'message' => 'Changing default sort from (' . implode(', ', $existingDefaultSorts) . ') to (' . implode(', ', $defaultSorts) . ')'
            ];
        }

        $uniqueness = array_key_exists('uniqueConstraint', $module) && is_array($module['uniqueConstraint']) && count($module['uniqueConstraint']) > 0;
        $existingUniqueness = array_key_exists('uniqueConstraint', $existingModule) && is_array($existingModule['uniqueConstraint']) && count($existingModule['uniqueConstraint']) > 0;
        $columns = [];
        if ($uniqueness) {
            $columns = $module['uniqueConstraint'][0][$module['type'] . '_unique']['columns'];
        }
        $existingColumns = [];
        if ($existingUniqueness) {
            $existingColumns = $existingModule['uniqueConstraint'][0][$existingModule['type'] . '_unique']['columns'];
        }
        if (count(array_diff($columns, $existingColumns)) > 0 || count(array_diff($existingColumns, $columns)) > 0) {
            if (!$uniquenessAllowed) {
                $changes[] = [
                    'field' => 'uniqueConstraint',
                    'message' => 'Preventing change to unique constraints on fields (' . implode(', ', $columns) . ') because existing data is not unique',
                    'error' => true
                ];
            } else if ($uniqueness && $existingUniqueness) {
                $changes[] = [
                    'field' => 'uniqueConstraint',
                    'message' => 'Changing unique constraints from fields (' . implode(', ', $existingColumns) . ') to fields (' . implode(', ', $columns) . ')'
                ];
            } else if ($uniqueness) {
                $changes[] = [
                    'field' => 'uniqueConstraint',
                    'message' => 'Adding unique constraints on fields (' . implode(', ', $columns) . ')'
                ];
            } else if ($existingUniqueness) {
                $changes[] = [
                    'field' => 'uniqueConstraint',
                    'message' => 'Removing unique constraints from fields (' . implode(', ', $existingColumns) . ')'
                ];
            }
        }
        
        unset($diffs['defaultSort']);
        unset($diffs['uniqueConstraint']);
        unset($diffs['@id']);
        unset($diffs['@type']);
        unset($diffs['sattrib']);
        unset($diffs['orderIndex']);
        unset($diffs['uuid']);
        foreach ($diffs as $key => $value) {
            $existingValue = array_key_exists($key, $existingModule) ? $existingModule[$key] : null;
            if (is_array($value)) {
                foreach ($value as $subKey => $subValue) {
                    $changes[] = [
                        'field' => $key . ' -> ' . $subKey,
                        'message' => ucfirst($key) . ' -> ' . $subKey . ' changed from ' . $existingValue[$subKey] . ' to ' . $subValue
                    ];
                }
                continue;
            }
            $changes[] = [
                'field' => $key,
                'message' => $key . ' changed from ' . $existingValue . ' to ' . $module[$key]
            ];
        }
        return $changes;
    }

    /**
     * Determines changes and conflicts between attributes
     *
     * @param array $attribute
     * @param array $existingAttribute
     * @return array
     */
    private function compareModuleAttributes(array $attribute, array $existingAttribute)
    {
        $conflicts = [];
        $changes = [];
        $newAttributeDiffs = array_diff($attribute, $existingAttribute);
        unset($newAttributeDiffs['@id']);
        unset($newAttributeDiffs['@type']);
        unset($newAttributeDiffs['sattrib']);
        unset($newAttributeDiffs['orderIndex']);
        foreach ($newAttributeDiffs as $key => $value) {
            $existingValue = array_key_exists($key, $existingAttribute) ? $existingAttribute[$key] : null;
            $changes[$attribute['name']][] = [
                'field' => $key,
                'old' => $existingValue,
                'new' => $value
            ];
        }
        $supportedTypeConversions = [
            ['integer', 'string'],
            ['float', 'string'],
            ['manyToMany', 'manyToOne']
        ];
        // If the type conversion is not valid, issue a conflict
        $isDateTimeConversion = $existingAttribute['formType'] != $attribute['formType'] && $existingAttribute['formType'] == 'datetime';
        if ($isDateTimeConversion || $existingAttribute['type'] != $attribute['type'] && !in_array([$existingAttribute['type'], $attribute['type']], $supportedTypeConversions)) {
            $conflicts[] = [
                'field' => 'type',
                'existingValue' => $existingAttribute['type'],
                'newValue' => $attribute['type'],
                'useExisting' => true,
                'message' => 'Data type conversion from "' . $existingAttribute['type'] . '"('  . $existingAttribute['formType'] . ') to "' . $attribute['type'] . '" (' . $attribute['formType'] . ') not supported'
            ];
        }
        return [$changes, $conflicts];
    }

    /**
     * Imports the config data using the specified options
     *
     * @param array $data
     * @param array $options
     * @return mixed
     */
    public function import(array $data, array $options, $tokenType)
    {
        $modulesToAdd = [];
        $modulesToUpdate = [];
        $moduleTypes = array_column($options, 'type');
        $existingModules = $this->requestMmds($moduleTypes, false);
        foreach ($options as $moduleOption) {
            if ((array_key_exists('isValid', $moduleOption) && !$moduleOption['isValid']) || (array_key_exists('include', $moduleOption) && !$moduleOption['include'])) {
                // Do not merge skipped modules
                continue;
            }
            $module = null;
            foreach ($data as $item) {
                if ($item['type'] == $moduleOption['type']) {
                    $module = $item;
                    break;
                }
            }
            if (!$moduleOption['exists']) {
                // If it does not exist, just put it on the list to import
                unset($module['uuid']);
                $existingModule = null;
            } else {
                $existingModule = self::findBy($existingModules, 'type', $moduleOption['type'], true);
                $module['@id'] = $moduleOption['@id'];

                // We don't allow changing these on import
                $module['peerReplicable'] = $existingModule['peerReplicable'];
                $module['trackable'] = $existingModule['trackable'];
                $module['ownable'] = $existingModule['ownable'];
                if (!$this->isUniqueConstraintAllowed($module, $existingModule)) {
                    $module['uniqueConstraint'] = $existingModule['uniqueConstraint'] ? $existingModule['uniqueConstraint'] : [];
                }
            }

            $moduleAttributes = [];

            $attributeList = [];
            foreach ($moduleOption['attributes'] as $attributeOption) {
                if (!$attributeOption['exists'] && array_key_exists('include', $attributeOption) && !$attributeOption['include']) {
                    // Doesn't exist and user chose not to include it
                    continue;
                }
                $attribute = self::findBy($module['attributes'], 'name', $attributeOption['name']);
                if ($existingModule && (!$attribute || (array_key_exists('include', $attributeOption) && !$attributeOption['include']))) {
                    // !$attribute means the field is in existing mmd but not in imported one
                    // !$attributeOption['include'] means the imported one 
                    $attribute = self::findBy($existingModule['attributes'], 'name', $attributeOption['name']);
                }
                $conflicts = array_key_exists('conflicts', $attributeOption) ? $attributeOption['conflicts'] : [];
                foreach ($conflicts as $conflict) {
                    if (!$conflict['useExisting']) {
                        continue;
                    }
                    $attribute[$conflict['field']] = $conflict['existingValue'];
                }
                if ($moduleOption['exists']) {
                    $attribute['@id'] = $attributeOption['@id'];
                }
                unset($attribute['sattrib']);
                unset($attribute['uuid']);
                $moduleAttributes[] = $attribute;

                if (in_array(strtolower($attributeOption['name']), $attributeList)) {
                    throw new \Exception(sprintf("Duplicate field detected on module %s with attribute name %s", $moduleOption['type'], $attributeOption['name']));
                } else {
                    array_push($attributeList, \strtolower($attributeOption['name']));
                }
            }
            $module['attributes'] = $moduleAttributes;
            if ($moduleOption['exists']) {
                if (isset($module['uuid'])) {
                    unset($module['uuid']);
                }
                $modulesToUpdate[] = $module;
            } else {
                $modulesToAdd[] = $module;
            }
        }
        // Once the lists of modules to add and update are there, we need to fix invalid references
        foreach ($modulesToAdd as $index => $module) {
            $modulesToAdd[$index] = $this->removeInvalidReferences($module, array_merge($modulesToAdd, $modulesToUpdate));
        }
        foreach ($modulesToUpdate as $index => $module) {
            $modulesToUpdate[$index] = $this->removeInvalidReferences($module, array_merge($modulesToAdd, $modulesToUpdate));
        }

        $results = [];
        // Import modules without attributes
        if (count($modulesToUpdate) > 0) {
            $fullURI = sprintf('%s/api/3/update/staging_model_metadatas?$export=true&$relationships=true', $this->container->getParameter('crudhub.uri'));
            $body = [
                'data' => $modulesToUpdate
            ];
            $results[] = !!$this->sendRequest($fullURI, 'put', $body, $tokenType);
        }
        if (count($modulesToAdd) > 0) {
            $fullURI = sprintf('%s/api/3/insert/staging_model_metadatas?$export=true&$relationships=true', $this->container->getParameter('crudhub.uri'));
            $body = [
                'data' => $modulesToAdd
            ];
            $results[] = !!$this->sendRequest($fullURI, 'post', $body, $tokenType);
        }
        if (count($modulesToUpdate) > 0 || count($modulesToAdd) > 0) {
            [$publishSuccess, $errors] = $this->publish();
            if (!$publishSuccess) {
                throw new \Exception(sprintf('Publish failed with exception: %s', implode(', ', $errors)));
            }
        }
        
        if (in_array(false, $results)) {
            return false;
        }
        return true;
    }

    /**
     * Finds the value in the array by key
     *
     * @param array $array
     * @param string $key
     * @param string $value
     * @return mixed
     */
    private static function findBy(array $array, string $key, string $value, $caseInsensitive = false)
    {
        foreach ($array as $item) {
            if (!$caseInsensitive && $item[$key] == $value) {
                return $item;
            } else if ($caseInsensitive && strtolower($item[$key]) == strtolower($value)) {
                return $item;
            }
        }
        return null;
    }

    /**
     * Parses the module and removes invalid relationships, inversed fields, and attributes
     *
     * @param array $module
     * @param array $data
     * @return void
     */
    private function removeInvalidReferences(array $module, array $modulesBeingImported)
    {
        $moduleAttributesToKeep = [];
        foreach ($module['attributes'] as $attribute) {
            if ($attribute['dataSource'] && array_key_exists('model', $attribute['dataSource']) && ($attribute['formType'] == 'manyToMany' || $attribute['formType'] == 'oneToMany' || $attribute['formType'] == 'lookup')) {
                // Check inversedField
                // Get related module
                $relatedModule = self::findBy($modulesBeingImported, 'type', $attribute['dataSource']['model'], true);
                if (!$relatedModule) {
                    // Check existing modules
                    $relatedModule = end($this->requestMmds([$attribute['dataSource']['model']], false));
                }
                if (!$relatedModule) {
                    continue;
                }
                if (array_key_exists('inversedField', $attribute) && $attribute['inversedField']) {
                    $inversedAttribute = self::findBy($relatedModule['attributes'], 'name', $attribute['inversedField']);
                    // If it doesn't find the inversed attribute, clear it
                    if (!$inversedAttribute) {
                        $attribute['inversedField'] = null;
                        // Check backwards inversed field
                    } else if (array_key_exists('inversedField', $inversedAttribute) && $inversedAttribute['inversedField'] && !self::findBy($module['attributes'], 'name', $inversedAttribute['inversedField'])) {
                        // If there is no backwards referencing field, remove this inversed field
                        $attribute['inversedField'] = null;
                    }
                } else if ($attribute['dataSource']['model'] == $module['type']) {
                    // self referencing relationship - add an inversedField to help users out
                    $attribute['inversedField'] = $attribute['name'];
                }
            }
            $moduleAttributesToKeep[] = $attribute;
        }
        $module['attributes'] = $moduleAttributesToKeep;
        return $module;
    }

    /**
     * Exports the config using the specified options
     *
     * @param array $options list of MMDs to export
     * @return array
     */
    public function export(array $options, $tokenType='JWT')
    {
        return $this->requestMmds($options);
    }

    /**
     * Requests MMDs with the option to use $export tag or not
     *
     * @param array $modules
     * @param boolean $export
     * @return array
     */
    private function requestMmds(array $modules, $export = true)
    {
        /** @var ManagerRegistry $managerRegistry */
        $managerRegistry = $this->container->get('doctrine');
        $smmdManager = $managerRegistry->getManagerForClass(StagingModelMetadata::class);
        $smmdRepository = $smmdManager->getRepository(StagingModelMetadata::class);
        /** @var ResourceInterface $picklistResource */
        $smmdResource = $this->container->get('resource.stagingmodelmetadata');
        $context = $smmdResource->getDenormalizationContext();
        $context['export'] = $export;
        $context['relationships'] = true;
        $context["ignore_field_authorization"] = true;
        $context["ignore_authorization"] = true;
        $exportModules = [];
        foreach ($modules as $module) {
            $stagingMmd = $smmdRepository->findOneBy(['type' => $module]);
            $normalizedObject = $this->container->get('serializer')->normalize($stagingMmd, 'json-ld', $context);
            $exportModules[] = $normalizedObject;
        }
        return $exportModules;
    }

    /**
     * 
     *
     * @param array $config
     * @return array
     */
    public function publish() {
        $rootDir = $this->container->getParameter('kernel.root_dir');
        $file503 = sprintf('%s/../web/503.json', $rootDir);

        $systemUpdate = new SystemUpdate();
        $systemUpdate->setMessage('Publish queued.');
        $systemUpdate->setProgressPercent(0);
        $filesystem = $this->container->get('filesystem');
        $filesystem->dumpFile($file503, json_encode($systemUpdate));

        $command = sprintf('/usr/bin/php app/console cybersponse:system:update -vvv --env=%s -la', $this->container->getParameter('kernel.environment'));

        chdir($this->container->getParameter('kernel.root_dir') . '/../');
        $logFile = '/var/log/cyops/cyops-api/last_system_publish.log';
        exec(sprintf('bash -c "%s > %s 2>&1"', $command, $logFile));

        $searchfor = array("An exception occurred", "ERROR:", "ORMException");
        $errors = array();
        $contents = file_get_contents($logFile);

        foreach ($searchfor as $searchString) {
            // escape special characters in the query
            $pattern = preg_quote($searchString, '/');
            // finalise the regular expression, matching the whole line
            $pattern = "/^.*$pattern.*\$/m";
            // search, and store all matching occurences in $matches
            if (preg_match_all($pattern, $contents, $matches)) {
                $errors = array_merge($errors, array_unique($matches[0]));
            }
        }
        if (count($errors) == 0) {
            return [true, null];
        } else {
            return [false, $errors];
        }
    }

    /**
     * @param $resourceShortName
     * @return string $name;
     */
    protected function getClassNameFromSingularDescription($singularDescription)
    {
        $escapedSingular = preg_replace('/[^a-z0-9\s]+/i', '', $singularDescription);
        $resourceClassName = \Doctrine\Common\Util\Inflector::classify($escapedSingular);
        /** @var ResourceCollectionInterface $resourceCollection */
        $resourceCollection = $this->container->get('api.resource_collection');
        $resource = $resourceCollection->getResourceForShortName($resourceClassName);
        return $resource->getEntityClass();
    }

}
