<?php
/* Copyright start
  Copyright (C) 2008 - 2020 Fortinet Inc.
  All rights reserved.
  FORTINET CONFIDENTIAL & FORTINET PROPRIETARY SOURCE CODE
  Copyright end */

namespace Cybersponse\CoreBundle\Service\Config;

use Cybersponse\AuthorizationBundle\Entity\Tag;
use Cybersponse\WorkflowBundle\Entity\Workflow;
use Cybersponse\WorkflowBundle\Entity\WorkflowCollection;
use Exception;
use GuzzleHttp\Psr7\Request as GuzzleRequest;

/**
 * Class PlaybookConfig
 * @package Cybersponse\CoreBundle\Service\Config
 */
class PlaybookConfig extends PortingConfig
{
    private const STEP_TYPES = [
        'MANUAL_INPUT' => 'fc04082a-d7dc-4299-96fb-6837b1baa0fe',
        'DECISION' => '12254cf5-5db7-4b1a-8cb1-3af081924b28'
    ];

    private $requiredMacros = [];

    /**
     *
     * @param array $data Playbook object containing collections & global vars
     * @return array
     */
    public function getImportOptions(array $data, $tokenType)
    {
        $authorizationChecker = $this->container->get('security.authorization_checker');
        $resourceCollection = $this->container->get('api.resource_collection');
        $resource = $resourceCollection->getResourceForEntity(Workflow::class);
        if (!$authorizationChecker->isGranted(['perm:workflows.update'], $resource) || !$authorizationChecker->isGranted(['perm:workflows.create'], $resource) || !$authorizationChecker->isGranted(['perm:workflows.read'], $resource)) {
            return [];
        }
        $options = [];
        if (array_key_exists('collections', $data) && count($data['collections']) > 0) {
            $subOptions = $this->generatePlaybookCollectionOptions($data['collections']);
            $options['collections'] = [
                "include" => count($subOptions) > 0,
                "values" => $subOptions
            ];
        }
        if (array_key_exists('globalVariables', $data) && count($data['globalVariables']) > 0) {
            $subOptions = $this->generateGlobalVariableOptions($data['globalVariables'], $tokenType);
            $options['globalVariables'] = [
                "include" => count($subOptions) > 0,
                "values" => $subOptions
            ];
        }
        return $options;
    }

    /**
     * @param array $config
     * @return array
     */
    private function generateGlobalVariableOptions(array $dynamicVariables, $tokenType)
    {
        $existingDynamicVariables = $this->getGlobalVariables($tokenType);
        $existingDynamicVariableNames = array_column($existingDynamicVariables, 'name');
        $options = [];
        foreach ($dynamicVariables as $dynamicVariable) {
            $existing = in_array($dynamicVariable['name'], $existingDynamicVariableNames);
            $individualOptions = [];
            $individualOptions['name'] = $dynamicVariable['name'];
            $individualOptions['include'] = !$existing;
            $individualOptions['required'] = !$existing && in_array($dynamicVariable['name'], $this->requiredMacros);
            $individualOptions['exists'] = $existing;

            $options[] = $individualOptions;
        }
        return $options;
    }

    private function getGlobalVariables($tokenType='JWT')
    {
        $tokenStorage = $this->container->get('security.token_storage');
        $token = $tokenStorage->getToken();
        $workflowMacroEndpoint = '/api/dynamic-variable/?offset=0&limit=2147483647';
        if ($token) {
            $uri = $this->container->getParameter('crudhub.uri') . '/api/wf' . $workflowMacroEndpoint;
            $response = $this->sendRequest($uri, 'get', null, $tokenType);
            $jsonDecodeResponse = json_decode($response, true);
            return array_key_exists('hydra:member', $jsonDecodeResponse) ? $jsonDecodeResponse['hydra:member'] : $jsonDecodeResponse;
        } else {
            $uri = $this->container->getParameter('workflow.uri') . $workflowMacroEndpoint;
            return $this->sealabRequest($uri, 'get');
        }
    }
    
    private function sealabRequest($uri, $method, $data = null)
    {
        $headers = array(
            "Content-type" => "application/json",
        );
        $workflowRequest = new GuzzleRequest($method, $uri, $headers, $data ? json_encode($data) : null);
        $proxyProvider = $this->container->get('cybersponse.core.provider.proxy');
        $client = $proxyProvider->getProxyClient();
        try {
            $response = $client->send($workflowRequest, ['verify' => false]);
            $statusCode = $response->getStatusCode();
            if ($statusCode < 400) {
                $jsonDecodeResponse = json_decode($response->getBody()->getContents(), true);
                return array_key_exists('hydra:member', $jsonDecodeResponse) ? $jsonDecodeResponse['hydra:member'] : $jsonDecodeResponse;
            } else {
                throw new Exception($uri.'\n'.$response->getBody());
            }
        } catch (\Exception $exception) {
            $errVar = json_decode($exception->getResponse()->getBody()->read(2048), true);
            $errMsg['message'] = $errVar['Error'] ? isset($errVar['Error']) : $errVar['error'];
            $errCode = $exception->getCode();
            throw (sprintf('%d: %s', $errCode, $errMsg));
        }
    }
    
    /**
     * @param array $config
     * @return array
     */
    private function generatePlaybookCollectionOptions(array $playbookCollections)
    {
        $repository = $this->loadRepository(WorkflowCollection::class);
        $options = [];
        foreach ($playbookCollections as $playbookCollection) {
            $existingByUuid = $repository->findOneBy(['uuid' => $playbookCollection['uuid']]);
            $existingByName = $repository->findOneBy(['name' => $playbookCollection['name']]);
            $individualOptions = [];
            $individualOptions['uuid'] = $playbookCollection['uuid'];
            $individualOptions['name'] = $playbookCollection['name'];
            $individualOptions['include'] = true;
            $individualOptions['rename'] = false;
            $individualOptions['exists_uuid'] = $existingByUuid ? true : false;
            $individualOptions['existing_uuid'] = $existingByName ? $existingByName->getUuid() : $playbookCollection['uuid'];
            $individualOptions['exists_name'] = $existingByName ? true : false;
            $individualOptions['exists'] = $individualOptions['exists_uuid'] || $individualOptions['exists_name'];
            $individualOptions['macros'] = $this->getMacrosFromPlaybooks($playbookCollection['workflows']);
            $this->requiredMacros = array_unique(array_merge($this->requiredMacros, $individualOptions['macros']));

            $options[] = $individualOptions;
        }
        return $options;
    }

    private function getMacrosFromPlaybooks($playbooks)
    {
        $macros = [];
        foreach ($playbooks as $playbook) {
            if (array_key_exists('steps', $playbook)) {
                foreach ($playbook['steps'] as $step) {
                    if (array_key_exists('arguments', $step)) {
                        $argsJson = json_encode($step['arguments']);
                        $regex = '/globalVars\.(\w+)/m';
                        $matches = [];
                        preg_match_all($regex, $argsJson, $matches, PREG_PATTERN_ORDER, 0);
                        $macros = array_merge($macros, $matches[1]);
                    }
                }
            }
            if ($playbook['versions'] && count($playbook['versions'])) {
                foreach($playbook['versions'] as $version) {
                    $version = json_decode($version['json'], true);
                    if (array_key_exists('steps', $version)) {
                        foreach ($version['steps'] as $step) {
                            if (array_key_exists('arguments', $step)) {
                                $argsJson = json_encode($step['arguments']);
                                $regex = '/globalVars\.(\w+)/m';
                                $matches = [];
                                preg_match_all($regex, $argsJson, $matches, PREG_PATTERN_ORDER, 0);
                                $macros = array_merge($macros, $matches[1]);
                            }
                        }
                    }
                }
            }
        }
        return $macros;
    }

    /**
     * Imports the config data using the specified options
     *
     * @param array $data
     * @param array $options
     * @return mixed
     */
    public function import(array $data, array $options, $tokenType)
    {
        $tagsResult = true;
        if (array_key_exists('exported_tags', $data)) {
            $tagsResult = $this->importTags($data['exported_tags'], $tokenType);
        }
        $collectionResult = true;
        if (array_key_exists('collections', $options) && $options['collections']['include']) {
            $collectionResult = $this->importPlaybookCollections($data['collections'], $options['collections']['values'], $tokenType);
        }
        $globalVariableResult = true;
        if (array_key_exists('globalVariables', $options) && $options['globalVariables']['include']) {
            $globalVariableResult = $this->importGlobalVariables($data['globalVariables'], $options['globalVariables']['values'], $tokenType);
        } elseif (array_key_exists('globalVariables', $options) && array_key_exists('collections', $options) && $options['collections']['include']) {
            // Import required global variables from playbooks
            $requiredMacros = [];
            foreach ($options['collections']['values'] as $collection) {
                $requiredMacros = array_merge($requiredMacros, $collection['macros']);
            }
            $uniqueRequiredMacros = array_unique($requiredMacros);
            $requiredMacrosToImport = [];
            foreach ($options['globalVariables']['values'] as $globalVariable) {
                if (in_array($globalVariable['name'], $uniqueRequiredMacros)) {
                    $requiredMacrosToImport[] = $globalVariable;
                }
            }
            $globalVariableResult = $this->importGlobalVariables($data['globalVariables'], $requiredMacrosToImport);
        }
        return $collectionResult && $globalVariableResult && $tagsResult;
    }

    private function importTags($tags, $tokenType)
    {
        $repository = $this->loadRepository(Tag::class);
        $tagObjects = [];
        foreach ($tags as $tag) {
            $existing = $repository->findOneBy(['uuid' => $tag]);
            if ($existing) {
                continue;
            }
            $tagObjects[] = [
                'uuid' => $tag
            ];
        }
        if (count($tagObjects) > 0) {
            $fullURI = sprintf('%s/api/3/bulkupsert/tags', $this->container->getParameter('crudhub.uri'));
            $body = [
              "__unique" => 'uuid',
              "__replace" => false,
              "__data" => $tagObjects
            ];
            return $this->sendRequest($fullURI, 'post', $body, $tokenType);
        }
        return true;
    }

    /**
     * Imports playbook collections
     *
     * @param array $collections
     * @param array $collectionOptions
     * @return boolean
     */
    private function importPlaybookCollections(array $collections, array $collectionOptions, $tokenType)
    {
        $toAdd = [];
        $deleteUuids = [];
        $playbookRepository = $this->loadRepository(Workflow::class);
        $collectionRepository = $this->loadRepository(WorkflowCollection::class);
        foreach ($collectionOptions as $playbookCollectionOption) {
            if (!$playbookCollectionOption['include']) {
                // Do not merge skipped items
                continue;
            }
            foreach ($collections as $collection) {
                if ($playbookCollectionOption['uuid'] == $collection['uuid']) {
                    $newPlaybooks = [];
                    foreach ($collection['workflows'] as $playbook) {
                        /** @var Workflow $existingPlaybook */
                        $existingPlaybook = $playbookRepository->findOneBy(['uuid' => $playbook['uuid']]);
                        if ($existingPlaybook && ($existingPlaybook->getCollection()->getUuid() != $collection['uuid'] || $playbookCollectionOption['rename'])) {
                            // If a uuid already exists in a different collection, or if we are renaming the current collection, generate a new one
                            unset($playbook['uuid']);
                        }
                        unset($playbook['collection']);
                        $playbook = $this->fixStepUuids($playbook);
                        $newPlaybooks[] = $playbook;
                    }
                    $collection['workflows'] = $newPlaybooks;
                    if (!$playbookCollectionOption['exists_uuid'] && !$playbookCollectionOption['exists_name']) {
                        $toAdd[] = $collection;
                        break;
                    }
                    if ($playbookCollectionOption['rename']) {
                        $number = 1;
                        while ($collectionRepository->findOneBy(['name' => $collection['name'] . ' (' . $number . ')'])) {
                            $number++;
                        }
                        $collection['name'] = $collection['name'] . ' (' . $number . ')';
                        if ($playbookCollectionOption['exists_uuid']) {
                            unset($collection['uuid']);
                        }
                        $toAdd[] = $collection;
                        break;
                    }
                    $deleteUuids[] = $playbookCollectionOption['existing_uuid'];
                    $toAdd[] = $collection;
                    break;
                }
            }
        }
        $results = [];
        if (count($deleteUuids) > 0) {
            $fullURI = sprintf('%s/api/3/delete/workflow_collections', $this->container->getParameter('crudhub.uri'));
            $body = [
              "ids" => $deleteUuids
            ];
            $result = $this->sendRequest($fullURI, 'delete', $body, $tokenType);
            $results[] = $result;
        }
        foreach ($toAdd as $item) {
            $fullURI = sprintf('%s/api/3/workflow_collections', $this->container->getParameter('crudhub.uri'));
            $result = $this->sendRequest($fullURI, 'post', $item, $tokenType);
            $results[] = $result;
        }
        if (in_array(false, $results)) {
            return false;
        }
        return true;
    }

    /**
     * Replaces the uuids for all playbook steps and points the routes to the new steps
     *
     * @param array $playbook
     * @return array
     */
    private function fixStepUuids(array $playbook)
    {
        $uuidMapping = [];
        // First generate the new uuids and keep track using $uuidMapping
        $steps = [];
        foreach ($playbook['steps'] as $step) {
            $newUuid = self::uuidv4(openssl_random_pseudo_bytes(16));
            $uuidMapping[$step['uuid']] = $newUuid;
            $step['uuid'] = $newUuid;
            $steps[] = $step;
        }
        $playbook['steps'] = $steps;
        // Next, find all step references and replace them
        if ($playbook['triggerStep']) {
            $playbook['triggerStep'] = '/api/3/workflow_steps/' . $uuidMapping[self::getUuidFromIri($playbook['triggerStep'])];
        }
        $routes = [];
        foreach ($playbook['routes'] as $route) {
            $route['uuid'] = self::uuidv4(openssl_random_pseudo_bytes(16));
            $route['targetStep'] = '/api/3/workflow_steps/' . $uuidMapping[self::getUuidFromIri($route['targetStep'])];
            $route['sourceStep'] = '/api/3/workflow_steps/' . $uuidMapping[self::getUuidFromIri($route['sourceStep'])];
            $routes[] = $route;
        }
        $playbook['routes'] = $routes;
        $steps = [];
        foreach ($playbook['steps'] as $step) {
            $stepTypeUuid = self::getUuidFromIri($step['stepType']);
            if ($stepTypeUuid == self::STEP_TYPES['MANUAL_INPUT']) {
                if (array_key_exists('timeout', $step['arguments']) && array_key_exists('step_iri', $step['arguments']['timeout'])) {
                    $step['arguments']['timeout']['step_iri'] = '/api/3/workflow_steps/' . $uuidMapping[self::getUuidFromIri($step['arguments']['timeout']['step_iri'])];
                }
                $options = [];
                foreach ($step['arguments']['response_mapping']['options'] as $responseMappingOption) {
                    $responseMappingOption['step_iri'] = '/api/3/workflow_steps/' . $uuidMapping[self::getUuidFromIri($responseMappingOption['step_iri'])];
                    $options[] = $responseMappingOption;
                }
                $step['arguments']['response_mapping']['options'] = $options;
            } elseif ($stepTypeUuid == self::STEP_TYPES['DECISION']) {
                $conditions = [];
                foreach ($step['arguments']['conditions'] as $condition) {
                    $condition['step_iri'] = '/api/3/workflow_steps/' . $uuidMapping[self::getUuidFromIri($condition['step_iri'])];
                    $conditions[] = $condition;
                }
                $step['arguments']['conditions'] = $conditions;
            }
            $steps[] = $step;
        }
        $playbook['steps'] = $steps;
        return $playbook;
    }

    /**
     * Converts IRIs to UUIds
     *
     * @param string $iri
     * @return string
     */
    private static function getUuidFromIri($iri)
    {
        $iriParts = explode('/', $iri);
        return end($iriParts);
    }

    /**
     * @see http://stackoverflow.com/a/15875555
     *
     * @param $data
     * @return string
     */
    private static function uuidv4($data)
    {
        assert(strlen($data) == 16);

        $data[6] = chr(ord($data[6]) & 0x0f | 0x40); // set version to 0100
        $data[8] = chr(ord($data[8]) & 0x3f | 0x80); // set bits 6-7 to 10

        return vsprintf('%s%s-%s-%s-%s-%s%s%s', str_split(bin2hex($data), 4));
    }
    
    private function importGlobalVariables(array $globalVariables, array $globalVariableOptions, $tokenType)
    {
        $toAdd = [];
        foreach ($globalVariableOptions as $globalVariableOption) {
            if (!$globalVariableOption['include']) {
                // Do not merge skipped items
                continue;
            }
            foreach ($globalVariables as $globalVariable) {
                if ($globalVariableOption['name'] == $globalVariable['name']) {
                    $toAdd[] = $globalVariable;
                    break;
                }
            }
        }
        $results = [];
        foreach ($toAdd as $item) {
            $tokenStorage = $this->container->get('security.token_storage');
            $token = $tokenStorage->getToken();
            if ($token) {
                $workflowMacroEndpoint = '/api/wf/api/dynamic-variable/';
                $uri = $this->container->getParameter('crudhub.uri') . $workflowMacroEndpoint;
                $results[] = $this->sendRequest($uri, 'post', $item, $tokenType);
            } else {
                $workflowMacroEndpoint = '/api/dynamic-variable/';
                $uri = $this->container->getParameter('workflow.uri') . $workflowMacroEndpoint;
                $results[] = $this->sealabRequest($uri, 'post', $item, $tokenType);
            }
        }
        
        if (in_array(false, $results)) {
            return false;
        }
        return true;
    }

    /**
     * Exports the config using the specified options
     *
     * @param array $options
     * @return array
     */
    public function export(array $options, $tokenType='JWT')
    {
        $authorizationChecker = $this->container->get('security.authorization_checker');
        $resourceCollection = $this->container->get('api.resource_collection');
        $resource = $resourceCollection->getResourceForEntity(Workflow::class);
        if (!$authorizationChecker->isGranted(['perm:workflows.read'], $resource)) {
            return [];
        }
        $usedMacros = [];
        if (array_key_exists('collections', $options)) {
            $data['collections'] = $this->getResourceFromIRI($options['collections'], WorkflowCollection::class, 'resource.workflowcollection', true);
            if (array_key_exists('includeVersions', $options) && $options['includeVersions']) {
                $newCollections = [];
                $recordTags = [];
                foreach ($data['collections'] as $collection) {
                    $iris = [];
                    foreach ($collection['workflows'] as $playbook) {
                        $iris[] = '/api/3/workflows/' . $playbook['uuid'];
                        foreach ($playbook['recordTags'] as $recordTag) {
                            $parts = explode('/', $recordTag);
                            $recordTags[] = end($parts);
                        }
                    }
                    $collection['workflows'] = $this->getResourceFromIRI($iris, Workflow::class, 'resource.workflow', true, true, true);
                    $newCollections[] = $collection;
                    foreach ($collection['recordTags'] as $recordTag) {
                        $parts = explode('/', $recordTag);
                        $recordTags[] = end($parts);
                    }
                }
                $data['collections'] = $newCollections;
            }
            $recordTags = [];
            $newCollections = [];
            foreach ($data['collections'] as $collection) {
                foreach ($collection['recordTags'] as $recordTag) {
                    $parts = explode('/', $recordTag);
                    $recordTags[] = end($parts);
                }
                foreach ($collection['workflows'] as $index => $playbook) {
                    $collection['workflows'][$index]['isPrivate'] = false;
                    $collection['workflows'][$index]['owners'] = [];
                    foreach ($playbook['recordTags'] as $recordTag) {
                        $parts = explode('/', $recordTag);
                        $recordTags[] = end($parts);
                    }
                }
                $usedMacros = array_merge($usedMacros, $this->getMacrosFromPlaybooks($collection['workflows']));
                $newCollections[] = $collection;
            }
            $data['collections'] = $newCollections;
            $data['exported_tags'] = array_values(array_unique($recordTags));
        }
        if (array_key_exists('globalVariables', $options) || count($usedMacros) > 0) {
            $data['globalVariables'] = [];
            $existingDynamicVariables = $this->getGlobalVariables($tokenType);
            $macrosToImport = array_unique(array_merge($options['globalVariables'], $usedMacros));
            foreach ($macrosToImport as $dynamicVariableName) {
                foreach ($existingDynamicVariables as $existingDynamicVariable) {
                    if ($existingDynamicVariable['name'] == $dynamicVariableName) {
                        $data['globalVariables'][] = $existingDynamicVariable;
                        break;
                    }
                }
            }
        }
        return $data;
    }
}
