<?php
/* Copyright start
  Copyright (C) 2008 - 2020 Fortinet Inc.
  All rights reserved.
  FORTINET CONFIDENTIAL & FORTINET PROPRIETARY SOURCE CODE
  Copyright end */

namespace Cybersponse\CoreBundle\Service\Config;

use Cybersponse\CoreBundle\Provider\ProxyProvider;
use Symfony\Component\DependencyInjection\ContainerInterface;
use Doctrine\ORM\EntityRepository;
use GuzzleHttp\Client;
use GuzzleHttp\Psr7\Request as GuzzleRequest;

/**
 * Interface PortingConfig
 * @package Cybersponse\CoreBundle\Service\Config
 */
abstract class PortingConfig
{

    const METHOD_MAP = [
        "post" => "create",
        "put" => "update",
        "delete" => "delete"
    ];

    /** @var ContainerInterface */
    protected $container;

    /**
     * @param ContainerInterface $container
     */
    public function setContainer($container)
    {
        $this->container = $container;
    }

    /**
     * Generates import options from import data
     *
     * @param array $data
     * @return array
     */
    abstract public function getImportOptions(array $data, $tokenType);

    /**
     * Imports the config data using the specified options
     *
     * @param array $data
     * @param array $options
     * @return mixed
     */
    abstract public function import(array $data, array $options, $tokenType);

    /**
     * Exports the config using the specified options
     *
     * @param array $options
     * @return array
     */
    abstract public function export(array $options, $tokenType='JWT');

    /**
     * Loads the doctrine repo for the given class
     *
     * @param string $className
     * @return EntityRepository
     */
    protected function loadRepository($className) {
        $entityManager = $this->container->get('doctrine')->getManagerForClass($className);
        $repository = $entityManager->getRepository($className);
        return $repository;
    }

    protected function getResourceFromIRI($iris, $className, $resourceName, $relationships = false, $export = true, $versions = false) {
        $repository = $this->loadRepository($className);
        /** @var ResourceInterface $picklistResource */
        $resource = $this->container->get($resourceName);
        $context = $resource->getDenormalizationContext();
        $context['export'] = $export;
        $context['relationships'] = $relationships;
        if ($versions) {
            $_GET['$versions'] = true;
            $_GET['$includeData'] = true;
        }
        $context["ignore_field_authorization"] = true;
        $context["ignore_authorization"] = true;
        $exportRecords = [];
        foreach ($iris as $iri) {
            $parts = explode('/', $iri);
            $record = $repository->findOneBy(['uuid' => end($parts)]);
            $normalizedObject = $this->container->get('serializer')->normalize($record, 'json-ld', $context);
            $exportRecords[] = $normalizedObject;
        }
        return $exportRecords;
    }

    protected function sendRequest($uri, $method, $data = null, $tokenType='JWT') {
        /** @var ProxyProvider $proxyProvider */

        $proxyProvider = $this->container->get('cybersponse.core.provider.proxy');
        $tokenStorage = $this->container->get('security.token_storage');
        $token = $tokenStorage->getToken();
        if ($token && $tokenType == 'JWT') {
            $client = new Client();
        } else {
            $client = $proxyProvider->getCrudhubClient();
        }
        $headers = array(
            "Content-type" => "application/json",
            "Authorization" => $token ? sprintf("Bearer %s", $token->getCredentials()) : null
        );
        $request = new GuzzleRequest($method, $uri, $headers, $data ? json_encode($data) : null);
        try {
            $response = $client->send($request, ['verify' => false]);
            $statusCode = $response->getStatusCode();
            if ($statusCode < 400) {
                //$jsonDecodeResponse = json_decode($response->getBody()->getContents(), true);
                return $response->getBody()->getContents();
            } else {
                throw new \Exception($uri.'\n'.$response->getBody()->getContents());
            }
        } catch (\Exception $exception) {
            $errCode = $exception->getCode();
            if ($errCode == 403) {
                $iri = str_replace($this->container->getParameter('crudhub.uri'), '', $uri);
                $module = explode('/', $iri)[4];
                throw new \Exception(sprintf('You do not have %s access on %s module', self::METHOD_MAP[$method], $module));
            }
            elseif ($errCode == 404) {
                $iri = str_replace($this->container->getParameter('crudhub.uri'), '', $uri);
                $message = $data ? json_encode($data) : "{}";
                throw new \Exception(sprintf('The Requested Resource %s (%s) Does not exists ', $iri, $method));
            }
            elseif ($errCode == 409) {
                $message = json_decode($exception->getResponse()->getBody()->getContents(), true);
                throw new \Exception(sprintf($message['hydra:description']));
            }
            $message = $exception->getResponse() ? $exception->getResponse()->getBody()->getContents() : $exception->getMessage();
            throw new \Exception(sprintf('%d: %s', $errCode, $message));
        }
    }

}
