<?php
/* Copyright start
  Copyright (C) 2008 - 2020 Fortinet Inc.
  All rights reserved.
  FORTINET CONFIDENTIAL & FORTINET PROPRIETARY SOURCE CODE
  Copyright end */

namespace Cybersponse\CoreBundle\Tests\Functional\Controller;


use Cybersponse\CoreBundle\Api\Bulk\BulkApiResults;
use Cybersponse\CoreBundle\Controller\TypeAgnosticResourceController;
use Cybersponse\CoreBundle\Entity;
use Cybersponse\TestBundle\Functional\ApiTestCase;
use Symfony\Bundle\FrameworkBundle\Client;

class TypeAgnosticResourceControllerTest extends ApiTestCase
{
    /** @var Client */
    protected $client;

    protected function setUp()
    {
        parent::setUp();
        $this->requestVerb = 'POST';
        $this->requestUri = '/api/3/';
    }

    public function test_CreateAction_GivenNonJsonInput_ShouldReturn400()
    {
        $this->requestContent = 'definitely not json';
        $this->executeSystemUnderTest();

        $this->assertEquals(400, $this->client->getResponse()->getStatusCode());
    }

    public function test_CreateAction_GivenNonJsonInput_ShouldReturnHydraError()
    {
        $this->requestContent = 'definitely not json';
        $this->executeSystemUnderTest();

        $responseContent = $this->client->getResponse()->getContent();
        $deserializedContent = json_decode($responseContent, true);
        $this->assertEquals('/api/3/contexts/Error', $deserializedContent['@context']);
    }

    public function test_CreateAction_GivenEmptyList_ShouldReturn200()
    {
        $this->setJsonRequestContent([]);
        $this->executeSystemUnderTest();

        $this->assertEquals(200, $this->client->getResponse()->getStatusCode());
    }

    public function test_CreateAction_GivenEmptyList_ShouldReturnEmptyList()
    {
        $this->setJsonRequestContent([]);
        $this->executeSystemUnderTest();

        $this->assertEquals('[]', $this->client->getResponse()->getContent());
    }

    public function test_CreateAction_GivenSingleUnauthorizedObject_ShouldReturn403()
    {
        $expectedType = 'Picklist';
        $input = [
            ['@type' => $expectedType]
        ];

        $this->setJsonRequestContent($input);
        $this->executeSystemUnderTest();

        $this->assertEquals(403, $this->client->getResponse()->getStatusCode());
    }

    public function test_CreateAction_GivenSingleUnauthorizedObject_ShouldReturnHydraError()
    {
        $input = [
            ['@type' => 'Picklist']
        ];

        $this->setJsonRequestContent($input);
        $this->executeSystemUnderTest();

        $responseContent = $this->client->getResponse()->getContent();
        $deserializedContent = json_decode($responseContent, true);
        $this->assertEquals($deserializedContent['@type'], 'Error');
    }

    public function test_CreateAction_GivenSingleUnauthorizedObject_ShouldReturnDescriptionWithUnauthorizedType()
    {
        $expectedType = 'Picklist';
        $input = [
            ['@type' => $expectedType]
        ];

        $this->setJsonRequestContent($input);
        $this->executeSystemUnderTest();

        $responseContent = $this->client->getResponse()->getContent();
        $deserializedContent = json_decode($responseContent, true);

        $expectedErrorMessage = sprintf(TypeAgnosticResourceController::ACCESS_DENIED_ERROR_FORMAT, $expectedType);
        $this->assertEquals($deserializedContent['hydra:description'], $expectedErrorMessage);
    }

    public function test_CreateAction_GivenMultipleUnauthorizedObject_ShouldReturnDescriptionWithAllUnauthorizedTypes()
    {
        $unauthorizedType1 = 'Picklist';
        $unauthorizedType2 = 'PicklistNames';
        $input = [
            ["@type" => $unauthorizedType1],
            ["@type" => $unauthorizedType2]
        ];

        $this->setJsonRequestContent($input);
        $this->executeSystemUnderTest();

        $responseContent = $this->client->getResponse()->getContent();
        $deserializedContent = json_decode($responseContent, true);

        $expectedErrorMessage = sprintf(TypeAgnosticResourceController::ACCESS_DENIED_ERROR_FORMAT, implode(', ', [$unauthorizedType1, $unauthorizedType2]));
        $this->assertEquals($deserializedContent['hydra:description'], $expectedErrorMessage);
    }

    public function test_CreateAction_GivenInvalidConstraint_ShouldReturn400()
    {
        $this->setRoles([
            $this->createModulePermission('application', true)
        ]);

        $expectedType = 'PicklistNames';
        $input = [
            [
                '@type' => $expectedType,
                'name' => null
            ]
        ];

        $this->setJsonRequestContent($input);
        $this->executeSystemUnderTest();

        $this->assertEquals(400, $this->client->getResponse()->getStatusCode());
    }

    public function test_CreateAction_GivenInvalidConstraint_ShouldReturnObjectInViolationResponse()
    {
        $this->setRoles([
            $this->createModulePermission('application', true)
        ]);

        $expectedType = 'PicklistNames';
        $input = [
            [
                '@type' => $expectedType,
                'name' => null
            ]
        ];

        $this->setJsonRequestContent($input);
        $this->executeSystemUnderTest();

        $responseContent = $this->client->getResponse()->getContent();
        $deserializedError = json_Decode($responseContent, true);
        $this->assertEquals(current($input), current($deserializedError)['object']);
    }

    public function test_CreateAction_GivenInvalidConstraint_ShouldReturnConstraintViolationsInViolationResponse()
    {
        $this->setRoles([
            $this->createModulePermission('application', true)
        ]);

        $expectedType = 'PicklistNames';
        $input = [
            [
                '@type' => $expectedType,
                'name' => null
            ]
        ];

        $this->setJsonRequestContent($input);
        $this->executeSystemUnderTest();

        $responseContent = $this->client->getResponse()->getContent();
        $deserializedError = json_decode($responseContent, true);
        $this->assertEquals('ConstraintViolationList', current($deserializedError)['violations']['@type']);
    }

    public function test_CreateAction_GivenSingleValidObject_ShouldReturn201()
    {
        $this->setRoles([
            $this->createModulePermission('application', true)
        ]);

        $type = sprintf('tester_%s', time());
        $input = [
            $this->createModelMetadata($type)
        ];

        $this->setJsonRequestContent($input);
        $this->executeSystemUnderTest();

        $response = $this->client->getResponse();
        $this->assertEquals(201, $response->getStatusCode());
    }

    public function test_CreateAction_GivenSingleValidObject_ShouldReturnObjectInPersistedCollection()
    {
        $this->setRoles([
            $this->createModulePermission('application', true)
        ]);

        $expectedType = sprintf('tester_%s', time());
        $input = [
            $this->createModelMetadata($expectedType)
        ];
        $this->setJsonRequestContent($input);
        $this->executeSystemUnderTest();

        $response = $this->client->getResponse();
        $deserializedResponse = json_decode($response->getContent(), true);
        $this->assertEquals($expectedType, current($deserializedResponse[BulkApiResults::PERSISTED_COLLECTION_NAME])['type']);
    }

    public function test_CreateAction_GivenMultipleValidObject_ShouldReturnObjectsInPersistedCollection()
    {
        $this->setRoles([
            $this->createModulePermission('application', true)
        ]);

        $firstExpectedValue = sprintf('first_tester_%s', time());
        $secondExpectedValue = sprintf('second_tester_%s', time());
        $input = [
            $this->createModelMetadata($firstExpectedValue),
            $this->createModelMetadata($secondExpectedValue)
        ];
        $this->setJsonRequestContent($input);
        $this->executeSystemUnderTest();

        $response = $this->client->getResponse();
        $deserializedResponse = json_decode($response->getContent(), true);

        $actualTypes = array_map(function ($persistedObject) {
            return array_key_exists('type', $persistedObject) ? $persistedObject['type'] : null;
        }, $deserializedResponse[BulkApiResults::PERSISTED_COLLECTION_NAME]);
        foreach ([$firstExpectedValue, $secondExpectedValue] as $expectedType) {
            if (!in_array($expectedType, $actualTypes)) {
                $this->fail(sprintf('Type "%s" was expected in persisted objects response.', $expectedType));
            }
        }
    }

    /**
     * @param $type
     * @return array
     */
    protected function createModelMetadata($type)
    {
        return [
            '@type' => 'ModelMetadata',
            'type' => $type,
            'tableName' => $type,
            'descriptions' => [
                'singular' => $type
            ]
        ];
    }
}
