<?php
/* Copyright start
  Copyright (C) 2008 - 2020 Fortinet Inc.
  All rights reserved.
  FORTINET CONFIDENTIAL & FORTINET PROPRIETARY SOURCE CODE
  Copyright end */

namespace Cybersponse\CustomerTenantBundle\Controller;

use Cybersponse\AuthorizationBundle\Permissions\PermissionsDictionaryBuilder;
use Cybersponse\CoreBundle\Api\ApiEvents;
use Cybersponse\UtilsBundle\ARCUtils;
use Cybersponse\QueryBundle\Query\ExpressionBuilder;
use Cybersponse\CoreBundle\Controller\AuthorizationResourceController;
use Cybersponse\UtilsBundle\Utils;
use Cybersponse\WorkflowBundle\Subscriber\DataEvent;
use Dunglas\ApiBundle\Event\Events;
use Dunglas\ApiBundle\Exception\DeserializationException;
use Symfony\Component\HttpKernel\Exception\HttpException;
use Dunglas\ApiBundle\JsonLd\Response;
use Symfony\Component\HttpFoundation\Request;
use Symfony\Component\Serializer\Serializer;

class TenantController extends AuthorizationResourceController
{
    const MODULE_TYPE = "tenants";
    const MASTER_ROLE_NAME = 'master';
    const REPLICATION_ROUTINGKEY = "route.crudhub.tenant.request.tenantconfig";
    const REMOTE_MMD_MODIFICATION_NOT_ALLOWED = 'Modifying Remote MMD Value is not allowed';
    const ALLOW_REMOTE_MMD = 'allowRemoteMMDModification';

    protected function appendAdditionalFilters()
    {
        $utilsObj = new Utils();
        if($utilsObj->getInstanceRole($this) == self::TENANT_ROLE_NAME){
            $queryProvider = $this->container->get('cybersponse.query.provider.query_provider');
            $accessibleQueryFilter =
                ExpressionBuilder::logicOr(
                    ExpressionBuilder::field('role')->equals('self')
                );
            $queryProvider->setAdditionalFilters($accessibleQueryFilter->toArray());
        }
    }

    /**
     * @param Request $request
     * @return Response|static
     */
    public function getAction(Request $request)
    {
        $this->appendAdditionalFilters();
        return parent::cgetAction($request);
    }

    /**
     * @param Request $request
     * @return Response
     * @throws DeserializationException
     */
    public function postAction(Request $request)
    {
        $utilsObj = new Utils();
        $resource = $this->identifyResourceByShortName(self::MODULE_TYPE);
        $this->denyAccessUnlessGranted($resource, PermissionsDictionaryBuilder::PERM_CRUD_CREATE);
        $this->get(self::EVENT_DISPATCHER)->dispatch(ApiEvents::PRE_API_CREATE, new DataEvent($resource, null));

        $blockPassword = Utils::isPasswordBlock($request);
        $decodedRequestData = json_decode($request->getContent(), true);
        $this->prePostCheck($decodedRequestData, $resource, $utilsObj);
        $request_data = $this->prePostAdd($resource, $decodedRequestData, $utilsObj);

        $object = $this->createObject($request_data, $resource, $resource->getDenormalizationContext());

        $this->get(self::EVENT_DISPATCHER)->dispatch(Events::PRE_CREATE_VALIDATION, new DataEvent($resource, $object));

        $violations = $this->get('validator')->validate($object, null, $resource->getValidationGroups());
        if (0 === count($violations)) {
            foreach ($this->getUser()->getTeams() as $owner) {
                $object->addOwner($owner);
            }
            $existing_relationship = ARCUtils::setRelationShip($request, false);
            $existing_set_asJson = ARCUtils::setAsJson($request, false);
            // Validation succeed
            $this->get(self::EVENT_DISPATCHER)->dispatch(Events::PRE_CREATE, new DataEvent($resource, $object));
            /** @var Serializer $serializer */
            $serializer = $this->get('serializer');
            $updatedObject = $serializer->normalize(
                $object,
                self::NORMALIZER_METHOD,
                $resource->getNormalizationContext() + ['block_password' => false]);

            ARCUtils::setRelationShip($request, $existing_relationship);
            ARCUtils::setAsJson($request, $existing_set_asJson);

            $data = $this->prepareChangeData($updatedObject, (object)null, 'Create');
            $data['esoperation'] = 'index';
            $utilsObj->sendData($this, [$data]);

            $tenantDiffObj = $utilsObj->prepareReplicationDiffData($updatedObject, null);
            $tenantData = $utilsObj->prepareReplicationData($tenantDiffObj, 'POST', null, null);

            $utilsObj->sendReplicationData($this, $tenantData, self::REPLICATION_ROUTINGKEY);

            return $this->getSuccessResponse($resource, $object, Response::HTTP_CREATED, [], [], [], $blockPassword);
        }
        return $this->getErrorResponse($violations);
    }

    /**
     * @param $decodedRequestData
     * @return bool|Response
     * @throws HttpException
     */
    public function prePostCheck($decodedRequestData, $resource, $utilsObj)
    {
        $tenantId = isset($decodedRequestData['tenantId']) ? $decodedRequestData['tenantId'] : null;
        if ($tenantId) {
            $existingTenantObj = $utilsObj->findWithWhere($this, $resource, ['tenantId' => $tenantId]);
            if ($existingTenantObj) {
                throw new HttpException(Response::HTTP_BAD_REQUEST, "TenantId: '" . $tenantId . "' already exists, use different TenantId");
            }
        }
        $tenantName = isset($decodedRequestData['name']) ? $decodedRequestData['name'] : null;
        if ($tenantName) {
            $existingTenantObj = $utilsObj->findWithWhere($this, $resource, ['name' => $tenantName]);
            if ($existingTenantObj) {
                throw new HttpException(Response::HTTP_BAD_REQUEST, "Tenant Name: '" . $tenantName . "' already exists, use different Tenant Name");
            }
        }
        if ($decodedRequestData['role'] == 'self') {
            throw new HttpException(Response::HTTP_FORBIDDEN, 'UnAuthorized Access');
        }        
    }

    public function prePostAdd($resource, $decodedRequestData, $utilsObj)
    {        
        $configStatus = isset($decodedRequestData['configstatus']) ? $decodedRequestData['configstatus'] : null;
        if (!$configStatus) {
            if (isset($decodedRequestData['isDedicated']) && !$decodedRequestData['isDedicated']) {
                $decodedRequestData[self::ALLOW_REMOTE_MMD] = false;
                if (!isset($decodedRequestData['tenantId'])) {
                    $decodedRequestData['tenantId'] = bin2hex(random_bytes(16));
                }
            } else {
                $decodedRequestData[self::ALLOW_REMOTE_MMD] = true;
            }
            if (!isset($decodedRequestData['role'])) {
                $decodedRequestData['role'] = 'tenant';
            }
        }

        return $decodedRequestData;
    }

    public function prePutValidate($instanceRole, $objectData, $isMmdModificationNotAllowed, $preUpdatedObject) {
        if ($instanceRole == self::MASTER_ROLE_NAME) {
            if (array_key_exists(self::ALLOW_REMOTE_MMD, $objectData) && $isMmdModificationNotAllowed && $objectData[self::ALLOW_REMOTE_MMD] != $preUpdatedObject[self::ALLOW_REMOTE_MMD]) {
                throw new HttpException(Response::HTTP_FORBIDDEN, self::REMOTE_MMD_MODIFICATION_NOT_ALLOWED);
            }
        }
    }

    public function putAction(Request $request, $id)
    {
        $resource = $this->identifyResourceByShortName(self::MODULE_TYPE);
        $utilsObj = new Utils();
        $this->denyAccessUnlessGranted($resource, PermissionsDictionaryBuilder::PERM_CRUD_UPDATE);
        $this->get(self::EVENT_DISPATCHER)->dispatch(ApiEvents::PRE_API_UPDATE, new DataEvent($resource, null));
        $objectData = json_decode($request->getContent(), true);
        $blockPassword = Utils::isPasswordBlock($request);

        $instanceRole = $utilsObj->getInstanceRole($this);

        $object = $this->findOrThrowNotFound($resource, $id, $request);
        $this->denyAccessUnlessGranted($resource, PermissionsDictionaryBuilder::PERM_CRUD_UPDATE, $object);
        $this->get(self::EVENT_DISPATCHER)->dispatch(ApiEvents::PRE_API_UPDATE_DESERIALIZE, new DataEvent($resource, $object));

        $context = $resource->getDenormalizationContext();
        $context['object_to_populate'] = $object;
        /** @var Serializer $serializer */
        $serializer = $this->get('serializer');
        
        $existing_set_asJson = ARCUtils::setAsJson($request, false);
        $existing_relationship = ARCUtils::setRelationShip($request, false);
        $preUpdatedObject = $serializer->normalize($object, self::NORMALIZER_METHOD, $resource->getNormalizationContext() + ['block_password' => false]);
        $this->prePutValidate($instanceRole, $objectData, $blockPassword, $preUpdatedObject);


        if (array_key_exists("__link", $objectData) || array_key_exists("__unlink", $objectData)) {
            $objectData = ARCUtils::mergeRequest($objectData, $preUpdatedObject);
        }

        $object = $this->createObject($objectData, $resource, $context);
        $this->get(self::EVENT_DISPATCHER)->dispatch(Events::PRE_UPDATE_VALIDATION, new DataEvent($resource, $object));

        $violations = $this->get('validator')->validate($object, null, $resource->getValidationGroups());
        if (0 === count($violations)) {
            // Validation succeed
            $this->get(self::EVENT_DISPATCHER)->dispatch(Events::PRE_UPDATE, new DataEvent($resource, $object, $preUpdatedObject));
            $updatedObject = $serializer->normalize($object, self::NORMALIZER_METHOD, $resource->getNormalizationContext() + ['block_password' => false]);
            $data = $this->prepareChangeData($updatedObject, $preUpdatedObject, 'Update', $objectData);
            $data['esoperation'] = 'modify';
            $utilsObj->sendData($this, [$data]);

            $tenantDiffObj = $utilsObj->prepareReplicationDiffData($updatedObject, $preUpdatedObject);
            $tenantData = $utilsObj->prepareReplicationData($tenantDiffObj, 'PUT', null, null);
            $utilsObj->sendReplicationData($this, $tenantData, self::REPLICATION_ROUTINGKEY);

            ARCUtils::setRelationShip($request, $existing_relationship);
            ARCUtils::setAsJson($request, $existing_set_asJson);

            return $this->getSuccessResponse($resource, $object, Response::HTTP_OK, [], [], [], $blockPassword);
        }
        return $this->getErrorResponse($violations);
    }

    public function deleteTenant($request, $resource, $object){
        /** @var Serializer $serializer */
        $serializer = $this->get('serializer');
        $existing_relationship = ARCUtils::setRelationShip($request, false);
        $existing_set_asJson = ARCUtils::setAsJson($request, false);
        $previousObject = $serializer->normalize($object, self::NORMALIZER_METHOD, $resource->getNormalizationContext());
        ARCUtils::setRelationShip($request, $existing_relationship);
        ARCUtils::setAsJson($request, $existing_set_asJson);

        if (isset($previousObject['role']) && $previousObject['role'] == 'self') {
            throw new HttpException(Response::HTTP_BAD_REQUEST, "Deletion of self tenant is not allowed");
        }

        $this->get(self::EVENT_DISPATCHER)->dispatch(ApiEvents::PRE_API_DELETE, new DataEvent($resource, null));
        $this->denyAccessUnlessGranted($resource, PermissionsDictionaryBuilder::PERM_CRUD_DELETE, $object);
        $this->get(self::EVENT_DISPATCHER)->dispatch(Events::PRE_DELETE, new DataEvent($resource, $object));
        $this->publishDeleteData($previousObject);
    }

    public function deleteAction(Request $request, $id)
    {
        $resource = $this->identifyResourceByShortName(self::MODULE_TYPE);
        $this->denyAccessUnlessGranted($resource, PermissionsDictionaryBuilder::PERM_CRUD_DELETE);
        $object = $this->findOrThrowNotFound($resource, $id, $request);
        $this->deleteTenant($request, $resource, $object);
        return new Response(null, Response::HTTP_NO_CONTENT, []);
    }

    public function bulkDeleteAction(Request $request)
    {
        $resource = $this->identifyResourceByShortName(self::MODULE_TYPE);
        $this->denyAccessUnlessGranted($resource, PermissionsDictionaryBuilder::PERM_CRUD_DELETE);
        $request_body = json_decode($request->getContent());
        $ids = $request_body->ids;
        foreach ($ids as $id) {
            $object = $this->findOrThrowNotFound($resource, $id, $request);
            $this->deleteTenant($request, $resource, $object);
        }
        return new Response(null, Response::HTTP_NO_CONTENT, []);
    }

    public function publishDeleteData($previousObject)
    {
        $data = $this->prepareChangeData((object)null, $previousObject, 'Delete');
        $data['esoperation'] = 'delete';
        $utilsObj = new Utils();
        $utilsObj->sendData($this, [$data]);

        $tenantDiffObj = $utilsObj->prepareReplicationDiffData((object)null, $previousObject);
        $tenantData = $utilsObj->prepareReplicationData($tenantDiffObj, 'DELETE', null, null);
        $utilsObj->sendReplicationData($this, $tenantData, self::REPLICATION_ROUTINGKEY);
    }

    /**
     * @param Request $request
     * @param $id
     * @return Response
     */
    public function workflowPublishAction(Request $request)
    {
        $resource = $this->identifyResourceByShortName('workflow_collections');
        $request_body = json_decode($request->getContent(), true);
        $this->denyAccessUnlessGranted($resource, PermissionsDictionaryBuilder::PERM_CRUD_UPDATE);
        $this->get(self::EVENT_DISPATCHER)->dispatch(ApiEvents::PRE_API_UPDATE, new DataEvent($resource, null));

        $collectionIds = $request_body['collectionIds'];
        $collections = [];
        $_collectionsIds = [];
        foreach ($collectionIds as $collectionId) {
            $explodeObj = explode('/', $collectionId);
            $_collectionId = end($explodeObj);

            $explodeCollectionObj = explode('/', $_collectionId);
            $collectionObject = $this->findOrThrowNotFound($resource, end($explodeCollectionObj), $request);

            $request->query->set("\$relationships", true);
            $request->query->set("\$export", true);

            /** @var Serializer $serializer */
            $serializer = $this->get('serializer');
            $normalizedCollectionData = $serializer->normalize($collectionObject, self::NORMALIZER_METHOD, $resource->getNormalizationContext());
            array_push($collections, $normalizedCollectionData);
            array_push($_collectionsIds, $_collectionId);
        }
        $request_body['collections'] = $collections;
        $request_body['collectionIds'] = $_collectionsIds;
        $request_body['method'] = 'POST';
        $utilsObj = new Utils();
        $utilsObj->sendReplicationData($this, $request_body, "route.crudhub.instructions.remoterequest.playbookdatapush");
        return new Response(
            null,
            Response::HTTP_NO_CONTENT,
            []
        );
    }

    /**
     * @param Request $request
     * @return Response
     */
    public function configureAction(Request $request)
    {
        $resource = $this->identifyResourceByShortName(self::MODULE_TYPE);
        $request_body = json_decode($request->getContent(), true);
        $this->denyAccessUnlessGranted($resource, PermissionsDictionaryBuilder::PERM_CRUD_CREATE);
        $tenantId = isset($request_body['tenantId']) ? $request_body['tenantId'] : null;

        $explodeTenantObj = explode('/', $tenantId);
        $request->query->set("\$relationships", true);
        $object = $this->findOrThrowNotFound($resource, end($explodeTenantObj), $request);

        /** @var Serializer $serializer */
        $serializer = $this->get('serializer');
        $normalizedTenantData = $serializer->normalize($object, self::NORMALIZER_METHOD, $resource->getNormalizationContext() + ['block_password' => false]);
        $utilsObj = new Utils();
        $normalizedTenantData['configstatus'] = '/api/3/picklists/0dff7239-38a4-4dc8-9fe6-97932d2bd35f';
        $tenantDiffObj = $utilsObj->prepareReplicationDiffData($normalizedTenantData);
        $tenantData = $utilsObj->prepareReplicationData($tenantDiffObj, 'POST', null, null);
        $utilsObj->sendReplicationData($this, $tenantData, self::REPLICATION_ROUTINGKEY);

        /*Updating Config to InProgress*/
        $request_data = $normalizedTenantData;
        $context = $resource->getDenormalizationContext();
        $context['object_to_populate'] = $object;

        $object = $this->createObject($request_data, $resource, $context);
        $this->get(self::EVENT_DISPATCHER)->dispatch(Events::PRE_UPDATE_VALIDATION, new DataEvent($resource, $object));

        $violations = $this->get('validator')->validate($object, null, $resource->getValidationGroups());
        if (0 === count($violations)) {
            $this->get(self::EVENT_DISPATCHER)->dispatch(Events::PRE_UPDATE, new DataEvent($resource, $object, $normalizedTenantData));
            return new Response(
                null,
                Response::HTTP_NO_CONTENT
            );
        }
        return $this->getErrorResponse($violations);
    }
}
