<?php
/* Copyright start
  Copyright (C) 2008 - 2020 Fortinet Inc.
  All rights reserved.
  FORTINET CONFIDENTIAL & FORTINET PROPRIETARY SOURCE CODE
  Copyright end */
namespace Cybersponse\FilesBundle\Serializer;

use Cybersponse\CoreBundle\Entity\BaseEntity;
use Cybersponse\FilesBundle\Entity\File;
use Cybersponse\FilesBundle\Entity\PgFile;
use Cybersponse\MetadataBundle\Generation\Metadata\ModelAttributeMetadata;
use Cybersponse\MetadataBundle\Loader\ModelClassMetadataFactory;
use Dunglas\ApiBundle\Api\IriConverterInterface;
use Dunglas\ApiBundle\Api\ResourceCollectionInterface;
use Dunglas\ApiBundle\Api\ResourceResolverTrait;
use Dunglas\ApiBundle\Exception\InvalidArgumentException;
use Dunglas\ApiBundle\JsonLd\Serializer\ItemNormalizer;
use Symfony\Component\Serializer\Normalizer\DenormalizerInterface;
use Symfony\Component\Serializer\Normalizer\NormalizerInterface;
use Dunglas\ApiBundle\Api\ResourceInterface;
use Cybersponse\AuthorizationBundle\Entity\Team;
use Cybersponse\AuthorizationBundle\Ownership\OwnershipProvider;

/**
 * Class FileFormTypeNormalizer
 *
 * @todo This class needs to be less File specific and more generic.
 * @package Cybersponse\FilesBundle\Serializer
 */
class FileFormTypeNormalizer implements NormalizerInterface, DenormalizerInterface
{
    use ResourceResolverTrait;
    const FORMAT_JSON_LD = 'json-ld';
    const FORMAT_JSON = 'json';
    /** @var ModelClassMetadataFactory */
    protected $modelClassMetadataFactory;
    /** @var IriConverterInterface */
    protected $iriConverter;
    /** @var ItemNormalizer */
    protected $itemNormalizer;
    /** @var ResourceInterface */
    protected $teamResource;
    /** @var OwnershipProvider */
    protected $ownershipProvider;
    protected $container;

    /**
     * {@inheritdoc}
     */
    public function supportsNormalization($data, $format = null)
    {
        $isSupported = (self::FORMAT_JSON_LD === $format || self::FORMAT_JSON === $format) && is_object($data) && $data instanceof BaseEntity;
        if ($isSupported) {
            $fileAttributes = $this->getFileAttributes(get_class($data));
        }
        return !empty($fileAttributes);
    }

    /**
     * @param $className
     * @return ModelAttributeMetadata[]
     */
    protected function getFileAttributes($className)
    {
        $fileAttributes = [];
        try {
            $classMetadata = $this->modelClassMetadataFactory->getMetadataFor($className);
            foreach ($classMetadata->getAttributes() as $modelAttributeMetadata) {
                $isSupported = $modelAttributeMetadata->getFormType() == 'file';
                if ($isSupported) {
                    $fileAttributes[] = $modelAttributeMetadata;
                }
            }
        } catch (\Exception $e) { }
        return $fileAttributes;
    }

    /**
     * {@inheritdoc}
     */
    public function supportsDenormalization($data, $type, $format = null)
    {
        return false;
    }

    /**
     * {@inheritdoc}
     */
    public function normalize($object, $format = null, array $context = [])
    {
        $className = get_class($object);
        $fileAttributes = $this->getFileAttributes($className);
        $normalizedObject = $this->itemNormalizer->normalize($object, $format, $context);
        if (array_key_exists('as_iri', $context) && $context['as_iri']) {
            return $normalizedObject;
        }
        foreach ($fileAttributes as $fileAttribute) {
            $attributeName = $fileAttribute->getName();
            $value = $normalizedObject[$attributeName];
            if (!is_string($value)) {
                continue;
            }
            try {

                /** @var File $file */
                $file = $this->iriConverter->getItemFromIri($value, true);
                if ($file instanceof PgFile) {
                    $normalizedObject[$attributeName] = $value;
                    continue;
                }
            } catch (InvalidArgumentException $exception) {
                $normalizedObject[$attributeName] = null;
                break;
            }
            // Add parent object teams so that authorization will get validated as per inheritance
            if ($this->ownershipProvider->isObjectOwnable($object)) {
                $teamIris = $this->getTeamIris($object);
                if ($teamIris != null && $file != null) {
                    /**@var Team $team */
                    foreach ($teamIris as $team) {
                        $file->addOwner($this->iriConverter->getItemFromIri($team, true));
                    }
                }
            }
            $fileResource = $this->guessResource($file);
            $fileContext = $fileResource->getNormalizationContext();
            $fileContext['resource'] = $fileResource;
            $normalizedFile = $this->itemNormalizer->normalize($file, $format, $fileContext);
            $normalizedObject[$attributeName] = $normalizedFile;
        }
        return $normalizedObject;
    }

    /**
     * @param $object object to normalize
     * @return string[] Array of teamIris
     */
    protected function getTeamIris($object)
    {
        if ($object != null) {
            $owners = $object->getOwners();
        }
        $teamIds = array();
        if ($owners != null) {
            foreach ($owners->toArray() as $team) {
                array_push($teamIds, $team->getUuid());
            }
        }
        $teamBaseIri = $this->iriConverter->getIriFromResource($this->teamResource);
        $teamIris = array_map(function ($teamId) use ($teamBaseIri) {
            return sprintf('%s/%s', $teamBaseIri, $teamId);
        }, $teamIds);
        return $teamIris;
    }

    /**
     * @param File $data
     * {@inheritdoc}
     */
    public function denormalize($data, $class, $format = null, array $context = [])
    {
        throw new \Exception('Denormalizing form types are not supported with this method.');
    }

    /**
     * @param ModelClassMetadataFactory $modelClassMetadataFactory
     */
    public function setModelClassMetadataFactory($modelClassMetadataFactory)
    {
        $this->modelClassMetadataFactory = $modelClassMetadataFactory;
    }

    /**
     * @param IriConverterInterface $iriConverter
     */
    public function setIriConverter($iriConverter)
    {
        $this->iriConverter = $iriConverter;
    }

    /**
     * @param ItemNormalizer $itemNormalizer
     */
    public function setItemNormalizer($itemNormalizer)
    {
        $this->itemNormalizer = $itemNormalizer;
    }

    /**
     * @param ResourceCollectionInterface $resourceCollection
     */
    public function setResourceCollection($resourceCollection)
    {
        $this->resourceCollection = $resourceCollection;
    }

    /**
     * @param ResourceInterface $teamResource
     */
    public function setTeamResource($teamResource)
    {
        $this->teamResource = $teamResource;
    }

    /**
     * @param OwnershipProvider $ownershipProvider
     */
    public function setOwnershipProvider($ownershipProvider)
    {
        $this->ownershipProvider = $ownershipProvider;
    }


    /**
     * Get the value of container
     */
    public function getContainer()
    {
        return $this->container;
    }
    /**
     * Set the value of container
     *
     * @return  self
     */
    public function setContainer($container)
    {
        $this->container = $container;
        return $this;
    }
}
