<?php
/* Copyright start
  Copyright (C) 2008 - 2020 Fortinet Inc.
  All rights reserved.
  FORTINET CONFIDENTIAL & FORTINET PROPRIETARY SOURCE CODE
  Copyright end */

namespace Cybersponse\MetadataBundle\Command;

use Symfony\Component\Console\Input\InputOption;
use Symfony\Component\Console\Style\SymfonyStyle;
use Symfony\Component\Console\Input\InputInterface;
use Symfony\Component\Console\Output\OutputInterface;
use Symfony\Component\Console\Output\ConsoleOutputInterface;
use Symfony\Bundle\FrameworkBundle\Command\ContainerAwareCommand;
use CyberSponse\MetadataBundle\Entity\ClusterCommand;

class ClusterPublishCommand extends ContainerAwareCommand
{
    protected function configure()
    {
        $this
            ->setName('cybersponse:cluster:publish')
            ->setDescription('Updates the system by clearing cache, and generating models.')
            ->setDefinition([
                new InputOption(
                    'nodeId',
                    'nodeId',
                    InputOption::VALUE_OPTIONAL | InputOption::VALUE_IS_ARRAY,
                    'Arguments for self nodeId.'
                ),
                new InputOption(
                    'sequenceId',
                    'sId',
                    InputOption::VALUE_OPTIONAL | InputOption::VALUE_IS_ARRAY,
                    'Arguments for the command sequence id.'
                ),
            ]);
    }

    protected function execute(InputInterface $input, OutputInterface $output)
    {
        $io = new SymfonyStyle($input, $output);
        $errorIo = $io;
        if ($output instanceof ConsoleOutputInterface) {
            // If it's available, get stdErr output
            $errorOutput = $output->getErrorOutput();
            $errorIo = new SymfonyStyle($input, $errorOutput);
        }

        $globalOptions = '--env=' . $input->getOption('env');

        $nodeId = $input->getOption('nodeId');
        $sequenceId = $input->getOption('sequenceId');

        $defaultCommandData = [
            'runCondition' => true,
            'insertNewline' => false
        ];

        $commandsToRun = [
            'System Backup' => [
                'command' => 'cybersponse:backup:all'
            ],
            'Generating Models' => [
                'command' => 'cybersponse:metadata:generate:models'
            ],
            'Clearing Redis Cache' => [
                'command' => 'redis:flushdb --no-interaction'
            ],
            'Clearing Cache' => [
                'command' => 'cache:clear'
            ],
            'Cleaning Up Old Backups' => [
                'command' => 'cybersponse:cleanup:all',
            ],
        ];
        $errorCode = 0;

        try {
            //Run all commands
            $current = 0;

            foreach ($commandsToRun as $processingText => $commandData) {
                $commandData = $commandData + $defaultCommandData;

                $io->writeln(sprintf('<fg=blue>%s) %s</fg=blue>', $current + 1, $processingText));
                if ($commandData['runCondition'] === true) {
                    if ($commandData['insertNewline'] === true) {
                        $io->writeln('');
                    }
                    $command = $commandData['command'];
                    list($returnVal, $returnString) = $this->executeCommand($command, $globalOptions);
                    $io->writeln($returnString);

                    if ($returnVal !== 0) {
                        throw new \Exception(sprintf('Command `%s` finished with code: %s', $command, $returnVal));
                    }
                    $io->write(' -> <fg=green>Finished</fg=green>', true);
                } else {
                    $io->write(' -> <fg=yellow>Skipped</fg=yellow>', true);
                }

                $current++;
            }
        } catch (\Exception $e) {
            //Perform rollback
            $errorIo->error(sprintf('Failed: %s, Restoring from backup...', $e->getMessage()));
            $errorCode = 1;
            $commands = ['cybersponse:restore:all', 'doctrine:decrypt:database --no-interaction', 'doctrine:encrypt:database --no-interaction'];
            foreach($commands as $command){
                list($returnVal, $returnString) = $this->executeCommand($command, $globalOptions);
                if ($returnVal !== 0) {
                    //Todo What happens when the system is not restored properly?
                    throw new \Exception(sprintf('Error Code: %s, Could not restore system: %s', $command, $returnVal, $returnString));
                }
            }
            $io->success('Successfully restored up the system. Please view the above errors and fix.');
        } finally {
            if ($nodeId and $sequenceId){
                $container = $this->getContainer();
                $entityManager = $container->get('doctrine.orm.entity_manager');
                $clusterCommandRepository = $entityManager->getRepository(\Cybersponse\MetadataBundle\Entity\ClusterCommand::class);
                $clusterCommand = $clusterCommandRepository->findOneBy([
                    'nodeId' => $nodeId,
                    'sequenceId' => $sequenceId]);
                if ($clusterCommand && $errorCode == 0){
                    $entityManager->remove($clusterCommand);
                    $entityManager->flush();
                } else if ($clusterCommand){
                    $clusterCommand->setStatus('failed');
                    $entityManager->update($clusterCommand);
                    $entityManager->flush();
                }   
            }
            $io->success('Finished with system update!');
        }
        return $errorCode;
    }

    /**
     * @param $command
     * @param $globalOptions
     * @return array
     */
    private function executeCommand($command, $globalOptions, $indexOption = "")
    {
        $rootDir = $this->getContainer()->getParameter('kernel.root_dir');
        $fullCommand =sprintf('/usr/bin/php %s/console %s %s', $rootDir,$command, $globalOptions);
        $execOutput = null;
        exec($fullCommand, $execOutput, $returnVal);
        return array($returnVal, $execOutput);
    }
}
