<?php
/* Copyright start
  Copyright (C) 2008 - 2020 Fortinet Inc.
  All rights reserved.
  FORTINET CONFIDENTIAL & FORTINET PROPRIETARY SOURCE CODE
  Copyright end */

namespace Cybersponse\MetadataBundle\Command;

use Cybersponse\AuthorizationBundle\Traits\Ownable;
use Cybersponse\CoreBundle\Traits\Indexable;
use Cybersponse\CoreBundle\Traits\Trackable;
use Cybersponse\AuthorizationBundle\Traits\Taggable;
use Cybersponse\CoreBundle\Traits\PeerReplicable;

use Cybersponse\MetadataBundle\Entity\AttributeMetadata;
use Cybersponse\MetadataBundle\Entity\StagingModelMetadata;
use Cybersponse\MetadataBundle\Generation\Mapping\ORM\UuidNamingStrategy;
use Cybersponse\MetadataBundle\Mapping\ModelMetadataFactory;
use Cybersponse\MetadataBundle\Service\EntityGenerator;

use Doctrine\Common\Util\Inflector;
use Doctrine\ORM\EntityManager;
use Doctrine\ORM\EntityRepository;
use Doctrine\ORM\Mapping\ClassMetadataInfo;
use Dunglas\ApiBundle\Api\ResourceCollectionInterface;
use Symfony\Bridge\Doctrine\ManagerRegistry;
use Symfony\Bundle\FrameworkBundle\Command\ContainerAwareCommand;
use Symfony\Component\Console\Input\InputInterface;
use Symfony\Component\Console\Input\InputOption;
use Symfony\Component\Console\Output\OutputInterface;
use Symfony\Component\Console\Style\SymfonyStyle;
use Symfony\Component\Routing\RequestContext;
use Symfony\Component\Routing\RouterInterface;

/**
 * Class GenerateModelsCommand
 */
class GenerateModelsCommand extends ContainerAwareCommand
{

    /** @var StagingModelMetadata[] */
    protected $stagingModelMetadataMap;

    /** @var RouterInterface */
    protected $router;

    /** @var string */
    protected $customEntityDirectory;

    /** @var string */
    protected $customEntityNamespace;

    /** @var array */
    protected $typeToClassNameMap;

    /** @var UuidNamingStrategy */
    protected $uuidNamingStrategy;

    /** @var ResourceCollectionInterface */
    protected $resourceCollection;

    protected function configure()
    {
        $this
            ->setName('cybersponse:metadata:generate:models')
            ->setDescription('Process current model metadata & perform code generation')
            ->setDefinition([
                new InputOption(
                    'type',
                    't',
                    InputOption::VALUE_REQUIRED | InputOption::VALUE_IS_ARRAY,
                    'The type of metadata, activites, alerts, etc.'
                )
            ]);
    }

    protected function execute(InputInterface $input, OutputInterface $output)
    {
        $io = new SymfonyStyle($input, $output);

        $progress = $io->createProgressBar();

        $this->uuidNamingStrategy = new UuidNamingStrategy();

        $container = $this->getContainer();

        /** @var EntityGenerator $entityGenerator */
        $entityGenerator = $container->get('cybersponse.entity_generator');
        $entityGenerator->setBackupExisting(false);
        $entityGenerator->setGenerateAnnotations(true);
        $entityGenerator->setGenerateStubMethods(true);
        $entityGenerator->setRegenerateEntityIfExists(true);
        $entityGenerator->setUpdateEntityIfExists(true);
        $entityGenerator->setNumSpaces(4);
        $entityGenerator->setAnnotationPrefix('Doctrine\ORM\Mapping\\');
        $entityGenerator->setFieldVisibility(EntityGenerator::FIELD_VISIBLE_PROTECTED);

        $metadataEntityManager = $container->get('doctrine.orm.entity_manager');

        $this->router = $container->get('router');
        $requestContext = new RequestContext();
        $requestContext->setMethod('GET');
        $this->router->setContext($requestContext);

        $this->resourceCollection = $container->get('api.resource_collection');

        $sourcePath = sprintf('%s/../src', $container->getParameter('kernel.root_dir'));

        $this->customEntityNamespace = $container->getParameter('cybersponse_metadata.custom_types_namespace');

         /** @var EntityRepository $modelMetaRepository */
        $stagingModelMetadataRepository = $metadataEntityManager->getRepository(StagingModelMetadata::class);

        /** @var StagingModelMetadata[] $stagingMetadatas */
        $stagingMetadatas = $stagingModelMetadataRepository->findAll();

        $this->stagingModelMetadataMap = [];
        $this->typeToClassNameMap = $container->getParameter('cybersponse_metadata.type_map');

        /** @var ManagerRegistry $managerRegistry */
        $managerRegistry = $container->get('doctrine');

        //Create fake model metadata from existing entities in the system:
        foreach ($this->typeToClassNameMap as $type => $className) {
            /** @var EntityManager $manager */
            $manager = $managerRegistry->getManagerForClass($className);
            $doctrineClassMetadataInfo = $manager->getClassMetadata($className);

            $stagingModelMetadata = $this->getModelMetadataFromDoctrineClassMetadataInfo($type, $doctrineClassMetadataInfo);
            $this->stagingModelMetadataMap[$type] = $stagingModelMetadata;
        }

        foreach ($stagingMetadatas as $stagingData) {
            $type = $stagingData->getType();
            $classShortName = ModelMetadataFactory::getClassShortNameForModelMetadata($stagingData);
            $this->typeToClassNameMap[$type] = sprintf('%s\%s', $this->customEntityNamespace, $classShortName);

            $attributeMetadatas = [];
            foreach ($stagingData->getAttributes() as $attributeMetadata) {
                $attributeMetadatas[$attributeMetadata->getName()] = $attributeMetadata;
            }
            foreach($attributeMetadatas as $attributeMetadata){
                $stagingData->addAttribute($attributeMetadata);
            }

            $this->stagingModelMetadataMap[$type] = $stagingData;
        }

        $types = $input->getOption('type');

        /** @var StagingModelMetadata[] $modelMetadatas */
        if (count($types) > 0) {
            $queryBuilder = $stagingModelMetadataRepository->createQueryBuilder('o');
            $stagingMetadatas = $queryBuilder->andWhere('o.type in (:types)')->setParameter('types', $types)->getQuery()->execute();
        } else {
            $stagingMetadatas = $stagingModelMetadataRepository->findAll();
        }

        $progress->setFormat(' %current%/%max% [%bar%] Working on %message%...');
        $progress->setBarWidth(50);
        $progress->start(count($stagingMetadatas));

        foreach ($stagingMetadatas as $stagingData) {

            $progress->setMessage($stagingData->getDescriptions()['singular']);

            //Build doctrine class metadata info
            $doctrineClassMetadataInfo = $this->buildDoctrineMetadata($stagingData);
            $parentClass = $this->getParentClassFromModelMetadata($stagingData);

            $entityGenerator->setClassToExtend($parentClass);
            $entityGenerator->setFieldVisibility(EntityGenerator::FIELD_VISIBLE_PROTECTED);
            $entityGenerator->setModelMetadata($stagingData);

            // Write entity class to a file
            $entityGenerator->writeEntityClass($doctrineClassMetadataInfo, $sourcePath);
            $progress->advance();
        }
        $progress->finish();

        $io->success(sprintf('Wrote all files to %s by namespace', $sourcePath));
    }

    private function getAttributeType($stagingModelData, $columnName) {
        foreach ($stagingModelData->getAttributes() as $attribute) {
            if($attribute->getName() == $columnName) {
                return $attribute->getFormType();
            }
        }
    }

    private function buildDoctrineMetadata(StagingModelMetadata $stagingModelData)
    {
        $attributeType = $stagingModelData->getType();
        $entityClass = $this->typeToClassNameMap[$attributeType];
        $parentClasses = [];

        $parentType = $stagingModelData->getParentType();
        if (array_key_exists($parentType, $this->typeToClassNameMap)) {
            $parentClasses[] = $this->typeToClassNameMap[$parentType];
        }

        $classMetadataInfo = new ClassMetadataInfo($entityClass, $this->uuidNamingStrategy);
        $uniqeConstraint = $stagingModelData->getUniqueConstraint();
        if(count($uniqeConstraint)){
            $uniqeConstraint = $uniqeConstraint[0];
            $constraintName = array_keys($uniqeConstraint)[0];
        }

        $uniqeConst = array();
        foreach($uniqeConstraint as $uniqeConstraintName) {
            foreach($uniqeConstraintName['columns'] as $column) {
                if (in_array($this->getAttributeType($stagingModelData, $column), ['lookup', 'picklist'])) {
                    $uniqeConst[] = Inflector::tableize($column);
                } else {
                    $uniqeConst[] = $column;
                }
            }
            $uniqeConstraint[$constraintName]['columns'] = $uniqeConst;
        }

        if ($stagingModelData->isTrackable()) {
            //Can be change to generic by exposing same in UI.
            $modifyDateIndexName = $stagingModelData->getType() . '_search_idx';
            $indexList[$modifyDateIndexName]['columns'] = ["modifydate"];
        }
        if ($stagingModelData->isIndexable()) {
            //Can be change to generic by exposing same in UI.
            $idIndexName = $stagingModelData->getType() . '_id_search_idx';
            $indexList[$idIndexName]['columns'] = ["id"];
        }

        $classMetadataInfo->setPrimaryTable(['name' => sprintf('`%s`', $stagingModelData->getTableName()), 'uniqueConstraints' => $uniqeConstraint, 'indexes' => $indexList]);
        $classMetadataInfo->setCustomRepositoryClass('Doctrine\ORM\EntityRepository');
        $classMetadataInfo->setParentClasses($parentClasses);

        foreach ($stagingModelData->getAttributes() as $attribute) {
            $attributeType = $attribute->getType();

            if (array_key_exists($attributeType, $this->typeToClassNameMap)) {
                $relationshipMapping = [
                    'fieldName' => $attribute->getName(),
                    'isOwningSide' => $attribute->isOwnsRelationship(),
                    'targetEntity' => $this->typeToClassNameMap[$attributeType],
                    'cascade' => ['persist']
                ];

                /**
                 * The inverse side has to use the mappedBy attribute of the OneToOne, OneToMany, or ManyToMany mapping declaration.
                 * The mappedBy attribute contains the name of the association-field on the owning side.
                 * The owning side has to use the inversedBy attribute of the OneToOne, ManyToOne, or ManyToMany mapping declaration.
                 * The inversedBy attribute contains the name of the association-field on the inverse-side.
                 */

                //Determine relationship type automatically
                $inversedField = $attribute->getInversedField();
                $relatedModelMetadata = $this->stagingModelMetadataMap[$attributeType];

                //Default to manyTo
                $relationshipType = 'manyTo';

                if (!is_null($inversedField)) {
                    $relatedAttribs = $relatedModelMetadata->getAttributes()->getValues();
                    foreach($relatedAttribs as $relatedAttribute){
                        $relatedAttributes[$relatedAttribute->getName()] = $relatedAttribute;
                    }

                    if (array_key_exists($inversedField, $relatedAttributes)) {
                        $inversedFieldAttribute = $relatedAttributes[$inversedField];

                        $relationshipType = 'oneTo';

                        if ($inversedFieldAttribute->isCollection()) {
                            $relationshipType = 'manyTo';
                        }
                    }
                }

                if ($attribute->isCollection()) {
                    $relationshipType .= 'Many';
                } else {
                    $relationshipType .= 'One';
                }

                //TODO Might need to make this more functional
                if ($relationshipType == 'manyToMany') {
                    $relationshipMapping['fetch'] = ClassMetadataInfo::FETCH_EXTRA_LAZY;
                    $relationshipMapping['orphanRemoval'] = $attribute->getOrphanRemoval();
                    if ($attribute->getOrphanRemoval()) {
                        $relationshipMapping['cascade'] = ['persist', 'remove', 'detach'];
                    }

                    if (!empty($inversedField)) {
                        if ($attribute->isOwnsRelationship()) {
                            $relationshipMapping['inversedBy'] = $attribute->getInversedField();
                        } else {
                            $relationshipMapping['mappedBy'] = $attribute->getInversedField();
                        }
                    }
                    if ($attribute->getFormType() == "multiselectpicklist") {
                        $relationshipMapping['joinTable'] = [
                            'name' => Inflector::tableize($stagingModelData->getType()) . '_picklist_' . Inflector::tableize($attribute->getName())
                        ];
                    }
                } else if ($relationshipType == 'oneToMany') {
                    if (!empty($inversedField)) {
                        $relationshipMapping['mappedBy'] = $inversedField;
                    }
                    $relationshipMapping['orphanRemoval'] = $attribute->getOrphanRemoval();
                } else if ($relationshipType == 'manyToOne') {
                    if (!empty($inversedField)) {
                        $relationshipMapping['inversedBy'] = $inversedField;
                    }

                    $relationshipMapping['joinColumns'] = [
                        'joinColumn' => [
                            'name' => Inflector::tableize($attribute->getName()),
                            'onDelete' => 'SET NULL'
                        ]
                    ];
                } else if ($relationshipType == 'oneToOne') {
                    if (!empty($inversedField)) {
                        $relationshipMapping['inversedBy'] = $inversedField;
                    }
                }

                $method = 'map' . ucfirst($relationshipType);
                $classMetadataInfo->$method($relationshipMapping);
            } else {
                $length = null;
                $validation = $attribute->getValidation();
                if ($validation !== null) {
                    if (array_key_exists('maxlength', $validation)) {
                        $length = $validation['maxlength'];
                        if ($length < 0) {
                            $length = null;
                        }
                        else if ($length > 65535 && $attributeType=='string') {
                            $attributeType = 'text';
                        }
                    }
                }
                $fieldMapping = [
                    'fieldName' => $attribute->getName(),
                    'type' => $attributeType,
                    'nullable' => !$attribute->isIdentifier(),
                    'unique' => $attribute->isUnique(),
                    'length' => $length
                ];

                $formType = $attribute->getFormType();

                //Json primitives to doctrine based on formtype
                if ($formType == 'textarea') {
                    $fieldMapping['type'] = 'text';
                }
                if ($formType == 'string' || $formType == 'richtext' ||  $formType == 'html') {
                    if (is_null($attribute->getLength()) || $attribute->getLength() > 255 ||  !$length) {
                        $fieldMapping['type'] = 'text';
                    }
                }
                if ($attributeType == 'array' || $attributeType == 'object') {
                    $fieldMapping['type'] = 'jsonb';
                }
                if ($formType == 'datetime' || $formType == 'livesync') {
                    $fieldMapping['type'] = 'datetime';
                }
                if ($formType == 'decimal') {
                    $fieldMapping['type'] = 'float';
                    $fieldMapping['precision'] = 10;
                    $fieldMapping['scale'] = 2;
                }

                $classMetadataInfo->mapField($fieldMapping);
            }
        }

        return $classMetadataInfo;
    }

    private function getParentClassFromModelMetadata(StagingModelMetadata $modelMetadata)
    {
        $parentType = $modelMetadata->getParentType();
        return $this->typeToClassNameMap[$parentType];
    }


    /**
     * @param $type
     * @param $doctrineClassMetadataInfo
     * @return StagingModelMetadata
     */
    private function getModelMetadataFromDoctrineClassMetadataInfo($type, $doctrineClassMetadataInfo)
    {
        $traits = $doctrineClassMetadataInfo->getReflectionClass()->getTraitNames();

        $parentClass = $doctrineClassMetadataInfo->getReflectionClass()->getParentClass();

        $parentType = null;
        if (in_array($parentClass, $this->typeToClassNameMap)) {
            $parentType = array_search($parentClass, $this->typeToClassNameMap);
        }

        $stagingModelMetadata = new StagingModelMetadata();
        $stagingModelMetadata->setType($type);
        $stagingModelMetadata->setOwnable(array_key_exists(Ownable::class, $traits));
        $stagingModelMetadata->setTaggable(array_key_exists(Taggable::class, $traits));
        $stagingModelMetadata->setIndexable(array_key_exists(Indexable::class, $traits));
        $stagingModelMetadata->setTrackable(array_key_exists(Trackable::class, $traits));
        $stagingModelMetadata->setPeerReplicable(array_key_exists(PeerReplicable::class, $traits));
        $stagingModelMetadata->setTableName($doctrineClassMetadataInfo->getTableName());
        $stagingModelMetadata->setParentType($parentType);

        $attributeMetadatas = [];
        foreach ($doctrineClassMetadataInfo->getFieldNames() as $fieldName) {
            $attributeMetadata = new AttributeMetadata();
            $attributeMetadata->setCollection(false);
            $attributeMetadata->setName($fieldName);
            $attributeMetadatas[$fieldName] = $attributeMetadata;
        }

        foreach ($doctrineClassMetadataInfo->getAssociationMappings() as $assocationFieldName => $associationMapping) {
            $attributeMetadata = new AttributeMetadata();
            $attributeMetadata->setCollection(true);
            $attributeMetadata->setName($assocationFieldName);

            $inversedField = $associationMapping['isOwningSide'] ? $associationMapping['inversedBy'] : $associationMapping['mappedBy'];
            $attributeMetadata->setInversedField($inversedField);
            $attributeMetadata->setOwnsRelationship($associationMapping['isOwningSide']);

            $sourceEntity = $associationMapping['sourceEntity'];
            if (!is_null($sourceEntity) && ($attributeType = array_search($sourceEntity, $this->typeToClassNameMap)) !== false) {
                $attributeMetadata->setType($attributeType);
            }
            $attributeMetadatas[$assocationFieldName] = $attributeMetadata;
        }

        if($attributeMetadatas){
            foreach($attributeMetadatas as $attrMetaData){
                $stagingModelMetadata->addAttribute($attrMetaData);
            }
        }
        return $stagingModelMetadata;
    }
}
