<?php
/* Copyright start
  Copyright (C) 2008 - 2020 Fortinet Inc.
  All rights reserved.
  FORTINET CONFIDENTIAL & FORTINET PROPRIETARY SOURCE CODE
  Copyright end */

namespace Cybersponse\MetadataBundle\Command;

use Doctrine\ORM\EntityManager;
use Doctrine\ORM\EntityRepository;
use PhpAmqpLib\Message\AMQPMessage;
use Symfony\Component\Filesystem\Filesystem;
use Symfony\Component\Serializer\Serializer;
use GuzzleHttp\Psr7\Request as GuzzleRequest;
use PhpAmqpLib\Connection\AMQPStreamConnection;
use Symfony\Component\Console\Input\InputOption;
use Symfony\Component\Console\Style\SymfonyStyle;
use Symfony\Component\Console\Input\InputInterface;
use Cybersponse\MetadataBundle\Command\SystemUpdate;
use Cybersponse\MetadataBundle\Entity\ModelMetadata;
use Symfony\Component\Console\Output\OutputInterface;
use Symfony\Component\Serializer\Encoder\JsonEncoder;
use Cybersponse\MetadataBundle\Entity\StagingModelMetadata;
use Symfony\Component\Console\Output\ConsoleOutputInterface;
use Symfony\Component\Serializer\Normalizer\ObjectNormalizer;
use Symfony\Bundle\FrameworkBundle\Command\ContainerAwareCommand;

class SystemUpdateCommand extends ContainerAwareCommand
{
    const COMPONENT = 'component';
    const STATUS = 'status';
    const ENTITY_TYPE = 'entityType';
    const PUBLISH_CREATE = 'publish_create';
    const PUBLISH_MODIFY = 'publish_modify';
    const PUBLISH_DELETE = 'publish_delete';
    const ADDED_MODEL = 'created';
    const MODIFIED_MODEL='modified';
    const DELETED_MODEL='deleted';
    const META_INFO = 'metainfo';
    const DELTA_DATA = 'data';
    const TRANSACTION_DATE = 'transactionDate';
    const ACTION = 'action';
    const COMPAREARRAY = ['ownable','taggable','trackable','peerreplicable','name','type'];
    /** @var Filesystem */
    protected $filesystem;

    /** @var EntityManager */
    protected $entityManager;

    /**
     * @param EntityManager $entityManager
     */
    public function setEntityManager($entityManager)
    {
        $this->entityManager = $entityManager;
    }


    protected function configure()
    {
        $this
            ->setName('cybersponse:system:update')
            ->setDescription('Updates the system by clearing cache, generating models, and generating modules.')
            ->setDefinition([
                new InputOption(
                    'log-to-503',
                    'l',
                    InputOption::VALUE_NONE,
                    'Reports status to ./web/{clientName}/503.json file.'
                ),
                new InputOption(
                    'last-log',
                    'a',
                    InputOption::VALUE_NONE,
                    'Will remove the log file at the end of this command.'
                ),
                new InputOption(
                    'force',
                    'f',
                    InputOption::VALUE_NONE,
                    'Will force the regeneration regardless if there were no metadata updates.'
                ),
                new InputOption(
                    'index',
                    'i',
                    InputOption::VALUE_OPTIONAL | InputOption::VALUE_IS_ARRAY,
                    'Arguments for indexing modules in ES.'
                ),
                new InputOption(
                    'sync',
                    's',
                    InputOption::VALUE_OPTIONAL,
                    'Arguments for synchronous indexing of elasticsearch'
                ),
            ]);
        $this->filesystem = new Filesystem();
    }

    protected function execute(InputInterface $input, OutputInterface $output)
    {
        $io = new SymfonyStyle($input, $output);
        $errorIo = $io;
        $indexObject = [];
        if ($output instanceof ConsoleOutputInterface) {
            // If it's available, get stdErr output
            $errorOutput = $output->getErrorOutput();
            $errorIo = new SymfonyStyle($input, $errorOutput);
        }

        $globalOptions = '--env=' . $input->getOption('env');

        if ($input->getOption('no-ansi') === true) {
            $globalOptions .= ' --no-ansi';
        } else {
            $globalOptions .= ' --ansi';
        }

        $modelMetadatasWithKeys = $this->getAllMetadata(ModelMetadata::class);
        $stagingModelMetadatasWithKeys = $this->getAllMetadata(StagingModelMetadata::class);

        $encoder = new JsonEncoder();
        $normalizer = new ObjectNormalizer();

        $normalizer->setCircularReferenceHandler(function ($modelMetadatasWithKeys) {
            return $modelMetadatasWithKeys->getUuid();
        });

        $serializer = new Serializer([$normalizer], [$encoder]);

        $normalizedMetadata = $serializer->serialize($modelMetadatasWithKeys, 'json');
        $normalizedMetadata = json_decode($normalizedMetadata, true);

        $normalizer->setCircularReferenceHandler(function ($stagingModelMetadatasWithKeys) {
            return $stagingModelMetadatasWithKeys->getUuid();
        });

        $normalizedStagingMetadata = $serializer->serialize($stagingModelMetadatasWithKeys, 'json');
        $normalizedStagingMetadata = json_decode($normalizedStagingMetadata, true);
        $metadataUpdatesRequired = true === $input->getOption('force');

        if ($normalizedMetadata !== $normalizedStagingMetadata) {
            $metadataUpdatesRequired = true;
        }
        
        $indexObjectAll = $this->getIndexOptions($normalizedStagingMetadata, $normalizedMetadata);
        $indexObject = $indexObjectAll[self::ADDED_MODEL];
        $indexUpdateObject = $indexObjectAll[self::MODIFIED_MODEL];
        $indexMetaInfo = $indexObjectAll[self::META_INFO];

        $defaultCommandData = [
            'runCondition' => true,
            'insertNewline' => false
        ];

        $commandsToRun = [
            'Decrypt Database' => [
                'command' => 'doctrine:decrypt:database --no-interaction'
            ],
            'System Backup' => [
                'command' => 'cybersponse:backup:all'
            ],
            'Generating Models' => [
                'command' => 'cybersponse:metadata:generate:models'
            ],
            'Clearing Redis Cache' => [
                'command' => 'redis:flushdb --no-interaction'
            ],
            'Clearing Cache' => [
                'command' => 'cache:clear'
            ],
            'Schema Update' => [
                'command' => 'doctrine:schema:update --em=default --force',
                'runCondition' => $metadataUpdatesRequired
            ],
            'Activating Latest ModelMetadata' => [
                'command' => 'cybersponse:metadata:activate',
                'runCondition' => $metadataUpdatesRequired
            ],
            'Generating Modules' => [
                'command' => 'cybersponse:metadata:generate:modules',
                'runCondition' => $metadataUpdatesRequired
            ],
            'Cleaning Up Old Backups' => [
                'command' => 'cybersponse:cleanup:all',
            ],
            'Encrypt Database' => [
                'command' => 'doctrine:encrypt:database --no-interaction'
            ],
            'Wait For Update Across Cluster' => [
                'command' => 'cybersponse:cluster:publish:wait'
            ]
        ];
        $errorCode = 0;

        try {
            //Run all commands
            $total = count($commandsToRun);
            $current = 0;
            $systemUpdate = new SystemUpdate();
            $systemUpdate
                ->setMessage('Starting system update.')
                ->setProgressPercent(0);

            $this->log($input, $systemUpdate);
            # TODO: if return is not success, do not proceed
            $this->broadcastLockUnlock('cluster_publish_lock', 'cybersponse:publish:lock');
            foreach ($commandsToRun as $processingText => $commandData) {
                $commandData = $commandData + $defaultCommandData;

                $io->writeln(sprintf('<fg=blue>%s) %s</fg=blue>', $current + 1, $processingText));
                $systemUpdate
                    ->setMessage($processingText)
                    ->setProgressPercent((int) (($current / $total) * 100));
                $this->log($input, $systemUpdate);

                if ($commandData['runCondition'] === true) {
                    if ($commandData['insertNewline'] === true) {
                        $io->writeln('');
                    }
                    $command = $commandData['command'];
                    list($returnVal, $returnString) = $this->executeCommand($command, $globalOptions);
                    $io->writeln($returnString);

                    if ($returnVal !== 0) {
                        throw new \Exception(sprintf('Command `%s` finished with code: %s', $command, $returnVal));
                    }
                    $io->write(' -> <fg=green>Finished</fg=green>', true);
                } else {
                    $io->write(' -> <fg=yellow>Skipped</fg=yellow>', true);
                }

                $current++;
            }
            $this->log($input, $systemUpdate);
            $this->broadcastLockUnlock('cluster_publish_unlock', 'cybersponse:publish:unlock');
            $body = $this->prepareChangeData('Success');
            $this->notifyAuditingComponent([$body]);
        } catch (\Exception $e) {
            $this->broadcastLockUnlock('cluster_publish_unlock', 'cybersponse:publish:unlock');
            $error = [];
            $error['error'] = $e->getMessage();
            $body = $this->prepareChangeData('Failure', $error);
            $this->notifyAuditingComponent([$body]);
            //Perform rollback
            $errorIo->error(sprintf('Failed: %s, Restoring from backup...', $e->getMessage()));
            $errorCode = 1;
            $commands = ['cybersponse:restore:all', 'doctrine:decrypt:database --no-interaction', 'doctrine:encrypt:database --no-interaction'];
            foreach ($commands as $command) {
                list($returnVal, $returnString) = $this->executeCommand($command, $globalOptions);
                if ($returnVal !== 0) {
                    //Todo What happens when the system is not restored properly?
                    throw new \Exception(sprintf('Error Code: %s, Could not restore system: %s', $command, $returnVal, $returnString));
                }
            }
            $io->success('Successfully restored up the system. Please view the above errors and fix.');
        } finally {
            $this->logFinish($input);
            if ($errorCode == 0) {
                $indexName = $input->getOption('index');
                $indexOption = "";
                $typeModule = (object) null;
                if (count($indexName) > 0) {
                    $indexName = json_decode($indexName[0], true);
                    $indexObject = array_merge($indexObject, $indexName['type']);
                    $typeModule = ['type' => $indexObject];
                    $indexOption = " --index='" . json_encode($typeModule) . "'";
                } elseif (count($indexObject)) {
                    $typeModule = ['type' => $indexObject,'meta' => $indexMetaInfo];
                    $indexOption = " --index='" . json_encode($typeModule) . "'";
                }

                if ($input->getOption('sync')) {
                    $indexOption .= " --sync=" . $input->getOption('sync');
                }
                $typeModule = (object) null;
                if (count($indexUpdateObject)) {
                    $typeModule = ['type' => $indexUpdateObject];
                    $updateOption = " --update='" . json_encode($typeModule) . "'";
                }
                if (count($indexObject)|| count($indexUpdateObject)) {
                    $this->executeCommand('cybersponse:elastic:create', $globalOptions, $indexOption, $updateOption);
                }
            }
            $io->success('Finished with system update!');
        }

        return $errorCode;
    }

    private function broadcastLockUnlock($commandId, $command)
    {
        $dasUri = $this->getContainer()->getParameter('das.uri');
        $proxyProvider = $this->getContainer()->get('cybersponse.core.provider.proxy');
        $dasClient = $proxyProvider->getDasClient();
        $objectData = array(
            "commandId" => $commandId,
            "command" => "sudo -u nginx php /opt/cyops-api/app/console " . $command . " --env=prod",
            "includeFiredrillNodes" => false
        );
        $headers = array(
            "Content-type" => "application/json"
        );
        $dasRequests = new GuzzleRequest('post', $dasUri . '/clustercommand/', $headers, json_encode($objectData));
        try {
            $dasResponse = $dasClient->send($dasRequests);
            $resCode = $dasResponse->getStatusCode();
            if ($resCode == 200) {
                return 0;
            }
        } catch (\Exception $error) {
            error_log($error->getMessage());
        }
        return -1;
    }

    /**
     * @param $class
     * @return array
     */
    private function getAllMetadata($class)
    {
        $metadataDocumentManager = $this->getContainer()->get('doctrine.orm.entity_manager');

        /** @var EntityRepository $modelMetadataRepository */
        $modelMetadataRepository = $metadataDocumentManager->getRepository($class);

        /** @var ModelMetadata[] $modelMetadatas */
        $modelMetadatas = $modelMetadataRepository->findAll();
        $modelMetadatasWithKeys = [];

        foreach ($modelMetadatas as $modelMetadata) {
            $modelMetadatasWithKeys[$modelMetadata->getType()] = $modelMetadata;
        }
        return $modelMetadatasWithKeys;
    }

    /**
     * @param InputInterface $input
     * @param SystemUpdate $systemUpdate
     */
    protected function log(InputInterface $input, SystemUpdate $systemUpdate)
    {
        $logTo503 = true === $input->getOption('log-to-503');
        $logDir = $this->getLogDir();

        if ($logTo503) {
            $this->filesystem->dumpFile(sprintf('%s/503.json', $logDir), json_encode($systemUpdate));
        }
    }

    private function getLogDir()
    {
        $container = $this->getContainer();

        $rootDir = $container->getParameter('kernel.root_dir');
        return sprintf('%s/../web', $rootDir);
    }

    /**
     * @param $command
     * @param $globalOptions
     * @return array
     */
    private function executeCommand($command, $globalOptions, $indexOption = "", $updateOption="")
    {
        if ($indexOption != "") {
            $globalOptions .= $indexOption;
        }
        if ($updateOption != "") {
            $globalOptions .=$updateOption;
        }
        $rootDir = $this->getContainer()->getParameter('kernel.root_dir');
        $fullCommand = sprintf('/usr/bin/php %s/console %s %s', $rootDir, $command, $globalOptions);
        $execOutput = null;
        exec($fullCommand, $execOutput, $returnVal);
        return array($returnVal, $execOutput);
    }


    /**
     * @param $status
     * @param array $error
     * @return array
     */
    protected function prepareChangeData($status, $error = [])
    {
        $error['status'] = $status;
        $error['message'] = (object) null;
        return [
            self::COMPONENT => 'crudhub',
            self::ENTITY_TYPE => 'ModelMetadata',
            self::DELTA_DATA => $error,
            self::TRANSACTION_DATE => (int) ($_SERVER['REQUEST_TIME_FLOAT'] * 1000),
            self::ACTION => 'STATUS',
        ];
    }

    /**
     * @param $body
     * @return bool
     */
    protected function notifyAuditingComponent($body)
    {
        $mqCredObj = parse_ini_file('/opt/cyops/configs/rabbitmq/rabbitmq_users.conf', false, INI_SCANNER_NORMAL);
        $rabbitMqUsername = $mqCredObj['cyops.rabbitmq.user'] ? $mqCredObj['cyops.rabbitmq.user'] : 'guest';
        $rabbitMqPassword = $mqCredObj['cyops.rabbitmq.password'] ? $mqCredObj['cyops.rabbitmq.password'] : 'guest';
        $rabbitMqHost = $mqCredObj['cyops.rabbitmq.host'] ? $mqCredObj['cyops.rabbitmq.host'] : 'localhost';
        $rabbitMqPort = $mqCredObj['cyops.rabbitmq.port'] ? $mqCredObj['cyops.rabbitmq.port'] : '5672';
        $rabbitMqVHost = $mqCredObj['cyops.rabbitmq.vhost'] ? $mqCredObj['cyops.rabbitmq.vhost'] : '/';

        $connection = new AMQPStreamConnection($rabbitMqHost, $rabbitMqPort, $rabbitMqUsername, $rabbitMqPassword, $rabbitMqVHost);
        $channel = $connection->channel();
        $channel->exchange_declare('cyops.crudhub.datanotify', 'topic', false, true, false);

        $msg = new AMQPMessage(json_encode($body), array('content_type' => 'application/json'));
        $channel->basic_publish($msg, 'cyops.crudhub.datanotify', 'cyops.crudhub.datanotify');
        $channel->basic_publish($msg, 'cyops.crudhub.datanotify', 'route.crudhub.instructions.remoterequest.modelmmdupdate');
        $channel->close();
        $connection->close();
    }

    /**
     * Finishes the log
     *
     * @param InputInterface $input
     */
    protected function logFinish(InputInterface $input)
    {
        $logTo503 = true === $input->getOption('log-to-503');
        $lastLog = true === $input->getOption('last-log');
        $logDir = $this->getLogDir();

        if ($logTo503 && $lastLog) {
            $this->filesystem->remove(sprintf('%s/503.json', $logDir));
        }
    }
    
    /** @return object
     * {
     * "created":[],
     * "deleted":[],
     * "modified":{
     *      "tasks":{
     *              "added":{"attributes":[{"newAttribute":{"name":"newAttribute","type":"string"}}]}
     *              },
     *      "indicators":{
     *              "modified":{"trackable":false,"taggable":false},
     *              "old":{"trackable":true,"taggable":true},
     *              "added":{"attributes":[{"newIndicatorField":{"name":"newIndicatorField","type":"string"}}]},
     *              "removed":{"attributes":[{"firstSeen":{"name":"firstSeen","type":"integer"}}]}
     *       }
     *  }
     *} */
    protected function getIndexOptions($stagingMetadata, $modelMetadata)
    {
        //use PHP in-built function to detect added/deleted/existing models
        $newModels = array_diff_key($stagingMetadata, $modelMetadata);
        $deletedModels = array_diff_key($modelMetadata, $stagingMetadata);
        $existingModels = array_intersect_key($stagingMetadata, $modelMetadata);
        $modelDiff=[];
        $modelDiff[self::ADDED_MODEL] = $newModels;
        $modelDiff[self::DELETED_MODEL] = $deletedModels ;

        //loop through data to find diff using != operator on whole model.will be looping below on each model.
        foreach ($existingModels as $key => $value) {
            $stagingValues = array_filter($stagingMetadata[$key], function ($varValue, $varKey) use ($stagingMetadata,$modelMetadata,$key) {
                if (\in_array($varKey, self::COMPAREARRAY) and $stagingMetadata[$key][$varKey] != $modelMetadata[$key][$varKey]) {
                    return true;
                } else {
                    return false;
                }
            }, ARRAY_FILTER_USE_BOTH);
            
            $stagingAttribute = [];
            $modelAttribute= [];
            $attrib =[];
            $attrib['added'] = [];
            $attrib['deleted'] = [];
            $attrib['modified'] = [];
           
            foreach ($stagingMetadata[$key]['attributes'] as $attribKey => $attribValue) {
                $stagingAttribute[$attribValue['name']]= ['name'=> $attribValue['name'],'type'=> $attribValue['type'],'formType'=> $attribValue['formType'],'encrypted'=> $attribValue['encrypted']];
            }
            foreach ($modelMetadata[$key]['attributes'] as $attribKey => $attribValue) {
                $modelAttribute[$attribValue['name']]= ['name'=> $attribValue['name'],'type'=> $attribValue['type'],'formType'=> $attribValue['formType'],'encrypted'=> $attribValue['encrypted']];
            }
        
            //array_multisort($stagingAttribute, $modelAttribute);
            $diffVarHolder = array_diff_key($stagingAttribute, $modelAttribute);
            if ($diffVarHolder) {
                $attrib['added'] = $diffVarHolder;
            }
            $diffVarHolder =  array_diff_key($modelAttribute, $stagingAttribute);
            if ($diffVarHolder) {
                $attrib['deleted'] = $diffVarHolder;
            }
            $existingModelsAttrib = array_intersect_key($modelAttribute, $stagingAttribute);
            foreach ($existingModelsAttrib as $existingModelsAttribkey => $existingModelsAttribvalue) {
                if ($stagingAttribute[$existingModelsAttribkey] !== $modelAttribute[$existingModelsAttribkey]) {
                    array_push($attrib['modified'], $stagingAttribute[$existingModelsAttribkey]);
                }
            }
            //check if we have any difference in mmd on exising model
            if (count($stagingValues) > 0) {
                $modelDiff[self::MODIFIED_MODEL][$key] =[];
                $modelDiff[self::MODIFIED_MODEL][$key]['modified']= $stagingValues ;
                foreach ($stagingValues  as $stagingKey => $satagingValue) {
                    $modelDiff[self::MODIFIED_MODEL][$key]['old'][$stagingKey] = $modelMetadata[$key][$stagingKey];
                }
            }
            if (count($attrib['modified'])>0) {
                if (!isset($modelDiff[self::MODIFIED_MODEL][$key])) {
                    $modelDiff[self::MODIFIED_MODEL][$key] =[];
                }
                $modelDiff[self::MODIFIED_MODEL][$key]['modified']['attributes']= $attrib['modified'] ;
                $modelDiff[self::MODIFIED_MODEL][$key]['old']['attributes']=[];
                foreach ($attrib['modified'] as $attribKey => $attribValue) {
                    array_push($modelDiff[self::MODIFIED_MODEL][$key]['old']['attributes'], $modelAttribute[$attribValue['name']]);
                }
            }
            if ($attrib['added']) {
                $modelDiff[self::MODIFIED_MODEL][$key]['added']['attributes']= $attrib['added'];
            }
            if ($attrib['deleted']) {
                $modelDiff[self::MODIFIED_MODEL][$key]['removed']['attributes']= $attrib['deleted'];
            }
        }

        $indexNames = [];
        if ($modelDiff[self::ADDED_MODEL] and array_keys($modelDiff[self::ADDED_MODEL])) {
            $indexNames[self::ADDED_MODEL] = array_keys($modelDiff[self::ADDED_MODEL]);
        } else {
            $indexNames[self::ADDED_MODEL]=[];
        }
        if (isset($modelDiff[self::DELETED_MODEL]) and array_keys($modelDiff[self::DELETED_MODEL])) {
            $indexNames[self::DELETED_MODEL] = array_keys($modelDiff[self::DELETED_MODEL]);
        } else {
            $indexNames[self::DELETED_MODEL] =[];
        }
        if (isset($modelDiff[self::MODIFIED_MODEL])) {
            $indexNames[self::MODIFIED_MODEL] =  $modelDiff[self::MODIFIED_MODEL];
            $getModifiedObject=[];
            // check of modified attrubutes if found send data to re index . only send --update the model if attributes are added .we can add model if attribute is dropped also
            foreach ($modelDiff[self::MODIFIED_MODEL] as $key => $value) {
                if (isset($value['modified']['attributes']) && count($value['modified']['attributes']) > 0) {
                    array_push($indexNames[self::ADDED_MODEL], $key);
                    $getModifiedObject[$key] = $value;
                    unset($indexNames[self::MODIFIED_MODEL][$key]);
                }
            }
            $indexNames[self::META_INFO] = $getModifiedObject;
        } else {
            $indexNames[self::MODIFIED_MODEL]=[];
        }
        return $indexNames;
    }
}
