<?php
/* Copyright start
  Copyright (C) 2008 - 2020 Fortinet Inc.
  All rights reserved.
  FORTINET CONFIDENTIAL & FORTINET PROPRIETARY SOURCE CODE
  Copyright end */

namespace Cybersponse\MetadataBundle\Generation\Loader;


use Cybersponse\MetadataBundle\Generation\Annotations\Classes\ClassAnnotationGeneratorInterface;
use Cybersponse\MetadataBundle\Generation\Annotations\Fields\FieldAnnotationGeneratorInterface;
use Cybersponse\MetadataBundle\Generation\Metadata\ModelAttributeMetadata;
use Cybersponse\MetadataBundle\Generation\Traits\TraitGeneratorInterface;
use Doctrine\Common\Annotations\Reader;
use Doctrine\ORM\Mapping\Id;
use Dunglas\ApiBundle\Mapping\ClassMetadata;
use Dunglas\ApiBundle\Mapping\Loader\AttributesLoader;
use Dunglas\ApiBundle\Util\ReflectionTrait;

class MetadataAnnotationLoader extends AttributesLoader
{
    use ReflectionTrait;

    /** @var ClassAnnotationGeneratorInterface[] */
    protected $classAnnotationGenerators;

    /** @var FieldAnnotationGeneratorInterface[] */
    protected $fieldAnnotationGenerators;

    /** @var TraitGeneratorInterface[] */
    protected $traitGenerators;

    /**
     * @var Reader
     */
    private $reader;

    public function __construct(Reader $reader)
    {
        $this->reader = $reader;
    }

    /**
     * @param ClassAnnotationGeneratorInterface[] $classAnnotationGenerators
     */
    public function setClassAnnotationGenerators($classAnnotationGenerators)
    {
        $this->classAnnotationGenerators = $classAnnotationGenerators;
    }

    /**
     * @param FieldAnnotationGeneratorInterface[] $fieldAnnotationGenerators
     */
    public function setFieldAnnotationGenerators($fieldAnnotationGenerators)
    {
        $this->fieldAnnotationGenerators = $fieldAnnotationGenerators;
    }

    /**
     * @param TraitGeneratorInterface[] $traitGenerators
     */
    public function setTraitGenerators($traitGenerators)
    {
        $this->traitGenerators = $traitGenerators;
    }

    /**
     * Loads metadata for the given class.
     *
     * @param ClassMetadata $classMetadata
     * @param string[]|null $normalizationGroups
     * @param string[]|null $denormalizationGroups
     * @param string[]|null $validationGroups
     *
     * @return bool
     */
    public function loadClassMetadata(
        ClassMetadata $classMetadata,
        array $normalizationGroups = null,
        array $denormalizationGroups = null,
        array $validationGroups = null
    ) {
        $this->loadFieldAnnotations($classMetadata);
        $this->loadClassAnnotations($classMetadata);
        $this->loadTraits($classMetadata);
        return true;
    }

    /**
     * @param ClassMetadata $classMetadata
     * @return array
     */
    private function loadClassAnnotations(ClassMetadata &$classMetadata)
    {
        $reflectionClass = $classMetadata->getReflectionClass();

        // Class annotation loading
        foreach ($this->classAnnotationGenerators as $classAnnotationGenerator) {
            // Read annotations for this property
            if ($annotation = $this->reader->getClassAnnotation($reflectionClass, $classAnnotationGenerator->getAnnotationClassName())) {
                $classAnnotationGenerator->loadClassMetadata($classMetadata, $annotation);
            }
        }
    }

    /**
     * @param ClassMetadata $classMetadata
     */
    private function loadFieldAnnotations(ClassMetadata $classMetadata)
    {
        $reflectionClass = $classMetadata->getReflectionClass();

        // Property annotation loading
        foreach ($classMetadata->getAttributes() as $attributeMetadata) {

            $attributeName = $attributeMetadata->getName();

            if (!$reflectionClass->hasProperty($attributeName)) {
                // Delete all non reflection properties
                $classMetadata->removeAttribute($attributeMetadata);
                continue;
            }

            $reflectionProperty = $reflectionClass->getProperty($attributeName);

            // Rebuild attribute metadata if it exists as a property
            $modelAttributeMetadata = new ModelAttributeMetadata($attributeMetadata);

            foreach ($modelAttributeMetadata->getTypes() as $type) {
                // Objects aren't represented in Nelmio correctly so we change it to a string
                if ($type->getType() == 'object' && is_null($type->getClass())) {
                    $type->setType('string');
                }
            }

            //Only let doctrine primary ID's be identifiers
            $modelAttributeMetadata->setIdentifier(false);
            if ($this->reader->getPropertyAnnotation($reflectionProperty, Id::class)) {
                $modelAttributeMetadata->setIdentifier(true);
            }

            foreach ($this->fieldAnnotationGenerators as $fieldAnnotationGenerator) {
                // Read annotations for this property
                if ($annotation = $this->reader->getPropertyAnnotation($reflectionProperty, $fieldAnnotationGenerator->getAnnotationClassName())) {
                    $fieldAnnotationGenerator->loadAttributeMetadata($modelAttributeMetadata, $annotation);
                }
            }

            // Add reflection properties back
            $classMetadata->addAttribute($modelAttributeMetadata);
        }
    }

    /**
     * @param ClassMetadata $classMetadata
     * @return array
     */
    private function loadTraits(ClassMetadata &$classMetadata)
    {
        // Class annotation loading
        foreach ($this->traitGenerators as $traitGenerator) {
            // Read annotations for this property
            if (in_array($traitGenerator->getTraitClassName(), class_uses($classMetadata->getName()))) {
                $traitGenerator->loadClassMetadata($classMetadata, null);
            }
        }
    }
}
