#### The Model Metadata ####
`Model Metadata` describes every piece of customizable data models in our system. We have two pieces of metadata: `Model` and `Attribute`. `Models` are comprised of any number of `Attributes`. Attributes can be of any type, ranging from any json primitive type, to relationships with existing entities. These relationships can be of type toMany or toOne.

### Changelog
Date | Name
--- | ---
09/08/2016 | Aaron Merryman

### ModelMetadata: `/api/3/apidoc/ModelMetadata` ###
* Type [`type`] __(required)__
    * Metadata type, json primitive for attributes or internal types for models. See `type` for `AttributeMetadata` for valid json primitives.
* Parent Type [`parentType`]
    * The type of the parent, if set to `null`, will default to the `BaseEntity` class.
* Table Name [`tableName`] __(required)__
    * The table name for ORM, this will be the table that is created in the database.
* Ownable [`ownable`]
    * Determines if this class should have the ownable fields (`owners`)
* Trackable [`trackable`]
    * Determines if this class should have the trackable fields (`createDate`, `createUser`, `modifyDate`, `modifyUser`)
* Indexable [`indexable`]
    * Determines if this class should have the indexable fields (`id`). This gives the object a global sequence id field.
* Attributes [`attributes`]
    * Key-value list of the attributes for this class. These attributes can be described here: `/api/3/apidoc/AttributeMetadata`
* Display Name [`displayName`]
    * Display name as an AngularJS template
* Descriptions [`descriptions`] __(required)__
    * Map of description type => description, example:
    ``` json
    {
        "singular" : "Alert",
        "plural" : "Alerts"
    }
    ```
    
### AttributeMetadata: `/api/3/apidoc/AttributeMetadata` ###
* Id [`id`]
    * The unique object id for this attribute, will be auto-generated if not specified
* Type [`type`] __(required)__
    * Metadata type, json primitive for attributes or iri/internal types for models. Supported primitives are:
    	`integer`,
    	`string`,
    	`boolean`,
    	`number`,
    	`null`,
    	`object`,
    	`array`
    * `object` and `array` are special types. See the Extended Data section below.
    * Note: `null` is supported in theory but may not make sense to use.
* Name [`name`] __(required, unique)__
    * The name of the attribute
    * Note: DO NOT use underscores in the field name. Must be Camel-Case.
* Length [`length`]
    * The length of the attribute. For anything less than 255, it will created in the SQL database as a VARCHAR with defined length.
* Form Type [`formType`]
    * The type of form to use to render this attribute. Supported types are:
    	* `integer`
    	* `text`
    	* `textarea`
    	* `checkbox`
    	* `manyToMany`
    	* `datetime`
    	* `picklist`
    	* `lookup`
    	* `oneToMany`
    	* ~~`hidden`~~ `(deprecated >=4.2, see visibility=false)`
    	* `file`
    	* `manyToOne`
    	* `richtext`
    	* `phone`
    	* `email`
    	* `decimal`
* Date Source [`dataSource`]
    * The source of the data, not defined explicitly. An example for picklist lookups:
    ``` json
    {
        "module" : "picklists",
        "query": {
            "logic": "AND",
            "filters": [
                {
                  "field": "listName__name",
                  "operator": "eq",
                  "value": "Criticality"
                }
            ],
            "sort": [
                {
                  "field": "orderIndex",
                  "direction": "ASC"
                }
            ]
        }
    }
    ```
    * An example for a relationships for all items in a collection, `query` is null because no query should be made to limit the lookup:
    ``` json
    {
        "module" : "alerts",
        "query" : null
    }
    ```
* Searchable [`searchable`]
    * Determines if this attribute should be searchable within a grid. This does not affect ElasticSearch.
* Grid Column [`gridColumn`]
    * Determines if this attribute should show up in a grid by default.
* Inversed Field [`inversedField`] __(validation enforced)__
    * Field name on the referenced type, if is collection
* Owns Relationship [`ownsRelationship`] __(validation enforced)__
    * Whether this model owns the relationships, if is collection
* Validation [`validation`]
    * Validation object, a list of validation attributes with boolean values or condition sets. Examples:
    ``` json
    { "required" : true }
    ```
    ``` json
    { "required" : false }
    ```
    * Example of conditional validation: Only require `resolved` if `status` equals `Resolved`: 
    ``` json
    {
        "required": [
            {
                "logic": "AND",
                "result": true,
                "filters": [
                    {
                        "field": "status",
                        "operator": "eq",
                        "value": "/api/3/picklists/97e9bc03-5a4c-43c6-b3a3-47422b42d288"
                    }
                ]
            }
        ]
    }
    ```

* Visibility [`visibility`]
    * Determines if this attribute should be visible during data entry.
    * This can be the following:
        * `true`, the default, will always be shown.
        * `false`, will not show up in the application.
        * Query object, see above, will be visible if the condition matches.
* Readable [`readable`]
    * Determines if this attribute should be returned during READ operations
* Writeable [`writeable`]
    * Determines if this attribute should be modified during UPDATE operations
* Identifier [`identifier`]
    * Determines if this attribute is a unique identifier for this class
* Unique [`unique`]
    * Determines if the attribute is unique
* Display Name [`displayName`]
    * Display name as a template using field names, example:
    ``` json
    {
        "displayName" : "{{ sourceId }}"
    }
    ```
* Descriptions [`descriptions`] __(required)__
    * Map of description type => description, example:
    ``` json
    {
        "singular" : "Source ID"
    }
    ```
    * `plural` may also be defined if needed, although it is currently not used in the system.

#### Examples ####
For the following examples, we will use `alerts`.

#### Viewing Model Metadata ####
To view a model, use `GET` on `/api/3/model_metadatas/7ab10952-b2ea-4910-9762-eb685bdd29eb`. You should receive a result like this:
``` json
{
  "@context": "/api/3/contexts/ModelMetadata",
  "@id": "/api/3/model_metadatas/7ab10952-b2ea-4910-9762-eb685bdd29eb",
  "@type": "ModelMetadata",
  "type": "alerts",
  "parentType": null,
  "tableName": "alerts",
  "ownable": true,
  "trackable": true,
  "indexable": true,
  "displayName": "{{ name }}",
  "descriptions": {
    "singular": "Alert",
    "plural": "Alerts"
  }
}
```

To view a model without defaults returned, use `GET` on `/api/3/model_metadatas/7ab10952-b2ea-4910-9762-eb685bdd29eb?$removeDefaults=true`. You should receive a result like this:
``` json
{
  "@context": "/api/3/contexts/ModelMetadata",
  "@id": "/api/3/model_metadatas/7ab10952-b2ea-4910-9762-eb685bdd29eb",
  "@type": "ModelMetadata",
  "type": "alerts",
  "tableName": "alerts",
  "displayName": "{{ name }}",
  "descriptions": {
    "singular": "Alert",
    "plural": "Alerts"
  }
}
```

##### Model Attributes ####
All models will have attributes to describe the fields on the json. These can be retrieved by using:

`GET /api/3/model_metadatas/7ab10952-b2ea-4910-9762-eb685bdd29eb?$relationships=true`
``` json
{
  "@context": "/api/3/contexts/ModelMetadata",
  "@id": "/api/3/model_metadatas/7ab10952-b2ea-4910-9762-eb685bdd29eb",
  "@type": "ModelMetadata",
  "type": "alerts",
  "parentType": null,
  "tableName": "alerts",
  "ownable": true,
  "trackable": true,
  "indexable": true,
  "attributes": [
    {
      "@id": "/api/3/attribute_metadatas/16c855d3-29b4-4d29-90a8-8f322b556067",
      "@type": "AttributeMetadata",
      "type": "string",
      "name": "name",
      "length": 255,
      "formType": "text",
      "dataSource": null,
      "searchable": true,
      "gridColumn": true,
      "collection": false,
      "inversedField": null,
      "ownsRelationship": false,
      "validation": {
        "required": true
      },
      "visibility": true,
      "readable": true,
      "writeable": true,
      "identifier": false,
      "unique": false,
      "displayName": "{{ name }}",
      "descriptions": {
        "singular": "Name"
      }
    },
    ...
  ]
}
```

#### Exporting a Model ####

To export a model, use `GET` on `/api/3/model_metadatas/7ab10952-b2ea-4910-9762-eb685bdd29eb?$relationships=true&$removeDefaults=true&$export=true`. You should receive a result like this:
``` json
{
  "@type": "ModelMetadata",
  "uuid": "7ab10952-b2ea-4910-9762-eb685bdd29eb",
  "type": "alerts",
  "tableName": "alerts",
  "attributes": [
    {
      "@type": "AttributeMetadata",
      "uuid": "16c855d3-29b4-4d29-90a8-8f322b556067",
      "name": "name",
      "length": 255,
      "searchable": true,
      "gridColumn": true,
      "validation": {
        "required": true
      },
      "displayName": "{{ name }}",
      "descriptions": {
        "singular": "Name"
      }
    },
    ...
  ]
}
```

#### Updating Model Metadata ####
To update a model, first you should `GET` the model you would like to update. Then, use `PUT` with the fields you want to change:

`PUT /api/3/model_metadatas/7ab10952-b2ea-4910-9762-eb685bdd29eb`
``` json
{
	"ownable": true
}
```

To update attributes on a model, first you should `GET` the model. Then use `PUT` on the `attributes` field with the modifications. You need to post all the attributes as a whole.

`PUT /api/3/model_metadatas/7ab10952-b2ea-4910-9762-eb685bdd29eb`
``` json
{
  "attributes": [
    {
      /* Existing with @id */
    },
    {
      /* New without @id */
    }
  ]
}
```

You cannot update an attribute directly right now. For instance, going to `/api/3/attribute_metadatas/56d50d6a5a4ef3f3348b4574` will not return a single metadata because it is an embedded relationship.

### Creating New Models ####
To create a new model, use the same syntax as the updating except use `POST` and specify a `type` and `parentType`.

`POST /api/3/model_metadatas`
``` json
{
	"type": "malwares",
    "descriptions": {
        "singular": "Malware",
        "plural": "Malware"
    },
    "tableName": "malware"
}
```

The plural description will be used as the Module description. TableName is required for Postgres implementations of this model.

### Deleting Models ###

Deleting a model can be done, but it is not recommended as it has not been fully tested. When deleting a model, the corresponding Module will be deleted. When renaming a Model too, the Module will be deleted and a new Module will be created. There is no migration for this currently.

### Extended Data ###

The attribute types that are type: `object` or `array` are often referred to as: Extended Data. These are schema-less json blobs that have minimal input validation.

Attribute metadata for these types will look like this: 
``` json
[
    {
        "@type": "AttributeMetadata",
        "uuid": "60e00773-2de6-44aa-8af0-174e4fbdec17",
        "name": "extendedData",
        "type": "object",
        "formType": "object",
        "displayName": "{{ extendedData }}",
        "descriptions": {
            "singular": "Extended Data"
        }
    },
    {
        "@type": "AttributeMetadata",
        "uuid": "338040a1-6965-4e56-b71d-80cc83ecbeb7",
        "name": "extendedDatas",
        "type": "array",
        "formType": "array",
        "displayName": "{{ extendedDatas }}",
        "descriptions": {
            "singular": "Extended Data (Array)"
        }
    }
]
```

#### Examples of Valid Data

``` json
{
    "extendedData": {
        "test": 1234,
        "key2": [
            "randomdata"
        ]
    }
}
```

``` json
{
    "extendedDatas": [
        1234,
        {
            "nested": "object"
        }
    ]
}
```

#### Examples of Bad Data

``` json
{
    "extendedData": [
        "array",
        "on object",
        "is bad"
    ]
}
```

``` json
{
    "extendedDatas": {
        "do not": "post object to array"
    }
}
```

#### Indexing of Extended Data

Because of the limitations on ElasticSearch, extended data needs to be transformed in to a string before it is indexed. Searches performed on this data will act like a traditional text search.


### Known issues ###

* ~~ParentType should be required~~
* ~~Singular checking for description needs to have validation~~
* Deleting a model will not remove the entity from the system.
* `type` fields should not have `_`'s in them.

### View Templates ###

Beautiful view template documentation can be found on the Exo-UI GitHub page [here](https://github.com/CyberSponse-Dev-Corp/blob/develop/docs/view-templates/README.md#view-templates). 

#### System Update ####
After making changes to the metadata, it is time to publish. If you have the UPDATE permission to the Application module, make a PUT command to `/api/publish`. You will get a response like this:
``` json
{
    "@type": "Publish",
    "status": "started"
}
```

This will indicate that a system update has begun. This will make every request to the api return a 503 until the following Symfony commands have finished running:
  1. `cybersponse:metadata:generate:models`
  2. `redis:flushdb --no-interaction`
  3. `cache:clear`
  4. `doctrine:schema:update --em=default --force`
  5. `doctrine:query:sql "SELECT public.drop_all_triggers();"`
  6. `doctrine:query:sql "SELECT public.trigger_all_tables();"`
  7. `cybersponse:metadata:generate:modules`
  8. `fos:elastica:reset`
  9. `fos:elastica:populate`

The progress will look something like this:
``` json
{
  "@type": "SystemUpdate",
  "code": 503,
  "message": "Clearing cache",
  "progressPercent": 22,
  "startTime": 1460678007
}
```

#### Known Issues
Please refer to this page as it will be continuously updated: [Known Issues](https://github.com/CyberSponse-Dev-Corp/ProdTechDoc/blob/master/CRUD-HUB/Metadata/KnownIssues.md)