<?php
/* Copyright start
  Copyright (C) 2008 - 2020 Fortinet Inc.
  All rights reserved.
  FORTINET CONFIDENTIAL & FORTINET PROPRIETARY SOURCE CODE
  Copyright end */

namespace Cybersponse\MetadataBundle\Service;

use Cybersponse\MetadataBundle\Entity\AttributeMetadata;
use Cybersponse\MetadataBundle\Entity\ModelMetadata;
use Cybersponse\MetadataBundle\Generation\Annotations\Classes\ClassAnnotationGeneratorInterface;
use Cybersponse\MetadataBundle\Generation\Annotations\Fields\FieldAnnotationGeneratorInterface;
use Cybersponse\MetadataBundle\Generation\Traits\TraitGeneratorInterface;
use Doctrine\Common\Inflector\Inflector;
use Doctrine\DBAL\Types\Type;
use Doctrine\ORM\Mapping\ClassMetadata;
use Doctrine\ORM\Mapping\ClassMetadataInfo;
use Doctrine\ORM\Tools\EntityGenerator as DoctrineEntityGenerator;

class EntityGenerator extends DoctrineEntityGenerator
{
    /** @var ModelMetadata */
    protected $modelMetadata;

    /** @var AttributeMetadata[] */
    protected $attributeMetadataMap;

    /** @var ClassAnnotationGeneratorInterface[] */
    protected $classAnnotationGenerators;

    /** @var FieldAnnotationGeneratorInterface[] */
    protected $fieldAnnotationGenerators;

    /** @var TraitGeneratorInterface[] */
    protected $traitGenerators;

    /**
     * @var string
     */
    protected static $addMethodAndRelationshipTemplate =
        '/**
 * <description>
 *
 * @param <variableType>$<variableName>
 * @return <entity>
 */
public function <methodName>(<methodTypeHint>$<variableName>)
{
<spaces>if (!$this-><fieldName>->contains($<variableName>)) {
<spaces><spaces>$this-><fieldName>[] = $<variableName>;
<spaces><spaces>$<variableName>-><relationshipMethodName>($this);
<spaces>}
<spaces>return $this;
}';
    /**
     * @var string
     */
    protected static $removeMethodAndRelationshipTemplate =
        '/**
 * <description>
 *
 * @param <variableType>$<variableName>
 */
public function <methodName>(<methodTypeHint>$<variableName>)
{
<spaces>if ($this-><fieldName>->contains($<variableName>)) {
<spaces><spaces>$this-><fieldName>->removeElement($<variableName>);
<spaces><spaces>$<variableName>-><relationshipMethodName>(<relationshipMethodArgument>);
<spaces>}
}';

    /**
     * @var string
     */
    protected static $addMethodTemplate =
        '/**
 * <description>
 *
 * @param <variableType> $<variableName>
 *
 * @return <entity>
 */
public function <methodName>(<methodTypeHint>$<variableName>)
{
<spaces>if (!$this-><fieldName>->contains($<variableName>)) {
<spaces><spaces>$this-><fieldName>[] = $<variableName>;
<spaces>}
<spaces>return $this;
}';

    public function __construct()
    {
        parent::__construct();
        $this->typeAlias = array_merge([
            'jsonb' => 'mixed'
        ], $this->typeAlias);
    }

    /**
     * @param ModelMetadata $modelMetadata
     */
    public function setModelMetadata($modelMetadata)
    {
        $this->modelMetadata = $modelMetadata;
        $this->attributeMetadataMap = [];

        foreach ($modelMetadata->getAttributes() as $attribute) {
            $this->attributeMetadataMap[$attribute->getName()] = $attribute;
        }
    }

    /**
     * @param ClassAnnotationGeneratorInterface[] $classAnnotationGenerators
     */
    public function setClassAnnotationGenerators($classAnnotationGenerators)
    {
        $this->classAnnotationGenerators = $classAnnotationGenerators;
    }

    /**
     * @param FieldAnnotationGeneratorInterface[] $fieldAnnotationGenerators
     */
    public function setFieldAnnotationGenerators($fieldAnnotationGenerators)
    {
        $this->fieldAnnotationGenerators = $fieldAnnotationGenerators;
    }

    /**
     * @param TraitGeneratorInterface[] $traitGenerators
     */
    public function setTraitGenerators($traitGenerators)
    {
        $this->traitGenerators = $traitGenerators;
    }

    /**
     * @param array $associationMapping
     * @param ClassMetadataInfo $metadata
     * @return string
     */
    protected function generateAssociationMappingPropertyDocBlock(array $associationMapping, ClassMetadataInfo $metadata)
    {
        $linesString = parent::generateAssociationMappingPropertyDocBlock($associationMapping, $metadata);

        return $this->modifyMappingPropertyDocBlock($associationMapping, $metadata, $linesString);
    }

    /**
     * @param array $fieldMapping
     * @param ClassMetadataInfo $metadata
     * @return string
     */
    protected function generateFieldMappingPropertyDocBlock(array $fieldMapping, ClassMetadataInfo $metadata)
    {
        $fieldMapping['columnName'] = sprintf('`%s`', strtolower($fieldMapping['fieldName']));

        $linesString = parent::generateFieldMappingPropertyDocBlock($fieldMapping, $metadata);

        return $this->modifyMappingPropertyDocBlock($fieldMapping, $metadata, $linesString);
    }

    protected function generateEntityDocBlock(ClassMetadataInfo $metadata)
    {
        $linesString = parent::generateEntityDocBlock($metadata);

        return $this->modifyEntityDocBlock($metadata, $linesString);
    }


    /**
     * @param array $fieldMapping
     * @param ClassMetadataInfo $metadata
     * @param $linesString
     * @return string
     */
    protected function modifyMappingPropertyDocBlock(array $fieldMapping, ClassMetadataInfo $metadata, $linesString)
    {
        $lines = explode("\n", $linesString);

        $lastLine = array_pop($lines);

        foreach ($this->fieldAnnotationGenerators as $fieldAnnotationGenerator) {
            if (!is_null($this->attributeMetadataMap)) {
                $attributeMetadata = $this->attributeMetadataMap[$fieldMapping['fieldName']];
                if ($fieldAnnotationGenerator->supportsGenerationForMetadata($attributeMetadata)) {
                    if ($code = $fieldAnnotationGenerator->generateFieldAnnotationFromAttributeMetadata($attributeMetadata, $metadata)) {
                        $lines[] = $this->spaces . ' * ' . $code;
                    }
                }
                continue;
            }
        }

        $lines[] = $lastLine;

        return implode("\n", $lines);
    }

    protected function generateEntityConstructor(ClassMetadataInfo $metadata)
    {
        $collections = array();

        if ($this->hasMethod('__construct', $metadata)) {
            $collections[] = 'parent::__construct();';
        }

        foreach ($metadata->associationMappings as $mapping) {
            if ($mapping['type'] & ClassMetadataInfo::TO_MANY) {
                $collections[] = '$this->' . $mapping['fieldName'] . ' = new \Doctrine\Common\Collections\ArrayCollection();';
            }
        }

        if ($collections) {
            return $this->prefixCodeWithSpaces(str_replace("<collections>", implode("\n" . $this->spaces, $collections), self::$constructorMethodTemplate));
        }

        return '';
    }

    protected function generateEntityBody(ClassMetadataInfo $metadata)
    {
        $traitsBlock = $this->generateTraitBlock();
        $body = parent::generateEntityBody($metadata);

        return sprintf("%s\n\n%s", $traitsBlock, $body);
    }

    protected function generateTraitBlock()
    {
        $traits = [];

        foreach ($this->traitGenerators as $traitGenerator) {
            if (!is_null($this->modelMetadata)) {
                if ($traitGenerator->supportsGenerationForMetadata($this->modelMetadata)) {
                    $traits[] = $traitGenerator->getTraitClassName();
                }
                continue;
            }
        }

        $traitsLines = array_map(function ($trait) {
            return sprintf('%suse \\%s;', $this->spaces, $trait);
        }, $traits);
        $traitsBlock = implode("\n", $traitsLines);

        return $traitsBlock;
    }

    /**
     * @param $type
     * @param $fieldName
     * @return string
     */
    private function getMethodName($type, $fieldName)
    {
        $methodName = $type . Inflector::classify($fieldName);
        if (in_array($type, array("add", "remove"))) {
            return Inflector::singularize($methodName);
        }
        return $methodName;
    }

    protected function generateEntityStubMethod(ClassMetadataInfo $metadata, $type, $fieldName, $typeHint = null, $defaultValue = null)
    {
        $methodName = $this->getMethodName($type, $fieldName);

        $variableName = Inflector::camelize($fieldName);
        if (in_array($type, array("add", "remove"))) {
            $variableName = Inflector::singularize($variableName);
        }

        if ($this->hasMethod($methodName, $metadata)) {
            return '';
        }
        $this->staticReflection[$metadata->name]['methods'][] = $methodName;

        $var = sprintf('%sMethodTemplate', $type);

        $inverseMethodName = null;
        $relationshipMethodArgument = '$this';

        if ($metadata->isAssociationInverseSide($fieldName) || $this->isFieldSelfInversing($metadata, $fieldName)) {
            $associationMapping = $metadata->getAssociationMapping($fieldName);
            $inversedField = is_null($associationMapping['mappedBy']) ? $associationMapping['inversedBy'] : $associationMapping['mappedBy'];
            if ($associationMapping['type'] == ClassMetadata::MANY_TO_MANY && in_array($type, array('add', 'remove'))) {
                $inverseMethodName = $this->getMethodName($type, $inversedField);

                $var = sprintf('%sMethodAndRelationshipTemplate', $type);
            }
            if ($associationMapping['type'] == ClassMetadata::ONE_TO_MANY && in_array($type, array('add', 'remove'))) {
                $inverseMethodName = $this->getMethodName("set", $inversedField);

                $var = sprintf('%sMethodAndRelationshipTemplate', $type);
                if ($type == 'remove') {
                    $relationshipMethodArgument = 'null';
                }
            }
        }

        $template = self::$$var;

        $methodTypeHint = null;
        $types          = Type::getTypesMap();
        $variableType   = $typeHint ? $this->getType($typeHint) . ' ' : null;

        if ($typeHint && ! isset($types[$typeHint])) {
            $variableType   =  '\\' . ltrim($variableType, '\\');
            $methodTypeHint =  '\\' . $typeHint . ' ';
        }

        $replacements = array(
            '<description>'       => ucfirst($type) . ' ' . $fieldName,
            '<methodTypeHint>'    => $methodTypeHint,
            '<variableType>'      => $variableType,
            '<variableName>'      => $variableName,
            '<methodName>'        => $methodName,
            '<fieldName>'         => $fieldName,
            '<variableDefault>'   => ($defaultValue !== null ) ? (' = '.$defaultValue) : '',
            '<entity>'            => $this->getClassName($metadata),
            '<relationshipMethodArgument>' => $relationshipMethodArgument
        );

        if (!is_null($inverseMethodName)) {
            $replacements['<relationshipMethodName>'] = $inverseMethodName;
        }

        $method = str_replace(
            array_keys($replacements),
            array_values($replacements),
            $template
        );

        return $this->prefixCodeWithSpaces($method);
    }

    /**
     * @param ClassMetadataInfo $metadata
     * @param $linesString
     * @return string
     */
    private function modifyEntityDocBlock(ClassMetadataInfo $metadata, $linesString)
    {
        $lines = explode("\n", $linesString);

        $lastLine = array_pop($lines);

        foreach ($this->classAnnotationGenerators as $classAnnotationGenerators) {
            if (!is_null($this->modelMetadata)) {
                if ($classAnnotationGenerators->supportsGenerationForMetadata($this->modelMetadata)) {
                    if ($code = $classAnnotationGenerators->generateClassAnnotationFromModelMetadata($this->modelMetadata, $metadata)) {
                        $lines[] = ' * ' . $code;
                    }
                }
                continue;
            }
        }

        $lines[] = $lastLine;

        return implode("\n", $lines);
    }

    /**
     * Returns true if the given field name is self inversing
     *
     * @param ClassMetadataInfo $metadata
     * @param $fieldName
     *
     * @return bool
     */
    private function isFieldSelfInversing($metadata, $fieldName)
    {
        if (!$metadata->hasAssociation($fieldName)) {
            return false;
        }
        $associationMapping = $metadata->getAssociationMapping($fieldName);
        return $associationMapping['targetEntity'] === $associationMapping['sourceEntity'];
    }
}
