<?php
/* Copyright start
  Copyright (C) 2008 - 2020 Fortinet Inc.
  All rights reserved.
  FORTINET CONFIDENTIAL & FORTINET PROPRIETARY SOURCE CODE
  Copyright end */

namespace Cybersponse\MetadataBundle\Tests\Unit\Conditional;


use Cybersponse\MetadataBundle\Conditional\ConditionSet;
use Cybersponse\MetadataBundle\Conditional\ConditionSetEvaluator;
use Cybersponse\MetadataBundle\Conditional\QueryInterface;
use Cybersponse\MetadataBundle\Conditional\QueryMatcher;

class ConditionSetEvaluatorTest extends \PHPUnit_Framework_TestCase
{
    /** @var QueryMatcher|\PHPUnit_Framework_MockObject_MockObject */
    protected $mockQueryMatcher;

    /** @var ConditionSetEvaluator|\PHPUnit_Framework_MockObject_MockObject */
    protected $conditionSetEvaluator;

    /** @var ConditionSet|\PHPUnit_Framework_MockObject_MockObject */
    protected $conditionSet;

    /** @var array */
    protected $conditionSetFields;

    protected function setUp()
    {
        parent::setUp();
        $this->mockQueryMatcher = $this->getMockBuilder(QueryMatcher::class)->getMock();
        $this->conditionSetEvaluator = new ConditionSetEvaluator();
        $this->conditionSet = $this->getMockBuilder(ConditionSet::class)->getMock();
    }

    protected function executeSystemUnderTest()
    {
        $this->conditionSet->method('getFields')->willReturn($this->conditionSetFields);
        /** setup? */
        $this->conditionSetEvaluator->setQueryMatcher($this->mockQueryMatcher);
        return $this->conditionSetEvaluator->evaluate(null, $this->conditionSet);
    }

    protected function tearDown()
    {
        parent::tearDown();
        unset($this->mockQueryMatcher);
        unset($this->conditionSetEvaluator);
        unset($this->conditionSet);
        unset($this->conditionSetFields);
    }

    protected function createQueryWithExpectedResult($expectedResult)
    {
        $matchingQuery = $this->getMockBuilder(QueryInterface::class)->getMock();
        $matchingQuery->method('getResult')->willReturn($expectedResult);
        return $matchingQuery;
    }

    // region Single Field
    // region Primitives
    public function test_GivenStringCondition_ShouldReturnConditionSetAsResult()
    {
        $this->mockQueryMatcher->method('matches')->willReturn(true);

        $this->conditionSetFields = [
            'field' => [
                'maxLength' => 'asdf'
            ]
        ];
        $expectedResult = $this->conditionSetFields;

        $actualResult = $this->executeSystemUnderTest();

        $this->assertEquals($expectedResult, $actualResult);
    }

    public function test_GivenNullCondition_ShouldReturnConditionSetAsResult()
    {
        $this->mockQueryMatcher->method('matches')->willReturn(true);

        $this->conditionSetFields = [
            'field' => [
                'maxLength' => null
            ]
        ];
        $expectedResult = $this->conditionSetFields;

        $actualResult = $this->executeSystemUnderTest();

        $this->assertEquals($expectedResult, $actualResult);
    }

    public function test_GivenIntegerCondition_ShouldReturnConditionSetAsResult()
    {
        $this->mockQueryMatcher->method('matches')->willReturn(true);

        $this->conditionSetFields = [
            'field' => [
                'maxLength' => 5
            ]
        ];
        $expectedResult = $this->conditionSetFields;

        $actualResult = $this->executeSystemUnderTest();

        $this->assertEquals($expectedResult, $actualResult);
    }

    public function test_GivenBoolCondition_ShouldReturnConditionSetAsResult()
    {
        $this->mockQueryMatcher->method('matches')->willReturn(true);

        $this->conditionSetFields = [
            'field' => [
                'required' => true
            ]
        ];
        $expectedResult = $this->conditionSetFields;

        $actualResult = $this->executeSystemUnderTest();

        $this->assertEquals($expectedResult, $actualResult);
    }
    // endregion
    // region Arrays
    public function test_GivenMatchingQuery_ShouldReturnMatchingQueryResult()
    {
        $this->mockQueryMatcher->method('matches')->willReturn(true);

        $expectedResult = true;
        $this->conditionSetFields = [
            'field' => [
                'required' => [$this->createQueryWithExpectedResult($expectedResult)]
            ]
        ];
        $expectedResult = [
            'field' => [
                'required' => $expectedResult
            ]
        ];

        $actualResult = $this->executeSystemUnderTest();

        $this->assertEquals($expectedResult, $actualResult);
    }

    public function test_GivenNonMatchingQuery_ShouldReturnEmptyResult()
    {
        $this->mockQueryMatcher->method('matches')->willReturn(false);

        $this->conditionSetFields = [
            'field' => [
                'required' => [$this->createQueryWithExpectedResult(null), $this->createQueryWithExpectedResult(null)]
            ]
        ];
        $expectedResult = [];

        $actualResult = $this->executeSystemUnderTest();

        $this->assertEquals($expectedResult, $actualResult);
    }

    public function test_GivenEmptyQuery_ShouldReturnEmptyResult()
    {
        $this->conditionSetFields = [
            'field' => [
                'required' => []
            ]
        ];
        $expectedResult = [];

        $actualResult = $this->executeSystemUnderTest();

        $this->assertEquals($expectedResult, $actualResult);
    }

    public function test_GivenMultipleMatchingQueries_ShouldReturnFirstMatchingQueryResult()
    {
        $this->mockQueryMatcher->method('matches')->willReturn(true);

        $expectedResult = true;
        $unexpectedResult = false;
        $this->conditionSetFields = [
            'field' => [
                'required' => [$this->createQueryWithExpectedResult($expectedResult), $this->createQueryWithExpectedResult($unexpectedResult)]
            ]
        ];
        $expectedResult = [
            'field' => [
                'required' => $expectedResult
            ]
        ];

        $actualResult = $this->executeSystemUnderTest();

        $this->assertEquals($expectedResult, $actualResult);
    }

    public function test_GivenNonMatchingQuery_GivenMatchingQuery_ShouldReturnMatchingQueryResult()
    {
        $this->mockQueryMatcher->method('matches')->willReturnOnConsecutiveCalls(false, true);

        $expectedResult = false;
        $unexpectedResult = true;
        $this->conditionSetFields = [
            'field' => [
                'required' => [$this->createQueryWithExpectedResult($unexpectedResult), $this->createQueryWithExpectedResult($expectedResult)]
            ]
        ];
        $expectedResult = [
            'field' => [
                'required' => $expectedResult
            ]
        ];

        $actualResult = $this->executeSystemUnderTest();

        $this->assertEquals($expectedResult, $actualResult);
    }
    // endregion
    // endregion
    // region Multiple Fields
    public function test_GivenMultipleFieldsWithMatchingQueries_ShouldReturnAllMatchingQueryResults()
    {
        $this->mockQueryMatcher->method('matches')->willReturnOnConsecutiveCalls(true, true);

        $expectedResult = true;
        $this->conditionSetFields = [
            'field' => [
                'required' => [$this->createQueryWithExpectedResult($expectedResult)]
            ],
            'other field' => [
                'required' => [$this->createQueryWithExpectedResult($expectedResult)]
            ]
        ];
        $expectedResult = [
            'field' => [
                'required' => $expectedResult
            ],
            'other field' => [
                'required' => $expectedResult
            ]
        ];

        $actualResult = $this->executeSystemUnderTest();

        $this->assertEquals($expectedResult, $actualResult);
    }

    public function test_GivenFieldWithMatchingQuery_GivenFieldWithNonMatchingQuery_ShouldOnlyReturnMatchingQueryResult()
    {
        $this->mockQueryMatcher->method('matches')->willReturnOnConsecutiveCalls(true, false);

        $expectedResult = true;
        $this->conditionSetFields = [
            'field' => [
                'required' => [$this->createQueryWithExpectedResult($expectedResult)]
            ],
            'other field' => [
                'required' => [$this->createQueryWithExpectedResult(null)]
            ]
        ];
        $expectedResult = [
            'field' => [
                'required' => $expectedResult
            ]
        ];

        $actualResult = $this->executeSystemUnderTest();

        $this->assertEquals($expectedResult, $actualResult);
    }
    // endregion
}