<?php
/* Copyright start
  Copyright (C) 2008 - 2020 Fortinet Inc.
  All rights reserved.
  FORTINET CONFIDENTIAL & FORTINET PROPRIETARY SOURCE CODE
  Copyright end */

namespace Cybersponse\ProxyBundle\Controller;

use GuzzleHttp\Client;
use GuzzleHttp\Psr7\Request as GuzzleRequest;
use Symfony\Bundle\FrameworkBundle\Controller\Controller;
use Symfony\Component\HttpFoundation\JsonResponse;
use Symfony\Component\HttpFoundation\Request;
use Symfony\Component\Security\Core\Exception\AccessDeniedException;

class GatewayActionController extends Controller
{
    const USER_UUID = 'userId';
    const USER = 'user';
    const REMOTE_ADDR = 'source';

    public function sendRequest($client, $gatewayRequest)
    {
        try {
            $response = $client->send($gatewayRequest);
            $statusCode = $response->getStatusCode();
            if ($statusCode >= 400) {
                $errMsg = "Failed to get response from audit log";
                return JsonResponse::create($errMsg, $statusCode);
            } else {
                $jsonDecodeResponse = json_decode($response->getBody()->getContents(), true);
                return JsonResponse::create($jsonDecodeResponse, $statusCode);
            }
        } catch (\Exception $exception) {
            error_log($exception->getMessage());
            $errVar = json_decode($exception->getResponse()->getBody()->read(2048), true);
            $errMsg['message'] = $errVar['Error'] ? isset($errVar['Error']) : $errVar['error'];
            $errMsg = $errMsg['message'] ? $errMsg : $errVar;
            $errCode = $exception->getCode();
            return JsonResponse::create($errMsg, $errCode);
        }
    }

    public function gatewayRouteAction(Request $request)
    {
        $objectData = json_decode($request->getContent(), true);
        if (!isset($objectData['entityUuid'])) {
            $this->checkAuthorization($request->getMethod(), $objectData);
        }
        $container = $this->container;
        $gatewayUri = $container->getParameter('gateway.uri');
        $gatewayUri = $gatewayUri . '/audit/activities';
        $token = $this->get('security.token_storage')->getToken();
        $objectData['teams'] = $this->getAccessibleTeamIris();
        $headers = array(
            "Content-type" => "application/json",
        );
        $headers['Authorization'] = "Bearer " . $token->getCredentials();
        $gatewayRequest = new GuzzleRequest('post', $gatewayUri, $headers, json_encode($objectData));
        $client = new Client();
        return $this->sendRequest($client, $gatewayRequest);
    }

    public function checkAuthorization($method, $objectData = [])
    {
        $authorizationChecker = $this->container->get('security.authorization_checker');
        if (isset($objectData['userId']) && $objectData['userId'] == $this->getUser()->getUuid()) {
            if (!$authorizationChecker->isGranted(['perm:activities.read'])) {
                throw new AccessDeniedException();
            }
            return;
        }
        if ($method === 'POST' && !$authorizationChecker->isGranted(['perm:security.read'])) {
            throw new AccessDeniedException();
        }
    }

    /**
     * @return array
     */
    protected function getAccessibleTeamIris()
    {
        $accessibleTeamIdsProvider = $this->container->get('cybersponse.authorization.ownership.accessible_team_ids_provider');
        $iriConverter = $this->container->get('api.iri_converter');
        $teamResource = $this->container->get('resource.team');
        $teamBaseIri = $iriConverter->getIriFromResource($teamResource);
        $accessibleTeamIds = $accessibleTeamIdsProvider->getDataForCurrentUser();
        $accessibleTeamIris = array_map(function ($teamId) use ($teamBaseIri) {
            return sprintf('%s/%s', $teamBaseIri, $teamId);
        }, $accessibleTeamIds);
        return json_decode(json_encode($accessibleTeamIris), true);
    }

    public function gatewayRouteIdAction(Request $request, $auditId)
    {

        $this->checkAuthorization($request->getMethod());
        $container = $this->container;
        $gatewayUri = $container->getParameter('gateway.uri');
        $gatewayUri = $gatewayUri . '/audit/activities/';
        if (isset($auditId) && $auditId != null) {
            $gatewayUri = $gatewayUri . $auditId;
        } else {
            $errMsg['message'] = "Invalid ID";
            return JsonResponse::create($errMsg, 404);
        }
        $token = $this->get('security.token_storage')->getToken();
        $headers = array(
            "Content-type" => "application/json",
        );
        $headers['Authorization'] = "Bearer " . $token->getCredentials();
        $gatewayRequest = new GuzzleRequest('get', $gatewayUri, $headers, $request->getContent());
        $client = new Client();
        return $this->sendRequest($client, $gatewayRequest);
    }

    public function gatewayDeleteRouteAction(Request $request)
    {
        $this->checkAuthorization($request->getMethod());
        $this->checkAuditDeleteAuthorization($request->getMethod());
        $container = $this->container;
        $gatewayUri = $container->getParameter('gateway.uri');
        $gatewayUri = $gatewayUri . '/audit/activities?';
        $query = $request->query->all();
        foreach ($query as $key => $value) {
            $gatewayUri = $gatewayUri . $key . '=' . $value . '&';
        }
        $gatewayUri = substr($gatewayUri, 0, -1);
        $token = $this->get('security.token_storage')->getToken();
        $headers = array(
            "Content-type" => "application/json",
        );
        $headers['Authorization'] = "Bearer " . $token->getCredentials();
        $gatewayRequest = new GuzzleRequest('delete', $gatewayUri, $headers);
        $client = new Client();
        return $this->sendRequest($client, $gatewayRequest);
    }

    public function checkAuditDeleteAuthorization($method)
    {
        $authorizationChecker = $this->container->get('security.authorization_checker');
        if (!$authorizationChecker->isGranted(['perm:activities.delete'])) {
            throw new AccessDeniedException();
        }
    }

    public function gatewayDeleteActivitiesRouteAction(Request $request)
    {
        $this->checkAuthorization($request->getMethod());
        $this->checkAuditDeleteAuthorization($request->getMethod());
        $container = $this->container;
        $gatewayUri = $container->getParameter('gateway.uri');
        $gatewayUri = $gatewayUri . '/audit/activities/delete';
        $token = $this->get('security.token_storage')->getToken();
        $headers = array(
            "Content-type" => "application/json",
        );
        $headers['Authorization'] = "Bearer " . $token->getCredentials();
        $objectData = json_decode($request->getContent(), true);
        $objectData[self::USER_UUID] = $this->getUser()->getUuid();
        $objectData[self::USER] = $this->extractUserName();
        $objectData[self::REMOTE_ADDR] = $_SERVER['REMOTE_ADDR'];
        $gatewayRequest = new GuzzleRequest('post', $gatewayUri, $headers, json_encode($objectData));
        $client = new Client();
        return $this->sendRequest($client, $gatewayRequest);
    }

    public function gatewayGetRouteAction(Request $request)
    {

        $this->checkAuthorization($request->getMethod());
        $this->checkAuditDeleteAuthorization($request->getMethod());
        $container = $this->container;
        $gatewayUri = $container->getParameter('gateway.uri');
        $gatewayUri = $gatewayUri . '/audit/activities/ttl';
        $token = $this->get('security.token_storage')->getToken();
        $headers = array(
            "Content-type" => "application/json",
        );
        $headers['Authorization'] = "Bearer " . $token->getCredentials();
        $gatewayRequest = new GuzzleRequest('get', $gatewayUri, $headers);
        $client = new Client();
        return $this->sendRequest($client, $gatewayRequest);
    }

    public function gatewayPostRouteAction(Request $request)
    {
        $this->checkAuthorization($request->getMethod());
        $container = $this->container;
        $objectData = json_decode($request->getContent(), true);
        $gatewayUri = $container->getParameter('gateway.uri');
        $gatewayUri = $gatewayUri . '/audit/activities/ttl';
        $token = $this->get('security.token_storage')->getToken();
        $headers = array(
            "Content-type" => "application/json",
        );
        $objectData[self::USER_UUID] = $this->getUser()->getUuid();
        $objectData[self::USER] = $this->extractUserName();
        $objectData[self::REMOTE_ADDR] = $_SERVER['REMOTE_ADDR'];
        $headers['Authorization'] = "Bearer " . $token->getCredentials();
        $gatewayRequest = new GuzzleRequest('post', $gatewayUri, $headers, json_encode($objectData));
        $client = new Client();
        return $this->sendRequest($client, $gatewayRequest);
    }

    public function gatewayttlDeleteRouteAction(Request $request)
    {
        $this->checkAuthorization($request->getMethod());
        $this->checkAuditDeleteAuthorization($request->getMethod());
        $objectData = json_decode($request->getContent(), true);
        $container = $this->container;
        $gatewayUri = $container->getParameter('gateway.uri');
        $gatewayUri = $gatewayUri . '/audit/activities/ttl';
        $token = $this->get('security.token_storage')->getToken();
        $headers = array(
            "Content-type" => "application/json",
        );
        $objectData[self::USER_UUID] = $this->getUser()->getUuid();
        $objectData[self::USER] = $this->extractUserName();
        $objectData[self::REMOTE_ADDR] = $_SERVER['REMOTE_ADDR'];
        $headers['Authorization'] = "Bearer " . $token->getCredentials();
        $gatewayRequest = new GuzzleRequest('delete', $gatewayUri, $headers, json_encode($objectData));
        $client = new Client();
        return $this->sendRequest($client, $gatewayRequest);
    }
    /**
     * @return string
     */
    private function extractUserName()
    {
        if (method_exists($this->getUser(), "getfirstname")) {
            if (method_exists($this->getUser(), "getlastname")) {
                $userName = $this->getUser()->getfirstname() . ' ' . $this->getUser()->getlastname();
            } else {
                $userName = $this->getUser()->getfirstname();
            }
        } else {
            $userName = $this->getUser()->getname();
        }
        return $userName;
    }
  
    public function gatewaySyslogGetAction(Request $request)
    {
        $authorizationChecker = $this->container->get('security.authorization_checker');
        if (!$authorizationChecker->isGranted(['perm:security.read'])) {
            throw new AccessDeniedException();
        }
        $container = $this->container;
        $gatewayUri = $container->getParameter('gateway.uri');
        $gatewayUri = $gatewayUri . '/config/syslog';
        $token = $this->get('security.token_storage')->getToken();
        $headers = array(
            "Content-type" => "application/json",
        );
        $headers['Authorization'] = "Bearer " . $token->getCredentials();
        $gatewayRequest = new GuzzleRequest('get', $gatewayUri, $headers);
        $client = new Client();
        return $this->sendRequest($client, $gatewayRequest);
    }

    public function gatewaySyslogPostAction(Request $request)
    {
        $authorizationChecker = $this->container->get('security.authorization_checker');
        if (!$authorizationChecker->isGranted(['perm:security.create'])) {
            throw new AccessDeniedException();
        }
        
        $objectData = json_decode($request->getContent(), true);
        $objectData[self::USER_UUID] = $this->getUser()->getUuid();
        $objectData[self::USER] = $this->extractUserName();
        $objectData[self::REMOTE_ADDR] = $_SERVER['REMOTE_ADDR'];

        $container = $this->container;
        $gatewayUri = $container->getParameter('gateway.uri');
        $gatewayUri = $gatewayUri . '/config/syslog';
        $token = $this->get('security.token_storage')->getToken();
        $originalHeader = $request->headers->all();
        $headers = array(
            "Content-type" => "application/json",
            "is-cluster-call" => $originalHeader["is-cluster-call"]
        );
        $headers['Authorization'] = "Bearer " . $token->getCredentials();
        $gatewayRequest = new GuzzleRequest('post', $gatewayUri, $headers, json_encode($objectData));
        $client = new Client();
        return $this->sendRequest($client, $gatewayRequest);
    }

    public function gatewaySyslogPutAction(Request $request)
    {
        $authorizationChecker = $this->container->get('security.authorization_checker');
        if (!$authorizationChecker->isGranted(['perm:security.update'])) {
            throw new AccessDeniedException();
        }

        $objectData = json_decode($request->getContent(), true);
        $objectData[self::USER_UUID] = $this->getUser()->getUuid();
        $objectData[self::USER] = $this->extractUserName();
        $objectData[self::REMOTE_ADDR] = $_SERVER['REMOTE_ADDR'];

        $container = $this->container;
        $gatewayUri = $container->getParameter('gateway.uri');
        $gatewayUri = $gatewayUri . '/config/syslog';
        $token = $this->get('security.token_storage')->getToken();
        $originalHeader = $request->headers->all();
        $headers = array(
            "Content-type" => "application/json",
            "is-cluster-call" => $originalHeader["is-cluster-call"]
        );
        $headers['Authorization'] = "Bearer " . $token->getCredentials();
        $gatewayRequest = new GuzzleRequest('put', $gatewayUri, $headers, json_encode($objectData));
        $client = new Client();
        return $this->sendRequest($client, $gatewayRequest);
    }

    public function gatewaySyslogDeleteAction(Request $request, $uuid)
    {
        $authorizationChecker = $this->container->get('security.authorization_checker');
        if (!$authorizationChecker->isGranted(['perm:security.delete'])) {
            throw new AccessDeniedException();
        }

        $objectData = json_decode($request->getContent(), true);
        $objectData[self::USER_UUID] = $this->getUser()->getUuid();
        $objectData[self::USER] = $this->extractUserName();
        $objectData[self::REMOTE_ADDR] = $_SERVER['REMOTE_ADDR'];

        $container = $this->container;
        $gatewayUri = $container->getParameter('gateway.uri');
        $gatewayUri = $gatewayUri . '/config/syslog/';
        if (isset($uuid) && $uuid != null) {
            $gatewayUri = $gatewayUri . $uuid;
        } else {
            $errMsg['message'] = "Invalid UUId";
            return JsonResponse::create($errMsg, 404);
        }
        $token = $this->get('security.token_storage')->getToken();
        $originalHeader = $request->headers->all();
        $headers = array(
            "Content-type" => "application/json",
            "is-cluster-call" => $originalHeader["is-cluster-call"]
        );
        $headers['Authorization'] = "Bearer " . $token->getCredentials();
        $gatewayRequest = new GuzzleRequest('delete', $gatewayUri, $headers, json_encode($objectData));
        $client = new Client();
        return $this->sendRequest($client, $gatewayRequest);
    }
}
