<?php
/* Copyright start
  Copyright (C) 2008 - 2020 Fortinet Inc.
  All rights reserved.
  FORTINET CONFIDENTIAL & FORTINET PROPRIETARY SOURCE CODE
  Copyright end */

namespace Cybersponse\ProxyBundle\Controller;

use stdClass;
use Cybersponse\ProxyBundle\DependencyInjection\Configuration;
use Cybersponse\ProxyBundle\GuzzleHttp\ProxyClient;
use Cybersponse\SettingsBundle\Utility\Resolver;
use GuzzleHttp\Exception\BadResponseException;
use GuzzleHttp\Psr7\Request as GuzzleRequest;
use GuzzleHttp\Psr7\Response as GuzzleResponse;
use Symfony\Bundle\FrameworkBundle\Controller\Controller;
use Symfony\Component\HttpFoundation\Request;
use Symfony\Component\HttpFoundation\Response;
use Symfony\Component\HttpKernel\Exception\NotFoundHttpException;
use Symfony\Component\Security\Core\Authentication\Token\TokenInterface;
use Symfony\Component\Security\Core\Authorization\AuthorizationChecker;
use Symfony\Component\Security\Core\Authorization\Voter\VoterInterface;
use Symfony\Component\Security\Core\Exception\AccessDeniedException;
use Doctrine\Common\Inflector\Inflector;
use Cybersponse\ProxyBundle\Provider\SamlActorProvider;
use Cybersponse\ProxyBundle\Provider\JwtLogoutKeyProvider;

class ProxyController extends Controller
{
    const LOGIN = 'login';
    const LOGOUT = 'logout';
    const SAML = 'saml';
    const USERID = 'userId';
    const SAML_FAILURE_MESSAGE = 'We are sorry, but the server encountered an error while handling your request. Please contact your administrator for assistance.';

    public function proxyRouteAction($routeIdentifier, $route, Request $request, $customHeader = [], $payload=null)
    {
        /** @var AuthorizationChecker $authorizationChecker */
        $authorizationChecker = $this->get('security.authorization_checker');

        $token = $this->get('security.token_storage')->getToken();

        $client = $this->findHandlerForRoute($routeIdentifier);
        $permissionGroups = $this->findPermissionGroupsForRoute($routeIdentifier);

        $matchedPermissionGroups = 0;

        foreach ($permissionGroups as $permissionName => $permissionGroup) {
            /** @var VoterInterface[] $additionalVoters */
            $additionalVoters = [];
            $permissionsRequired = [];
            // Every User should be able to retrieve the config token section so that UI can handle it accordingly.
            $ignoreAuthorization = false;
            $operator = $permissionGroup['operator'];

            //Check method first
            if ($request->getMethod() != $permissionGroup['action']) {
                continue;
            }

            $uuid = null;

            //This was matched because it had no permissions required
            if (empty($permissionGroup['permissions'])) {
                $matchedPermissionGroups++;
            }

            //Check permissions
            foreach ($permissionGroup['permissions'] as $permission) {
                //Check incoming route
                $routeRegex = sprintf('/%s/', str_replace('/', '\/', $permissionGroup['route']));
                preg_match($routeRegex, $route, $matches);

                $ignoreAuthorization = false;
                if (array_key_exists('ignoreAuthorization', $permissionGroup)) {
                    $ignoreAuthorization = $permissionGroup['ignoreAuthorization'];
                }

                if (empty($matches) && !$ignoreAuthorization) {
                    continue;
                }

                //This was matched because it matched the regex
                $matchedPermissionGroups++;

                //Check query strings to make sure it matches
                foreach ($permissionGroup['params'] as $key => $value) {
                    if ($request->query->get($key) != $value) {
                        continue 2; //Continue to next permission group
                    }
                }

                //Build Permissions
                $permissionsRequired[] = $permission;

                //Determine where uuid is
                if (!array_key_exists('uuidLocation', $permissionGroup)) {
                    continue;
                }
                $uuidLocation = $permissionGroup['uuidLocation'];

                // Check if in the URL
                if (isset($uuidLocation[Configuration::URL])) {
                    $matchNumber = $uuidLocation[Configuration::URL]['matchNumber'];
                    if (count($matches) >= $matchNumber) {
                        $uuid = $matches[$matchNumber];
                    }
                }

                // Check if in the BODY
                if (isset($uuidLocation[Configuration::BODY])) {
                    $path = $uuidLocation[Configuration::BODY]['path'];
                    $bodyAsArray = json_decode($request->getContent(), true);
                    $uuid = Resolver::get($bodyAsArray, $path);
                }

                // Check if in the Query String
                if (isset($uuidLocation[Configuration::QUERY_STRING])) {
                    $queryKey = $uuidLocation[Configuration::QUERY_STRING]['key'];
                    $uuid = $request->query->get($queryKey);
                }

                //Build additional voters
                foreach ($permissionGroup['additionalVoters'] as $voter) {
                    $additionalVoters[] = $this->container->get($voter);
                }
            }

            //Keep checking for more matched permission groups
            if ($matchedPermissionGroups <= 0) {
                continue;
            }

            if ($operator && $operator == 'AND') {
                $permissionMatchCount = 0;
                foreach ($permissionsRequired as $perm) {
                    if ($authorizationChecker->isGranted($perm)) {
                        $permissionMatchCount ++;
                    }
                }
                if ($permissionMatchCount < count($permissionsRequired)) {
                    throw new AccessDeniedException();
                } else {
                    break;
                }
            } else {
                if ($authorizationChecker->isGranted($permissionsRequired)) {
                    //You're good to keep going through the system!
                    break;
                }
            }

            if (array_key_exists('ignoreAuthorization', $permissionGroup)) {
                $ignoreAuthorization = $permissionGroup['ignoreAuthorization'];
            }
            if ($ignoreAuthorization) {
                break;
            }

            if (!empty($additionalVoters)) {
                if ($this->checkAdditionalVoters($additionalVoters, $token, $uuid)) {
                    //You're good to keep going through the system!
                    break;
                }
            } else {
                throw new AccessDeniedException();
            }
        }

        if ($matchedPermissionGroups == 0) {
            throw new AccessDeniedException();
        }

        $userRequestUri = sprintf("%s/%s?%s", $client->getTargetUri(), $route, $request->getQueryString());

        if ($payload) {
            $content = $payload;
        } else {
            $content = $request->getContent();
        }

        //Remove host from the headers so guzzle rebuilds it with the port
        $headersExceptForHost = $request->headers->all();
        unset($headersExceptForHost['host']);
        $headersExceptForHost['X-REMOTE_ADDR'] = $_SERVER['REMOTE_ADDR'];
        $headersExceptForHost['X-USER'] = $this->getUser()->getUuid();
        $hmacRequest = new GuzzleRequest($request->getMethod(), $userRequestUri, array_merge($headersExceptForHost,$customHeader), $content);

        /** @var GuzzleResponse $response */
        try {
            $response = $client->send($hmacRequest);
        } catch (BadResponseException $exception) {
            $response = $exception->getResponse();
        }
        return Response::create($response->getBody()->getContents(), $response->getStatusCode(), ['Content-Type' => $response->getHeader('Content-Type')]);
    }

    public function wfProxyRouteAction($routeIdentifier, $route, Request $request)
    {
        $teams = $this->getAccessibleTeamIris();
        if (strpos($route,"api/workflows/log_list") !== false && $request->getMethod() == "POST") {
            $payload = json_decode($request->getContent(), true);
            $payload['PB_OWNERS'] = $teams;
            return $this->proxyRouteAction($routeIdentifier, $route, $request, [], json_encode($payload));
        } else if (strpos($route,"api/workflows/") !== false && $request->getMethod() == "GET") {
            $Customheader['X_PB_OWNERS'] = $teams;
            return $this->proxyRouteAction($routeIdentifier, $route, $request, $Customheader);
        }
        return $this->proxyRouteAction($routeIdentifier, $route,$request, []);
        
    }


    public function integrationProxyRouteAction($routeIdentifier, $route, Request $request)
    {
        if ($request->getMethod() == "POST" || $request->getMethod() == "PUT" || $request->getMethod() == "DELETE") {
            if (strpos($route,"import-connector") === false) {
                $teams = $this->getAccessibleTeamIris();
                $query = $request->getContent() ? json_decode($request->getContent()): new stdClass();
                $query->rbac_info = new stdClass();
                $query->rbac_info->{'teams'} = $teams;
                $query->rbac_info->{'source'} = $_SERVER['REMOTE_ADDR'];
                $query->rbac_info->{'user_iri'} = '/api/3/people/' . $this->getUser()->getUuid();
                return $this->proxyRouteAction($routeIdentifier, $route, $request, [], json_encode($query));
            }
        }
        return $this->proxyRouteAction($routeIdentifier, $route, $request);
    }

    public function postmanProxyRouteAction($routeIdentifier, $route, Request $request)
    {
        if ((strpos($route, "staging-mmd/") !== false && $request->getMethod() == "PUT") || (strpos($route, "publish/") !== false && $request->getMethod() == "POST")){
            $customheader['X-USER'] = $this->getUser()->getUuid();
            return $this->proxyRouteAction($routeIdentifier, $route, $request, $customheader);
        }
        return $this->proxyRouteAction($routeIdentifier, $route,$request);
    }

    /**
     * @return array
     */
    protected function getAccessibleTeamIris()
    {
        $accessibleTeamIdsProvider = $this->container->get('cybersponse.authorization.ownership.accessible_team_ids_provider');
        $accessibleTeamIds = $accessibleTeamIdsProvider->getDataForCurrentUser();
        return json_encode($accessibleTeamIds);
    }

    public function routePublicSAMLAction($routeIdentifier, $route, Request $request)
    {
        if ($route !== self::LOGIN && $route !== self::LOGOUT) {
            return new Response(null, Response::HTTP_NOT_FOUND);
        }
        $client = $this->findHandlerForRoute($routeIdentifier);
        $userRequestUri = sprintf("%s/%s?%s", $client->getTargetUri(), $route, $request->getQueryString());
        //Remove host from the headers so guzzle rebuilds it with the port
        $headersExceptForHost = $request->headers->all();
        unset($headersExceptForHost['host']);
        $hmacRequest = new GuzzleRequest($request->getMethod(), $userRequestUri, $headersExceptForHost, $request->getContent());

        /** @var GuzzleResponse $response */
        try {
            $response = $client->send($hmacRequest, array(self::SAML => true));
            if ($response->getStatusCode() === 200 && $request->getMethod() === 'POST') {
                $html = $this->validateSAMLUser($response, $route, $headersExceptForHost);
                return Response::create($html, $response->getStatusCode(), ['Content-Type' => 'text/html']);
            }
        } catch (BadResponseException $exception) {
            $response = $exception->getResponse();
            $respBody = json_decode($response->getBody()->getContents(), true);
            $error = $respBody ? $respBody['Error']: self::SAML_FAILURE_MESSAGE;
            return Response::create($this->getErrorHTML($error), $response->getStatusCode(), ['Content-Type' => $response->getHeader('Content-Type')]);
        }
        return Response::create($response->getBody()->getContents(), $response->getStatusCode(), ['Content-Type' => $response->getHeader('Content-Type')]);
    }

    public function routeAuthorizedSAMLAction($routeIdentifier, $route, Request $request)
    {
        $client = $this->findHandlerForRoute($routeIdentifier);
        $this->checkAuthorized($request->getMethod());
        $userRequestUri = sprintf("%s/%s?%s", $client->getTargetUri(), $route, $request->getQueryString());
        //Remove host from the headers so guzzle rebuilds it with the port
        $headersExceptForHost = $request->headers->all();
        unset($headersExceptForHost['host']);
        $hmacRequest = new GuzzleRequest($request->getMethod(), $userRequestUri, $headersExceptForHost, $request->getContent());

        /** @var GuzzleResponse $response */
        try {
            $response = $client->send($hmacRequest, array(self::SAML => true));
        } catch (BadResponseException $exception) {
            $response = $exception->getResponse();
        }
        return Response::create($response->getBody()->getContents(), $response->getStatusCode(), ['Content-Type' => $response->getHeader('Content-Type')]);
    }

    public function routeLogoutAction(Request $request)
    {
        $client = $this->findHandlerForRoute(self::SAML);
        $userRequestUri = sprintf("%s/%s?%s", $client->getTargetUri(), self::LOGOUT, $request->getQueryString());
        //Remove host from the headers so guzzle rebuilds it with the port
        $headersExceptForHost = $request->headers->all();
        unset($headersExceptForHost['host']);
        $hmacRequest = new GuzzleRequest($request->getMethod(), $userRequestUri, $headersExceptForHost, $request->getContent());

        /** @var GuzzleResponse $response */
        $response = null;
        try {
            $this->logoutJwtAction($request);
            if (!$request->query->has('loggedout')) {
                $response = $client->send($hmacRequest, array(self::SAML => true));
                return Response::create($response->getBody()->getContents(), $response->getStatusCode(), ['Content-Type' => $response->getHeader('Content-Type')]);
            }
        } catch (BadResponseException $exception) {
            $response = $exception->getResponse();
        }
        return new Response($response, Response::HTTP_OK);
    }

    /**
     * @param $resourceShortName
     * @return string $name;
     */
    protected function getResourceFromShortName($resourceShortName)
    {
        $resourceClassName = Inflector::singularize(Inflector::classify($resourceShortName));
        /** @var ResourceCollectionInterface $resourceCollection */
        $resourceCollection = $this->container->get('api.resource_collection');
        $resource = $resourceCollection->getResourceForShortName($resourceClassName);
        return $resource->getEntityClass();
    }

    /**
     * @param $routeIdentifier
     * @return ProxyClient
     */
    public function findHandlerForRoute($routeIdentifier)
    {
        $handler = $this->findHandlerMapForRoute($routeIdentifier);

        return $this->container->get($handler[$routeIdentifier]['handlerName']);
    }

    /**
     * @param $routeIdentifier
     * @return array
     */
    private function findHandlerMapForRoute($routeIdentifier)
    {
        $handlerMap = $this->container->getParameter('cybersponse.proxy.handler.map');

        if (!array_key_exists($routeIdentifier, $handlerMap)) {
            throw new NotFoundHttpException();
        }
        return $handlerMap;
    }

    private function logoutJwtAction(Request $request)
    {
        /** @var JwtLogoutKeyProvider $jwtLogoutProvider */
        $jwtLogoutProvider = $this->get('cybersponse.proxy.das_jwt_logout_key_provider');
        $jwtKey = $request->headers->get('authorization');
        $jwtLogoutProvider->setJwtLogoutKey($jwtKey);
    }

    public function checkAuthorized($method)
    {
        $authorizationChecker = $this->container->get('security.authorization_checker');
        if ($method === 'POST') {
            if (!$authorizationChecker->isGranted(['perm:security.create']) && !$authorizationChecker->isGranted(['perm:security.update'])) {
                throw new AccessDeniedException();
            }
        } else {
            if (!$authorizationChecker->isGranted(['perm:security.read'])) {
                throw new AccessDeniedException();
            }
        }
    }
    /**
     * @param $routeIdentifier
     * @return array
     */
    private function findPermissionGroupsForRoute($routeIdentifier)
    {
        $handler = $this->findHandlerMapForRoute($routeIdentifier);
        return $handler[$routeIdentifier]['permissions'];
    }

    /**
     * @param VoterInterface[] $additionalVoters
     * @param TokenInterface $token
     * @param $uuid
     * @return bool|void
     */
    private function checkAdditionalVoters($additionalVoters, $token, $uuid)
    {
        $deny = 0;
        foreach ($additionalVoters as $voter) {
            $result = $voter->vote($token, null, ['given_uuid' => $uuid]);
            switch ($result) {
                case VoterInterface::ACCESS_GRANTED:
                    return true;

                case VoterInterface::ACCESS_DENIED:
                    ++$deny;

                    break;

                default:
                    break;
            }
        }

        if ($deny > 0) {
            throw new AccessDeniedException();
        }

        return true;
    }

    /**
     * @param String $userId
     * @param array $headers
     * @return GuzzleResponse $response
     */
    private function authenticateSAMLUser($userId, $headers)
    {

        $client = $this->findHandlerForRoute('auth');
        $body = "{ \"credentials\" : {\"loginid\" : \"$userId\", \"token\" : \"\" }, \"domain\" : \"\" , \"type\" : 6, \"tenant\" : \"crudhub\", \"status\" : 1}";
        $userRequestUri = sprintf("%s/%s", $client->getTargetUri(), 'users');
        /** @var GuzzleResponse $response */
        $hmacRequest = new GuzzleRequest('POST', $userRequestUri, $headers, $body);
        return $client->send($hmacRequest);
    }


    /**
     * @param GuzzleResponse $response
     * @param String $route
     * @param array $headers
     * @return String $html
     */
    private function validateSAMLUser($response, $route, $headers)
    {
        $userId = null;
        $jwtToken = null;
        if ($route === self::LOGIN) {
            $actorBody = json_decode($response->getBody()->getContents(), true);
            if (array_key_exists(self::USERID, $actorBody)) {
                $userId = $actorBody[self::USERID];
                /** @var GuzzleResponse $validateResp */
                try {
                    $validateResp = $this->authenticateSAMLUser($userId, $headers);
                    if ($validateResp->getStatusCode() == 200 || $validateResp->getStatusCode() == 201) {
                        $validateRespBody = json_decode($validateResp->getBody()->getContents(), true);
                        $jwtToken = $validateRespBody['token'];
                        $actorBody[self::USERID] = $validateRespBody['uuid'];
                        /** @var SamlActorProvider $samlProvider */
                        $samlProvider = $this->get('cybersponse.saml_proxy.provider.samlactor');
                        $samlProvider->saveSamlUser($this->getResourceFromShortName('Person'), $actorBody);
                        return $this->getHTML($jwtToken, $route);
                    } else {
                        $respBody = json_decode($validateResp->getBody()->getContents(), true);
                        return $this->getErrorHTML($respBody['Error']);
                    }
                } catch (BadResponseException $exception) {
                    $response = $exception->getResponse();
                    $respBody = json_decode($response->getBody()->getContents(), true);
                    return $this->getErrorHTML($respBody['Error']);
                }
            } else {
                return $this->getErrorHTML('Invalid authentication request, no user found');
            }
        }
        return $this->getHTML($jwtToken, $route);
    }

    private function getHTML($jwtToken, $route)
    {
        $host = $_SERVER['HTTP_HOST'];
        $https = $_SERVER['HTTPS'] == 'on' ? 'https' : 'http';
        $html = "";
        if ($route === self::LOGIN) {
            $html = "<html><body onload=\"javascript:window.location.href = '$https://$host/login/$jwtToken?saml=true'\"></body></html>";
        } elseif ($route === self::LOGOUT) {
            $html = "<html><body onload=\"javascript:window.location.href = '$https://$host/logout?loggedout=true'\"></body></html>";
        }
        return $html;
    }

    private function getErrorHTML($error)
    {
        $error = urlencode($error);
        $host = $_SERVER['HTTP_HOST'];
        $https = $_SERVER['HTTPS'] == 'on' ? 'https' : 'http';
        return "<html><body onload=\"javascript:window.location.href = '$https://$host/login/?error=$error'\"></body></html>";
    }
}
