<?php
/* Copyright start
  Copyright (C) 2008 - 2020 Fortinet Inc.
  All rights reserved.
  FORTINET CONFIDENTIAL & FORTINET PROPRIETARY SOURCE CODE
  Copyright end */

namespace Cybersponse\ProxyBundle\Provider;


use Symfony\Component\HttpFoundation\Request;
use Symfony\Component\Security\Core\Exception\AuthenticationException;

class FingerprintProvider
{


    /**
     * @param $theirFingerprint
     * @param SecretKeyProviderInterface $secretKeyProvider
     * @param Request $request
     * @param int $timeout Seconds
     * @return bool
     */
    public static function validateFingerprint($theirFingerprint, SecretKeyProviderInterface $secretKeyProvider, Request $request=null, $timeout = 60)
    {
        $theirDecodedFingerprint = base64_decode($theirFingerprint);
        $theirDecodedFingerprintParts = explode(';', $theirDecodedFingerprint);

        if (count($theirDecodedFingerprintParts) != 4) {
            throw new AuthenticationException('Fingerprint has invalid number of parts.');
        }

        list(
            $theirAlgo,
            $theirTimestampString,
            $theirKeyIdentifier,
            $theirHashedFingerprint
        ) = $theirDecodedFingerprintParts;
        $theirTimestamp = new \DateTime($theirTimestampString, new \DateTimeZone('UTC'));
        $currentTimestamp = new \DateTime(null, new \DateTimeZone('UTC'));

        $timeGap = abs($theirTimestamp->getTimestamp() - $currentTimestamp->getTimestamp());
        if ($timeGap > $timeout) {
            throw new AuthenticationException('Fingerprint has expired.');
        }

        $theirSecretKey = $secretKeyProvider->getSecretKey($theirKeyIdentifier);

        if (!$request) {
            return true;
        }

        $content = $request->getContent();

        if (empty($request->getContent())) {
            $content = $theirKeyIdentifier;
        }

        $ourHashedFingerprint = static::createHashedFingerprint(
            $theirSecretKey,
            $theirAlgo,
            $request->getMethod(),
            $theirTimestamp->format('Y-m-d H:i:s'),
            $request->getSchemeAndHttpHost() . $request->getRequestUri(),
            $content
        );

        if (strcmp($ourHashedFingerprint, $theirHashedFingerprint) !== 0) {
            throw new AuthenticationException('Fingerprint does not match.');
        }

        return true;
    }

    /**
     * @param $keyIdentifier
     * @param $secretKey
     * @param $algo
     * @param $verb
     * @param $timestamp
     * @param $uri
     * @param $payload
     * @return string
     */
    public static function createFingerprint($keyIdentifier, $secretKey, $algo, $verb, $timestamp, $uri, $payload)
    {
        $hashedFingerprint = self::createHashedFingerprint($secretKey, $algo, $verb, $timestamp, $uri, $payload);

        return base64_encode(sprintf(
            "%s;%s;%s;%s",
            $algo,
            $timestamp,
            $keyIdentifier,
            $hashedFingerprint
        ));
    }

    /**
     * @param $secretKey
     * @param $algo
     * @param $verb
     * @param $timestamp
     * @param $uri
     * @param $payload
     * @return string
     */
    private static function createHashedFingerprint($secretKey, $algo, $verb, $timestamp, $uri, $payload)
    {
        $hashedPayload = hash(
            $algo,
            $payload,
            false
        );

        $rawFingerprint = sprintf(
            "%s.%s.%s.%s.%s",
            $algo,
            $verb,
            $timestamp,
            $uri,
            $hashedPayload
        );

        return hash_hmac(
            $algo,
            $rawFingerprint,
            $secretKey,
            false
        );
    }

    /**
     * We want to calculate $timeGap into seconds.
     *
     * $seconds = $timeGap->days*86400 + $timeGap->h*360 + $timeGap->i*60 + $timeGap->s;
     * Here
     * 86400 is the number of seconds in a day
     * 3600 is the number of seconds in an hour
     * 60 is the number of seconds in a minute
     * @param \DateInterval $timeGap
     */
    private static function getSeconds($timeGap)
    {
        return ($timeGap->days * 86400 + $timeGap->h * 3600
            + $timeGap->i * 60 + $timeGap->s);
    }
}
