<?php
/* Copyright start
  Copyright (C) 2008 - 2020 Fortinet Inc.
  All rights reserved.
  FORTINET CONFIDENTIAL & FORTINET PROPRIETARY SOURCE CODE
  Copyright end */

namespace Cybersponse\QueryBundle\Controller;


use Cybersponse\AuthorizationBundle\AuthorizationProvider;
use Cybersponse\AuthorizationBundle\Permissions\PermissionsDictionaryBuilder;
use Cybersponse\QueryBundle\Entity\Query;
use Cybersponse\QueryBundle\Provider\AggregateDataProviderInterface;
use Cybersponse\QueryBundle\Constants\AggregateOperators;
use Cybersponse\QueryBundle\Provider\QueryProvider;
use Doctrine\Common\Inflector\Inflector;
use Doctrine\Common\Persistence\ManagerRegistry;
use Dunglas\ApiBundle\Api\ResourceCollectionInterface;
use Dunglas\ApiBundle\Api\ResourceInterface;
use Dunglas\ApiBundle\JsonLd\Response;
use Cybersponse\MetadataBundle\Entity\StagingModelMetadata;
use Symfony\Bundle\FrameworkBundle\Controller\Controller;
use Symfony\Component\HttpFoundation\JsonResponse;
use Symfony\Component\HttpFoundation\Request;
use Symfony\Component\HttpKernel\Exception\HttpException;
use Cybersponse\UtilsBundle\Utils;

class AdvancedQueryController extends Controller
{
    const DEFAULT_KEYS = ['id', '@id', '@type'];
    const SELECTIVE_FIELD_KEY = '__selectFields';

    /**
     * @param string $resourceShortName
     * @param string $queryId
     * @param Request $request
     * @return \Symfony\Component\HttpFoundation\Response|static
     */
    public function applyAction($resourceShortName, $queryId, Request $request)
    {
        $resource = $this->getResourceFromShortName($resourceShortName);
        if (!$resource) {
            throw new HttpException(Response::HTTP_NOT_FOUND, sprintf('Resource "%s" not found.', $resourceShortName));
        }

        $blockPassword = false;
        if ($resourceShortName == 'routers' || $resourceShortName == 'tenants' || $resourceShortName == 'agents') {
            $blockPassword = Utils::isPasswordBlock($request);
        }

        $namedQuery = $this->getQueryById($queryId);

        // TODO: Validate $namedQuery->models

        if (!$namedQuery) {
            throw new HttpException(Response::HTTP_NOT_FOUND, sprintf('Query "%s" not found.', $queryId));
        }

        $query = $namedQuery->getQuery();

        $data = $this->applyQuery($request, $query, $resource, $blockPassword);
        return JsonResponse::create($data);
    }

    /**
     * @param string $resourceShortName
     * @param Request $request
     * @return \Symfony\Component\HttpFoundation\Response|static
     */
    public function submitAction($resourceShortName, Request $request)
    {
        $resource = $this->getResourceFromShortName($resourceShortName);
        if (!$resource) {
            throw new HttpException(Response::HTTP_NOT_FOUND, sprintf('Resource "%s" not found.', $resourceShortName));
        }

        $blockPassword = false;
        if ($resourceShortName == 'routers' || $resourceShortName == 'tenants' || $resourceShortName == 'agents' ) {
            $blockPassword = Utils::isPasswordBlock($request);
        }

        $content = $request->getContent();
        $query = json_decode($content, true);
        $data = $this->applyQuery($request, $query, $resource, $blockPassword);
        return JsonResponse::create($data);
    }

    /**
     * @param array $query
     * @return bool
     */
    protected function isAggregateQuery($query)
    {
        $isAggregateQuery = false;

        if (array_key_exists(Query::KEYWORD_AGGREGATES, $query)) {
            foreach ($query[Query::KEYWORD_AGGREGATES] as $aggregate) {
                if (!in_array($aggregate['operator'], [AggregateOperators::SELECT, AggregateOperators::FIELDS])) {
                    $isAggregateQuery = true;
                    break;
                }
            }
        }

        return $isAggregateQuery;
    }

    /**
     * @param $resourceShortName
     * @return ResourceInterface|null
     */
    protected function getResourceFromShortName($resourceShortName)
    {
        $resourceClassName = Inflector::singularize(Inflector::classify($resourceShortName));
        /** @var ResourceCollectionInterface $resourceCollection */
        $resourceCollection = $this->container->get('api.resource_collection');
        $resource = $resourceCollection->getResourceForShortName($resourceClassName);

        if ($resource === null) {
            $model = $this->getClassNameByType($resourceShortName);
            $singular = $model->getDescriptions()['singular'];
            $escapedSingular = preg_replace('/[^a-z0-9\s]+/i', '', $singular);
            $resourceName = \Doctrine\Common\Util\Inflector::classify($escapedSingular);
            $resource = $resourceCollection->getResourceForShortName($resourceName);
        }
        return $resource;
    }

    /**
     * @param string $queryId
     * @return Query|null
     */
    protected function getQueryById($queryId)
    {
        /** @var ManagerRegistry $managerRegistry */
        $managerRegistry = $this->container->get('doctrine');
        $queryManager = $managerRegistry->getManagerForClass(Query::class);
        $queryRepository = $queryManager->getRepository(Query::class);
        return $queryRepository->find($queryId);
    }

    /**
     * @param string $type
     * @return StagingModelMetadata|null
     */
    protected function getClassNameByType($type)
    {
        /** @var ManagerRegistry $managerRegistry */
        $managerRegistry = $this->container->get('doctrine');
        $queryManager = $managerRegistry->getManagerForClass(StagingModelMetadata::class);
        $queryRepository = $queryManager->getRepository(StagingModelMetadata::class);
        return $queryRepository->findOneBy(array('type' => $type));
    }

    /**
     * @param Request $request
     * @param $query
     * @param $resource
     * @return array|\Dunglas\ApiBundle\Model\PaginatorInterface|null|object|\Traversable
     */
    protected function applyQuery(Request $request, $query, $resource, $blockPassword=true)
    {
        /** @var AuthorizationProvider $authorizationProvider */
        $authorizationProvider = $this->get('cybersponse.authorization.authorization_provider');
        $authorizationProvider->denyAccessUnlessGranted($resource, PermissionsDictionaryBuilder::PERM_CRUD_READ);

        /** @var QueryProvider $queryProvider */
        $queryProvider = $this->container->get('cybersponse.query.provider.query_provider');
        $queryProvider->setInitialQuery($query);

        /** @var AggregateDataProviderInterface $aggregateDataProvider */
        $aggregateDataProvider = $this->container->get('cybersponse.query.data_provider');

        $data = $this->isAggregateQuery($query) ?
            $aggregateDataProvider->getAggregate($resource, $request) :
            $aggregateDataProvider->getCollection($resource, $request);

        $serializer = $this->container->get('serializer');

        $__selectFields = array();
        if (isset($query[self::SELECTIVE_FIELD_KEY]) && $query[self::SELECTIVE_FIELD_KEY]) {
            $__selectFields = array_unique(array_merge($query[self::SELECTIVE_FIELD_KEY], self::DEFAULT_KEYS), SORT_REGULAR);
        }
        return $serializer->normalize(
            $data,
            'json-ld',
            $resource->getNormalizationContext() + ['request_uri' => $request->getRequestUri(), 'required_attributes' => $__selectFields, 'block_password' => $blockPassword]
        );
    }
}
