<?php
/* Copyright start
  Copyright (C) 2008 - 2020 Fortinet Inc.
  All rights reserved.
  FORTINET CONFIDENTIAL & FORTINET PROPRIETARY SOURCE CODE
  Copyright end */

namespace Cybersponse\QueryBundle\Filter;


use Cybersponse\QueryBundle\Constants\AggregateOperators;
use Cybersponse\QueryBundle\Constants\AttributeLevelOperators;
use Cybersponse\QueryBundle\Constants\ModelLevelOperators;

class FilterArrayFactory
{
    /**
     * @param array $queryParts Example: ['field$operator' => 'value']
     * @return array
     *
     * Query Rules (Simple):
     *   single-query = [association__]field[$operator]=value
     *   full-query = single-query[&full-query]
     *
     * Query Rules (BNF):
     *   # <identifier> should actually has a more complex regex, but for simplicity...
     *   <identifier>            ::= [-_a-zA-Z0-9]+
     *
     *   <field>                 ::= <identifier>
     *   <association>           ::= <identifier>
     *   <operator>              ::= <identifier>
     *   <value>                 ::= .+
     *
     *   <field-query-lhs>       ::= <field> "$" <operator> | <field>
     *   <association-query-lhs> ::= <association> "__" <field-query-lhs>
     *   <single-query-lhs>      ::= <association-query-lhs> | <field-query-lhs>
     *   <single-query>          ::= <single-query-lhs> "=" <value>
     *   <full-query>            ::= <single-query> "&" <full-query> | <single-query>
     */
    public static function fromQueryParts($queryParts)
    {
        $filters = [];
        foreach ($queryParts as $queryPart) {
            $filterLeftHandSide = $queryPart['key'];
            $filterRightHandSide = $queryPart['value'];
            $operationParts = explode(AttributeLevelOperators::OPERATOR_TOKEN, urldecode($filterLeftHandSide), 2);
            $field = $operationParts[0];
            $operator = count($operationParts) === 1 ?
                AttributeLevelOperators::DEFAULT_OPERATOR : // $queryPart = '[association__]field=value'
            $operationParts[1];           // $queryPart = '[association__]field$operator=value'

            if (!empty($field)) {
                $value = $filterRightHandSide;

                $filter = [
                    'operator' => $operator,
                    'field' => self::convertPropertyToken($field),
                    'value' => $value
                ];
                $filters[] = $filter;
            } else {
                switch ($operator) {
                    case AggregateOperators::FIELDS:
                        $selects = explode(',', $filterRightHandSide);
                        foreach ($selects as $field) {
                            $filter = [
                                'operator' => AggregateOperators::SELECT,
                                'field' => self::convertPropertyToken($field)
                            ];
                            $filters[] = $filter;
                        }
                        break;
                    case ModelLevelOperators::ORDER_BY:
                        $orderings = explode(',', $filterRightHandSide);
                        foreach ($orderings as $ordering) {
                            if (strpos($ordering, '-') === 0) {
                                $value = 'DESC';
                                $field = substr($ordering, 1);
                            } else {
                                $value = 'ASC';
                                $field = strpos($ordering, '+') === 0 ?
                                    substr($ordering, 1) :
                                    $ordering;
                            }
                            $filter = [
                                'operator' => $operator,
                                'field' => self::convertPropertyToken($field),
                                'value' => $value
                            ];
                            $filters[] = $filter;
                        }
                        break;
                }
            }
        }
        return $filters;
    }

    /**
     * @param $rawQueryString string
     * @return array
     */
    public static function fromQueryString($rawQueryString)
    {
        $queryStringParameters = self::getQueryStringParameters($rawQueryString);
        return self::fromQueryParts($queryStringParameters);
    }

    /**
     * @param $rawQueryString string
     * @return array in the form of [['key'=>'field','value'=>'value'], ...]
     */
    protected static function getQueryStringParameters($rawQueryString)
    {
        $queryStringParts = explode('&', $rawQueryString);
        $queryStringParameters = [];
        foreach ($queryStringParts as $queryStringPart) {
            $queryStringPartPieces = explode('=', $queryStringPart);
            if (count($queryStringPartPieces) === 2) {
                $queryStringParameters[] = [
                    'key' => urldecode($queryStringPartPieces[0]),
                    'value' => urldecode($queryStringPartPieces[1])
                ];
            }
        }
        return $queryStringParameters;
    }

    /**
     * @param string $field
     * @return string
     */
    protected static function convertPropertyToken($field)
    {
        $replacementPattern = sprintf('/%s/', AttributeLevelOperators::PROPERTY_OPERATOR);
        return preg_replace($replacementPattern, '.', $field);
    }
}
