<?php
/* Copyright start
  Copyright (C) 2008 - 2020 Fortinet Inc.
  All rights reserved.
  FORTINET CONFIDENTIAL & FORTINET PROPRIETARY SOURCE CODE
  Copyright end */

namespace Cybersponse\QueryBundle\ORM;


use Cybersponse\AuthorizationBundle\Ownership\OwnershipProvider;
use Cybersponse\QueryBundle\Entity\Query;
use Cybersponse\CoreBundle\Provider\ActorDataProviderInterface;
use Cybersponse\QueryBundle\Filter\AbstractFilterQueryBuilder;
use Cybersponse\QueryBundle\Constants\AggregateOperators;
use Cybersponse\QueryBundle\Constants\AttributeLevelOperators;
use Doctrine\Common\Persistence\ManagerRegistry;
use Doctrine\ORM\Mapping\ClassMetadata;
use Doctrine\ORM\Query\Expr;
use Doctrine\ORM\Query\Expr\From;
use Doctrine\ORM\QueryBuilder;
use Dunglas\ApiBundle\JsonLd\Response;
use Cybersponse\ViewTemplateBundle\Entity\GenericTemplate;

use Symfony\Component\HttpKernel\Exception\HttpException;

class FilterQueryBuilder extends AbstractFilterQueryBuilder
{
    const AUTHORIZATION_TEAMS_PARAMETER = 'teams';
    const AUTHORIZATION_ROLES_PARAMETER = 'roles';
    const JSON_FIELD = 'jsonb';
    const GENERIC_TEMPLATE_CLASS =  ['Cybersponse\ViewTemplateBundle\Entity\GenericTemplate','Cybersponse\ViewTemplateBundle\Entity\Dashboard','Cybersponse\ViewTemplateBundle\Entity\Reporting'];

    /** @var int */
    protected $parameterCount = 0;
    /** @var array */
    protected $parameters = [];
    /** @var array */
    protected $selects = [];
    /** @var array */
    protected $orderings = [];
    /** @var array */
    protected $groupings = [];
    /** @var array */
    protected $joins = [];
    /** @var array */
    protected $associationAliases = [];
    /** @var string */
    protected $fromAlias;
    /** @var ClassMetadata */
    protected $metadata;

    /** @var QueryBuilder */
    protected $queryBuilder;

    /** @var OwnershipProvider */
    protected $ownershipProvider;

    /** @var ActorDataProviderInterface */
    protected $accessibleTeamIdsProvider;

    /** @var ManagerRegistry */
    protected $managerRegistry;

    protected $aggregateQuery = false;

     /** @var TokenStorageInterface */
     protected $tokenStorage;


    /**
     * FilterQueryBuilder constructor.
     * @param OwnershipProvider $ownershipProvider
     * @param ActorDataProviderInterface $accessibleTeamIdsProvider
     * @param ManagerRegistry $managerRegistry
     */
    public function __construct(OwnershipProvider $ownershipProvider, ActorDataProviderInterface $accessibleTeamIdsProvider, ManagerRegistry $managerRegistry)
    {
        $this->accessibleTeamIdsProvider = $accessibleTeamIdsProvider;
        $this->ownershipProvider = $ownershipProvider;
        $this->managerRegistry = $managerRegistry;
    }

    /**
     * @param QueryBuilder $queryBuilder
     * @return self
     */
    public function setQueryBuilder($queryBuilder)
    {
        $this->queryBuilder = $queryBuilder;

        return $this;
    }

    /**
     * @param array $query
     * @return self
     */
    public function fromQueryObject($query)
    {
        /** @var From $fromClause */
        $fromClause = current($this->queryBuilder->getDQLPart('from'));
        $fromEntityClass = $fromClause->getFrom();
        $this->fromAlias = $fromClause->getAlias();

        $this->metadata = $this->queryBuilder->getEntityManager()->getClassMetadata($fromEntityClass);
        // apply ownership filters
        // TODO Also apply UserOwnable authorization someday
        if ((isset($GLOBALS["__associatedFieldRequest__"]) ? $GLOBALS["__associatedFieldRequest__"] : null) != true) {
            if ($this->ownershipProvider->isClassOwnable($fromEntityClass)) {
                $ownershipExpression = $this->getOwnershipExpression($this->fromAlias);
                if ($ownershipExpression) {
                    $this->queryBuilder->andWhere($ownershipExpression);
                }
            }
        }
        if ( in_array($fromEntityClass,self::GENERIC_TEMPLATE_CLASS)){
            $userIdExpression = $this->getUserExpression($this->fromAlias);
            if ($userIdExpression) {
                $this->queryBuilder->orWhere($userIdExpression);
            }
            $ownershipExpression = $this->getRoleExpression($this->fromAlias);
            if ($ownershipExpression) {
                $this->queryBuilder->orWhere($ownershipExpression);
            }
           
        }
        if ($this->ownershipProvider->isClassConditionalOwnable($fromEntityClass)) {
            $conditionalOwnershipExpression = $this->getOwnershipExpression($this->fromAlias);
            if ($conditionalOwnershipExpression) {
                $this->queryBuilder->andWhere($conditionalOwnershipExpression);
            }
        }
        // apply query filters
        $filterExpression = $this->getFilterExpression($query);
        if ($filterExpression) {
            $this->queryBuilder->andWhere($filterExpression);
        }

        // apply parameters
        if (!empty($this->parameters)) {
            $this->queryBuilder->setParameters($this->parameters);
        }

        if (array_key_exists(Query::KEYWORD_AGGREGATES, $query)) {
            foreach ($query[Query::KEYWORD_AGGREGATES] as $aggregate) {
                $aggregateExpression = $this->getAggregateExpression($aggregate);
                if (!is_null($aggregateExpression)) {
                    $this->selects[] = $aggregateExpression;
                }
            }
        }

        if (array_key_exists(Query::KEYWORD_SORT, $query)) {
            foreach ($query[Query::KEYWORD_SORT] as $ordering) {
                $fieldAlias = $this->getFieldAlias($ordering['field']);
                $this->orderings[] = [
                    'field' => $fieldAlias,
                    'direction' => $ordering['direction']
                ];
            }
        }

        if (!$this->aggregateQuery && empty($this->groupings) && !empty($this->selects)) {
            $primaryIdentifier = $this->metadata->getIdentifier()[0];
            $this->selects[] = sprintf('o.%s', $primaryIdentifier);
            $this->queryBuilder->distinct();
        }

        if (!$this->aggregateQuery && empty($this->orderings) && empty($this->groupings)) {
            // Apply default ordering
            // Todo this should be set by the metadata
            $identifier = $this->metadata->getIdentifier();

            $orderByField = current($identifier);

            $this->orderings[] = [
                'field' => sprintf('o.%s', $orderByField),
                'direction' => 'ASC'
            ];
        }

        if ($this->orderings) {
            foreach ($this->orderings as $ordering) {
                $this->queryBuilder->addOrderBy($ordering['field'], $ordering['direction']);
            }
        }

        if (!empty($this->groupings)) {
            foreach ($this->groupings as $grouping) {
                $this->queryBuilder->addGroupBy($grouping);
            }
        }

        if (count($this->selects) > 0) {
            $this->queryBuilder->select(implode(', ', $this->selects));
        }

        // Apply join clauses
        // TODO: How does this sort 10+ joins (t0.asdf, t1.asdf, t2.asdf, ..., t10.asdf)
        usort($this->joins, function ($join1, $join2) {
            return strcasecmp($join1['alias'], $join2['alias']);
        });

        foreach ($this->joins as $join) {
            // TODO: This is out of order

            if ($join['type'] && $this->metadata::TO_ONE) {
                $this->queryBuilder->leftJoin($join['join'], $join['alias']);
            } else {
                $this->queryBuilder->join($join['join'], $join['alias']);
            }
        }

        return $this;
    }

    /**
     * @param array $aggregateFilter
     * @return null|string
     */
    protected function getAggregateExpression($aggregateFilter)
    {
        $aggregateExpression = null;
        $aggregateOperator = $aggregateFilter['operator'];
        $this->convertFilterFieldToDotNotation($aggregateFilter);
        $fieldAlias = $this->getFieldAlias($aggregateFilter['field'], isset($aggregateFilter['fieldOperator']) ? $aggregateFilter['fieldOperator'] : null);

        switch ($aggregateOperator) {
            case AggregateOperators::SUM:
                $this->aggregateQuery = true;
                $aggregateExpression = sprintf('SUM(%s)', $fieldAlias);
                break;
            case AggregateOperators::COUNT_DISTINCT:
                $this->aggregateQuery = true;
                $aggregateExpression = $this->getExpression()->countDistinct($fieldAlias);
                break;
            case AggregateOperators::COUNT:
                $this->aggregateQuery = true;
                $aggregateExpression = $this->getExpression()->count($fieldAlias);
                break;
            case AggregateOperators::GROUP_BY:
                $this->groupings[] = $fieldAlias;
                $aggregateExpression = $fieldAlias;
                break;
            case AggregateOperators::SELECT:
            case AggregateOperators::FIELDS:
                $aggregateExpression = $fieldAlias;
                break;
            case AggregateOperators::MAX:
                $this->aggregateQuery = true;
                $aggregateExpression = $this->getExpression()->max($fieldAlias);
                break;
            case AggregateOperators::MIN:
                $this->aggregateQuery = true;
                $aggregateExpression = $this->getExpression()->min($fieldAlias);
                break;
            case AggregateOperators::AVG:
                $this->aggregateQuery = true;
                $aggregateExpression = $this->getExpression()->avg($fieldAlias);
                break;
            case AggregateOperators::MEDIAN:
                $this->aggregateQuery = true;
                $aggregateExpression = sprintf('percentile_cont(0.5) within group (order by %s)', $fieldAlias);
                break;
            case AggregateOperators::DISTINCT:
                break;
        }

        if (array_key_exists('alias', $aggregateFilter)) {
            $resultAlias = $aggregateFilter['alias'];
            $aggregateExpression = sprintf('%s AS %s', $aggregateExpression, $resultAlias);
        }

        return $aggregateExpression;
    }

    /**
     * TODO: Move this to AuthorizationBundle
     * @param string $fromAlias
     * @return Expr
     */
    protected function getOwnershipExpression($fromAlias)
    {
        $accessibleTeams = $this->accessibleTeamIdsProvider->getDataForCurrentUser();

        $ownersAlias = $this->getAssociationAlias($this->getPropertyPath($fromAlias, 'owners'));

        $this->parameters[':' . self::AUTHORIZATION_TEAMS_PARAMETER] = $accessibleTeams;
        return $this->getExpression()->in($ownersAlias, ':' . self::AUTHORIZATION_TEAMS_PARAMETER) . ' or ' . $this->getExpression()->isNull($ownersAlias);
    }

     /**
     * TODO: Move this to AuthorizationBundle
     * @param string $fromAlias
     * @return Expr
     */
    protected function getRoleExpression($fromAlias)
    {
        $accessibleRole = $this->getCurrentUserRoles();

        // $userAlias = $this->getAssociationAlias($this->getPropertyPath($fromAlias, 'userid'));
        $roleAlias = $this->getAssociationAlias($this->getPropertyPath($fromAlias, 'roles'));
        $userUuid = $this->getCurrentUserAsIri();
        $this->parameters[':' . self::AUTHORIZATION_ROLES_PARAMETER] = $accessibleRole;
                   
        return $this->getExpression()->in($roleAlias, ':' . self::AUTHORIZATION_ROLES_PARAMETER);// . ' or ' . $this->getExpression()->eq($fromAlias.'.userid',$userUuid);
    }

     /**
     * TODO: Move this to AuthorizationBundle
     * @param string $fromAlias
     * @return Expr
     */
    protected function getUserExpression($fromAlias)
    {
        $userUuid = $this->getCurrentUserAsIri();

        // $userAlias = $this->getAssociationAlias($this->getPropertyPath($fromAlias, 'userid'));

        $this->parameters[':createuser'] = $userUuid;
        $this->parameters[':uuid']= '82bc8209-ef8d-463d-bcd8-db3ac08882cb-463d';
        return $this->getExpression()->eq($fromAlias.'.createUser',':createuser'). ' or ' . $this->getExpression()->eq($fromAlias.'.uuid',':uuid' );
    }
    


    /**
     * @return array string
     */
    protected function getCurrentUserRoles()
    {
         /** @var  $user Actor */
         $user = $this->tokenStorage->getToken()->getUser();
         $roles = $user->getRoles();
         $roleIds = array();
         if (!is_null($roles)) {
             foreach ($roles as $role) {
                 array_push($roleIds, $role->getUuid());
             }
         }
         return $roleIds;
    }

    /**
     * @param TokenStorageInterface $tokenStorage
     */
    public function setTokenStorage($tokenStorage)
    {
        $this->tokenStorage = $tokenStorage;
    }

    /**
     * @return string
     */
    private function getCurrentUserAsIri()
    {
        $user = $this->tokenStorage->getToken()->getUser()->getUuid();
        // $iri = $this->iriConverter->getIriFromItem($user);

        return $user;
    }

    /**
     * @param array $filter
     * @return mixed|null
     * @throws \HttpException
     */
    protected function getFilterExpression($filter)
    {
        $expression = null;

        if (array_key_exists(Query::KEYWORD_LOGIC, $filter) && array_key_exists(Query::KEYWORD_FILTERS, $filter)) {
            $logic = $filter[Query::KEYWORD_LOGIC];
            switch ($logic) {
                case 'AND':
                    $logicMethod = 'andX';
                    break;
                case 'OR':
                    $logicMethod = 'orX';
                    break;
                default:
                    $errormsg = (sprintf('Unknown logic type specified: "%s"', $logic));
                    throw new HttpException(Response::HTTP_BAD_REQUEST, $errormsg);
            }

            $filters = $filter[Query::KEYWORD_FILTERS];
            $expressions = [];

            foreach ($filters as $nestedFilter) {
                $nestedExpression = $this->getFilterExpression($nestedFilter);
                if ($nestedExpression) {
                    $expressions[] = $nestedExpression;
                }
            }

            if (count($expressions) > 1) {
                $expression = call_user_func_array([$this->getExpression(), $logicMethod], $expressions);
            } else if (count($expressions) === 1) {
                $expression = current($expressions);
            }
        } else {
            $expression = $this->getFieldExpression($filter);
        }

        return $expression;
    }


    /**
     * @param ClassMetadata $metadata
     * @param string $field
     * @return bool
     */
    protected function isAssociation($metadata, $field)
    {
        $fieldParts = explode('.', $field);
        $isAssociation = false;
        $currentField = array_shift($fieldParts);
        if ($currentField == $field and count($fieldParts) == 0) {
            $this->isField($metadata, $field);
        }
        if ($metadata->hasAssociation($currentField)) {
            if (count($fieldParts) > 0) {
                $nextField = implode('.', $fieldParts);
                $associationMapping = $metadata->getAssociationMapping($currentField);
                $targetEntity = $associationMapping['targetEntity'];
                $associatedMetadata = $this->getMetadataForClass($targetEntity);
                $isAssociation = $this->isAssociation($associatedMetadata, $nextField);
            } else {
                $isAssociation = true;
            }
        }

        return $isAssociation;
    }

    /**
     * @param string $field
     * @param string $value
     * @return array
     */
    protected function splitFieldFromJson($field, $value)
    {
        $queryParts = explode('.', $field);
        $currentField = array_shift($queryParts);
        $fieldParts = [$currentField];
        $metadata = $this->metadata;
        while ($metadata->hasAssociation($currentField)) {
            $metadata = $this->getMetadataForClass($metadata->associationMappings[$currentField]['targetEntity']);
            $currentField = array_shift($queryParts);
            if ($currentField) {
                $fieldParts[] = $currentField;
            }
        }

        $isJson = false;
        if ($metadata->hasField($currentField) && $metadata->fieldMappings[$currentField]['type'] == self::JSON_FIELD) {
            $isJson = true;
        }

        $jsonObject = $value;
        foreach (array_reverse($queryParts) as $part) {
            $jsonObject = [
                $part => $jsonObject
            ];
        }

        $fieldAlias = call_user_func_array([$this, 'getPropertyPath'], $fieldParts);
        return [$fieldAlias, $jsonObject, $isJson];
    }

    /**
     * @param string $field
     * @return array
     */
    protected function getJsonQuery($field)
    {
        $queryParts = explode('.', $field);
        $currentField = array_shift($queryParts);
        $metadata = $this->metadata;
        while ($metadata->hasAssociation($currentField)) {
            $metadata = $this->getMetadataForClass($metadata->associationMappings[$currentField]['targetEntity']);
            $currentField = array_shift($queryParts);
        }

        return $queryParts;
    }

    /**
     * @param $class
     * @return ClassMetadata
     */
    protected function getMetadataForClass($class)
    {
        $manager = $this->managerRegistry->getManagerForClass($class);
        return $manager->getClassMetadata($class);
    }

    /**
     * @param array $filter
     * @return Expr\Comparison|null
     */
    protected function getFieldExpression(array $filter)
    {
        if (!array_key_exists('field', $filter)) {
            throw new HttpException(Response::HTTP_BAD_REQUEST, 'Filter must have a field.');
        }
        if (!array_key_exists('operator', $filter)) {
            throw new HttpException(Response::HTTP_BAD_REQUEST, 'Filter must have an operator.');
        }

        $this->convertFilterFieldToDotNotation($filter);
        $field = $filter['field'];
        list($field, $json, $isJson) = $this->splitFieldFromJson($field, $filter['value']);
        $fieldAlias = $this->getFieldAlias($field);
        $isAssociation = $this->isAssociation($this->metadata, $field);
        $operator = strtolower($filter['operator']);

        $expression = null;

        if (array_key_exists('value', $filter)) {
            $value = $filter['value'];
            if ($isAssociation && is_string($value) && preg_match('/^\/api\//', $value)) {
                #For fixing error in php7
                #https://stackoverflow.com/questions/4636166/only-variables-should-be-passed-by-reference
                $tmp = explode('/', $value);
                $value = end($tmp);
            }

            switch ($operator) {
                case AttributeLevelOperators::EQUALS:
                    $value = !is_string($json) ? json_encode($json) : $value;
                    $expression = $this->getExpression()->eq($fieldAlias, $this->getNextParameterName($value));
                    break;
                case AttributeLevelOperators::NOT_EQUALS:
                    $value = !is_string($json) ? json_encode($json) : $value;
                    $expression = $this->getExpression()->neq($fieldAlias, $this->getNextParameterName($value));
                    break;
                case AttributeLevelOperators::LESS_THAN:
                    $expression = $this->getExpression()->lt($fieldAlias, $this->getNextParameterName($value));
                    break;
                case AttributeLevelOperators::LESS_THAN_EQUALS:
                    $expression = $this->getExpression()->lte($fieldAlias, $this->getNextParameterName($value));
                    break;
                case AttributeLevelOperators::GREATER_THAN:
                    $expression = $this->getExpression()->gt($fieldAlias, $this->getNextParameterName($value));
                    break;
                case AttributeLevelOperators::GREATER_THAN_EQUALS:
                    $expression = $this->getExpression()->gte($fieldAlias, $this->getNextParameterName($value));
                    break;
                case AttributeLevelOperators::IS_NULL:
                    $expressionBuilder = $this->getExpression();
                    $expression = filter_var($value, FILTER_VALIDATE_BOOLEAN)
                        ? $expressionBuilder->isNull($fieldAlias)
                        : $expressionBuilder->isNotNull($fieldAlias);
                    break;
                case AttributeLevelOperators::IN:
                    if (!is_array($value)) {
                        $value = explode('|', $value);
                    }

                    if ($isAssociation) {
                        foreach ($value as $key => $item) {
                            if (is_string($item) && preg_match('/^\/api\//', $item)) {
                                $value[$key] = end(explode('/', $item));
                            }
                        }
                    }

                    $expression = $this->getExpression()->in($fieldAlias, $this->getNextParameterName($value));
                    break;
                case AttributeLevelOperators::NIN:
                    if (!is_array($value)) {
                        $value = explode('|', $value);
                    }

                    if ($isAssociation) {
                        foreach ($value as $key => $item) {
                            if (is_string($item) && preg_match('/^\/api\//', $item)) {
                                $value[$key] = end(explode('/', $item));
                            }
                        }
                    }

                    $expression = $this->getExpression()->notIn($fieldAlias, $this->getNextParameterName($value));
                    break;
                case AttributeLevelOperators::LIKE:
                    $value = strtolower($value);
                    if ($isJson) {
                        $expression = $this->getExpression()->jsonLike($fieldAlias, $this->getJsonQuery($filter['field']), $this->getNextParameterName($value));
                    } else {
                        $expression = $this->getExpression()->like('TEXT_LOWER(' . $fieldAlias . ')', $this->getNextParameterName($value));
                    }
                    break;
                case AttributeLevelOperators::NOT_LIKE:
                    $value = strtolower($value);
                    # ToDo Add Support for JSON Not like
                    if (!$isJson) {
                        $expression = $this->getExpression()->notLike('TEXT_LOWER(' . $fieldAlias . ')', $this->getNextParameterName($value));
                    }
                    break;
                case AttributeLevelOperators::CONTAINS:
                    $value = json_encode($json);
                    $expression = $this->getExpression()->jsonContains($fieldAlias, $this->getNextParameterName($value));
                    break;
                case AttributeLevelOperators::EXISTS:
                    $expression = $this->getExpression()->jsonExists($fieldAlias, $this->getJsonQuery($filter['field']), $this->getNextParameterName($value));
                    break;
                default:
                    $this->handleUnsupportedOperator($operator);
            }
        }

        return $expression;
    }

    /**
     * @param string $association
     * @param boolean $addJoins
     * @return string
     */
    protected function getAssociationAlias($association, $addJoins = true)
    {
        if (count(explode('.', $association)) < 2) {
            return $association;
        }

        if ($addJoins) {
            $this->addAssociationJoins($association);
        }

        if (!array_key_exists($association, $this->associationAliases)) {
            $this->associationAliases[$association] = 't' . count($this->associationAliases);
        }

        return $this->associationAliases[$association];
    }

    /**
     * @param string $association Example: association.field OR association1.association2.field
     */
    protected function addAssociationJoins($association)
    {
        if (!array_key_exists($association, $this->joins)) {
            $associations = explode('.', $association);
            $parent = current($associations);
            $metadata = $this->metadata;
            foreach ($associations as $index => $partialAssociation) {
                if ($index === 0) {
                    continue;
                }

                $fullAssociation = $this->getPropertyPath($parent, $partialAssociation);

                $parentAlias = $this->getAssociationAlias($parent, false);
                $alias = $this->getAssociationAlias($fullAssociation, false);

                $this->joins[$fullAssociation] = [
                    'join' => $this->getPropertyPath($parentAlias, $partialAssociation),
                    'alias' => $alias,
                    'type' => $metadata->associationMappings[$partialAssociation]['type']
                ];

                // prep for next iteration
                $metadata = $this->getMetadataForClass($metadata->associationMappings[$partialAssociation]['targetEntity']);
                $parent = $fullAssociation;
            }
        }
    }

    /**
     * @param $value
     * @return string
     */
    protected function getNextParameterName($value)
    {
        $parameterName = ':param' . $this->parameterCount++;
        $this->parameters[$parameterName] = $value;
        return $parameterName;
    }

    /**
     * @param string $fieldPath
     * @return null|string
     */
    protected function getFieldAlias($fieldPath, $operation = null)
    {
        if (!$fieldPath) {
            return null;
        }

        if ($fieldPath === '*') {
            return $this->fromAlias;
        }

        $fieldPath = $this->getPropertyPath($this->fromAlias, $fieldPath);
        $aliasParts = explode('.', $fieldPath);

        if (count($aliasParts) > 2) {
            $associationParts = array_slice($aliasParts, 0, count($aliasParts) - 1);
            $field = end($aliasParts);
            $associationPath = call_user_func_array([$this, 'getPropertyPath'], $associationParts);
            $associationAlias = $this->getAssociationAlias($associationPath);
            $fieldAlias = $this->getPropertyPath($associationAlias, $field);
        } else {
            $field = end($aliasParts);
            if ($this->metadata->hasAssociation($field)) {
                $associationAlias = $this->getPropertyPath($this->fromAlias, $field);
                $fieldAlias = $this->getAssociationAlias($associationAlias);
            } else {
                $fieldAlias = $this->getPropertyPath($this->fromAlias, $field);
            }
        }
        //Computation Queries
        $fieldAlias = $this->checkForComputeQuery($fieldAlias, $operation);
        return $fieldAlias;
    }

    protected function checkForComputeQuery($fieldAlias, $fieldOperator = null)
    {
        $roiFieldExpr = explode(',', $fieldAlias);
        if (count($roiFieldExpr) > 1) {
            $op = '-';
            if ($fieldOperator == 'sum') {
                $op = '+';
            }
            $fieldAlias = str_replace(",", $op . 'o.', $fieldAlias);
        }
        return $fieldAlias;
    }

    /**
     * @return string
     */
    protected function getPropertyPath()
    {
        $propertyPathParts = func_get_args();
        return implode('.', $propertyPathParts);
    }

    /**
     * @return JsonExpressionBuilder
     */
    protected function getExpression()
    {
        return new JsonExpressionBuilder();
    }

    /**
     * @param $filter
     */
    protected function convertFilterFieldToDotNotation(&$filter)
    {
        $filter['field'] = preg_replace(sprintf('/%s/', AttributeLevelOperators::PROPERTY_OPERATOR), '.', $filter['field']);
    }


    /**
     * @param ClassMetadata $metadata
     * @param string $field
     * @return bool
     */
    private function isField(ClassMetadata $metadata, $field)
    {
        if (!($metadata->hasField($field) or
            $metadata->hasAssociation($field))) {
            throw new HttpException(Response::HTTP_BAD_REQUEST, 'Invalid query/body parameters [Field:' . $field . ' does not exist in module definition]');
        }
        return true;
    }
}
