<?php
/* Copyright start
  Copyright (C) 2008 - 2020 Fortinet Inc.
  All rights reserved.
  FORTINET CONFIDENTIAL & FORTINET PROPRIETARY SOURCE CODE
  Copyright end */

namespace Cybersponse\QueryBundle\Provider\ORM;


use Doctrine\ORM\QueryBuilder;
use Doctrine\ORM\EntityManager;
use Dunglas\ApiBundle\JsonLd\Response;
use Doctrine\ORM\Mapping\ClassMetadataInfo;
use Dunglas\ApiBundle\Api\ResourceInterface;
use Dunglas\ApiBundle\Doctrine\Orm\Paginator;
use Symfony\Component\HttpFoundation\Request;
use Doctrine\Common\Persistence\ManagerRegistry;
use Dunglas\ApiBundle\Doctrine\Orm\DataProvider;
use Dunglas\ApiBundle\Doctrine\Orm\Util\QueryChecker;
use Symfony\Component\HttpKernel\Exception\HttpException;
use Dunglas\ApiBundle\Doctrine\Orm\Filter\FilterInterface;
use Cybersponse\CoreBundle\Doctrine\Orm\EfficientPaginator;
use Cybersponse\QueryBundle\Provider\AggregateDataProviderInterface;
use Doctrine\ORM\Tools\Pagination\Paginator as DoctrineOrmPaginator;

class AggregateDataProvider extends DataProvider implements AggregateDataProviderInterface
{
    /** @var ManagerRegistry */
    protected $managerRegistry;

    /** @var string|null */
    protected $order;

    /** @var string */
    protected $pageParameter;

    /** @var int */
    protected $itemsPerPage;

    /** @var bool */
    protected $enableClientRequestItemsPerPage;

    /** @var string */
    protected $itemsPerPageParameter;

    /**
     * @param ManagerRegistry $managerRegistry
     * @param string|null     $order
     * @param string          $pageParameter
     * @param int             $itemsPerPage
     * @param bool            $enableClientRequestItemsPerPage
     * @param string          $itemsPerPageParameter
     */
    public function __construct(
        ManagerRegistry $managerRegistry,
        $order,
        $pageParameter,
        $itemsPerPage,
        $enableClientRequestItemsPerPage,
        $itemsPerPageParameter
    ) {
        $this->managerRegistry = $managerRegistry;
        $this->order = $order;
        $this->pageParameter = $pageParameter;
        $this->itemsPerPage = $itemsPerPage;
        $this->enableClientRequestItemsPerPage = $enableClientRequestItemsPerPage;
        $this->itemsPerPageParameter = $itemsPerPageParameter;
        parent::__construct($managerRegistry, $order, $pageParameter, $itemsPerPage, $enableClientRequestItemsPerPage, $itemsPerPageParameter);
    }

    /**
     * @inheritdoc
     */
    public function getAggregate(ResourceInterface $resource, Request $request)
    {
        $entityClass = $resource->getEntityClass();

        /** @var EntityManager $manager */
        $manager = $this->managerRegistry->getManagerForClass($resource->getEntityClass());
        $repository = $manager->getRepository($entityClass);

        $page = (int)$request->get($this->pageParameter, 1);

        $itemsPerPage = $this->itemsPerPage;
        if ($this->enableClientRequestItemsPerPage && $requestedItemsPerPage = $request->get($this->itemsPerPageParameter)) {
            $itemsPerPage = (int)$requestedItemsPerPage;
        }

        $queryBuilder = $repository
            ->createQueryBuilder('o')
            ->setFirstResult(($page - 1) * $itemsPerPage)
            ->setMaxResults($itemsPerPage);

        foreach ($resource->getFilters() as $filter) {
            if ($filter instanceof FilterInterface) {
                $filter->apply($resource, $queryBuilder, $request);
            }
        }

        $classMetaData = $manager->getClassMetadata($entityClass);
        $identifiers = $classMetaData->getIdentifier();

        foreach ($classMetaData->getAssociationNames() as $i => $association) {
            $mapping = $classMetaData->associationMappings[$association];

            if (ClassMetadataInfo::FETCH_EAGER === $mapping['fetch']) {
                $queryBuilder->leftJoin('o.' . $association, 'a' . $i);
                $queryBuilder->addSelect('a' . $i);
            }
        }

        if (null !== $this->order && 1 === count($identifiers)) {
            $identifier = $identifiers[0];
            $queryBuilder->addOrderBy('o.' . $identifier, $this->order);
        }

        try {
            return $queryBuilder->getQuery()->execute();
        } catch (\Throwable $th) {
            $errMsg = $th->getmessage();
            throw new HttpException(Response::HTTP_BAD_REQUEST, $errMsg);
        }
    }

    /**
     * Gets the paginator.
     *
     * @param QueryBuilder $queryBuilder
     *
     * @return Paginator
     */
    protected function getPaginator(QueryBuilder $queryBuilder)
    {
        $doctrineOrmPaginator = new DoctrineOrmPaginator($queryBuilder);
        // Disable output walkers by default (performance)
        $doctrineOrmPaginator->setUseOutputWalkers($this->useOutputWalkers($queryBuilder));

        return new EfficientPaginator($doctrineOrmPaginator);
    }

    /**
     * Determines whether output walkers should be used.
     *
     * @param QueryBuilder $queryBuilder
     *
     * @return bool
     */
    private function useOutputWalkers(QueryBuilder $queryBuilder)
    {
        /*
         * "Cannot count query that uses a HAVING clause. Use the output walkers for pagination"
         *
         * @see https://github.com/doctrine/doctrine2/blob/900b55d16afdcdeb5100d435a7166d3a425b9873/lib/Doctrine/ORM/Tools/Pagination/CountWalker.php#L50
         */
        if (QueryChecker::hasHavingClause($queryBuilder)) {
            return true;
        }
        /*
         * "Paginating an entity with foreign key as identifier only works when using the Output Walkers. Call Paginator#setUseOutputWalkers(true) before iterating the paginator."
         *
         * @see https://github.com/doctrine/doctrine2/blob/900b55d16afdcdeb5100d435a7166d3a425b9873/lib/Doctrine/ORM/Tools/Pagination/LimitSubqueryWalker.php#L87
         */
        if (QueryChecker::hasRootEntityWithForeignKeyIdentifier($queryBuilder, $this->managerRegistry)) {
            return true;
        }
        /*
         * "Cannot select distinct identifiers from query with LIMIT and ORDER BY on a column from a fetch joined to-many association. Use output walkers."
         *
         * @see https://github.com/doctrine/doctrine2/blob/900b55d16afdcdeb5100d435a7166d3a425b9873/lib/Doctrine/ORM/Tools/Pagination/LimitSubqueryWalker.php#L149
         */
        if (QueryChecker::hasMaxResults($queryBuilder)
            && QueryChecker::hasOrderByOnToManyJoin($queryBuilder, $this->managerRegistry)) {
            return true;
        }

        // Disable output walkers by default (performance)
        return false;
    }

    public function startsWith ($string, $startString)
    {
        $len = strlen($startString);
        return (substr($string, 0, $len) === $startString);
    }

    /**
     * {@inheritdoc}
     */
    public function getItemWithWhere(ResourceInterface $resource, $filters, $fetchData = false)
    {
        $entityClass = $resource->getEntityClass();
        $manager = $this->managerRegistry->getManagerForClass($entityClass);
        $repository = $manager->getRepository($entityClass);

        $queryBuilder = $repository->createQueryBuilder('o');

        foreach ($filters as $key => $value) {
            if (gettype($value) == 'array'){
                $queryBuilder->andwhere('o.'. $key . ' in (:' . $key . ')');
                $queryBuilder->setParameter($key, $value);
                continue;
            }
            if (gettype($value) != 'boolean') {
                if ($this->startsWith($value, '/api/3/')) {
                    $explodedValue = explode('/', $value);
                    $value = end($explodedValue);
                }
                $queryBuilder->andwhere('o.'. $key . '= :' . $key);
                $queryBuilder->setParameter($key, $value);
            } else {
                $queryBuilder->andwhere('o.'. $key . '= :value');
                $queryBuilder->setParameter('value', $value);
            }
        }

        return $queryBuilder->getQuery()->execute();
    }
}
