<?php
/* Copyright start
  Copyright (C) 2008 - 2020 Fortinet Inc.
  All rights reserved.
  FORTINET CONFIDENTIAL & FORTINET PROPRIETARY SOURCE CODE
  Copyright end */

namespace Cybersponse\QueryBundle\Tests\Unit\Filter\FilterArrayFactory;


use Cybersponse\QueryBundle\Constants\AttributeLevelOperators;
use Cybersponse\QueryBundle\Filter\FilterArrayFactory;

class FromQueryPartsMethodTest extends \PHPUnit_Framework_TestCase
{
    /** @var array */
    protected $queryParts;

    protected function executeSystemUnderTest()
    {
        return FilterArrayFactory::fromQueryParts($this->queryParts);
    }

    protected function tearDown()
    {
        unset($this->queryParts);
        parent::tearDown();
    }

    /**
     * Query: []
     *
     * Filter:
     * []
     */
    public function test_GivenEmptyQueryString_ShouldReturnEmptyArray()
    {
        $this->queryParts = [];
        $filters = $this->executeSystemUnderTest();
        $this->assertEquals([], $filters);
    }

    /**
     * Query: ["field" => "value"]
     *
     * Filter:
     * [
     *   {
     *     field: 'field',
     *     operator: 'eq',
     *     value: 'value'
     *   }
     * ]
     */
    public function test_GivenSimpleField_ShouldReturnFilterWithOperatorKey()
    {
        $this->queryParts = [
            [
                'key' => 'field',
                'value' => 'value'
            ]
        ];
        $filters = $this->executeSystemUnderTest();
        $this->assertArrayHasKey('operator', $filters[0]);
    }

    /**
     * Query: ["field" => "value"]
     *
     * Filter:
     * [
     *   {
     *     field: 'field',
     *     operator: 'eq',
     *     value: 'value'
     *   }
     * ]
     */
    public function test_GivenSimpleField_ShouldReturnFilterWithFieldKey()
    {
        $this->queryParts = [
            [
                'key' => 'field',
                'value' => 'value'
            ]
        ];
        $filters = $this->executeSystemUnderTest();
        $this->assertArrayHasKey('field', $filters[0]);
    }

    /**
     * Query: ["field" => "value"]
     *
     * Filter:
     * [
     *   {
     *     field: 'field',
     *     operator: 'eq',
     *     value: 'value'
     *   }
     * ]
     */
    public function test_GivenSimpleField_ShouldReturnFilterWithValueKey()
    {
        $this->queryParts = [
            [
                'key' => 'field',
                'value' => 'value'
            ]
        ];
        $filters = $this->executeSystemUnderTest();
        $this->assertArrayHasKey('value', $filters[0]);
    }

    /**
     * Query: ["field" => "value"]
     *
     * Filter:
     * [
     *   {
     *     field: 'field',
     *     operator: 'eq',
     *     value: 'value'
     *   }
     * ]
     */
    public function test_GivenSimpleField_ShouldReturnSingleFilter()
    {
        $this->queryParts = [
            [
                'key' => 'field',
                'value' => 'value'
            ]
        ];
        $filters = $this->executeSystemUnderTest();
        $this->assertCount(1, $filters);
    }

    /**
     * Query: ?field=value
     *
     * Filter:
     * [
     *   {
     *     field: 'field',
     *     operator: 'eq',
     *     value: 'value'
     *   }
     * ]
     */
    public function test_GivenSimpleField_ShouldReturnFilterWithDefaultOperator()
    {
        $this->queryParts = [
            [
                'key' => 'field',
                'value' => 'value'
            ]
        ];
        $filters = $this->executeSystemUnderTest();
        $this->assertEquals(AttributeLevelOperators::DEFAULT_OPERATOR, $filters[0]['operator']);
    }

    /**
     * Query: ["someField" => "value"]
     *
     * Filter:
     * [
     *   {
     *     field: 'someField',
     *     operator: 'eq',
     *     value: 'value'
     *   }
     * ]
     */
    public function test_GivenSimpleField_ShouldReturnFilterWithSpecifiedField()
    {
        $expectedField = 'someField';
        $this->queryParts = [
            [
                'key' => $expectedField,
                'value' => 'value'
            ]
        ];
        $filters = $this->executeSystemUnderTest();
        $this->assertEquals($expectedField, $filters[0]['field']);
    }

    /**
     * Query: ["field" => "someValue"]
     *
     * Filter:
     * [
     *   {
     *     field: 'field',
     *     operator: 'eq',
     *     value: 'someValue'
     *   }
     * ]
     */
    public function test_GivenSimpleField_ShouldReturnFilterWithSpecifiedValue()
    {
        $expectedValue = 'someValue';
        $this->queryParts = [
            [
                'key' => 'field',
                'value' => $expectedValue
            ]
        ];
        $filters = $this->executeSystemUnderTest();
        $this->assertEquals($expectedValue, $filters[0]['value']);
    }

    /**
     * Query: ["field1" => "value1", "field2" => "value2"]
     *
     * Filter:
     * [
     *   {
     *     field: 'field1',
     *     operator: 'eq',
     *     value: 'value1'
     *   },
     *   {
     *     field: 'field2',
     *     operator: 'eq',
     *     value: 'value2'
     *   }
     * ]
     */
    public function test_GivenTwoUniqueFields_ShouldReturnTwoFilters()
    {
        $this->queryParts = [
            [
                'key' => 'field1',
                'value' => 'value1'
            ],
            [
                'key' => 'field2',
                'value' => 'value2'
            ]
        ];
        $filters = $this->executeSystemUnderTest();
        $this->assertCount(2, $filters);
    }

    /**
     * Query: ["field$gt" => "value"]
     *
     * Filter:
     * [
     *   {
     *     field: 'field',
     *     operator: 'gt',
     *     value: 'value'
     *   }
     * ]
     */
    public function test_GivenOperator_ShouldReturnSpecifiedOperator()
    {
        $expectedOperator = AttributeLevelOperators::GREATER_THAN;
        $this->queryParts = [
            [
                'key' => "field$$expectedOperator",
                'value' => 'value'
            ]
        ];
        $filters = $this->executeSystemUnderTest();
        $this->assertEquals($expectedOperator, $filters[0]['operator']);
    }
}
