<?php
/* Copyright start
  Copyright (C) 2008 - 2020 Fortinet Inc.
  All rights reserved.
  FORTINET CONFIDENTIAL & FORTINET PROPRIETARY SOURCE CODE
  Copyright end */

namespace Cybersponse\QueryBundle\Tests\Unit\ORM;


use Cybersponse\AuthorizationBundle\Entity\Team;
use Cybersponse\AuthorizationBundle\Ownership\AccessibleTeamIdsProvider;
use Cybersponse\AuthorizationBundle\Ownership\OwnershipProvider;
use Cybersponse\QueryBundle\ORM\FilterQueryBuilder;
use Doctrine\Common\Persistence\ManagerRegistry;
use Doctrine\ORM\EntityManager;
use Doctrine\ORM\EntityManagerInterface;
use Doctrine\ORM\Mapping\ClassMetadata;
use Doctrine\ORM\Query;
use Doctrine\ORM\QueryBuilder;

class FilterQueryBuilderTest extends \PHPUnit_Framework_TestCase
{
    const CLASS_TO_QUERY = 'ClassToQuery';
    const ALIAS_TO_QUERY = 'o';

    /** @var FilterQueryBuilder */
    protected $filterQueryBuilder;

    /** @var array */
    protected $query;

    /** @var OwnershipProvider|\PHPUnit_Framework_MockObject_MockObject */
    protected $mockOwnershipProvider;

    /** @var AccessibleTeamIdsProvider|\PHPUnit_Framework_MockObject_MockObject */
    protected $mockAccessibleTeamIdsProvider;

    /** @var ManagerRegistry|\PHPUnit_Framework_MockObject_MockObject */
    protected $mockManagerRegistry;

    protected $invalidFilter = [
        'BAD' => 'FILTER'
    ];

    protected $validFilter = [
        'field' => 'age',
        'operator' => 'eq',
        'value' => 25
    ];

    protected function setUp()
    {
        $this->mockAccessibleTeamIdsProvider = $this->getMockBuilder(AccessibleTeamIdsProvider::class)->getMock();
        $this->mockOwnershipProvider = $this->getMockBuilder(OwnershipProvider::class)->getMock();
        $this->mockManagerRegistry = $this->getMockBuilder(ManagerRegistry::class)->getMock();
        $this->filterQueryBuilder = new FilterQueryBuilder($this->mockOwnershipProvider, $this->mockAccessibleTeamIdsProvider, $this->mockManagerRegistry);
    }

    /**
     * @return QueryBuilder
     */
    protected function executeSystemUnderTest()
    {
        $queryBuilder = $this->createQueryBuilder();
        $this->filterQueryBuilder->setQueryBuilder($queryBuilder);
        $this->filterQueryBuilder->fromQueryObject($this->query);

        return $queryBuilder;
    }

    protected function tearDown()
    {
        unset($this->filterQueryBuilder);
        unset($this->query);
        unset($this->mockOwnershipProvider);
        unset($this->mockAccessibleTeamIdsProvider);
        unset($this->mockManagerRegistry);
    }

    public function test_Constructor_ShouldNotReturnNull()
    {
        $this->assertNotNull($this->filterQueryBuilder);
    }

    public function test_FromQueryObjectMethod_GivenOwnableClass_ShouldAddAccessibleOwnerIdWhereClause()
    {
        $expectedAccessibleTeamIds = [1, 2, 3];
        $this->mockAccessibleTeamIdsProvider->method('getDataForCurrentUser')->willReturn($expectedAccessibleTeamIds);
        $this->mockOwnershipProvider->method('isClassOwnable')->willReturn(true);

        $this->query = [
            'logic' => 'AND',
            'filters' => []
        ];
        $queryBuilder = $this->executeSystemUnderTest();

        // assert teams parameter exists
        $parameter = $queryBuilder->getParameter(FilterQueryBuilder::AUTHORIZATION_TEAMS_PARAMETER);
        $this->assertEquals($expectedAccessibleTeamIds, $parameter->getValue());
        $this->assertContains('IN(:' . FilterQueryBuilder::AUTHORIZATION_TEAMS_PARAMETER . ')', $queryBuilder->getDQL());
    }

    public function test_FromQueryObjectMethod_GivenFilterWithoutOperator_ShouldThrowInvalidArgumentException()
    {
        $filterWithInvalidOperator = [
            'field' => 'someField',
            'value' => 'someValue'
        ];
        $this->query = [
            'logic' => 'AND',
            'filters' => [$filterWithInvalidOperator]
        ];

        try {
            $this->executeSystemUnderTest();
            $this->fail('Did not throw InvalidArgumentException.');
        } catch (\InvalidArgumentException $invalidArgumentException) {
            // do nothing
        }
    }

    public function test_FromQueryObjectMethod_GivenInvalidOperator_ShouldDescribeExceptionCause()
    {
        $invalidOperator = '0|-<';
        $filterWithInvalidOperator = [
            'field' => 'someField',
            'operator' => $invalidOperator,
            'value' => 'someValue'
        ];
        $this->query = [
            'logic' => 'AND',
            'filters' => [$filterWithInvalidOperator]
        ];

        try {
            $this->executeSystemUnderTest();
            $this->fail('Did not throw InvalidArgumentException.');
        } catch (\InvalidArgumentException $invalidArgumentException) {
            $this->assertEquals(sprintf(FilterQueryBuilder::UNSUPPORTED_OPERATOR_ERROR_MESSAGE, $invalidOperator), $invalidArgumentException->getMessage());
        }
    }

    public function test_FromQueryObjectMethod_GivenFilterWithoutField_ShouldThrowInvalidArgumentException()
    {
        $filterWithoutField = [
            'operator' => 'someOperator',
            'value' => 'someValue'
        ];
        $this->query = [
            'logic' => 'AND',
            'filters' => [$filterWithoutField]
        ];

        try {
            $this->executeSystemUnderTest();
            $this->fail('Did not throw InvalidArgumentException.');
        } catch (\InvalidArgumentException $exception) {
            // do nothing
        }
    }

    public function test_FromQueryObjectMethod_GivenFiveValidFilters_ShouldAddFiveWhereClauses()
    {
        $filters = [
            $this->validFilter,
            $this->validFilter,
            $this->validFilter,
            $this->validFilter,
            $this->validFilter
        ];
        $this->query = [
            'logic' => 'AND',
            'filters' => $filters
        ];
        $queryBuilder = $this->executeSystemUnderTest();

        $whereClauses = $queryBuilder->getDQLPart('where')->getParts();
        $firstClause = current($whereClauses);
        $this->assertEquals(5, count($firstClause->getParts()));
    }

    public function test_FromQueryObjectMethod_GivenLogicAnd_GivenSingleFilter_ShouldReturnFilterExpression()
    {
        $this->query = [
            'logic' => 'AND',
            'filters' => [
                ['field' => 'asdf', 'operator' => 'eq', 'value' => 'asdf']
            ]
        ];
        $queryBuilder = $this->executeSystemUnderTest();

        $whereClauses = $queryBuilder->getDQLPart('where')->getParts();
        $firstClause = current($whereClauses);
        $this->assertTrue($firstClause instanceof Query\Expr\Comparison);
    }

    public function test_FromQueryObjectMethod_GivenLogicAnd_GivenMultipleFilters_ShouldReturnAndxExpression()
    {
        $this->query = [
            'logic' => 'AND',
            'filters' => [
                ['field' => 'asdf', 'operator' => 'eq', 'value' => 'asdf'],
                ['field' => 'asdf', 'operator' => 'eq', 'value' => 'asdf']
            ]
        ];
        $queryBuilder = $this->executeSystemUnderTest();

        $whereClauses = $queryBuilder->getDQLPart('where')->getParts();
        $firstClause = current($whereClauses);
        $this->assertTrue($firstClause instanceof Query\Expr\Andx);
    }

    public function test_FromQueryObjectMethod_GivenLogicOr_GivenSingleFilter_ShouldReturnFilterExpression()
    {
        $this->query = [
            'logic' => 'OR',
            'filters' => [
                ['field' => 'asdf', 'operator' => 'eq', 'value' => 'asdf']
            ]
        ];
        $queryBuilder = $this->executeSystemUnderTest();

        $whereClauses = $queryBuilder->getDQLPart('where')->getParts();
        $firstClause = current($whereClauses);
        $this->assertTrue($firstClause instanceof Query\Expr\Comparison);
    }

    public function test_FromQueryObjectMethod_GivenLogicOr_GivenMultipleFilters_ShouldReturnOrxExpression()
    {
        $this->query = [
            'logic' => 'OR',
            'filters' => [
                ['field' => 'asdf', 'operator' => 'eq', 'value' => 'asdf'],
                ['field' => 'asdf', 'operator' => 'eq', 'value' => 'asdf']
            ]
        ];
        $queryBuilder = $this->executeSystemUnderTest();

        $whereClauses = $queryBuilder->getDQLPart('where')->getParts();
        $firstClause = current($whereClauses);
        $this->assertTrue($firstClause instanceof Query\Expr\Orx);
    }

    public function test_FromQueryObjectMethod_GivenUnknownLogic_ShouldThrowException()
    {
        $this->query = [
            'logic' => 'WAT',
            'filters' => [
                ['field' => 'asdf', 'operator' => 'eq', 'value' => 'asdf']
            ]
        ];
        try {
            $this->executeSystemUnderTest();
            $this->fail('Did not throw InvalidArgumentException');
        } catch (\InvalidArgumentException $invalidArgumentException) {
            // do nothing
        }
    }

    public function test_FromQueryObjectMethod_GivenNestedQueryObject_ShouldReturnNestedExpression()
    {
        $this->query = [
            'logic' => 'AND',
            'filters' => [
                [
                    'logic' => 'OR',
                    'filters' => [
                        [
                            'field' => 'status',
                            'operator' => 'eq',
                            'value' => 'Closed'
                        ],
                        [
                            'field' => 'status',
                            'operator' => 'eq',
                            'value' => 'Open'
                        ]
                    ]
                ]
            ]
        ];
        $queryBuilder = $this->executeSystemUnderTest();
        $queryExpression = $queryBuilder->getDQLPart('where')->getParts();
        $nestedExpression = current($queryExpression);
        $this->assertTrue($nestedExpression instanceof Query\Expr\Orx);
    }

    /**
     * @return QueryBuilder
     */
    protected function createQueryBuilder()
    {
        /** @var \PHPUnit_Framework_MockObject_MockObject|EntityManagerInterface $mockEntityManager */
        $mockEntityManager = $this->getMockBuilder(EntityManager::class)->disableOriginalConstructor()->getMock();
        $mockEntityManager->method('getIdentifier')->willReturn('id');
        $mockEntityManager->method('getExpressionBuilder')->willReturn(new Query\Expr());

        $this->mockManagerRegistry->method('getManagerForClass')->willReturn($mockEntityManager);

        /** @var \PHPUnit_Framework_MockObject_MockObject|ClassMetadata $mockClassMetadata */
        $mockClassMetadata = $this->getMockBuilder(ClassMetadata::class)->disableOriginalConstructor()->getMock();
        $mockClassMetadata->method('getIdentifier')->willReturn(['id']);
        $mockClassMetadata->associationMappings = [
            'owners' => [
                'type' => 'many-to-many',
                'targetEntity' => Team::class
            ]
        ];
        $mockEntityManager->method('getClassMetadata')->willReturn($mockClassMetadata);

        $queryBuilder = new QueryBuilder($mockEntityManager);
        $queryBuilder
            ->select(self::ALIAS_TO_QUERY)
            ->from(self::CLASS_TO_QUERY, self::ALIAS_TO_QUERY, null);

        return $queryBuilder;
    }
}
