<?php
/* Copyright start
  Copyright (C) 2008 - 2020 Fortinet Inc.
  All rights reserved.
  FORTINET CONFIDENTIAL & FORTINET PROPRIETARY SOURCE CODE
  Copyright end */

namespace Cybersponse\SettingsBundle\Controller;


use Cybersponse\ActorsBundle\Entity\Actor;
use Cybersponse\AuthorizationBundle\Permissions\PermissionsDictionaryBuilder;
use Cybersponse\CoreBundle\Controller\AuthorizationResourceController;
use Cybersponse\SettingsBundle\Entity\SystemSettings;
use Cybersponse\SettingsBundle\Entity\UserSettings;
use Cybersponse\SettingsBundle\Utility\Resolver;
use Cybersponse\SettingsBundle\Utility\UndefinedIndexException;
use Symfony\Component\HttpFoundation\JsonResponse;
use Symfony\Component\HttpFoundation\Request;
use Symfony\Component\HttpKernel\Exception\NotFoundHttpException;

/**
 * Class SettingsController
 */
class SettingsController extends AuthorizationResourceController
{
    /**
     * @param $path
     * @return JsonResponse
     */
    public function getCurrentAction($path)
    {
        $settingsProvider = $this->get('cybersponse.settings.provider.settings');

        list($userSettings, $systemSettings) = $settingsProvider->getDataForCurrentUser();
        return $this->getSettings($path, $userSettings, $systemSettings);
    }

    /**
     * @param $path
     * @return JsonResponse
     */
    public function setCurrentAction($path, Request $request)
    {
        $settingsProvider = $this->get('cybersponse.settings.provider.settings');
        $newValues = json_decode($request->getContent(), true);

        /** @var UserSettings $userSettings */
        list($userSettings, $systemSettings) = $settingsProvider->getDataForCurrentUser();

        return $this->setSettings($path, $userSettings, $systemSettings, $newValues);
    }

    /**
     * @param $id
     * @param $path
     * @param Request $request
     *
     * @return JsonResponse
     *
     * @throws UndefinedIndexException
     */
    public function unsetCurrentAction($path, Request $request)
    {
        $resource = $this->getResource($request);
        $this->denyAccessUnlessGranted($resource, [PermissionsDictionaryBuilder::PERM_CRUD_UPDATE]);

        $settingsProvider = $this->get('cybersponse.settings.provider.settings');

        /** @var UserSettings $userSettings */
        list($userSettings, $systemSettings) = $settingsProvider->getDataForCurrentUser();

        return $this->unsetSettings($path, $userSettings, $systemSettings);
    }

    /**
     * @param $id
     * @param $path
     * @return JsonResponse
     */
    public function getAction($id, $path, Request $request)
    {
        $resource = $this->getResource($request);
        $this->denyAccessUnlessGranted($resource, [PermissionsDictionaryBuilder::PERM_CRUD_READ]);

        $settingsProvider = $this->get('cybersponse.settings.provider.settings');

        /** @var Actor $actor */
        $actor = $this->findOrThrowNotFound($this->get('resource.actor'), $id);

        list($userSettings, $systemSettings) = $settingsProvider->getDataForActor($actor);
        return $this->getSettings($path, $userSettings, $systemSettings);
    }

    /**
     * @param $id
     * @param $path
     * @param Request $request
     * @return JsonResponse
     */
    public function setAction($id, $path, Request $request)
    {
        $resource = $this->getResource($request);
        $this->denyAccessUnlessGranted($resource, [PermissionsDictionaryBuilder::PERM_CRUD_UPDATE]);

        $settingsProvider = $this->get('cybersponse.settings.provider.settings');
        $newValues = json_decode($request->getContent(), true);

        /** @var Actor $actor */
        $actor = $this->findOrThrowNotFound($this->get('resource.actor'), $id);

        list($userSettings, $systemSettings) = $settingsProvider->getDataForActor($actor);

        return $this->setSettings($path, $userSettings, $systemSettings, $newValues);
    }

    /**
     * @param $id
     * @param $path
     * @param Request $request
     *
     * @return JsonResponse
     *
     * @throws UndefinedIndexException
     */
    public function unsetAction($id, $path, Request $request)
    {
        $resource = $this->getResource($request);
        $this->denyAccessUnlessGranted($resource, [PermissionsDictionaryBuilder::PERM_CRUD_DELETE]);

        $settingsProvider = $this->get('cybersponse.settings.provider.settings');

        /** @var Actor $actor */
        $actor = $this->findOrThrowNotFound($this->get('resource.actor'), $id);

        list($userSettings, $systemSettings) = $settingsProvider->getDataForActor($actor);

        return $this->unsetSettings($path, $userSettings, $systemSettings);
    }

    /**
     * @param string $path
     * @param UserSettings $userSettings
     * @param SystemSettings $systemSettings
     * @param $newValues
     *
     * @return JsonResponse
     */
    protected function setSettings($path, $userSettings, $systemSettings, $newValues)
    {
        $settingsProvider = $this->get('cybersponse.settings.provider.settings');

        $userSettingsValues = $userSettings->getPublicValues();

        Resolver::set($userSettingsValues, $path, $newValues, '/');

        $settingsProvider->saveUserSettings($userSettings, $userSettingsValues);
        $userSettingsValues = $settingsProvider->getMergedSettings($userSettings, $systemSettings);

        return JsonResponse::create($userSettingsValues);
    }

    /**
     * @param string $path
     * @param UserSettings $userSettings
     * @param SystemSettings $systemSettings
     *
     * @return JsonResponse
     */
    protected function getSettings($path, $userSettings, $systemSettings)
    {
        $settingsProvider = $this->get('cybersponse.settings.provider.settings');

        $userSettingsValues = $settingsProvider->getMergedSettings($userSettings, $systemSettings);

        if (empty($path)) {
            //Empty path means return entire object and not the empty string key
            return JsonResponse::create($userSettingsValues);
        }

        try {
            $resolvedValues = Resolver::get($userSettingsValues, $path, '/');
        } catch (UndefinedIndexException $e) {
            throw new NotFoundHttpException('Not Found');
        }

        return JsonResponse::create($resolvedValues);
    }

    /**
     * @param string $path
     * @param UserSettings $userSettings
     * @param SystemSettings $systemSettings
     *
     * @return JsonResponse
     */
    private function unsetSettings($path, $userSettings, $systemSettings)
    {
        $settingsProvider = $this->get('cybersponse.settings.provider.settings');

        $userSettingsValues = $userSettings->getPublicValues();

        if (empty($path)) {
            //Empty path means return entire object and not the empty string key
            return JsonResponse::create($userSettingsValues);
        }

        try {
            Resolver::uset($userSettingsValues, $path, '/');
        } catch (UndefinedIndexException $e) {
            throw new NotFoundHttpException('Not Found');
        }

        $settingsProvider->saveUserSettings($userSettings, $userSettingsValues);
        $userSettingsValues = $settingsProvider->getMergedSettings($userSettings, $systemSettings);

        return JsonResponse::create($userSettingsValues);
    }

    public function cgetAction(Request $request)
    {
        throw $this->createNotFoundException();
    }

    public function cpostAction(Request $request)
    {
        throw $this->createNotFoundException();
    }
}
