<?php
/* Copyright start
  Copyright (C) 2008 - 2020 Fortinet Inc.
  All rights reserved.
  FORTINET CONFIDENTIAL & FORTINET PROPRIETARY SOURCE CODE
  Copyright end */

namespace Cybersponse\TenantBundle\DataFixtures;


use Cybersponse\ActorsBundle\Entity\Actor;
use Cybersponse\CoreBundle\Api\Bulk\DenormalizationResult;
use Cybersponse\CoreBundle\Serializer\TypeAgnosticLinkedDataSerializer;
use Cybersponse\TenantBundle\DataFixtures\Populator\Methods\DunglasAdder;
use Doctrine\Common\DataFixtures\AbstractFixture;
use Doctrine\Common\DataFixtures\FixtureInterface;
use Doctrine\Common\Persistence\ObjectManager;
use Dunglas\ApiBundle\Mapping\ClassMetadataFactory;
use Nelmio\Alice\Fixtures;
use Nelmio\Alice\Fixtures\Loader;
use Nelmio\Alice\Persister\Doctrine;
use Nelmio\Alice\PersisterInterface;
use Nelmio\Alice\Util\TypeHintChecker;
use Symfony\Component\DependencyInjection\ContainerInterface;
use Symfony\Component\Finder\Finder;
use Symfony\Component\Finder\SplFileInfo;
use Symfony\Component\HttpFoundation\File\File;
use Symfony\Component\Serializer\Serializer;
use Symfony\Component\Yaml\Parser;

abstract class AbstractFixtureLoader extends AbstractFixture implements FixtureLoaderInterface, FixtureInterface
{
    /** @var ContainerInterface */
    protected $container;

    /**
     * @param ObjectManager $manager
     * @throws \Exception
     */
    public function load(ObjectManager $manager)
    {
        $this->persistFixtureFiles($manager);
        $manager->flush();
    }

    /**
     * @param ObjectManager $manager
     */
    public function persistFixtureFiles($manager)
    {
        $persister = new Doctrine($manager);
        $loader = $this->getLoader();
        $loader->setPersister($persister);

        $rootDir = $this->container->getParameter('kernel.root_dir');
        $environment = $this->container->getParameter('kernel.environment');

        $coreDir = sprintf('%s/Fixtures/%s', $rootDir, $environment);

        $yamlParser = new Parser();

        $yamlString = file_get_contents(sprintf('%s/config.yml', $coreDir));
        $defaultConfig = $yamlParser->parse($yamlString);

        $managerRegistryMap = [
            'ORM' => 'doctrine',
        ];

        $configName = $this->getConfigName();

        $managers = $this->container->get($managerRegistryMap[$configName])->getManagers();
        $currentManagerName = 'default';
        foreach($managers as $managerObjName => $managerObj) {
            if ($manager == $managerObj) {
                $currentManagerName = $managerObjName;
                break;
            }
        }

        $paths = $this->flatten($defaultConfig[$configName][$currentManagerName]);

        foreach ($paths as $path) {
            $fullPath = sprintf('%s/%s/%s/%s', $coreDir, $configName, $currentManagerName, $path);
            $finder = new Finder();

            if (is_file($fullPath)) {
                $objects = $this->loadFixtureFile($loader, $fullPath, $persister);
                $persister->persist($objects);
                continue;
            }

            if (is_dir($fullPath)) {
                $finder->files()->in($fullPath)->sortByName();
                /** @var SplFileInfo $file */
                foreach ($finder as $file) {
                    $objects = $this->loadFixtureFile($loader, $file->getPathname(), $persister);
                    $persister->persist($objects);
                }
            }
        }
    }

    /**
     * Returns the loader with the given providers
     *
     * @return Loader
     */
    public function getLoader()
    {
        $providers = $this->getProviders();

        /** @var ClassMetadataFactory $classMetadataFactory */
        $classMetadataFactory = $this->container->get('api.mapping.class_metadata_factory');
        /** @var Serializer $serializer */
        $serializer = $this->container->get('serializer');

        $dunglasAdder = new DunglasAdder(new TypeHintChecker());
        $dunglasAdder->setClassMetadataFactory($classMetadataFactory);
        $dunglasAdder->setSerializer($serializer);

        $loader = new Loader('en_US', $providers);
        $loader->addPopulator($dunglasAdder);
        return $loader;
    }

    /**
     * @param Loader $loader
     * @param string $fullPath
     * @param PersisterInterface $persister
     * @return array
     */
    protected function loadFixtureFile($loader, $fullPath, $persister)
    {
        $file = new File($fullPath);
        $objects = [];

        if ($file->getExtension() == 'yml') {
            $objects = $loader->load($fullPath);
            foreach ($objects as $object) {
                if ($object instanceof Actor) {
                    $this->setReference($object->getUserId(), $object);
                }
            }
        }

        if ($file->getExtension() == 'json') {
            $jsonString = file_get_contents($fullPath);
            /** @var Serializer $serializer */
            $serializer = $this->container->get('serializer');
            $items = $serializer->decode($jsonString, 'json');
            if (is_array($items) && !array_key_exists(0, $items)) {
                $items = [$items];
            }

            /** @var TypeAgnosticLinkedDataSerializer $typeAgnosticLinkedDataDenormalizer */
            $typeAgnosticLinkedDataDenormalizer = $this->container->get('cybersponse.core.serializer.type_agnostic_linked_data');

            $unlinkedDenormalizationResults = $typeAgnosticLinkedDataDenormalizer->denormalizeUnlinked($items);

            $persister->persist(array_map(function (DenormalizationResult $denormalizationResult) use ($persister) {
                return $denormalizationResult->getResult();
            }, $unlinkedDenormalizationResults));

            $linkedDenormalizationResults = $typeAgnosticLinkedDataDenormalizer->denormalizeLinked($unlinkedDenormalizationResults);

            $objects = array_map(function (DenormalizationResult $denormalizationResult) {
                return $denormalizationResult->getResult();
            }, $linkedDenormalizationResults);
        }

        return $objects;
    }

    private function flatten($data, $parentStr = '')
    {
        $values = [];

        foreach ($data as $key => $value) {
            if (empty($value)) {
                $result = [$key];
            } else if (is_array($value)) {
                $newStr = $parentStr;
                if (!is_int($key)) {
                    $newStr .= $key . '/';
                }
                $result = $this->flatten($value, $newStr);
            } else {
                $result = [$parentStr . $value];
            }
            $values = array_merge($values, $result);
        }
        return $values;
    }

    /**
     * Sets the Container.
     *
     * @param ContainerInterface|null $container A ContainerInterface instance or null
     */
    public function setContainer(ContainerInterface $container = null)
    {
        $this->container = $container;
    }
}
