<?php
/* Copyright start
  Copyright (C) 2008 - 2020 Fortinet Inc.
  All rights reserved.
  FORTINET CONFIDENTIAL & FORTINET PROPRIETARY SOURCE CODE
  Copyright end */

namespace Cybersponse\WorkflowBundle\Serializer;


use Cybersponse\WorkflowBundle\Api\RouteResource;
use Cybersponse\WorkflowBundle\Subscriber\DataEvent;
use Dunglas\ApiBundle\Api\IriConverterInterface;
use Dunglas\ApiBundle\Event\DataEvent as DunglasDataEvent;
use Symfony\Component\HttpFoundation\RequestStack;
use Symfony\Component\Security\Core\Authentication\Token\Storage\TokenStorageInterface;
use Symfony\Component\Serializer\Normalizer\NormalizerInterface;
use Symfony\Component\Serializer\Serializer;
use Symfony\Component\Serializer\SerializerAwareInterface;
use Symfony\Component\Serializer\SerializerInterface;

class DataEventNormalizer implements NormalizerInterface, SerializerAwareInterface
{
    const FORMAT_JSON = 'json';

    /** @var Serializer */
    protected $serializer;

    /** @var RequestStack */
    protected $requestStack;

    /** @var TokenStorageInterface */
    protected $tokenStorage;

    /** @var IriConverterInterface */
    protected $iriConverter;

    /**
     * {@inheritdoc}
     */
    public function supportsNormalization($data, $format = null)
    {
        return self::FORMAT_JSON === $format && ($data instanceof DataEvent || $data instanceof DunglasDataEvent);
    }

    /**
     * @param DataEvent|DunglasDataEvent $dataEvent
     * {@inheritdoc}
     */
    public function normalize($dataEvent, $format = null, array $context = [])
    {
        $request = $this->requestStack->getCurrentRequest();

        $resource = $dataEvent->getResource();

        $normalizationContext = $resource->getNormalizationContext();
        $resourceShortName = null;
        $format = 'json';

        if (!($resource instanceof RouteResource)) {
            $resourceShortName = $resource->getShortName();
            $format = 'json-ld';
        }

        $result = [
            'request' => [
                'data' => $this->serializer->normalize(
                    $dataEvent->getData(),
                    $format,
                    $normalizationContext
                ),
                'method' => $request->getMethod(),
                'uri' => $request->getUri(),
                'baseUri' => $request->getUriForPath(null),
                'headers' => array_map(function ($item) {
                    return $item[0];
                }, $request->headers->all()),
                'body' => $request->getContent(),
                'query' => $request->query->all()
            ],
            'resource' => $resourceShortName
        ];

        if ($dataEvent instanceof DataEvent && !is_null($dataEvent->getPrevious())) {
            $result['previous'] = [
                'data' => $dataEvent->getPrevious()
            ];
        }

        $currentUserIri = $this->getCurrentUserAsIri();

        if (!is_null($currentUserIri)) {
            $result['currentUser'] = $currentUserIri;
        }

        return $result;
    }

    /**
     * Sets the owning Serializer object.
     *
     * @param SerializerInterface $serializer
     */
    public function setSerializer(SerializerInterface $serializer)
    {
        $this->serializer = $serializer;
    }

    /**
     * @param RequestStack $requestStack
     */
    public function setRequestStack($requestStack)
    {
        $this->requestStack = $requestStack;
    }

    /**
     * @param TokenStorageInterface $tokenStorage
     */
    public function setTokenStorage($tokenStorage)
    {
        $this->tokenStorage = $tokenStorage;
    }

    /**
     * @param IriConverterInterface $iriConverter
     */
    public function setIriConverter($iriConverter)
    {
        $this->iriConverter = $iriConverter;
    }

    /**
     * @return string
     */
    private function getCurrentUserAsIri()
    {
        $token = $this->tokenStorage->getToken();
        if (is_null($token)) {
            return null;
        }

        $user = $token->getUser();
        $iri = $this->iriConverter->getIriFromItem($user);

        return $iri;
    }
}
