<?php
/* Copyright start
  Copyright (C) 2008 - 2020 Fortinet Inc.
  All rights reserved.
  FORTINET CONFIDENTIAL & FORTINET PROPRIETARY SOURCE CODE
  Copyright end */

namespace Cybersponse\WorkflowBundle\Subscriber;


use Cybersponse\QueryBundle\ORM\JsonExpressionBuilder;
use Cybersponse\WorkflowBundle\Constants\Triggers;
use Cybersponse\WorkflowBundle\Entity\Rule;
use Cybersponse\WorkflowBundle\Rule\RuleExecutionEngine;
use Doctrine\Common\Inflector\Inflector;
use Doctrine\Common\Persistence\ManagerRegistry;
use Doctrine\ORM\EntityManagerInterface;
use Doctrine\ORM\EntityRepository;
use Dunglas\ApiBundle\Api\ResourceCollectionInterface;
use Dunglas\ApiBundle\Event\Events;
use Psr\Log\LoggerInterface;
use Symfony\Component\EventDispatcher\EventSubscriberInterface;
use Symfony\Component\Serializer\Normalizer\NormalizerInterface;


class RuleTriggerSubscriber implements EventSubscriberInterface
{
    const RULE_PRIORITY = 2;

    const TRIGGER_WHEN_POST = 'post';
    const TRIGGER_ACTION_CREATE = 'create';
    const TRIGGER_ACTION_UPDATE = 'update';

    /** @var ResourceCollectionInterface */
    protected $resourceCollection;

    /** @var ManagerRegistry */
    protected $managerRegistry;

    /** @var LoggerInterface */
    protected $logger;

    /** @var RuleExecutionEngine */
    protected $ruleExecutionEngine;

    /** @var NormalizerInterface */
    protected $normalizer;

    /**
     * @inheritdoc
     */
    public static function getSubscribedEvents()
    {
        return [
            Events::POST_CREATE => ['handlePostCreate', self::RULE_PRIORITY],
            Events::POST_UPDATE => ['handlePostUpdate', self::RULE_PRIORITY],
        ];
    }

    /**
     * @param ResourceCollectionInterface $resourceCollection
     */
    public function setResourceCollection($resourceCollection)
    {
        $this->resourceCollection = $resourceCollection;
    }

    /**
     * @param ManagerRegistry $managerRegistry
     */
    public function setManagerRegistry($managerRegistry)
    {
        $this->managerRegistry = $managerRegistry;
    }

    /**
     * @param LoggerInterface $logger
     */
    public function setLogger($logger)
    {
        $this->logger = $logger;
    }

    /**
     * @param RuleExecutionEngine $ruleExecutionEngine
     */
    public function setRuleExecutionEngine($ruleExecutionEngine)
    {
        $this->ruleExecutionEngine = $ruleExecutionEngine;
    }

    /**
     * @param NormalizerInterface $normalizer
     */
    public function setNormalizer($normalizer)
    {
        $this->normalizer = $normalizer;
    }

    /**
     * Handlers create triggers
     *
     * @param DataEvent $event
     */
    public function handlePostCreate(DataEvent $event)
    {
        $this->handleTriggerType(Triggers::POST_CREATE, $event);
    }

    /**
     * Handles update triggers
     *
     * @param DataEvent $event
     */
    public function handlePostUpdate(DataEvent $event)
    {
        $this->handleTriggerType(Triggers::POST_UPDATE, $event);
    }

    /**
     * Handles all triggers based on trigger type
     *
     * @param string $triggerType
     * @param DataEvent $event
     */
    public function handleTriggerType($triggerType, DataEvent $event)
    {
        $when = self::TRIGGER_WHEN_POST;
        switch ($triggerType) {
            case Triggers::POST_CREATE:
                $action = self::TRIGGER_ACTION_CREATE;
                break;
            case Triggers::POST_UPDATE:
                $action = self::TRIGGER_ACTION_UPDATE;
                break;
            default:
                return;
        }
        $resource = $this->getTriggeredResource($event->getData());
        $rules = $this->getRulesByTrigger($when, $action, $resource);

        $arguments = $this->normalizer->normalize(
            $event,
            'json'
        );

        $this->ruleExecutionEngine->executeRules($rules, $arguments);
    }

    /**
     * @param mixed $object
     * @return string
     */
    protected function getTriggeredResource($object)
    {
        $resource = $this->resourceCollection->getResourceForEntity($object);
        return $resource->getModelType();
    }

    /**
     * @param string $when
     * @param string $action
     * @param string $resource
     * @return Rule[]
     */
    protected function getRulesByTrigger($when, $action, $resource)
    {
        /** @var EntityManagerInterface $ruleManager */
        $ruleManager = $this->managerRegistry->getManagerForClass(Rule::class);
        /** @var EntityRepository $ruleRepository */
        $ruleRepository = $ruleManager->getRepository(Rule::class);
        $queryBuilder = $ruleRepository->createQueryBuilder('r');

        $expressionBuilder = new JsonExpressionBuilder();
        $triggerFilter = strtolower(sprintf('{"when":"%s","action":["%s"],"resource":"%s"}', $when, $action, $resource));
        $queryBuilder->setParameter(':triggerFilter', $triggerFilter);
        $queryBuilder->setParameter(':isActive', true);

        $queryBuilder->where($expressionBuilder->eq('r.isActive', ':isActive'))
            ->andWhere($expressionBuilder->jsonContains('r.trigger', ':triggerFilter'));
        return $queryBuilder->getQuery()->execute();
    }
}
