<?php
/* Copyright start
  Copyright (C) 2008 - 2020 Fortinet Inc.
  All rights reserved.
  FORTINET CONFIDENTIAL & FORTINET PROPRIETARY SOURCE CODE
  Copyright end */

namespace Cybersponse\WorkflowBundle\Subscriber;


use Cybersponse\AuthorizationBundle\AuthorizationProvider;
use Cybersponse\AuthorizationBundle\Permissions\PermissionsDictionaryBuilder;
use Cybersponse\ProxyBundle\GuzzleHttp\ProxyClient;
use Cybersponse\WorkflowBundle\Api\RouteResource;
use Cybersponse\WorkflowBundle\Constants\Triggers;
use Cybersponse\WorkflowBundle\Repository\WorkflowRepository;
use Doctrine\ORM\EntityManager;
use Ramsey\Uuid\Uuid;
use Dunglas\ApiBundle\Api\ResourceCollectionInterface;
use Dunglas\ApiBundle\Api\ResourceInterface;
use Dunglas\ApiBundle\Event\Events;
use GuzzleHttp\Exception\BadResponseException;
use GuzzleHttp\Psr7\Request;
use Monolog\Logger;
use Psr\Http\Message\ResponseInterface;
use Symfony\Component\EventDispatcher\EventDispatcherInterface;
use Symfony\Component\EventDispatcher\EventSubscriberInterface;
use Symfony\Component\HttpFoundation\JsonResponse;
use Symfony\Component\HttpFoundation\Response as HttpResponse;
use Symfony\Component\Security\Core\Exception\AccessDeniedException;
use Symfony\Component\Serializer\Serializer;
use DateTime;


class WorkflowEventSubscriber implements EventSubscriberInterface
{
    const AUTH_METHOD_HEADER = 'X-CS-Authentication-Method';
    const ROUTE_PARAM = 'route';
    const RESOURCE_PARAM = 'resource';

    const FIELD_BASED_TRIGGER = 'fieldbasedtrigger';
    const FIELD_OR_OPERATOR = 'OR';
    const FIELD_AND_OPERATOR = 'AND';
    const LIKE_OPERATOR = 'like';
    const NOT_LIKE_OPERATOR = 'notlike';
    const IN_OPERATOR = 'in';
    const NIN_OPERATOR = 'nin';
    const ADDED_VALUE_IS = 'addedvalueis';
    const EQUALS_OPERATOR = 'eq';
    const NOT_EQUALS_OPERATOR = 'neq';
    const GREATER_THAN_EQUALS_OPERATOR = 'gte';
    const LESS_THAN_EQUALS_OPERATOR = 'lte';
    const GREATER_THAN_OPERATOR = 'gt';
    const LESS_THAN_OPERATOR = 'lt';
    const NULL_OPERATOR = 'isnull';
    const CHANGED_OPERATOR = 'changed';
  
    const NON_DATETIME_CONVERT_OPERATOR = [self::NULL_OPERATOR, self::CHANGED_OPERATOR];
    // field Constants
    const FIELD = 'field';
    const OPERATOR = 'operator';
    const VALUE = 'value';
    const FIELD_TYPE = 'type';
    const WORKFLOW_PRIORITY = 2;
    /** @var EventDispatcherInterface */
    protected $eventDispatcher;

    /** @var EntityManager */
    protected $entityManager;

    /** @var Serializer */
    protected $serializer;

    /** @var ProxyClient */
    protected $proxyClient;

    /** @var string */
    protected $endpoint;

    /** @var Logger */
    protected $monolog;

    /** @var WorkflowRepository */
    protected $workflowRepository;

    /** @var ResourceInterface */
    protected $workflowResource;

    /** @var AuthorizationProvider */
    protected $authorizationProvider;

    /** @var ResourceCollectionInterface */
    protected $resourceCollection;

    /** @var int */
    protected $timeout;
    /** @var array */
    private $eventContext;

    public function __construct(
        EventDispatcherInterface $eventDispatcher,
        EntityManager $entityManager
    ) {
        $this->eventDispatcher = $eventDispatcher;
        $this->entityManager = $entityManager;
    }

    /**
     * {@inheritdoc}
     */
    public static function getSubscribedEvents()
    {
        return [
            Events::PRE_CREATE => ['handlePreCreate', self::WORKFLOW_PRIORITY],
            Events::PRE_UPDATE => ['handlePreUpdate', self::WORKFLOW_PRIORITY],
            Events::PRE_DELETE => ['handlePreDelete', self::WORKFLOW_PRIORITY],
            Events::POST_CREATE => ['handlePostCreate', self::WORKFLOW_PRIORITY],
            Events::POST_UPDATE => ['handlePostUpdate', self::WORKFLOW_PRIORITY],
            Events::POST_DELETE => ['handlePostDelete', self::WORKFLOW_PRIORITY],
            Triggers::API_CALL => ['handleApiCall', self::WORKFLOW_PRIORITY],
            Triggers::ACTION => ['handleAction', self::WORKFLOW_PRIORITY],
        ];
    }

    /**
     * @param ProxyClient $proxyClient
     */
    public function setProxyClient($proxyClient)
    {
        $this->proxyClient = $proxyClient;
    }

    public function in_array_all($needles, $haystack) {
        return empty(array_diff($needles, $haystack));
     }

    /**
     * @param string $endpoint
     */
    public function setEndpoint($endpoint)
    {
        $this->endpoint = $endpoint;
    }

    /**
     * @param Serializer $serializer
     */
    public function setSerializer($serializer)
    {
        $this->serializer = $serializer;
    }

    /**
     * @param Logger $monolog
     */
    public function setMonolog($monolog)
    {
        $this->monolog = $monolog;
    }

    /**
     * @param WorkflowRepository $workflowRepository
     */
    public function setWorkflowRepository($workflowRepository)
    {
        $this->workflowRepository = $workflowRepository;
    }

    /**
     * @param ResourceInterface $workflowResource
     */
    public function setWorkflowResource($workflowResource)
    {
        $this->workflowResource = $workflowResource;
    }

    /**
     * @param AuthorizationProvider $authorizationProvider
     */
    public function setAuthorizationProvider($authorizationProvider)
    {
        $this->authorizationProvider = $authorizationProvider;
    }

    /**
     * @param ResourceCollectionInterface $resourceCollection
     */
    public function setResourceCollection($resourceCollection)
    {
        $this->resourceCollection = $resourceCollection;
    }

    /**
     * @param int $timeout
     */
    public function setTimeout($timeout)
    {
        $this->timeout = $timeout;
    }

    /**
     * Handles api call triggers
     *
     * @param DataEvent $event
     */
    public function handleApiCall(DataEvent $event)
    {
        $this->handleTriggerType(Triggers::API_CALL, $event);
    }

    /**
     * Handles all triggers based on triggerType
     *
     * @param $triggerType
     * @param DataEvent $event
     */
    public function handleTriggerType($triggerType, DataEvent $event)
    {
        $this->eventContext = ['eventLogId' => spl_object_hash($event)];

        $shortName = $event->getResource()->getShortName();

        $triggerResource = $shortName;
        $triggerParameter = self::ROUTE_PARAM;
        if (!($event->getResource() instanceof RouteResource)) {
            $triggerParameter = self::RESOURCE_PARAM;
            $triggerResource = $event->getResource()->getModelType();
        }

        //set default response for event
        $event->setResponse(JsonResponse::create('Not Found', HttpResponse::HTTP_NOT_FOUND));

        $this->monolog->addInfo(sprintf(
            'Searching for playbooks with trigger: "%s" caused by %s: %s...',
            $triggerType,
            $event->getResource() instanceof RouteResource ? 'route' : 'entity',
            $shortName
        ), $this->eventContext);

        $owners = [];
        if (method_exists($event->getData(), 'getOwners') && $event->getData()->getOwners() != null) {
            foreach($event->getData()->getOwners() as $owner){
                array_push($owners, $owner->getUuid());
            }
        }

        $workflows = $this->workflowRepository->getWorkflowsForTriggerType($triggerType, $owners);

        if ($shortName != 'Workflow') {
            foreach ($workflows as $workflow) {
                $triggerStep = $workflow->getTriggerStep();
                $arguments = $triggerStep->getArguments();

                //Validate trigger step
                if (!array_key_exists($triggerParameter, $arguments)) {
                    continue;
                }

                if ($arguments[$triggerParameter] != $triggerResource) {
                    continue;
                }

                if (($triggerType !== Triggers::API_CALL && $triggerType !== Triggers::ACTION) && $this->isFieldBasedTrigger($arguments)) {
                    if (!$this->evaluateTriggerFields($arguments, $event)) {
                        continue;
                    }
                }

                //Check authorization on action trigger
                if ($triggerType == Triggers::ACTION) {
                    $authorized = false;
                    $objects = $event->getData();
                    if ($objects !== null && array_key_exists('records', $objects)) {
                        $objects = $objects['records'];
                    }
                    foreach ($objects as $object) {
                        if ($object !== null && array_key_exists('@type', $object)) {
                            $type = $object['@type'];
                        }
                        $resource = $this->resourceCollection->getResourceForShortName($type);
                        if ($this->authorizationProvider->isGranted($resource, PermissionsDictionaryBuilder::PERM_CRUD_UPDATE)) {
                            $authorized = true;
                        }
                    }
                    if (!$authorized && count($objects) > 0) {
                        throw new AccessDeniedException();
                    }
                }

                //Normalize workflow
                $normalizedBody = $this->serializer->normalize(
                    $workflow,
                    'json-ld',
                    array_merge(
                        [
                            'show_to_many' => true,
                            'show_to_one' => true,
                            'ignore_authorization' => true,
                            'ignore_field_authorization' => true,
                            'ignore_relationship_authorization' => true
                        ],
                        $this->workflowResource->getNormalizationContext()
                    )
                );

                //Normalize Arguments
                $defaultArguments = $this->serializer->normalize(
                    $event,
                    'json'
                );
                $headerVars = $event->getWfHeaderObject();
                if (isset($headerVars['X-RUNBYUSER'])) {
                    $defaultArguments['request']['headers']['X-RUNBYUSER'] = $headerVars['X-RUNBYUSER'];
                }
                if (isset($headerVars['X-TRIGGERTYPE'])) {
                    $defaultArguments['request']['headers']['X-TRIGGERTYPE'] = $headerVars['X-TRIGGERTYPE'];
                }

                $authenticationMethods = [];
                //Update arguments on normalized workflow so they are not persisted
                unset($arguments[$triggerParameter]);

                foreach ($normalizedBody['steps'] as &$normalizedStep) {
                    $iriParts = explode('/', $normalizedStep['@id']);
                    if (end($iriParts) == $triggerStep->getUuid()) {
                        if (isset($arguments['authentication_methods'])) {
                            $authenticationMethods = $arguments['authentication_methods'];
                            unset($arguments['authentication_methods']);
                        }
                        $normalizedStep['arguments'] = $defaultArguments + $arguments;
                        break;
                    }
                }

                //Encode workflow
                $body = $this->serializer->encode(
                    $normalizedBody,
                    'json'
                );
                $this->monolog->addInfo(sprintf(
                    "Making %s request to %s:",
                    $workflow->getSynchronous() ? 'synchronous' : 'asynchronous',
                    $this->endpoint
                ), $this->eventContext);
                $this->monolog->addInfo($body, $this->eventContext);

                //Build authentication method header if set
                $headers = ['Content-Type' => 'application/json'];
                if (!empty($authenticationMethods)) {
                    $headers[self::AUTH_METHOD_HEADER] = implode(',', $authenticationMethods);
                }

                $this->handleRequest($body, $event, $headers);

                $triggerLimit = $workflow->getTriggerLimit();

                if (!is_null($triggerLimit)) {
                    $triggerLimit--;
                    $workflow->setTriggerLimit($triggerLimit);
                    if ($triggerLimit <= 0) {
                        $this->entityManager->remove($workflow);
                        $this->entityManager->flush($workflow);
                    }
                }
            }
        }
    }

    /**
     * @param $arguments
     * @return boolean
     */
    private function isFieldBasedTrigger($arguments)
    {
        if (array_key_exists(self::FIELD_BASED_TRIGGER, $arguments)) {
            return true;
        }
        return false;
    }

    /**
     * @param $arguments
     * @param DataEvent $event
     * @return boolean
     */
    private function evaluateTriggerFields($arguments, DataEvent $event)
    {
        $previousData = $event->getPrevious();
        $currentData = $this->serializer->normalize($event->getData(), 'json-ld');
        return $this->evaluateObject($arguments[self::FIELD_BASED_TRIGGER], $currentData, $previousData);
    }

    /**
     * @param $val
     * @return string
     */
    private function parseDateField($val) {
        if (strpos($val, '{{') !== false) {
          $regex = '/{{\s*getRelativeDate\(\s*(?\'years\'[^,]*)\s*,\s*(?\'months\'[^,]*)\s*,\s*(?\'days\'[^,]*)\s*,\s*(?\'hours\'[^,]*)\s*,\s*(?\'minutes\'[^,]*)\s*,\s*(?\'seconds\'[^,]*)\s*\)\s*}}/m';
          preg_match_all($regex, $val, $matches, PREG_SET_ORDER, 0);
          if (count($matches) != 1) {
              return null;
          }
          $years = $matches[0]['years'];
          $relativeDate = new DateTime();
          $relativeDate->modify($years . ' years');

          $months = $matches[0]['months'];
          if ($months == '"start"') {
            $relativeDate->setDate($relativeDate->format('Y'), 1, $relativeDate->format('d'));
          } else if ($months == '"end"') {
            $relativeDate->setDate($relativeDate->format('Y'), 12, $relativeDate->format('d'));
          } else if (intval($months)) {
            $relativeDate->modify($months . ' months');
          }

          $days = $matches[0]['days'];
          if ($days == '"start"') {
            $relativeDate->setDate($relativeDate->format('Y'), $relativeDate->format('m'), 1);
          } else if ($days == '"end"') {
            $relativeDate->modify('last day of this month');
          } else if (intval($days)) {
            $relativeDate->modify($days . ' days');
          }

          $hours = $matches[0]['hours'];
          if ($hours == '"start"') {
            $relativeDate->setTime(0, $relativeDate->format('i'));
          } else if ($hours == '"end"') {
            $relativeDate->setTime(23, $relativeDate->format('i'));
          } else if (intval($hours)) {
            $relativeDate->modify($hours . ' hours');
          }

          $minutes = $matches[0]['minutes'];
          if ($minutes == '"start"') {
            $relativeDate->setTime($relativeDate->format('H'), 0);
          } else if ($minutes == '"end"') {
            $relativeDate->setTime($relativeDate->format('H'), 59);
          } else if (intval($minutes)) {
            $relativeDate->modify($minutes . ' minutes');
          }

          $seconds = $matches[0]['seconds'];
          if ($seconds == '"start"') {
            $relativeDate->setTime($relativeDate->format('H'), $relativeDate->format('i'), 0);
          } else if ($seconds == '"end"') {
            $relativeDate->setTime($relativeDate->format('H'), $relativeDate->format('i'), 59);
          } else if (intval($seconds)) {
            $relativeDate->modify($seconds . ' seconds');
          }
          return $relativeDate->getTimestamp();
        }
        $valDate = new DateTime($val);
        return $valDate->getTimestamp();
    }

    /**
     * @param queryObject
     * @param $currentData
     * @param $previousData
     * @return boolean
     */
    private function evaluateObject($queryObject, $currentData, $previousData) {
      $fieldExpressions = $queryObject['filters'];
      $fieldOperator = $queryObject['logic'] ? $queryObject['logic'] : self::FIELD_AND_OPERATOR;
      $evaluation = false;
      if ($fieldOperator === self::FIELD_AND_OPERATOR) {
          $evaluation = true;
      }
      foreach ($fieldExpressions as $fieldExpression) {
          if (array_key_exists('filters', $fieldExpression)) {
              $currEval = $this->evaluateObject($fieldExpression, $currentData, $previousData);
              $evaluation = $this->evaluateOperator($fieldOperator, $currEval, $evaluation);
              continue;
          }
          if (!array_key_exists($fieldExpression[self::FIELD], $currentData) && !array_key_exists($fieldExpression[self::FIELD], $previousData)) {
            // This will be the case for M2M, For performance optimization, if M2M fields is not changed, it is not normalized.
            // M2M only supports 'isChanged' and 'added' operator. So field not present means there has been no change,
            // and condition should evaluate to False
            if ($fieldOperator === self::FIELD_AND_OPERATOR) {
                return false;
            }
            continue;
          }
          $currDataVal = $currentData[$fieldExpression[self::FIELD]];
          $fieldExprVal = $fieldExpression[self::VALUE];
          $fieldType = isset($fieldExpression[self::FIELD_TYPE]) ? $fieldExpression[self::FIELD_TYPE]: null;
          $fieldExpressionOperator = $fieldExpression['operator'];

          if ($fieldType == 'datetime' && !in_array($fieldExpressionOperator, self::NON_DATETIME_CONVERT_OPERATOR)) {
            $fieldExprVal = $this->parseDateField($fieldExprVal);
          }
          if ($fieldExpression[self::FIELD] == 'recordTags') {
              $tags = [];
              foreach($fieldExprVal as $item) {
                  $exploded_item = explode('/', $item);
                  array_push($tags, end($exploded_item));
              }
              $fieldExprVal = $tags;
          }
          $prevDataVal = $previousData[$fieldExpression[self::FIELD]];
          if (is_array($currDataVal) && isset($currDataVal['@id'])) {
              $currDataVal = $currDataVal['@id'];
          } else if (is_array($currDataVal)) {
              $finalVal = [];
              foreach($currDataVal as $val) {
                  $item = isset($val['@id']) ? $val['@id']: $val;
                  array_push($finalVal, $item);
              }
              $currDataVal = $finalVal;
          }
          if (is_array($prevDataVal) && isset($prevDataVal['@id'])) {
              $prevDataVal = $prevDataVal['@id'];
          } else if (is_array($prevDataVal)) {
              $finalVal = [];
              foreach($prevDataVal as $val) {
                  $item = isset($val['@id']) ? $val['@id']: $val;
                  array_push($finalVal, $item);
              }
              $prevDataVal = $finalVal;
          }
          switch ($fieldExpression['operator']) {

              case self::EQUALS_OPERATOR:
                  $currEval = strtolower($currDataVal) == strtolower($fieldExprVal);
                  $evaluation = $this->evaluateOperator($fieldOperator, $currEval, $evaluation);
                  break;
              case self::NOT_EQUALS_OPERATOR:
                  $currEval = strtolower($currDataVal) !== strtolower($fieldExprVal);
                  $evaluation = $this->evaluateOperator($fieldOperator, $currEval, $evaluation);
                  break;
              case self::LIKE_OPERATOR:
                  $currEval = false;
                  $fieldExprVal = preg_quote($fieldExprVal);
                  $expression = str_replace('%', '.*', $fieldExprVal);
                  $expression = str_replace('_', '.{1}', $expression);
                  $regex = sprintf('/^%s$/mi', $expression);
                  preg_match_all($regex, $currDataVal, $matches, PREG_SET_ORDER, 0);
                  $currEval = count($matches) > 0;
                  $evaluation = $this->evaluateOperator($fieldOperator, $currEval, $evaluation);
                  break;
              case self::NOT_LIKE_OPERATOR:
                  $currEval = false;
                  $fieldExprVal = preg_quote($fieldExprVal);
                  $expression = str_replace('%', '.*', $fieldExprVal);
                  $expression = str_replace('_', '.{1}', $expression);
                  $regex = sprintf('/%s/mi', $expression);
                  preg_match_all($regex, $currDataVal, $matches, PREG_SET_ORDER, 0);
                  $currEval = count($matches) == 0;
                  $evaluation = $this->evaluateOperator($fieldOperator, $currEval, $evaluation);
                  break;
              case self::NULL_OPERATOR:
                  $currEval = (is_null($currDataVal) ? 'true' : 'false') === $fieldExprVal;
                  $evaluation = $this->evaluateOperator($fieldOperator, $currEval, $evaluation);
                  break;
              case self::IN_OPERATOR:
                  $currDataVal = $currDataVal ? $currDataVal: [];
                  $currEval = $this->in_array_all($fieldExprVal, $currDataVal);
                  $evaluation = $this->evaluateOperator($fieldOperator, $currEval, $evaluation);
                  break;
              case self::NIN_OPERATOR:
                  $result = array_intersect($fieldExprVal, $currDataVal);
                  $currEval = count($result) ? false : true;
                  $evaluation = $this->evaluateOperator($fieldOperator, $currEval, $evaluation);
                  break;
              case self::CHANGED_OPERATOR:
              if (is_array($currDataVal) || is_array($prevDataVal)) {
                  $prevDataVal = $prevDataVal ? $prevDataVal: [];
                  if (!is_null($currDataVal)) {
                      sort($currDataVal);
                      sort($prevDataVal);
                      $currEval = $currDataVal !== $prevDataVal;
                  } else {
                      $currEval = !is_null($prevDataVal);
                  }
                  $evaluation = $this->evaluateOperator($fieldOperator, $currEval, $evaluation);
                  break;
              } else {
                  $currEval = strtolower($currDataVal) !== strtolower($prevDataVal);
                  $evaluation = $this->evaluateOperator($fieldOperator, $currEval, $evaluation);
                  break;
              }
              case self::ADDED_VALUE_IS:
              if (is_array($currDataVal)) {
                  $prevDataVal = $prevDataVal ? $prevDataVal: [];
                  $fieldExprVal = is_array($fieldExprVal) ? $fieldExprVal: [$fieldExprVal];
                  $currentMatch = array_intersect($fieldExprVal, $currDataVal);
                  $prevMatch = array_intersect($fieldExprVal, $prevDataVal);
                  if (count($currentMatch) > 0 && count($prevMatch) == 0) {
                      $currEval = true;
                  } else {
                      $currEval = false;
                  }
                  $evaluation = $this->evaluateOperator($fieldOperator, $currEval, $evaluation);
                  break;
              } else {
                  $evaluation = false;
                  break;
              }
              case self::GREATER_THAN_EQUALS_OPERATOR:
                  if(is_int($fieldExprVal) && is_int($currDataVal) )
                  {
                      $currEval = $currDataVal >= $fieldExprVal;
                      $evaluation = $this->evaluateOperator($fieldOperator, $currEval, $evaluation);
                  } else {
                    $evaluation = false;
                  }
                  break;
              case self::LESS_THAN_EQUALS_OPERATOR:
                  if(is_int($fieldExprVal) && is_int($currDataVal) )
                  {
                      $currEval = $currDataVal <= $fieldExprVal;
                      $evaluation = $this->evaluateOperator($fieldOperator, $currEval, $evaluation);
                  } else {
                    $evaluation = false;
                  }
                  break;
              case self::GREATER_THAN_OPERATOR:
                  if(is_int($fieldExprVal) && is_int($currDataVal) )
                  {
                      $currEval = $currDataVal > $fieldExprVal;
                      $evaluation = $this->evaluateOperator($fieldOperator, $currEval, $evaluation);
                  } else {
                    $evaluation = false;
                  }
                  break;
              case self::LESS_THAN_OPERATOR:
                  if(is_int($fieldExprVal) && is_int($currDataVal) )
                  {
                      $currEval = $currDataVal < $fieldExprVal;
                      $evaluation = $this->evaluateOperator($fieldOperator, $currEval, $evaluation);
                  } else {
                    $evaluation = false;
                  }
                  break;
              default:
                  return $this->evaluateOperator($fieldOperator, false, $evaluation);
          }
      }
      return $evaluation;
    }

    /**
     * @param $operator
     * @param $prevEval
     * @param $currEval
     * @return boolean
     */
    private function evaluateOperator($operator, $prevEval, $currEval)
    {
        if ($operator === self::FIELD_OR_OPERATOR) {
            return $prevEval || $currEval;
        } elseif ($operator === self::FIELD_AND_OPERATOR) {
            return $prevEval && $currEval;
        } else {
            return false;
        }
    }

    /**
     * @param $body
     * @param DataEvent $event
     * @param array $headers
     */
    private function handleRequest($body, DataEvent $event, $headers = [])
    {
        $workflowRequestUri = $this->proxyClient->getTargetUri() . $this->endpoint;
        $request = new Request('POST', $workflowRequestUri, $headers, $body);

        $this->monolog->addInfo(sprintf(
            'Waiting for up to %ds for a playbook response...',
            $this->timeout
        ), $this->eventContext);

        try {
            $response = $this->proxyClient->send($request, [
                'timeout' => $this->timeout
            ]);
            $httpResponse = $this->logResponse($response);
        } catch (BadResponseException $exception) {
            $httpResponse = $this->logResponse($exception->getResponse());
        }

        $event->setResponse($httpResponse);
    }

    /**
     * @param ResponseInterface $response
     * @return HttpResponse
     */
    private function logResponse(ResponseInterface $response)
    {
        $httpResponse = new HttpResponse($response->getBody(), $response->getStatusCode(), $response->getHeaders());

        $success = $httpResponse->isSuccessful();

        $this->monolog->addRecord(
            $success ? Logger::INFO : Logger::ALERT,
            sprintf(
                'Playbook engine returned %s code: %s',
                $success ? 'successful' : 'non successful',
                $httpResponse->getStatusCode()
            ),
            $this->eventContext
        );

        return $httpResponse;
    }

    /**
     * Handles api call triggers
     *
     * @param DataEvent $event
     */
    public function handleAction(DataEvent $event)
    {
        $this->handleTriggerType(Triggers::ACTION, $event);
    }

    /**
     * Handlers create triggers
     *
     * @param DataEvent $event
     */
    public function handlePreCreate(DataEvent $event)
    {
        $data = $event->getData();
        if ($data != null && method_exists($data, 'getUUID') && $data->getUUID() == null) {
            $data->setUUID(Uuid::uuid4()->toString());
        }
        $this->handleTriggerType(Triggers::PRE_CREATE, $event);
    }

    /**
     * Handles update triggers
     *
     * @param DataEvent $event
     */
    public function handlePreUpdate(DataEvent $event)
    {
        $this->handleTriggerType(Triggers::PRE_UPDATE, $event);
    }

    /**
     * Handles delete triggers
     *
     * @param DataEvent $event
     */
    public function handlePreDelete(DataEvent $event)
    {
        $this->handleTriggerType(Triggers::PRE_DELETE, $event);
    }

    /**
     * Handlers create triggers
     *
     * @param DataEvent $event
     */
    public function handlePostCreate(DataEvent $event)
    {
        $this->handleTriggerType(Triggers::POST_CREATE, $event);
    }

    /**
     * Handles update triggers
     *
     * @param DataEvent $event
     */
    public function handlePostUpdate(DataEvent $event)
    {
        $this->handleTriggerType(Triggers::POST_UPDATE, $event);
    }

    /**
     * Handles delete triggers
     *
     * @param DataEvent $event
     */
    public function handlePostDelete(DataEvent $event)
    {
        $this->handleTriggerType(Triggers::POST_DELETE, $event);
    }
}
